-----------------------------------------------------------------------------
Implementation of FIRST

(c) 1993-2001 Andy Gill, Simon Marlow
-----------------------------------------------------------------------------

> module First ( mkFirst ) where

> import GenUtils
> import NameSet ( NameSet )
> import qualified NameSet as Set
> import Grammar
> import Data.IntSet (IntSet)

> import Data.Map ( Map )
> import qualified Data.Map as Map hiding ( Map )

> import Ranges
> import Debug.Trace

\subsection{Utilities}

> joinSymSets :: (Show a, Show b) => b -> (b -> b -> b) -> (a -> Map Name b) -> [a] -> Map Name b
> joinSymSets def union f = foldr 
>       (\ h b -> let
>                   h' = f h
>                 in
>                    if incEmpty h'
>                    then {-trace ("EMPTY:"++show h ++ show h') $-} Map.unionWith union b (Map.filterWithKey (\k _ -> not (isEmpty k)) h')
>                    else {-trace ("NOT_EMPTY:"++show h++show h')-} h')
>        (Map.singleton epsilonTok def) -- This is how we detect epsilon rules (epsilonTok is a dummy token that isn't used anywhere else)
> --       (Map.empty) -- TODO: why was epsilonTok here? (Set.singleton epsilonTok)

Does the Set include the $\epsilon$ symbol ?

> incEmpty :: Map Name b -> Bool
> incEmpty set = any (isEmpty . fst) (Map.toList set)

\subsection{Implementation of FIRST}

> mkFirst :: Grammar -> [(Name, LookaheadRel)] -> Map Name LookaheadRel
> mkFirst (Grammar { first_term = fst_term
>		   , lookupProdNo = prodNo
>		   , lookupProdsOfName = prodsOfName
>		   , non_terminals = nts
>		   })
>       = trace ("mkFirst:"++show (length (show env))) $ joinSymSets lookaheadRelE unionLookaheadRel (\ (name,ranges) -> case lookup name env of
>                               Nothing -> Map.singleton name ranges
>                               Just ix -> Map.map (composeLookaheadAndRange ranges) ix)
>   where
>       env :: [(Name, Map Name RangeRel)]
>       env = mkClosure (==) (getNext fst_term prodNo prodsOfName)
>               [ (name, Map.empty) | name <- nts ]

> getNext :: Name -> (a -> (b, [(Name, RangeRel)], c, d)) -> (Name -> [a])
>         -> [(Name, Map Name RangeRel)] -> [(Name, Map Name RangeRel)]
> getNext fst_term prodNo prodsOfName env = 
>		[ (nm, next nm) | (nm,_) <- env ]
>    where
>       fn :: (Name, RangeRel) -> Map Name RangeRel
>    	fn (t, ranges) | t == errorTok || t >= fst_term = Map.singleton t ranges
>    	fn (x, ranges) = case lookup x env of
>           	        Just t -> Map.map (composeRelOut ranges) t
>                       Nothing -> error "attempted FIRST(e) :-("

> 	next :: Name -> Map Name RangeRel
> 	next t | t >= fst_term = Map.singleton t rangeRelE
> 	next n =
>       	foldb (Map.unionWith unionRangeRel)
>               	[ joinSymSets rangeRelE unionRangeRel fn ([(x, {-Why did I do this?-} composeRelOut r rangeRelE) | (x,r) <- snd4 (prodNo rl)]) | 
>				rl <- prodsOfName n ]

My little hack

> snd4 :: (a, b, c, d) -> b
> snd4 (_,b,_,_) = b
