-----------------------------------------------------------------------------
Generation of LALR parsing tables.

(c) 1993-1996 Andy Gill, Simon Marlow
(c) 1997-2001 Simon Marlow
-----------------------------------------------------------------------------

> module LALR
>	(genActionTable, genGotoTable, genLR0items, precalcClosure0,
>	 propLookaheads, calcLookaheads, mergeLookaheadInfo, countConflicts,
>	 Lr0Item(..), Lr1Item)
>	where

> import GenUtils
> import Data.Set ( Set )
> import qualified Data.Set as Set hiding ( Set )
> import Data.Map ( Map )
> import qualified Data.Map as Map hiding ( Map )
> import qualified NameSet
> import NameSet ( NameSet )
> import Grammar

> import Control.Monad.ST
> import Data.Array.ST
> import Data.Array as Array
> import Data.List (nub)

> import Ranges

> import Debug.Trace

> unionNameMap :: (Name -> NameSet) -> NameSet -> NameSet
> unionNameMap f = NameSet.fold (NameSet.union . f) NameSet.empty

This means rule $a$, with dot at $b$ (all starting at 0)

> data Lr0Item = Lr0 {-#UNPACK#-}!Int {-#UNPACK#-}!Int			-- (rule, dot)
>       deriving (Eq,Ord

#ifdef DEBUG

>	,Show

#endif

>       )

> data Lr1Item = Lr1 {-#UNPACK#-}!Int {-#UNPACK#-}!Int (Map Name LookaheadRel)  -- (rule, dot, lookahead).  Note that the lookahead reals are relative to the item not to the item set
>  deriving (Show)
> {-instance Show Lr1Item where
>   show (Lr1 rule dot lookahead) = "(Lr1 " ++ show rule ++ " " ++ show dot ++ " " ++ show (Map.filterWithKey (\k _ -> k /= dummyTok) lookahead) ++ ")"-}

> type RuleList = [(Lr0Item, RangeRel)]

-----------------------------------------------------------------------------
Generating the closure of a set of LR(0) items

Precalculate the rule closure for each non-terminal in the grammar,
using a memo table so that no work is repeated.

> type PrecalcMap = Map Name {-Rule Name-} RangeRel

> precalcClosure0 :: Grammar -> Name -> RangeRel -> RuleList
> precalcClosure0 g = 
>      trace ("PRECALC:"++show (length (show info'))) $
>	\n rel -> case lookup n info' of
>		Nothing -> []
>		Just c -> {-trace ("PRECALCED:"++show n ++ "=>"++show c) $-} map (\(i,r) -> (i, composeRelOut rel r)) c
>  where

>       -- Simply converts "info" to "RuleList" instead of "NameSet"
>	info' :: [(Name, [(Lr0Item, RangeRel)])]
>	info' = map (\(n,rules) -> (n,map (\(rule, rel) -> (Lr0 rule 0, rel)) (Map.toList rules))) info

>	info :: [(Name, PrecalcMap)]
>	info = mkClosure (==) (\f -> map (follow f) f)
>              -- Initial info is [(non_term, prods of non-term)]
>			(map (\nt -> (nt,Map.fromList [(rule, rangeRelE) | rule <- lookupProdsOfName g nt])) nts)

>	follow :: [(Name, PrecalcMap)] -> (Name, PrecalcMap) -> (Name, PrecalcMap)
>	follow f (nt,rules) = (nt, Map.unionWith unionRangeRel rules
>                              (Map.foldrWithKey (\k v s -> {-trace ("Follow:"++show nt++show rules++"/"++show k++show v++show s++show f) $-} Map.unionWith unionRangeRel (followNT f k v) s)
>                               Map.empty rules))

>	followNT :: [(Name, PrecalcMap)] -> Name -> RangeRel -> PrecalcMap
>	followNT f rule item_rel = 
>		case findRule g rule 0 of
>			Just (nt, nt_rel) | nt >= firstStartTok && nt < fst_term ->
>				case lookup nt f of
>					Just rs -> {-trace ("COMPOSE:"++show rule++"/"++show (nt,nt_rel)++show rs) $-} Map.map (composeRelOut (composeRelOut item_rel nt_rel)) rs
>					Nothing -> error "followNT"
>			_ -> Map.empty

>	nts = non_terminals g
>	fst_term = first_term g

> closure0 :: Grammar -> (Name -> RangeRel -> RuleList) -> Map Lr0Item RangeRel -> Map Lr0Item RangeRel
> closure0 g closureOfNT set = trace ("CLOSURE0:"++show set ++ show res) res
>    where
>       res = Map.foldrWithKey addRules Map.empty set
> 	fst_term = first_term g
>       addRules :: Lr0Item -> RangeRel -> Map Lr0Item RangeRel -> Map Lr0Item RangeRel
>	addRules rule ranges set' =
>         Map.unionsWith unionRangeRel (set' : map (uncurry Map.singleton) ((rule, ranges) : closureOfRule rule ranges))
> 
>       closureOfRule :: Lr0Item -> RangeRel -> [(Lr0Item, RangeRel)]
>	closureOfRule (Lr0 rule dot) ranges = 
>           case findRule g rule dot of 
>           	(Just (nt, rel)) | nt >= firstStartTok && nt < fst_term 
>		   -> closureOfNT nt (composeRelOut ranges rel)
>               _  -> []

-----------------------------------------------------------------------------
Generating the closure of a set of LR(1) items

> closure1 :: Grammar -> ([(Name, LookaheadRel)] -> Map Name LookaheadRel) -> [(Lr1Item, RangeRel)] -> [(Lr1Item, RangeRel)]
> closure1 g first set
>       = {-trace ("CLOSURE1:"++show set) $ trace ("CLOSURE1':"++show (fst res)) $-}  fst res
>	where
>       res = (mkClosure (\(_,new) _ -> null new) addItems ([],set))
>	fst_term = first_term g

>	addItems :: ([(Lr1Item, RangeRel)],[(Lr1Item, RangeRel)]) -> ([(Lr1Item, RangeRel)],[(Lr1Item, RangeRel)])
>	addItems (old_items, new_items) = (new_old_items, new_new_items)
>	  where
>		new_old_items = new_items `union_items` old_items
>		new_new_items = subtract_items 
>				   (foldr union_items [] (map fn new_items))
>					new_old_items

>		fn :: (Lr1Item, RangeRel) -> [(Lr1Item, RangeRel)]
>		fn (Lr1 rule dot lookahead, ranges) =
>		    case lookupProdNo g rule of { (_name,lhs,_,_) ->
>		    case drop dot lhs of
>			(b:beta) | fst b >= firstStartTok && fst b < fst_term ->
>			    let lookahead' = Map.unionsWith unionLookaheadRel [ first ([(x, composeRangeAndLookahead (snd b) (composeLookaheadAndRange lookaheadRelE r))| (x,r) <- beta] ++ [(a, composeRangeAndLookahead (snd b) aLookahead)]) | (a, aLookahead) <- Map.toList lookahead]
>			    in [ (Lr1 rule' 0 lookahead', composeRelOut ranges (snd b)) | rule' <- lookupProdsOfName g (fst b) ]
>			_ -> []
>		    }

Subtract the first set of items from the second.

> subtract_items :: [(Lr1Item, RangeRel)] -> [(Lr1Item, RangeRel)] -> [(Lr1Item, RangeRel)]
> subtract_items items1 items2 = foldr (subtract_item items2) [] items1

These utilities over item sets are crucial to performance.

Stamp on overloading with judicious use of type signatures...

> subtract_item :: [(Lr1Item, RangeRel)] -> (Lr1Item, RangeRel) -> [(Lr1Item, RangeRel)] -> [(Lr1Item, RangeRel)]
> subtract_item [] i result = i : result
> subtract_item ((Lr1 rule dot as, ranges):items) i@(Lr1 rule' dot' as', ranges') result =
>	case compare rule' rule of
>		LT -> i : result
>		GT -> carry_on
>		EQ -> case compare dot' dot of
>			LT -> i : result
>			GT -> carry_on
>			EQ -> if ranges /= ranges' then error "subtract_item"
>                             else case {-trace ("Map.diff:"++show as'++show as) $-} Map.differenceWith maybeDiffRel as' as of
>				bs | Map.null bs -> result
>				   | otherwise -> (Lr1 rule dot bs, ranges) : result
>  where
>	carry_on = subtract_item items i result
>       maybeDiffLookahead (LookaheadRel x) (LookaheadRel y) = if Set.null z then Nothing else Just (LookaheadRel z) where
>          z = Set.filter (`notIn` y) x
>          notIn x' y = not (any (\y' -> unionLookaheadRel1 x' y' == y') (Set.toList y))
>          unionLookaheadRel1 (s1,l1) (s2, l2) = (unionRangeRel s1 s2, unionRangeRel l1 l2)

>       maybeDiffRel = maybeDiffLookahead
> {-
>       maybeDiffRel x y = if x /= y then trace ("maybeDiffRel:"++show x ++ show y) $ Just x -- error $ "maybeDiffRel:" ++ show x ++ show y
>                          else Nothing -}

Union two sets of items.

> union_items :: [(Lr1Item, RangeRel)] -> [(Lr1Item, RangeRel)] -> [(Lr1Item, RangeRel)]
> union_items is [] = is
> union_items [] is = is
> union_items (i@(Lr1 rule dot as, ranges):is) (i'@(Lr1 rule' dot' as', ranges'):is') =
>	case compare rule rule' of
>		LT -> drop_i
>		GT -> drop_i'
>		EQ -> case compare dot dot' of
>			LT -> drop_i
>			GT -> drop_i'
>			EQ -> (Lr1 rule dot (Map.unionWith unionLookaheadRel as as'), (ranges `unionRangeRel` ranges')) : union_items is is'
>  where
>	drop_i  = i  : union_items is (i':is')
>	drop_i' = i' : union_items (i:is) is'

-----------------------------------------------------------------------------
goto(I,X) function

The input should be the closure of a set of kernel items I together with
a token X (terminal or non-terminal.  Output will be the set of kernel
items for the set of items goto(I,X) (i.e., the "derivative" of I by X
constructed by stepping each element of I over an X and which
is the state that we should goto from state I after seeing an X).
This function does not compute the closure of the resulting derivative
but assumes that the incoming I is closed.

This code keeps the ranges the same b/c we are skipping *over*
not *into* the given name.

> gotoClosure :: Grammar -> Map Lr0Item RangeRel -> Name -> Map EdgeRel (Map Lr0Item RangeRel)
> gotoClosure gram i x = Map.foldrWithKey (\item ranges result -> Map.unionWith (Map.unionWith unionRangeRel) (fn item ranges) result) Map.empty i
>    where
>       fn :: Lr0Item -> RangeRel -> Map EdgeRel (Map Lr0Item RangeRel)
>       fn (Lr0 rule_no dot) ranges =
>          case findRule gram rule_no dot of
>               Just (t,rel) | x == t -> Map.singleton (EdgeRel { edgeRelNT = composeLookaheadAndRange lookaheadRelE rel, edgeRelItem = ranges }) (Map.singleton (Lr0 rule_no (dot+1)) rangeRelE)
>               _ -> Map.empty

-----------------------------------------------------------------------------
Generating LR0 Item sets

The item sets are generated in much the same way as we find the
closure of a set of items: we use two sets, those which have already
generated more sets, and those which have just been generated.  We
keep iterating until the second set is empty.

The addItems function is complicated by the fact that we need to keep
information about which sets were generated by which others.

> type ItemSetWithGotos = (Map Lr0Item RangeRel, [(Name, (Map EdgeRel Int))])

> genLR0items :: Grammar -> (Name -> RangeRel -> RuleList) -> [ItemSetWithGotos]
> genLR0items g precalcClosures
>	= fst (mkClosure (\(_old,new) _ -> null new)
>               addItems
>                 (([],startRules)))
>  where

>    n_starts = length (starts g)
>    startRules :: [Map Lr0Item RangeRel]
>    startRules = [ Map.unionWith unionRangeRel (Map.singleton (Lr0 rule 0) rangeRelE) (closure0 g precalcClosures ((Map.singleton (Lr0 rule 0) rangeRelE))) | rule <- [0..n_starts] ]

>    tokens = non_terminals g ++ terminals g

>    addItems :: ([ItemSetWithGotos], [Map Lr0Item RangeRel])
>	      -> ([ItemSetWithGotos], [Map Lr0Item RangeRel])
>	      
>    addItems (oldSets,newSets) = {-trace ("addItems:"++show (oldSets,newSets)++show (newOldSets,newNewSets)++"\ngotos:"++show gotos ++ show intgotos)-} (newOldSets, reverse newNewSets)
>     where
>	
>	newOldSets = oldSets ++ (zip newSets (map Map.toAscList intgotos))

>	itemSets = map fst oldSets ++ newSets

First thing to do is for each set in I in newSets, generate goto(I,X)
for each token (terminals and nonterminals) X.

>	gotos :: [[(Name, Map EdgeRel (Map Lr0Item RangeRel))]] -- after going to (Name,Ranges) get to state (Map Lr0Item Ranges)
>	gotos = map (filter (not . Map.null . snd)) -- Inner map can't be empty but we must remove empty outer maps
>	    (map (\i -> let i' = closure0 g precalcClosures i in
> --                      [ (x, [(newR, newI) | (newI, newR) <- Map.toList (gotoRanges (gotoClosure g i' x))]) |
>                         trace ("GOTOS CLOSURE:"++show i++show i') [ (x, gotoClosure g i' x) | x <- tokens ]) newSets)

>       {-gotoRanges :: Map Lr0Item RangeRel -> Map (Map Lr0Item RangeRel) EdgeRel
>       gotoRanges i = Map.union gotoInf (foldr addGoto Map.empty [0..rangeMax]) where
>           rangeMax = maximum (-1 : map rangeMaxNotInf (Map.elems i)) :: Int
>           addGoto :: Int -> Map (Map Lr0Item Ranges) Ranges -> Map (Map Lr0Item Ranges) Ranges
>           addGoto indent sets =
>               Map.union sets (Map.singleton (gotoAt indent) (singletonRanges indent))
>           gotoInf = if Map.null s then Map.empty else Map.singleton s (Ranges [(rangeMax+1, Inf)])
>           s = (gotoAt (rangeMax+1))
>           gotoAt :: Int -> Map Lr0Item Ranges
>           gotoAt indent = Map.filterWithKey (filterRel indent) i
>           filterRel :: Int -> Lr0Item -> Ranges -> Bool -- returns true iff the next token being at "indent" will work with (Lr0Item, Ranges)
>           filterRel indent (Lr0 rule dot) ranges = case findRule g rule (dot - 1) of
>                                                           -- -1 b/c gotoClosure already did +1
>               Nothing -> error "foo"
>               Just (_, rel) -> indent `elemRanges` (applyRangeRel ranges rel) -}

Next, we assign each new set a number, which is the index of this set
in the list of sets comprising all the sets generated so far plus
those generated in this iteration.  We also filter out those sets that
are new, i.e. don't exist in the current list of sets, so that they
can be added.

We also have to make sure that there are no duplicate sets in the
*current* batch of goto(I,X) sets, as this could be disastrous.  I
think I've squished this one with the '++ reverse newSets' in
numberSets.

numberSets is built this way so we can use it quite neatly with a foldr.
Unfortunately, the code's a little opaque.


>       -- Numbers new sets and eliminates duplicates
>	numberSets 
>		:: [(Name, Map EdgeRel (Map Lr0Item RangeRel))] -- element from gotos
>		-> (Int, -- gensym for sets
>		    [Map Name (Map EdgeRel Int)], -- Goto edges
>		    [Map Lr0Item RangeRel]) -- sets of items already constructed
>		-> (Int, [Map Name (Map EdgeRel Int)], [Map Lr0Item RangeRel])
>
>	numberSets [] (i,gotos',newSets') = (i,(Map.empty:gotos'),newSets')
>	numberSets ((x,gotoix):rest) (i,g':gotos',newSets')
>	   = numberSets rest (foo (Map.toList gotoix) i (g':gotos') newSets')
>            where foo :: [(EdgeRel, Map Lr0Item RangeRel)] -- edge from gotos for name "x"
>                      -> Int --gensym
>                      -> [Map Name (Map EdgeRel Int)] --Goto edges
>                      -> [Map Lr0Item RangeRel] --sets alreay constructed
>                      -> (Int, -- gensym
>                          [Map Name (Map EdgeRel Int)], --Goto edges
>                          [Map Lr0Item RangeRel])
>                  foo [] i gotos' newSets' = (i, gotos', newSets')
>                  foo ((range,ix):rest) i (g':gotos') newSets'=
>                    case indexInto 0 ix (itemSets ++ reverse newSets') of
>			Just j  -> foo rest i     ((Map.insertWith (Map.unionWith (error "numberSets")) x (Map.singleton range j) g'):gotos') newSets'
>			Nothing -> foo rest (i+1) ((Map.insertWith (Map.unionWith (error "numberSets1")) x (Map.singleton range i) g'):gotos') (ix:newSets')
>	numberSets _ _ = error "genLR0items/numberSets: Unhandled case"

Finally, do some fiddling around to get this all in the form we want.

>	intgotos :: [Map Name (Map EdgeRel Int)] -- Goto edges
>	newNewSets  :: [Map Lr0Item RangeRel]
>	(_, (_:intgotos), newNewSets) =
>		foldr numberSets (length newOldSets, [Map.empty], []) gotos

> indexInto :: Eq a => Int -> a -> [a] -> Maybe Int
> indexInto _ _ []		   = Nothing
> indexInto i x (y:ys) | x == y    = Just i
>		       | otherwise = indexInto (i+1) x ys

-----------------------------------------------------------------------------
Computing propagation of lookaheads

ToDo: generate this info into an array to be used in the subsequent
calcLookaheads pass.

> propLookaheads 
>	:: Grammar
>	-> [(Map Lr0Item RangeRel, [(Name, Map EdgeRel Int)])]		-- LR(0) kernel sets
>	-> ([(Name, LookaheadRel)] -> Map Name LookaheadRel)			-- First function
>	-> (
>		[(Int{-set-}, (Lr0Item, RangeRel){-rule-}, Map Name LookaheadRel{-lookahead-})],	-- spontaneous lookaheads
>		Array Int{-set-} [((Lr0Item, RangeRel){-rule-}, Int{-origin set-}, (Lr0Item, RangeRel){-origin rule-}, RangeRel{-TODO:origin item indent-})]	-- propagated lookaheads
>	   )

> propLookaheads gram sets first = (concat s, array (0,length sets - 1) 
>			[ (a,b) | (a,b) <- p ])
>   where

>     (s,p) = unzip (zipWith propLASet sets [0..])

>     propLASet :: (Map Lr0Item RangeRel, [(Name, Map EdgeRel Int)]) -- LR(0) kernel sets
>               -> Int -- self index
>               -> ([(Int, (Lr0Item, RangeRel), Map Name LookaheadRel)], -- spontaneous
>                   (Int,[((Lr0Item, RangeRel), Int, (Lr0Item, RangeRel), RangeRel)])) -- self index + propagated
>     propLASet (set,goto) i = {-trace ("propLASet:"++show i++show set++show goto) $-} (start_spont ++ concat s', (i, concat p'))
>	where

>	  (s',p') = unzip (map propLAItem (Map.toList set))

>	  -- spontaneous EOF lookaheads for each start state & rule...
>	  start_info :: [(String, Name, Name, Bool)]
>	  start_info = starts gram	

>	  start_spont :: [(Int, (Lr0Item, RangeRel), Map Name LookaheadRel)]
>	  start_spont	= [ (start, (Lr0 start 0, rangeRelE), 
>                                 Map.singleton (startLookahead gram partial) lookaheadRelE)
>			  | (start, (_,_,_,partial)) <- 
>				zip [ 0 .. length start_info - 1] start_info]

>	  propLAItem :: (Lr0Item, RangeRel) -> ([(Int, (Lr0Item, RangeRel), Map Name LookaheadRel)], [((Lr0Item, RangeRel), Int, (Lr0Item, RangeRel), RangeRel)])
>	  propLAItem item@(Lr0 rule dot, ranges) = (spontaneous, propagated)
>	    where

>               lookaheads = Map.unions $ (Map.singleton dummyTok {-(error "propLAItem/j")-}{-(LookaheadRel (Set.empty))-} lookaheadRelE) :
>                 [ Map.singleton (startLookahead gram partial) lookaheadRelE
>                   | (start, (_,_,_,partial)) <- zip [ 0 .. length start_info - 1] start_info
>                 , start == rule]
>		j = closure1 gram first [(Lr1 rule dot lookaheads, ranges)]

>		spontaneous :: [(Int{-set-}, (Lr0Item, RangeRel){-item-}, Map Name LookaheadRel {-lookahead-})]
>		spontaneous = concat [ 
>		 (case findRule gram rule' dot' of
>		     Nothing -> []
>		     Just (x,_rel)  -> case lookup x goto of
>			 	  Nothing -> error "spontaneous"
>				  Just ks  ->
>					case Map.filterWithKey (\name _ -> name /= dummyTok) ts of
>					   ts' | Map.null ts' -> []
>					       | otherwise -> [(k, (Lr0 rule' (dot' + 1), Map.findWithDefault (error $ "propLAItem/spont") (Lr0 rule' (dot' + 1)) (fst $ sets !! k)), ts') | (_r,k) <- Map.toList ks, LookaheadRel r' <- [edgeRelNT _r], Set.member (rangeRelE, _rel) r'])
>			| (Lr1 rule' dot' ts, ranges') <- j ]

>		propagated :: [((Lr0Item, RangeRel), Int, (Lr0Item, RangeRel), RangeRel)]
>		propagated = concat [
>		 (case findRule gram rule' dot' of
>		     Nothing -> []
>		     Just (x,_rel)  -> case lookup x goto of
>				  Nothing -> error "propagated"
>				  Just ks  -> [(item, k, (Lr0 rule' (dot' + 1), Map.findWithDefault (error "propLAItem/prop") (Lr0 rule' (dot' + 1)) (fst $ sets !! k)), dummyRel) | (_r, k) <- Map.toList ks, LookaheadRel r' <- [edgeRelNT _r], Set.member (rangeRelE, _rel) r'])
>			| (Lr1 rule' dot' ts, ranges) <- j, (dummy, LookaheadRel dummyLookahead) <- Map.toList ts, dummyTok == dummy, (dummyRel,next) <- Set.toList dummyLookahead, next == rangeRelE || error "bad dummyLookahead"]

The lookahead for a start rule depends on whether it was declared
with %name or %partial: a %name parser is assumed to parse the whole
input, ending with EOF, whereas a %partial parser may parse only a
part of the input: it accepts when the error token is found.

> startLookahead :: Grammar -> Bool -> Name
> startLookahead gram partial = if partial then errorTok else eof_term gram

-----------------------------------------------------------------------------
Calculate lookaheads

Special version using a mutable array:

> calcLookaheads
>	:: Int					-- number of states
>	-> [(Int, (Lr0Item, RangeRel), Map Name LookaheadRel)]		-- spontaneous lookaheads
>	-> Array Int [((Lr0Item, RangeRel), Int, (Lr0Item, RangeRel), RangeRel)]	-- propagated lookaheads
>	-> Array Int [((Lr0Item, RangeRel), Map Name LookaheadRel)]

> calcLookaheads n_states spont prop
>	= runST (do
>	    arr <- newArray (0,n_states) []
>           trace ("START_CALCLOOK" ++ show spont) $ return ()
>	    propagate arr (foldr fold_lookahead [] spont)
>           trace "END_CALCLOOK" $ return ()
>	    freeze arr
>	)

>   where
>	propagate :: STArray s Int [((Lr0Item, RangeRel), Map Name LookaheadRel)]
>			 -> [(Int, (Lr0Item, RangeRel), Map Name LookaheadRel)] -> ST s ()
>	propagate _   []  = trace ("PROP_END:") $ return ()
>	propagate arr new = trace ("PROP_NEW:") $ do 
>               return ()
>		let
>		   items = [ (i,item'',Map.map (composeRangeAndLookahead range) s) | (j,item,s) <- new, 
>				            (item',i,item'',range) <- prop ! j,
>				            item == item'
>  --                          if i == 348 then trace ("BANG!:"++show j++show s) True else True]
>  --                          if i == 134 then trace ("BANG!:"++show j++show s) True else True]
>  --                          if i == 61 then trace ("BANG!:"++show j++show s) True else True]
>                                           ]
>               items `seq` return ()
>		new_new <- get_new arr items []
>               --trace ("Get new:"++show items ++ show new_new) $ return ()
>		add_lookaheads arr new
>               --trace ("PROP_NEW':"++show (length new_new)++"/"++show (length items)) $ return ()
>		propagate arr new_new

This function is needed to merge all the (set_no,item,name) triples
into (set_no, item, set name) triples.  It can be removed when we get
the spontaneous lookaheads in the right form to begin with (ToDo).

> add_lookaheads :: STArray s Int [((Lr0Item, RangeRel), Map Name LookaheadRel)]
>                -> [(Int, (Lr0Item, RangeRel), Map Name LookaheadRel)]
>                -> ST s ()
> add_lookaheads _      [] = return ()
> add_lookaheads arr ((i,item,s) : lookaheads) = do
>	las <- readArray arr i
>	writeArray arr i (add_lookahead item s las)
>	add_lookaheads arr lookaheads

> get_new :: STArray s Int [((Lr0Item, RangeRel), Map Name LookaheadRel)]
>         -> [(Int, (Lr0Item, RangeRel), Map Name LookaheadRel)]
>         -> [(Int, (Lr0Item, RangeRel), Map Name LookaheadRel)]
>         -> ST s [(Int, (Lr0Item, RangeRel), Map Name LookaheadRel)]
> get_new _   []                   new = return new
> get_new arr (l@(i,_item,_s):las) new = do
>	state_las <- readArray arr i
>	get_new arr las (get_new' l [(item, lookahead) | (item, lookahead) <- state_las] new)

> add_lookahead :: (Lr0Item, RangeRel) -> Map Name LookaheadRel -> [((Lr0Item, RangeRel), Map Name LookaheadRel)] ->
> 			[((Lr0Item, RangeRel), Map Name LookaheadRel)]
> add_lookahead item s [] = [(item,s)]
> add_lookahead item s (m@(item',s') : las)
>	| item == item' = (item, Map.unionWith unionLookaheadRel s s') : las
>	| otherwise     = m : add_lookahead item s las

> get_new' :: (Int,(Lr0Item, RangeRel), Map Name LookaheadRel) -> [((Lr0Item, RangeRel), Map Name LookaheadRel)]
>          -> [(Int,(Lr0Item, RangeRel), Map Name LookaheadRel)]
>          -> [(Int,(Lr0Item, RangeRel), Map Name LookaheadRel)]
> get_new' l [] new = l : new
> get_new' l@(i,item,s) ((item',s') : las) new
>	| item == item' =
>		let s'' = Map.differenceWith maybeDiffLookahead s s' in
>		if Map.null s'' then new else
>		((i,item,s''):new)
>	| otherwise = 
>		get_new' l las new
>  where
>    maybeDiffLookahead (LookaheadRel x) (LookaheadRel y) = if Set.null z then Nothing else Just (LookaheadRel z) where
>      z = Set.filter (`notIn` y) x
>      notIn x' y = not (any (\y' -> unionLookaheadRel1 x' y' == y') (Set.toList y))
>      unionLookaheadRel1 (s1,l1) (s2, l2) = (unionRangeRel s1 s2, unionRangeRel l1 l2)

> --       maybeDiffLookahead x@(x1,x2) y@(y1,y2)
> --        | x == y = Nothing
> --        | unionRangeRel x1 y1 == y1 && unionRangeRel x2 y2 == y2 = Nothing
> --        | otherwise = Just x

= if x /= y then Just $ diffLookaheadRel x y --unionLookaheadRel x y --error $ "maybeDiffRel2" ++ show x ++ show y
                          else Nothing


> fold_lookahead :: (Int,(Lr0Item, RangeRel), Map Name LookaheadRel)
>               -> [(Int,(Lr0Item, RangeRel), Map Name LookaheadRel)]
>		-> [(Int,(Lr0Item, RangeRel), Map Name LookaheadRel)]
> fold_lookahead l [] = [l]
> fold_lookahead l@(i,item,s) (m@(i',item',s'):las)
>  	| i == i' && item == item' = (i,item, Map.unionWith unionLookaheadRel s s'):las
>	| i < i' = (i,item,s):m:las
>	| otherwise = m : fold_lookahead l las

Normal version:

calcLookaheads
      :: Int					-- number of states
      -> [(Int, Lr0Item, Set Name)]		-- spontaneous lookaheads
      -> Array Int [(Lr0Item, Int, Lr0Item)]	-- propagated lookaheads
      -> Array Int [(Lr0Item, Set Name)]

calcLookaheads n_states spont prop
      = rebuildArray $ fst (mkClosure (\(_,new) _ -> null new) propagate
         ([], foldr addLookahead [] spont))
      where

        rebuildArray :: [(Int, Lr0Item, Set Name)] -> Array Int [(Lr0Item, Set Name)]
        rebuildArray xs = accumArray (++) [] (0,n_states-1)
      		      [ (a, [(b,c)]) | (a,b,c) <- xs ]

        propagate (las,new) = 
      	let
      	   items = [ (i,item'',s) | (j,item,s) <- new, 
      			       (item',i,item'') <- prop ! j,
      			       item == item' ]
      	   new_new = foldr (\i new -> getNew i las new) [] items
      	   new_las = foldr addLookahead las new
      	in
      	(new_las, new_new)

addLookahead :: (Int,Lr0Item,Set Name) -> [(Int,Lr0Item,Set Name)]
      	-> [(Int,Lr0Item,Set Name)]
addLookahead l [] = [l]
addLookahead l@(i,item,s) (m@(i',item',s'):las)
 	| i == i' && item == item' = (i,item, s `Set.union` s'):las
      | i < i' = (i,item,s):m:las
      | otherwise = m : addLookahead l las

getNew :: (Int,Lr0Item,Set Name) -> [(Int,Lr0Item,Set Name)]
      -> [(Int,Lr0Item,Set Name)] -> [(Int,Lr0Item,Set Name)]
getNew l [] new = l:new
getNew l@(i,item,s) (m@(i',item',s'):las) new
 	| i == i' && item == item' = 
      	let s'' = filter (`notElem` s') s in
      	if null s'' then new else
      	((i,item,s''):new)
      | i < i'    = (i,item,s):new
      | otherwise = getNew l las new

-----------------------------------------------------------------------------
Merge lookaheads

Stick the lookahead info back into the state table.

> mergeLookaheadInfo
>	:: Array Int [((Lr0Item, RangeRel), Map Name LookaheadRel)] 	-- lookahead info
>	-> [(Map Lr0Item RangeRel, [(Name, Map EdgeRel Int)])] 	-- state table
>	-> [ ([(Lr1Item, RangeRel)], [(Name, Map EdgeRel Int)]) ]

> mergeLookaheadInfo lookaheads sets
>	= zipWith mergeIntoSet sets [0..]
>	where

>	  mergeIntoSet :: (Map Lr0Item RangeRel, [(Name, Map EdgeRel Int)]) -> Int -> ([(Lr1Item, RangeRel)], [(Name, Map EdgeRel Int)])
>	  mergeIntoSet (items, goto) i
>		= (concat (map mergeIntoItem (Map.toList items)), goto)
>		where

>	  	  mergeIntoItem :: (Lr0Item, RangeRel) -> [(Lr1Item, RangeRel)]
>	  	  mergeIntoItem item@(Lr0 rule dot, ranges)
>		     = [(Lr1 rule dot la, ranges)]
>		     where la = case [ s | (item',s) <- lookaheads ! i,
>					    item == item' ] of
>					[] -> Map.empty
>					[x] -> x
>					_ -> error "mergeIntoItem"

-----------------------------------------------------------------------------
Generate the goto table

This is pretty straightforward, given all the information we stored
while generating the LR0 sets of items.

Generating the goto table doesn't need lookahead info.

> genGotoTable :: Grammar -> [ItemSetWithGotos] -> GotoTable
> genGotoTable g sets = {-trace ("GOTO TABLE:"++show gotoTable)-} gotoTable
>   where
>	Grammar{ first_nonterm = fst_nonterm,
>		 first_term    = fst_term,
>		 non_terminals = non_terms } = g
>
>	-- goto array doesn't include %start symbols
>       gotoTable  = listArray (0,length sets-1)
>         [
>           (array (fst_nonterm, fst_term-1) [ 
>		(n, case lookup n goto of
>			Nothing -> Map.empty
>			Just s  -> s)
>                             | n <- non_terms,
>			        n >= fst_nonterm, n < fst_term ])
>                 | (_set,goto) <- sets  ]

-----------------------------------------------------------------------------
Generate the action table

> genActionTable :: Grammar -> ([(Name, LookaheadRel)] -> Map Name LookaheadRel) ->
>		 [([(Lr1Item, RangeRel)],[(Name, Map EdgeRel Int)])] -> ActionTable
> genActionTable g first sets = actionTable
>   where
>	Grammar { first_term = fst_term,
>		  terminals = terms,
>		  starts = starts',
>       	  priorities = prios } = g

>	n_starts = length starts'
>	isStartRule rule = rule < n_starts -- a bit hacky, but it'll do for now

>       term_lim = (head terms,last terms)
>       actionTable = array (0,length sets-1)
>             [ (set_no, accumArray res'
>				 Map.empty {-LR'Fail-} term_lim 
>				(possActions goto set))
>                   | ((set,goto),set_no) <- zip sets [0..] ]

>       possAction :: [(Name, Map EdgeRel Int)] -> [(Lr1Item, RangeRel)] -> (Lr1Item, RangeRel) -> [(Name, (EdgeRel, LRAction))]
>       possAction goto _set (Lr1 rule pos la, range) = 
>          case findRule g rule pos of
>               Just (t, rel) | t >= fst_term || t == errorTok -> 
>			case lookup t goto of
>                       	Nothing -> []
>                               Just j  ->
>                                 [ (t,(edgeRel, LR'Shift k{-'-} (case lookup t prios of
>                                                                   Nothing -> No
>                                                                   Just p -> p)))
>                                    | (edgeRel, k) <- Map.toList j ]
>               Nothing
>		   | isStartRule rule
>		   -> let (_,_,_,partial) = starts' !! rule in
>		      [ (startLookahead g partial, (EdgeRel { edgeRelNT = composeLookaheadAndRange lookaheadRelE rangeRelZ, edgeRelItem = range}, LR'Accept{-'-})) ]
>                  | otherwise   
>		   -> case lookupProdNo g rule of
>                          (_,_,_,p) -> map (\(n,rel) -> (n, (EdgeRel { edgeRelNT = rel, edgeRelItem = range}, (LR'Reduce rule p)))) (Map.toAscList la) {-(repeat (LR'Reduce rule p))-}
>               _ -> []

>	possActions goto coll = 
>		(concat [ possAction goto coll item |
>				item <- closure1 g first coll ])

These comments are now out of date! /JS

Here's how we resolve conflicts, leaving a complete record of the
conflicting actions in an LR'Multiple structure for later output in
the info file.

Shift/reduce conflicts are always resolved as shift actions, and
reduce/reduce conflicts are resolved as a reduce action using the rule
with the lowest number (i.e. the rule that comes first in the grammar
file.)

NOTES on LR'MustFail: this was introduced as part of the precedence
parsing changes.  The problem with LR'Fail is that it is a soft
failure: we sometimes substitute an LR'Fail for an LR'Reduce (eg. when
computing default actions), on the grounds that an LR'Fail in this
state will also be an LR'Fail in the goto state, so we'll fail
eventually.  This may not be true with precedence parsing, though.  If
there are two non-associative operators together, we must fail at this
point rather than reducing.  Hence the use of LR'MustFail.


NOTE: on (LR'Multiple as a) handling
      PCC [sep04] has changed this to have the following invariants:
        * the winning action appears only once, in the "a" slot
	* only reductions appear in the "as" list
	* there are no duplications
      This removes complications elsewhere, where LR'Multiples were 
      building up tree structures... 

>       res' :: Map EdgeRel LRAction -> (EdgeRel, LRAction) -> Map EdgeRel LRAction
>       res' m (edgeRel, action) = Map.insertWith res edgeRel action m
>       res :: LRAction -> LRAction -> LRAction
>       res LR'Fail x = x
>       res x LR'Fail = x
>	res LR'MustFail _ = LR'MustFail
>	res _ LR'MustFail = LR'MustFail
>	res x x' | x == x' = x
>       res (LR'Accept) _ = LR'Accept
>       res _ (LR'Accept) = LR'Accept

>	res (LR'Multiple as x) (LR'Multiple bs x')
>        | x == x' = LR'Multiple (nub $ as ++ bs) x
>		-- merge dropped reductions for identical action

>	res (LR'Multiple as x) (LR'Multiple bs x')
>	       = case res x x' of 
>		   LR'Multiple cs a
>		     | a == x    -> LR'Multiple (nub $ x' : as ++ bs ++ cs) x
>		     | a == x'   -> LR'Multiple (nub $ x  : as ++ bs ++ cs) x'
>		     | otherwise -> error "failed invariant in resolve"
>		       		-- last means an unexpected change
>		   other -> other
>		-- merge dropped reductions for clashing actions, but only 
>		-- if they were S/R or R/R

>	res a@(LR'Multiple _ _) b = res a (LR'Multiple [] b)
>	res a b@(LR'Multiple _ _) = res (LR'Multiple [] a) b 
>	  -- leave cases above to do the appropriate merging

>       res a@(LR'Shift {}) b@(LR'Reduce {}) = res b a
>       res a@(LR'Reduce _ p) b@(LR'Shift _ p')
>		= case (p,p') of
>                      (No,_) -> LR'Multiple [a] b	-- shift wins
>                      (_,No) -> LR'Multiple [a] b	-- shift wins
>                      (Prio c i, Prio _ j)
>               		| i < j     -> b
>               		| i > j     -> a
>			        | otherwise ->
>				   case c of
>                                     LeftAssoc  -> a
>                                     RightAssoc -> b
>                                     None       -> LR'MustFail
>       res a@(LR'Reduce r p) b@(LR'Reduce r' p')
>		= case (p,p') of
>                      (No,_) -> LR'Multiple [a] b	-- give to earlier rule?
>                      (_,No) -> LR'Multiple [a] b
>                      (Prio _ i, Prio _ j)
>               		| i < j     -> b
>               		| j > i     -> a
>				| r < r'    -> LR'Multiple [b] a
>				| otherwise -> LR'Multiple [a] b
>       res _ _ = error "confict in resolve"

-----------------------------------------------------------------------------
Count the conflicts

> countConflicts :: ActionTable -> (Array Int (Int,Int), (Int,Int))
> countConflicts action
>   = (conflictArray, foldr (\(a,b) (c,d) -> (a+c, b+d)) (0,0) conflictList)
>   
>   where
>	   
>	conflictArray = listArray (Array.bounds action) (map (const (0, 0)) (assocs action)) --listArray (Array.bounds action) conflictList
>       conflictList :: [(Int, Int)]
>	conflictList  = map (const (0, 0)) (assocs action) --map countConflictsState (assocs action)
>
>	countConflictsState (_state, actions)
>	  = foldr countMultiples (0,0) (elems actions)
>	  where
>	    countMultiples (LR'Multiple (_:_) (LR'Shift{})) (sr,rr) 
>	    	= (sr + 1, rr)
>	    countMultiples (LR'Multiple (_:_) (LR'Reduce{})) (sr,rr) 
>	    	= (sr, rr + 1)
>	    countMultiples (LR'Multiple _ _) _
>	    	= error "bad conflict representation"
>	    countMultiples _ c = c

-----------------------------------------------------------------------------

> findRule :: Grammar -> Int -> Int -> Maybe (Name, RangeRel)
> findRule g rule dot = 
>	case lookupProdNo g rule of
>	   (_,lhs,_,_) -> case drop dot lhs of
>		            (a:_) -> Just a
>      			    _     -> Nothing
