module ParamRules(expand_rules) where

import AbsSyn
import Control.Monad.Writer
import Control.Monad.Error
import Control.Monad.Instances()  -- mtl is broken, so we use Either monad
import Data.List(partition,intersperse)
import qualified Data.Set as S
import qualified Data.Map as M    -- XXX: Make it work with old GHC.
import GenUtils
import Data.List

import Ranges

import Debug.Trace

expand_rules :: [String] -> [Rule] -> Either String [Rule1]
expand_rules tokens rs = do
                     let (funs,rs1) = split_rules rs
                     (as,is) <- runM2 (mapM (\r -> inst_rule tokens r []) rs1)
                     bs <- make_insts tokens rs funs (S.toList is) S.empty
                     let rules_with_first_eq = gen_first_eq tokens (as++bs) -- map (make_first_eq tokens) (as++bs)
                     --trace ("GEN_FIRST_EQ"++show rules_with_first_eq) $ return ()
                     let rules_with_null = gen_null tokens rules_with_first_eq -- map (make_nulls tokens) (as++bs)
                     --trace ("GEN_NULL:"++show rules_with_null) $ return ()
                     let rs' = rules_with_null -- as++bs++first_eq++nulls
                     --trace ("Finishing expand_rules" ++ show rs' ) $ return ()
                     let rs'' = loop tokens (remove_empty_prods (map ("NULL:"++) tokens) rs')
                     {-trace ("Final rules:"++show rs''++show tokens) $-}
                     return rs''
--                     return rs

gen_first_eq :: [String] -> [Rule1] -> [Rule1]
gen_first_eq tokens rules = --trace "GEN_FIRST_EQ" $
  case nub $ sort $ need_first_eq_names tokens rules of
    [] -> trace ("GEN_FIRST_EQ_DONE") rules
    need -> gen_first_eq tokens (rules ++ map (make_first_eq tokens) [r | r@(name', _, _) <- rules, name' `elem` need])

need_first_eq_names :: [String] -> [Rule1] -> [RuleName]
need_first_eq_names tokens rules = --trace ("NEED_FIRST_EQ_NAMES:"++show rules)
  [ t' | (_, prods, _) <- rules, (terms, _, _, _) <- prods, (t, _) <- terms, null [() | (r, _, _) <- rules, r == t], take (length "FIRST_EQ:") t == "FIRST_EQ:", let t' = drop (length "FIRST_EQ:") t, t' `notElem` tokens]


gen_null :: [String] -> [Rule1] -> [Rule1]
gen_null tokens rules =
  case nub $ sort $ need_null_names tokens rules of
    [] -> rules
    need -> trace ("GEN_NULL:"++show need++show rules) $ gen_null tokens (rules ++ map (make_null tokens) [r | r@(name', _, _) <- rules, name' `elem` need])

need_null_names :: [String] -> [Rule1] -> [RuleName]
need_null_names tokens rules = --trace ("NEED_NULL_NAMES:"++show rules)
  [ t' | (_, prods, _) <- rules, (terms, _, _, _) <- prods, (t, _) <- terms, null [() | (r, _, _) <- rules, r == t], take (length "NULL:") t == "NULL:", let t' = drop (length "NULL:") t, t' `notElem` tokens]

loop tokens rules =
  let empty = [name | (name, prods, _) <- rules, null prods] in
  --trace ("NullLoop:"++show empty ++show rules) $
  if null empty then rules else loop tokens (remove_empty_prods empty rules)

{-
  let rules' = [(name,
        [ p | p@(terms, code, line, prec) <- prods, not (any (`elem` empty') (map fst terms)) ],
        code) | (name, prods, code) <- rules, name `notElem` empty'] in
-}

remove_empty_prods empty rules = [(name,
        [ p | p@(terms, code, line, prec) <- prods, not (any (`elem` empty) (map fst terms)) ],
        code) | (name, prods, code) <- rules, name `notElem` empty]


make_null :: [String] -> Rule1 -> Rule1
make_null terminals (name, prods, code) = ("NULL:"++name, concatMap (make_null_prod terminals) prods, code)

make_null_prod :: [String] -> Prod1 -> [Prod1]
make_null_prod terminals (terms, code, line, prec)
  | and [t `notElem` terminals | (t, _) <- terms] =
    [(map make_null_term terms, code, line, prec)]
  | otherwise = []

make_null_term (name, rel)
  | take (length "FIRST_EQ:") name == "FIRST_EQ:" = ("NULL:" ++ drop (length "FIRST_EQ:") name, rangeRelE)
  | otherwise = ("NULL:"++name, rangeRelE)
  

make_first_eq :: [String] -> Rule1 -> Rule1
make_first_eq terminals (name, prods, code) = ("FIRST_EQ:"++name, concatMap (make_first_eq_prod terminals []) prods, code)

make_first_eq_prod :: [String] -> [(RuleName, RangeRel)] -> Prod1 -> [Prod1]
make_first_eq_prod terminals pre_terms ([], code, line, prec) = []
make_first_eq_prod terminals pre_terms (term1:terms, code, line, prec) = 
  (pre_terms ++ make_first_eq_term terminals term1 ++ terms, code, line, prec) :
  (make_first_eq_prod terminals (pre_terms ++ [make_null_term term1]) (terms, code, line, prec))

make_first_eq_term terminals (name, rel)
  | name `elem` terminals = [(name, rangeRelE)]
  | take (length "FIRST_EQ:") name == "FIRST_EQ:" = [(name, rangeRelE)]
  | otherwise = [("FIRST_EQ:"++name, rangeRelE)]


{-
nullableRules nullables rules =
  nullables' 

nullableRule nullables (_, _, prod, _)

nullableProd ([], _, _, _) = True
nullableProd (ts, _, _, _) = all nullableTerm ts

nullableTerm (App t args) = nullableRule 
-}

--data IndentRel = IndentRelZ | IndentRelE | IndentRelI deriving (Eq, Ord)

type RuleName = String
type Inst     = (RuleName, [(RuleName, RangeRel)])
type Funs     = M.Map RuleName Rule
type Rule1    = (RuleName,[Prod1],Maybe String)
type Prod1    = ([(RuleName, RangeRel)],String,Int,Maybe String)

inst_name :: Inst -> RuleName
inst_name (f,[])  = f
inst_name (f,xs)  = f ++ "(" ++ concat (intersperse "," (map show xs)) ++ ")"


-- | A renaming substitution used when we instantiate a parameterized rule.
type Subst    = [(RuleName,(RuleName, RangeRel))]
type M1       = Writer (S.Set Inst)
type M2       = ErrorT String M1

-- | Collects the instances arising from a term.
from_term :: [String] -> Subst -> Term -> M1 (RuleName, RangeRel)
from_term tokens s (App f [])  = return $ case lookup f s of
                            Just g  -> g
                            Nothing -> (f, if f `elem` tokens then rangeRelI else rangeRelE)

from_term tokens s (App "INDENT" [t]) = do [(x, r)] <- from_terms tokens s [t]
                                           trace ("INDENT:"++show x ++ show r) $ return ()
                                           return (x, composeRelOut rangeRelI r)

from_term tokens s (App "EQUAL" [t]) = do [(x, r)] <- from_terms tokens s [t]
                                          trace ("Equal:"++show x ++ show r) $ return ()
                                          return (x, rangeRelE)

from_term tokens s (App "FIRST_EQ" [t])
  | App t' [] <- t, t' `elem` tokens = from_term tokens s (App "EQUAL" [t])
  | App "FIRST_EQ" _ <- t = trace ("SKIPPING:" ++ show t) $ from_term tokens s t
  | otherwise = do
      (x, _) <- from_term tokens s t
      --let i = ("FIRST_EQ", xs)
      --tell (S.singleton i)
      trace ("First Equal:"++show t ++ show x) $ return ()
      return $ ("FIRST_EQ:" ++ x, rangeRelE)

from_term tokens s (App "ZERO" [t]) = do [(x, r)] <- from_terms tokens s [t]
                                         trace ("ZERO:"++show x ++ show r ++ "=>" ++ show (composeRelOut rangeRelZ rangeRelI)) $ return ()
                                         return (x, composeRelOut rangeRelZ r)

from_term tokens s (App f ts)  = do xs <- from_terms tokens s ts
                                    let i = (f, xs)
                                    tell (S.singleton i)
                                    return $ (inst_name i, rangeRelE)

-- | Collects the instances arising from a list of terms.
from_terms :: [String] -> Subst -> [Term] -> M1 [(RuleName, RangeRel)]
from_terms tokens s ts = mapM (from_term tokens s) ts

-- XXX: perhaps change the line to the line of the instance
inst_prod :: [String] -> Subst -> Prod -> M1 Prod1
inst_prod tokens s (ts,c,l,p)  = do xs <- from_terms tokens s ts
                                    return (xs,c,l,p)

inst_rule :: [String] -> Rule {-rule-} -> [(RuleName, RangeRel)] {-actual arguments-}-> M2 Rule1
inst_rule tokens (x,xs,ps,t) ts  = do
                               s <- build xs ts []
                               ps1 <- lift $ mapM (inst_prod tokens s) ps
                               let y = inst_name (x,ts)
                               return (y,ps1,t)    -- XXX: type?
  where build :: [String] -> [(RuleName, RangeRel)] -> Subst -> M2 Subst
        build (x':xs') (t':ts') m = build xs' ts' ((x',t'):m)
        build [] [] m  = return m
        build xs' [] _  = err ("Need " ++ show (length xs') ++ " more arguments")
        build _ ts' _   = err (show (length ts') ++ " arguments too many.")

        err m = throwError ("In " ++ inst_name (x,ts) ++ ": " ++ m)

{-
make_first_eq_prod :: [String] -> Prod -> M1 [Prod1]
make_first_eq_prod tokens ([], code, line, prec) = return []
make_first_eq_prod tokens (term1:terms, code, line, prec) = do
  term1' <- from_term tokens [] (App "FIRST_EQ" [term1])
  terms' <- from_terms tokens [] terms
  return [(term1':terms', code, line, prec)]
-}

make_rule :: [String] -> [Rule] -> Funs -> Inst -> M2 Rule1
{-
make_rule tokens nonFuns funs ("FIRST_EQ",[(name, range)])
  | [(_, [], prods, prodType)] <- [r | r@(x,_,_,_) <- nonFuns, x == name] = do
      prods' <- lift $ liftM concat $ mapM (make_first_eq_prod tokens) prods
      return (inst_name ("FIRST_EQ", [(name, range)]), prods', prodType)
-}

make_rule tokens nonFuns funs (f,xs) =
  case M.lookup f funs of
    Just r  -> inst_rule tokens r xs
    Nothing -> throwError ("Undefined rule: " ++ f ++ show xs)

runM2 :: ErrorT e (Writer w) a -> Either e (a, w)
runM2 m = case runWriter (runErrorT m) of
            (Left e,_)   -> Left e
            (Right a,xs) -> Right (a,xs)

make_insts :: [String] -> [Rule] -> Funs -> [Inst] -> S.Set Inst -> Either String [Rule1]
make_insts _ _ _ [] _ = return []
make_insts tokens nonFuns funs is done =
  do (as,ws) <- runM2 (mapM (make_rule tokens nonFuns funs) is)
     let done1 = S.union (S.fromList is) done
     let is1 = filter (not . (`S.member` done1)) (S.toList ws)
     bs <- make_insts tokens nonFuns funs is1 done1
     return (as++bs)


split_rules :: [Rule] -> (Funs,[Rule])
split_rules rs = let (xs,ys) = partition has_args rs
                 in (M.fromList [ (x,r) | r@(x,_,_,_) <- xs ],ys)
  where has_args (_,xs,_,_) = not (null xs)



