> module Ranges where

> --import Grammar
> import Data.Map ( Map )
> import qualified Data.Map as Map hiding ( Map )
> import Data.Set ( Set )
> import qualified Data.Set as Set hiding ( Set )

> import Data.List

> unionMap :: (Ord b) => (a -> Ranges -> Map b Ranges) -> Map a Ranges -> Map b Ranges
> unionMap f = Map.foldrWithKey (\item ranges set -> (Map.unionWith unionRanges (f item ranges) set)) Map.empty

> data Nat = Fin Int | Inf deriving (Ord, Eq, Show)

> data Ranges = Ranges [(Int, Nat)] deriving (Ord, Eq, Show) -- Bounds are inclusive
> type RangeMap a = [(Ranges, a)]

> {-
> lowerBound (RangeRel eq indent zero) ranges =
>   relE' = if eq then range else emptyRanges
>   indent' = case indent of
>               Nothing -> emptyRanges
>               Just x -> Ranges [(Fin $ minRanges ranges + x, Inf)]
>   zero' = case zero of
>             Nothing -> emptyRanges
>             Just x -> Ranges [(Fin x, Inf)]
> upperBound (RangeRel eq indent zero) ranges =
>   eq' = if eq then ragnes else emptyRanges
>   indent' = case indent of
>               Nothing -> emptyRanges
>               Just x -> case maxRanges of
>                           Nothing -> Ranges [(Fin 0, Inf)]
>                           Just y -> Ranges [(Fin 0, Fin (y - x))]
>   zero' = case zero of
>             Nothing -> emptyRanges
>             Just x -> if maxRanges ranges > x then Ranges [(Fin 0, Inf)] else emptyRanges
> bound low hi c x = lowerBound low c `rangeIntersect` upperBound hi x
> -}

> --data RangeRel = RangeRel Bool{-RelE-} (Maybe Int{-RelI^n RelI-})  (Maybe Int{-RelI^n RelZ-}) deriving (Eq, Ord)

> data EdgeRel = EdgeRel { edgeRelNT :: LookaheadRel{-nonterm-}, edgeRelItem :: RangeRel{-item-} }
>   deriving (Ord, Eq, Show)

> unionRanges :: Ranges -> Ranges -> Ranges
> unionRanges (Ranges x) (Ranges y) = Ranges $ fn (sort (x ++ y)) where
>   fn [] = []
>   fn ((x_low, Inf):_) = [(x_low, Inf)]
>   fn [x] = [x]
>   fn ((x_low, Fin x_high):(y_low, y_high):ys)
>      | y_low > x_high + 1 = (x_low, Fin x_high) : fn ((y_low, y_high):ys)
>      | otherwise = fn ((x_low, max (Fin x_high) y_high) : ys)
> differenceRanges :: Ranges -> Ranges -> Ranges
> differenceRanges (Ranges x) (Ranges y) = Ranges $ fn (sort x) (sort y) where
>   fn x [] = x
>   fn [] _ = []
> {-
>   fn [(x_low, Inf)] [(y_low, Inf)]
>      | y_low > x_low = (x_low, Fin (y_low - 1))
>      | otherwise = []
>   fn ((x_low, Fin x_high):xs) [(y_low, Inf y_high)]
>      | y_low > x_high = (x_low, x_high) : fn xs ((y_low, y_high):ys)
>      | y_low > x_low  = [(x_low, y_low-1)]
>      | otherwise      = []
>   fn [(x_low, Inf)] ((y_low, Fin y_high):ys)
>      | y_low  >  x_low  = (x_low, y_low-1) : fn [(y_high+1, Inf)] ys
>      | y_high >= x_low  = fn [y_high+1, Inf] ys
>      | otherwise = fn [(x_low, Inf)] ys
> -}
>   fn ((x_low, x_high):xs) ((y_low, y_high):ys)
>      | Fin y_low  >  x_high = (x_low, x_high ) : fn xs ((y_low, y_high):ys)
>      | y_high >= x_high = 
>         if y_low > x_low
>         then (x_low, Fin (y_low-1)) : fn xs ((y_low, y_high):ys)
>         else fn xs ((y_low, y_high):ys)
>      | y_low  >  x_low  = (x_low, Fin (y_low-1)) : fn ((y_low,x_high):xs) ((y_low,y_high):ys)
>      | y_high >= Fin x_low  =
>         case y_high of
>           Inf -> []
>           Fin y_high' -> fn ((y_high'+1, x_high):xs) ys
>      | otherwise        = fn ((x_low, x_high):xs) ys
> maybeDifferenceRanges :: Ranges -> Ranges -> Maybe Ranges
> maybeDifferenceRanges x y = case differenceRanges x y of Ranges [] -> Nothing; z -> Just z

> singletonRanges :: Int -> Ranges
> singletonRanges x = Ranges [(x, Fin x)]
> greaterThanRanges x = Ranges [(x+1, Inf)]
> rangeMaxNotInf (Ranges [(x, Inf)]) = x - 1
> rangeMaxNotInf (Ranges [(x, Fin y)]) = y
> rangeMaxNotInf (Ranges (x:xs)) = rangeMaxNotInf (Ranges xs)
> elemRanges :: Int -> Ranges -> Bool
> elemRanges x (Ranges []) = False
> elemRanges x (Ranges [(y, Inf)]) = x >= y
> elemRanges x (Ranges ((y_low, Fin y_high):ys)) 
>   | x < y_low = False
>   | x <= y_high = True
>   | otherwise = elemRanges x (Ranges ys)
> minRanges (Ranges rs) = minimum (map fst rs)

> --adjustRange IndentRelZ range = range

> -- TODO: eq is not being represented correctly here (e.g. union E I should not be E)
> data RangeRel = RangeRel Bool{-RelE-} Bool{-RelZ-} (Maybe Int{-RelI^n (n>=1)-})  (Maybe Int{-RelI^n RelZ (n>=1)-}) deriving (Eq, Ord)

> instance Show RangeRel where
>   showsPrec _ (RangeRel e z i iz) s = "[" ++ (if e then "=" else "") ++ (if z then "Z" else "") ++ (case i of Just n -> "I"++show n; Nothing -> "") ++ (case iz of Just n -> "Z"++show n; Nothing -> "") ++ "]" ++ s

> rangeRelE = RangeRel True  False Nothing Nothing
> rangeRelZ = RangeRel False True  Nothing Nothing
> rangeRelI = RangeRel False False (Just 1) Nothing

> incMaybeInt = maybe Nothing (Just . (+1))

> --adjustRelOut :: IndentRel -> RangeRel -> RangeRel
> --adjustRelOut IndentRelZ (RangeRel woZ wZ) = RangeRel Nothing (minMaybe woZ wZ)
> --adjustRelOut IndentRelE (RangeRel woZ wZ) = RangeRel woZ wZ
> --adjustRelOut IndentRelI (RangeRel woZ wZ) = RangeRel (incMaybeInt woZ) wZ

Needed to account for multipaths to the same item

> unionRangeRel :: RangeRel -> RangeRel -> RangeRel
> unionRangeRel (RangeRel e1 z1 i1 iz1) (RangeRel e2 z2 i2 iz2) =
>   RangeRel (e1 || e2) (z1 || z2) (minMaybe i1 i2) (minMaybe iz1 iz2)

> composeRelOut :: RangeRel{-outside-} -> RangeRel{-Inside-} -> RangeRel
> composeRelOut (RangeRel e1 z1 i1 iz1) (RangeRel e2 z2 i2 iz2) =
>  RangeRel (e1 && e2) (z2 || z1 && e2)
>             (addMaybe i1 i2 `minMaybe`
>              (if e1 then i2 else Nothing) `minMaybe`
>              (if e2 then i1 else Nothing))
>             (addMaybe i2 iz1 `minMaybe`
>              iz2 `minMaybe`
>              (if e2 then iz1 else Nothing) `minMaybe`
>              (if z1 then i2 else Nothing))

> minMaybe Nothing (Just x) = Just x
> minMaybe (Just x) Nothing = Just x
> minMaybe Nothing Nothing = Nothing
> minMaybe (Just x) (Just y) = Just (min x y)

> addMaybe Nothing (Just x) = Nothing
> addMaybe (Just x) Nothing = Nothing
> addMaybe Nothing Nothing = Nothing
> addMaybe (Just x) (Just y) = Just (x + y)

> {-
> applyRangeRel ranges (RangeRel e woZ wZ) = (if e then ranges else Ranges []) `unionRanges` (applyWithoutZ ranges woZ) `unionRanges` (applyWithZ ranges wZ) where
>       applyWithoutZ ranges Nothing = Ranges []
>       applyWithoutZ ranges (Just n) = Ranges [(minRanges ranges + n, Inf)]
>       applyWithZ ranges Nothing = Ranges []
>       applyWithZ ranges (Just n) = Ranges [(n, Inf)]
> -}

> newtype LookaheadRel = LookaheadRel (Set (RangeRel {-parent to self-}, RangeRel {- parent to lookahead -})) deriving (Eq, Ord)
> instance Show LookaheadRel where
>   show (LookaheadRel r) = "{" ++ concatMap show (Set.toList r) ++ "}"
> lookaheadRelE = LookaheadRel $ Set.singleton (rangeRelE, rangeRelE)
> unionLookaheadRel :: LookaheadRel -> LookaheadRel -> LookaheadRel
> unionLookaheadRel (LookaheadRel r1) (LookaheadRel r2) = LookaheadRel (Set.fromList r3) where
>  looks = Set.toList (Set.union r1 r2) -- TODO: fix and to the right way
>  looks' = map (foldr1 (\x y -> (fst x, unionRangeRel (snd x) (snd y)))) $ groupBy (\x y -> fst x == fst y) looks
>  looks'' = map (foldr1 (\x y -> (unionRangeRel (fst x) (fst y), snd x))) $ groupBy (\x y -> snd x == snd y) looks'
>  looks''' = case looks'' of
>               [(x1,y1),(x2,y2)] | x1 == unionRangeRel x1 x2 && y1 == unionRangeRel y1 y2 -> [(x1,y1)]
>               [(x1,y1),(x2,y2)] | x2 == unionRangeRel x1 x2 && y2 == unionRangeRel y1 y2 -> [(x2,y2)]
>               _ -> looks''
>  r3 = looks'''
> composeLookaheadAndRange :: LookaheadRel -> RangeRel -> LookaheadRel
> composeLookaheadAndRange (LookaheadRel lookahead) range = LookaheadRel $ Set.map f lookahead where
>   f (self,lookahead) = (self, composeRelOut lookahead range)
> composeRangeAndLookahead :: RangeRel -> LookaheadRel -> LookaheadRel
> composeRangeAndLookahead  range (LookaheadRel lookahead) = LookaheadRel $ Set.map f lookahead where
>   f (self,lookahead) = (composeRelOut self range, lookahead)
