{-# LANGUAGE FlexibleInstances, RecordWildCards #-}
import Criterion
import Criterion.Main
import Criterion.Monad
import Criterion.IO
import Criterion.Measurement
import Criterion.Analysis
import Criterion.Types
import Data.Monoid
import Criterion.Environment
import System.Environment
import Statistics.Resampling.Bootstrap (Estimate(..))
import Control.Monad.Trans
import Criterion.Config
import Control.DeepSeq
import Data.Generics
import Data.List
import qualified Data.Vector.Unboxed as V (filter, toList, length)
import Control.Monad
import Language.Haskell.Syntax
import Language.Haskell.Parser
import Language.Haskell.Lexer
import Language.Haskell.ParseMonad
import Statistics.Types
import Statistics.Types (Sample)
import Criterion.Report
import Text.Printf
import Data.Digest.SHA2

import Debug.Trace

class IsEOF x where
  isEOF :: x -> Bool

instance IsEOF Token where
  isEOF EOF = True
  isEOF _ = False

instance IsEOF (Token, a) where
  isEOF (EOF, _) = True
  isEOF _ = False

runLexer str = runParser (lexer f) str where
  f x = if isEOF x
        then return 0
        else do x <- lexer f
                return (x + 1)

getFile :: Bool -> Bool -> String -> IO ()
getFile printAST printSHA file = do
  pre_str <- readFile file
  let str = unlines (map (\line -> case line of '#':' ':_ -> ""; _ -> line) (lines pre_str))
  trace (file) $ return ()
  case parseModule str of
    ParseFailed srcLoc string -> do
      putStrLn $ file ++ ": Parse Error at " ++ show srcLoc
    ParseOk ast -> do
      case runLexer str of
        ParseFailed srcLoc string -> do
          putStrLn $ file ++ ": Lex error: " ++ string ++ " at " ++ show srcLoc
        ParseOk numTokens -> do
          putStrLn $ file ++ ": bytes="++(show (length str))++" toks="++(show numTokens)++" lines="++(show $ length $ lines str)++" nodes="++(show (everything (+) (const (1::Int)) ast))++(if printSHA then " SHA1="++(show $ sha256Ascii (show ast)) else "")
          when printAST $ putStrLn (show ast)

main :: IO ()
main = do
  args <- getArgs
  case args of
    "--ast" : rest -> mapM_ (getFile True True) rest
    "--nosha" : rest -> mapM_ (getFile False False) rest
    _ -> mapM_ (getFile False True) args
