{-# OPTIONS_GHC -w #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Language.Haskell.Parser
-- Copyright   :  (c) Simon Marlow, Sven Panne 1997-2000
-- License     :  BSD-style (see the file libraries/base/LICENSE)
--
-- Maintainer  :  libraries@haskell.org
-- Stability   :  experimental
-- Portability :  portable
--
-- Haskell parser.
--
-----------------------------------------------------------------------------
module Language.Haskell.Parser (
		parseModule, parseModuleWithMode,
		ParseMode(..), defaultParseMode, ParseResult(..)) where

import Language.Haskell.Syntax
import Language.Haskell.ParseMonad
import Language.Haskell.Lexer
import Language.Haskell.ParseUtils
import Debug.Trace
import Data.List (sort)
import Language.Haskell.Ranges

-- parser produced by Happy Version 1.18.9

data HappyAbsSyn 
	= HappyTerminal ((Token,Ranges))
	| HappyErrorToken (Int, Ranges)
	| HappyAbsSyn4 (HsModule)
	| HappyAbsSyn5 (SrcLoc)
	| HappyAbsSyn6 (([HsImportDecl],[HsDecl]))
	| HappyAbsSyn9 (())
	| HappyAbsSyn11 (Maybe [HsExportSpec])
	| HappyAbsSyn12 ([HsExportSpec])
	| HappyAbsSyn15 (HsExportSpec)
	| HappyAbsSyn16 ([HsImportDecl])
	| HappyAbsSyn19 (Bool)
	| HappyAbsSyn20 (Maybe Module)
	| HappyAbsSyn21 (Maybe (Bool, [HsImportSpec]))
	| HappyAbsSyn22 ((Bool, [HsImportSpec]))
	| HappyAbsSyn24 ([HsImportSpec])
	| HappyAbsSyn25 (HsImportSpec)
	| HappyAbsSyn26 ([HsCName])
	| HappyAbsSyn27 (HsCName)
	| HappyAbsSyn28 (HsDecl)
	| HappyAbsSyn29 (Int)
	| HappyAbsSyn30 (HsAssoc)
	| HappyAbsSyn31 ([HsOp])
	| HappyAbsSyn32 ([HsDecl])
	| HappyAbsSyn37 ([HsType])
	| HappyAbsSyn46 ([HsName])
	| HappyAbsSyn48 (HsSafety)
	| HappyAbsSyn49 (String)
	| HappyAbsSyn50 (HsName)
	| HappyAbsSyn51 (HsType)
	| HappyAbsSyn54 (HsQName)
	| HappyAbsSyn55 (HsQualType)
	| HappyAbsSyn56 (HsContext)
	| HappyAbsSyn58 ((HsName, [HsName]))
	| HappyAbsSyn60 ([HsConDecl])
	| HappyAbsSyn61 (HsConDecl)
	| HappyAbsSyn62 ((HsName, [HsBangType]))
	| HappyAbsSyn64 (HsBangType)
	| HappyAbsSyn66 ([([HsName],HsBangType)])
	| HappyAbsSyn68 (([HsName],HsBangType))
	| HappyAbsSyn70 ([HsQName])
	| HappyAbsSyn81 (HsRhs)
	| HappyAbsSyn82 ([HsGuardedRhs])
	| HappyAbsSyn83 (HsGuardedRhs)
	| HappyAbsSyn84 (HsExp)
	| HappyAbsSyn91 ([HsPat])
	| HappyAbsSyn92 (HsPat)
	| HappyAbsSyn97 ([HsExp])
	| HappyAbsSyn100 ([HsStmt])
	| HappyAbsSyn101 (HsStmt)
	| HappyAbsSyn102 ([HsAlt])
	| HappyAbsSyn107 (HsAlt)
	| HappyAbsSyn108 (HsGuardedAlts)
	| HappyAbsSyn109 ([HsGuardedAlt])
	| HappyAbsSyn110 (HsGuardedAlt)
	| HappyAbsSyn118 ([HsFieldUpdate])
	| HappyAbsSyn120 (HsFieldUpdate)
	| HappyAbsSyn131 (HsOp)
	| HappyAbsSyn132 (HsQOp)
	| HappyAbsSyn146 (HsLiteral)
	| HappyAbsSyn148 (Module)

{- to allow type-synonyms as our monads (likely
 - with explicitly-specified bind and return)
 - in Haskell98, it seems that with
 - /type M a = .../, then /(HappyReduction M)/
 - is not allowed.  But Happy is a
 - code-generator that can just substitute it.
type HappyReduction m = 
	   (Int, Ranges) 
	-> ((Token,Ranges))
	-> HappyState ((Token,Ranges)) (HappyStk HappyAbsSyn -> m HappyAbsSyn)
	-> [HappyState ((Token,Ranges)) (HappyStk HappyAbsSyn -> m HappyAbsSyn)] 
	-> HappyStk HappyAbsSyn 
	-> m HappyAbsSyn
-}

action_0,
 action_1,
 action_2,
 action_3,
 action_4,
 action_5,
 action_6,
 action_7,
 action_8,
 action_9,
 action_10,
 action_11,
 action_12,
 action_13,
 action_14,
 action_15,
 action_16,
 action_17,
 action_18,
 action_19,
 action_20,
 action_21,
 action_22,
 action_23,
 action_24,
 action_25,
 action_26,
 action_27,
 action_28,
 action_29,
 action_30,
 action_31,
 action_32,
 action_33,
 action_34,
 action_35,
 action_36,
 action_37,
 action_38,
 action_39,
 action_40,
 action_41,
 action_42,
 action_43,
 action_44,
 action_45,
 action_46,
 action_47,
 action_48,
 action_49,
 action_50,
 action_51,
 action_52,
 action_53,
 action_54,
 action_55,
 action_56,
 action_57,
 action_58,
 action_59,
 action_60,
 action_61,
 action_62,
 action_63,
 action_64,
 action_65,
 action_66,
 action_67,
 action_68,
 action_69,
 action_70,
 action_71,
 action_72,
 action_73,
 action_74,
 action_75,
 action_76,
 action_77,
 action_78,
 action_79,
 action_80,
 action_81,
 action_82,
 action_83,
 action_84,
 action_85,
 action_86,
 action_87,
 action_88,
 action_89,
 action_90,
 action_91,
 action_92,
 action_93,
 action_94,
 action_95,
 action_96,
 action_97,
 action_98,
 action_99,
 action_100,
 action_101,
 action_102,
 action_103,
 action_104,
 action_105,
 action_106,
 action_107,
 action_108,
 action_109,
 action_110,
 action_111,
 action_112,
 action_113,
 action_114,
 action_115,
 action_116,
 action_117,
 action_118,
 action_119,
 action_120,
 action_121,
 action_122,
 action_123,
 action_124,
 action_125,
 action_126,
 action_127,
 action_128,
 action_129,
 action_130,
 action_131,
 action_132,
 action_133,
 action_134,
 action_135,
 action_136,
 action_137,
 action_138,
 action_139,
 action_140,
 action_141,
 action_142,
 action_143,
 action_144,
 action_145,
 action_146,
 action_147,
 action_148,
 action_149,
 action_150,
 action_151,
 action_152,
 action_153,
 action_154,
 action_155,
 action_156,
 action_157,
 action_158,
 action_159,
 action_160,
 action_161,
 action_162,
 action_163,
 action_164,
 action_165,
 action_166,
 action_167,
 action_168,
 action_169,
 action_170,
 action_171,
 action_172,
 action_173,
 action_174,
 action_175,
 action_176,
 action_177,
 action_178,
 action_179,
 action_180,
 action_181,
 action_182,
 action_183,
 action_184,
 action_185,
 action_186,
 action_187,
 action_188,
 action_189,
 action_190,
 action_191,
 action_192,
 action_193,
 action_194,
 action_195,
 action_196,
 action_197,
 action_198,
 action_199,
 action_200,
 action_201,
 action_202,
 action_203,
 action_204,
 action_205,
 action_206,
 action_207,
 action_208,
 action_209,
 action_210,
 action_211,
 action_212,
 action_213,
 action_214,
 action_215,
 action_216,
 action_217,
 action_218,
 action_219,
 action_220,
 action_221,
 action_222,
 action_223,
 action_224,
 action_225,
 action_226,
 action_227,
 action_228,
 action_229,
 action_230,
 action_231,
 action_232,
 action_233,
 action_234,
 action_235,
 action_236,
 action_237,
 action_238,
 action_239,
 action_240,
 action_241,
 action_242,
 action_243,
 action_244,
 action_245,
 action_246,
 action_247,
 action_248,
 action_249,
 action_250,
 action_251,
 action_252,
 action_253,
 action_254,
 action_255,
 action_256,
 action_257,
 action_258,
 action_259,
 action_260,
 action_261,
 action_262,
 action_263,
 action_264,
 action_265,
 action_266,
 action_267,
 action_268,
 action_269,
 action_270,
 action_271,
 action_272,
 action_273,
 action_274,
 action_275,
 action_276,
 action_277,
 action_278,
 action_279,
 action_280,
 action_281,
 action_282,
 action_283,
 action_284,
 action_285,
 action_286,
 action_287,
 action_288,
 action_289,
 action_290,
 action_291,
 action_292,
 action_293,
 action_294,
 action_295,
 action_296,
 action_297,
 action_298,
 action_299,
 action_300,
 action_301,
 action_302,
 action_303,
 action_304,
 action_305,
 action_306,
 action_307,
 action_308,
 action_309,
 action_310,
 action_311,
 action_312,
 action_313,
 action_314,
 action_315,
 action_316,
 action_317,
 action_318,
 action_319,
 action_320,
 action_321,
 action_322,
 action_323,
 action_324,
 action_325,
 action_326,
 action_327,
 action_328,
 action_329,
 action_330,
 action_331,
 action_332,
 action_333,
 action_334,
 action_335,
 action_336,
 action_337,
 action_338,
 action_339,
 action_340,
 action_341,
 action_342,
 action_343,
 action_344,
 action_345,
 action_346,
 action_347,
 action_348,
 action_349,
 action_350,
 action_351,
 action_352,
 action_353,
 action_354,
 action_355,
 action_356,
 action_357,
 action_358,
 action_359,
 action_360,
 action_361,
 action_362,
 action_363,
 action_364,
 action_365,
 action_366,
 action_367,
 action_368,
 action_369,
 action_370,
 action_371,
 action_372,
 action_373,
 action_374,
 action_375,
 action_376,
 action_377,
 action_378,
 action_379,
 action_380,
 action_381,
 action_382,
 action_383,
 action_384,
 action_385,
 action_386,
 action_387,
 action_388,
 action_389,
 action_390,
 action_391,
 action_392,
 action_393,
 action_394,
 action_395,
 action_396,
 action_397,
 action_398,
 action_399,
 action_400,
 action_401,
 action_402,
 action_403,
 action_404,
 action_405,
 action_406,
 action_407,
 action_408,
 action_409,
 action_410,
 action_411,
 action_412,
 action_413,
 action_414,
 action_415,
 action_416,
 action_417,
 action_418,
 action_419,
 action_420,
 action_421,
 action_422,
 action_423,
 action_424,
 action_425,
 action_426,
 action_427,
 action_428,
 action_429,
 action_430,
 action_431,
 action_432,
 action_433,
 action_434,
 action_435,
 action_436,
 action_437,
 action_438,
 action_439,
 action_440,
 action_441,
 action_442,
 action_443,
 action_444,
 action_445,
 action_446,
 action_447,
 action_448,
 action_449,
 action_450,
 action_451,
 action_452,
 action_453,
 action_454,
 action_455,
 action_456,
 action_457,
 action_458,
 action_459,
 action_460,
 action_461,
 action_462,
 action_463,
 action_464,
 action_465,
 action_466,
 action_467,
 action_468,
 action_469,
 action_470,
 action_471,
 action_472,
 action_473,
 action_474,
 action_475,
 action_476,
 action_477,
 action_478,
 action_479,
 action_480,
 action_481,
 action_482,
 action_483,
 action_484,
 action_485,
 action_486,
 action_487,
 action_488,
 action_489,
 action_490,
 action_491,
 action_492,
 action_493,
 action_494,
 action_495,
 action_496,
 action_497,
 action_498,
 action_499,
 action_500,
 action_501,
 action_502,
 action_503,
 action_504,
 action_505,
 action_506,
 action_507,
 action_508,
 action_509,
 action_510,
 action_511,
 action_512,
 action_513,
 action_514,
 action_515,
 action_516,
 action_517,
 action_518,
 action_519,
 action_520,
 action_521,
 action_522,
 action_523,
 action_524,
 action_525,
 action_526,
 action_527,
 action_528,
 action_529,
 action_530,
 action_531,
 action_532,
 action_533,
 action_534,
 action_535,
 action_536,
 action_537,
 action_538,
 action_539,
 action_540,
 action_541,
 action_542,
 action_543,
 action_544,
 action_545,
 action_546,
 action_547,
 action_548,
 action_549,
 action_550,
 action_551,
 action_552,
 action_553,
 action_554,
 action_555,
 action_556,
 action_557,
 action_558,
 action_559,
 action_560,
 action_561,
 action_562,
 action_563,
 action_564,
 action_565,
 action_566,
 action_567,
 action_568,
 action_569,
 action_570,
 action_571,
 action_572,
 action_573,
 action_574,
 action_575,
 action_576,
 action_577,
 action_578,
 action_579,
 action_580,
 action_581,
 action_582,
 action_583,
 action_584,
 action_585,
 action_586,
 action_587,
 action_588,
 action_589,
 action_590,
 action_591,
 action_592,
 action_593,
 action_594,
 action_595,
 action_596,
 action_597,
 action_598,
 action_599,
 action_600,
 action_601,
 action_602,
 action_603,
 action_604,
 action_605,
 action_606,
 action_607,
 action_608,
 action_609,
 action_610,
 action_611,
 action_612,
 action_613,
 action_614,
 action_615,
 action_616,
 action_617,
 action_618,
 action_619,
 action_620,
 action_621,
 action_622,
 action_623,
 action_624,
 action_625,
 action_626,
 action_627,
 action_628,
 action_629,
 action_630,
 action_631,
 action_632,
 action_633,
 action_634,
 action_635,
 action_636,
 action_637,
 action_638,
 action_639,
 action_640,
 action_641,
 action_642,
 action_643,
 action_644,
 action_645,
 action_646,
 action_647,
 action_648,
 action_649,
 action_650,
 action_651,
 action_652,
 action_653,
 action_654,
 action_655,
 action_656,
 action_657,
 action_658,
 action_659,
 action_660,
 action_661,
 action_662,
 action_663,
 action_664,
 action_665,
 action_666,
 action_667,
 action_668,
 action_669,
 action_670,
 action_671,
 action_672,
 action_673,
 action_674,
 action_675,
 action_676,
 action_677,
 action_678,
 action_679,
 action_680,
 action_681,
 action_682,
 action_683,
 action_684,
 action_685,
 action_686,
 action_687,
 action_688,
 action_689,
 action_690,
 action_691,
 action_692,
 action_693,
 action_694,
 action_695,
 action_696,
 action_697,
 action_698,
 action_699,
 action_700,
 action_701,
 action_702,
 action_703,
 action_704,
 action_705,
 action_706,
 action_707,
 action_708,
 action_709,
 action_710,
 action_711,
 action_712,
 action_713,
 action_714,
 action_715,
 action_716,
 action_717,
 action_718,
 action_719,
 action_720,
 action_721,
 action_722,
 action_723,
 action_724,
 action_725,
 action_726,
 action_727,
 action_728,
 action_729,
 action_730,
 action_731,
 action_732,
 action_733,
 action_734,
 action_735,
 action_736,
 action_737,
 action_738,
 action_739,
 action_740,
 action_741,
 action_742,
 action_743,
 action_744,
 action_745 :: () => Ranges -> (Int, Ranges) -> ({-HappyReduction (P) = -}
	   (Int, Ranges) 
	-> ((Token,Ranges))
	-> HappyState ((Token,Ranges)) (HappyStk HappyAbsSyn -> (P) HappyAbsSyn)
	-> [HappyState ((Token,Ranges)) (HappyStk HappyAbsSyn -> (P) HappyAbsSyn)] 
	-> HappyStk HappyAbsSyn 
	-> (P) HappyAbsSyn)

happyReduce_1,
 happyReduce_2,
 happyReduce_3,
 happyReduce_4,
 happyReduce_5,
 happyReduce_6,
 happyReduce_7,
 happyReduce_8,
 happyReduce_9,
 happyReduce_10,
 happyReduce_11,
 happyReduce_12,
 happyReduce_13,
 happyReduce_14,
 happyReduce_15,
 happyReduce_16,
 happyReduce_17,
 happyReduce_18,
 happyReduce_19,
 happyReduce_20,
 happyReduce_21,
 happyReduce_22,
 happyReduce_23,
 happyReduce_24,
 happyReduce_25,
 happyReduce_26,
 happyReduce_27,
 happyReduce_28,
 happyReduce_29,
 happyReduce_30,
 happyReduce_31,
 happyReduce_32,
 happyReduce_33,
 happyReduce_34,
 happyReduce_35,
 happyReduce_36,
 happyReduce_37,
 happyReduce_38,
 happyReduce_39,
 happyReduce_40,
 happyReduce_41,
 happyReduce_42,
 happyReduce_43,
 happyReduce_44,
 happyReduce_45,
 happyReduce_46,
 happyReduce_47,
 happyReduce_48,
 happyReduce_49,
 happyReduce_50,
 happyReduce_51,
 happyReduce_52,
 happyReduce_53,
 happyReduce_54,
 happyReduce_55,
 happyReduce_56,
 happyReduce_57,
 happyReduce_58,
 happyReduce_59,
 happyReduce_60,
 happyReduce_61,
 happyReduce_62,
 happyReduce_63,
 happyReduce_64,
 happyReduce_65,
 happyReduce_66,
 happyReduce_67,
 happyReduce_68,
 happyReduce_69,
 happyReduce_70,
 happyReduce_71,
 happyReduce_72,
 happyReduce_73,
 happyReduce_74,
 happyReduce_75,
 happyReduce_76,
 happyReduce_77,
 happyReduce_78,
 happyReduce_79,
 happyReduce_80,
 happyReduce_81,
 happyReduce_82,
 happyReduce_83,
 happyReduce_84,
 happyReduce_85,
 happyReduce_86,
 happyReduce_87,
 happyReduce_88,
 happyReduce_89,
 happyReduce_90,
 happyReduce_91,
 happyReduce_92,
 happyReduce_93,
 happyReduce_94,
 happyReduce_95,
 happyReduce_96,
 happyReduce_97,
 happyReduce_98,
 happyReduce_99,
 happyReduce_100,
 happyReduce_101,
 happyReduce_102,
 happyReduce_103,
 happyReduce_104,
 happyReduce_105,
 happyReduce_106,
 happyReduce_107,
 happyReduce_108,
 happyReduce_109,
 happyReduce_110,
 happyReduce_111,
 happyReduce_112,
 happyReduce_113,
 happyReduce_114,
 happyReduce_115,
 happyReduce_116,
 happyReduce_117,
 happyReduce_118,
 happyReduce_119,
 happyReduce_120,
 happyReduce_121,
 happyReduce_122,
 happyReduce_123,
 happyReduce_124,
 happyReduce_125,
 happyReduce_126,
 happyReduce_127,
 happyReduce_128,
 happyReduce_129,
 happyReduce_130,
 happyReduce_131,
 happyReduce_132,
 happyReduce_133,
 happyReduce_134,
 happyReduce_135,
 happyReduce_136,
 happyReduce_137,
 happyReduce_138,
 happyReduce_139,
 happyReduce_140,
 happyReduce_141,
 happyReduce_142,
 happyReduce_143,
 happyReduce_144,
 happyReduce_145,
 happyReduce_146,
 happyReduce_147,
 happyReduce_148,
 happyReduce_149,
 happyReduce_150,
 happyReduce_151,
 happyReduce_152,
 happyReduce_153,
 happyReduce_154,
 happyReduce_155,
 happyReduce_156,
 happyReduce_157,
 happyReduce_158,
 happyReduce_159,
 happyReduce_160,
 happyReduce_161,
 happyReduce_162,
 happyReduce_163,
 happyReduce_164,
 happyReduce_165,
 happyReduce_166,
 happyReduce_167,
 happyReduce_168,
 happyReduce_169,
 happyReduce_170,
 happyReduce_171,
 happyReduce_172,
 happyReduce_173,
 happyReduce_174,
 happyReduce_175,
 happyReduce_176,
 happyReduce_177,
 happyReduce_178,
 happyReduce_179,
 happyReduce_180,
 happyReduce_181,
 happyReduce_182,
 happyReduce_183,
 happyReduce_184,
 happyReduce_185,
 happyReduce_186,
 happyReduce_187,
 happyReduce_188,
 happyReduce_189,
 happyReduce_190,
 happyReduce_191,
 happyReduce_192,
 happyReduce_193,
 happyReduce_194,
 happyReduce_195,
 happyReduce_196,
 happyReduce_197,
 happyReduce_198,
 happyReduce_199,
 happyReduce_200,
 happyReduce_201,
 happyReduce_202,
 happyReduce_203,
 happyReduce_204,
 happyReduce_205,
 happyReduce_206,
 happyReduce_207,
 happyReduce_208,
 happyReduce_209,
 happyReduce_210,
 happyReduce_211,
 happyReduce_212,
 happyReduce_213,
 happyReduce_214,
 happyReduce_215,
 happyReduce_216,
 happyReduce_217,
 happyReduce_218,
 happyReduce_219,
 happyReduce_220,
 happyReduce_221,
 happyReduce_222,
 happyReduce_223,
 happyReduce_224,
 happyReduce_225,
 happyReduce_226,
 happyReduce_227,
 happyReduce_228,
 happyReduce_229,
 happyReduce_230,
 happyReduce_231,
 happyReduce_232,
 happyReduce_233,
 happyReduce_234,
 happyReduce_235,
 happyReduce_236,
 happyReduce_237,
 happyReduce_238,
 happyReduce_239,
 happyReduce_240,
 happyReduce_241,
 happyReduce_242,
 happyReduce_243,
 happyReduce_244,
 happyReduce_245,
 happyReduce_246,
 happyReduce_247,
 happyReduce_248,
 happyReduce_249,
 happyReduce_250,
 happyReduce_251,
 happyReduce_252,
 happyReduce_253,
 happyReduce_254,
 happyReduce_255,
 happyReduce_256,
 happyReduce_257,
 happyReduce_258,
 happyReduce_259,
 happyReduce_260,
 happyReduce_261,
 happyReduce_262,
 happyReduce_263,
 happyReduce_264,
 happyReduce_265,
 happyReduce_266,
 happyReduce_267,
 happyReduce_268,
 happyReduce_269,
 happyReduce_270,
 happyReduce_271,
 happyReduce_272,
 happyReduce_273,
 happyReduce_274,
 happyReduce_275,
 happyReduce_276,
 happyReduce_277,
 happyReduce_278,
 happyReduce_279,
 happyReduce_280,
 happyReduce_281,
 happyReduce_282,
 happyReduce_283,
 happyReduce_284,
 happyReduce_285,
 happyReduce_286,
 happyReduce_287,
 happyReduce_288,
 happyReduce_289,
 happyReduce_290,
 happyReduce_291,
 happyReduce_292,
 happyReduce_293,
 happyReduce_294,
 happyReduce_295,
 happyReduce_296,
 happyReduce_297,
 happyReduce_298,
 happyReduce_299,
 happyReduce_300,
 happyReduce_301,
 happyReduce_302,
 happyReduce_303,
 happyReduce_304,
 happyReduce_305,
 happyReduce_306,
 happyReduce_307,
 happyReduce_308,
 happyReduce_309,
 happyReduce_310,
 happyReduce_311,
 happyReduce_312,
 happyReduce_313,
 happyReduce_314,
 happyReduce_315,
 happyReduce_316,
 happyReduce_317,
 happyReduce_318,
 happyReduce_319,
 happyReduce_320,
 happyReduce_321,
 happyReduce_322,
 happyReduce_323,
 happyReduce_324,
 happyReduce_325,
 happyReduce_326,
 happyReduce_327,
 happyReduce_328,
 happyReduce_329,
 happyReduce_330,
 happyReduce_331,
 happyReduce_332,
 happyReduce_333,
 happyReduce_334,
 happyReduce_335,
 happyReduce_336,
 happyReduce_337,
 happyReduce_338,
 happyReduce_339,
 happyReduce_340,
 happyReduce_341,
 happyReduce_342,
 happyReduce_343,
 happyReduce_344,
 happyReduce_345,
 happyReduce_346,
 happyReduce_347,
 happyReduce_348,
 happyReduce_349,
 happyReduce_350,
 happyReduce_351,
 happyReduce_352,
 happyReduce_353,
 happyReduce_354,
 happyReduce_355,
 happyReduce_356,
 happyReduce_357,
 happyReduce_358,
 happyReduce_359,
 happyReduce_360,
 happyReduce_361,
 happyReduce_362,
 happyReduce_363,
 happyReduce_364,
 happyReduce_365,
 happyReduce_366,
 happyReduce_367,
 happyReduce_368,
 happyReduce_369,
 happyReduce_370,
 happyReduce_371,
 happyReduce_372,
 happyReduce_373,
 happyReduce_374,
 happyReduce_375,
 happyReduce_376,
 happyReduce_377,
 happyReduce_378,
 happyReduce_379,
 happyReduce_380,
 happyReduce_381,
 happyReduce_382,
 happyReduce_383,
 happyReduce_384,
 happyReduce_385,
 happyReduce_386,
 happyReduce_387,
 happyReduce_388,
 happyReduce_389,
 happyReduce_390,
 happyReduce_391,
 happyReduce_392,
 happyReduce_393,
 happyReduce_394,
 happyReduce_395,
 happyReduce_396,
 happyReduce_397,
 happyReduce_398,
 happyReduce_399,
 happyReduce_400,
 happyReduce_401,
 happyReduce_402,
 happyReduce_403,
 happyReduce_404,
 happyReduce_405,
 happyReduce_406,
 happyReduce_407,
 happyReduce_408,
 happyReduce_409,
 happyReduce_410,
 happyReduce_411,
 happyReduce_412,
 happyReduce_413,
 happyReduce_414,
 happyReduce_415,
 happyReduce_416,
 happyReduce_417,
 happyReduce_418,
 happyReduce_419,
 happyReduce_420,
 happyReduce_421,
 happyReduce_422,
 happyReduce_423,
 happyReduce_424,
 happyReduce_425,
 happyReduce_426,
 happyReduce_427,
 happyReduce_428,
 happyReduce_429,
 happyReduce_430,
 happyReduce_431,
 happyReduce_432,
 happyReduce_433,
 happyReduce_434,
 happyReduce_435,
 happyReduce_436,
 happyReduce_437,
 happyReduce_438,
 happyReduce_439,
 happyReduce_440,
 happyReduce_441,
 happyReduce_442,
 happyReduce_443,
 happyReduce_444,
 happyReduce_445,
 happyReduce_446,
 happyReduce_447,
 happyReduce_448,
 happyReduce_449,
 happyReduce_450,
 happyReduce_451,
 happyReduce_452,
 happyReduce_453,
 happyReduce_454,
 happyReduce_455,
 happyReduce_456,
 happyReduce_457,
 happyReduce_458,
 happyReduce_459,
 happyReduce_460,
 happyReduce_461,
 happyReduce_462,
 happyReduce_463,
 happyReduce_464,
 happyReduce_465,
 happyReduce_466,
 happyReduce_467,
 happyReduce_468,
 happyReduce_469,
 happyReduce_470,
 happyReduce_471,
 happyReduce_472,
 happyReduce_473,
 happyReduce_474,
 happyReduce_475,
 happyReduce_476,
 happyReduce_477,
 happyReduce_478,
 happyReduce_479,
 happyReduce_480,
 happyReduce_481,
 happyReduce_482,
 happyReduce_483,
 happyReduce_484,
 happyReduce_485,
 happyReduce_486,
 happyReduce_487,
 happyReduce_488,
 happyReduce_489,
 happyReduce_490,
 happyReduce_491,
 happyReduce_492,
 happyReduce_493,
 happyReduce_494,
 happyReduce_495,
 happyReduce_496,
 happyReduce_497,
 happyReduce_498,
 happyReduce_499,
 happyReduce_500,
 happyReduce_501,
 happyReduce_502,
 happyReduce_503,
 happyReduce_504,
 happyReduce_505,
 happyReduce_506,
 happyReduce_507,
 happyReduce_508,
 happyReduce_509,
 happyReduce_510,
 happyReduce_511,
 happyReduce_512,
 happyReduce_513,
 happyReduce_514,
 happyReduce_515,
 happyReduce_516,
 happyReduce_517,
 happyReduce_518,
 happyReduce_519,
 happyReduce_520,
 happyReduce_521,
 happyReduce_522,
 happyReduce_523,
 happyReduce_524,
 happyReduce_525,
 happyReduce_526,
 happyReduce_527,
 happyReduce_528,
 happyReduce_529,
 happyReduce_530,
 happyReduce_531,
 happyReduce_532,
 happyReduce_533,
 happyReduce_534,
 happyReduce_535,
 happyReduce_536,
 happyReduce_537,
 happyReduce_538,
 happyReduce_539,
 happyReduce_540,
 happyReduce_541,
 happyReduce_542,
 happyReduce_543,
 happyReduce_544,
 happyReduce_545,
 happyReduce_546,
 happyReduce_547,
 happyReduce_548,
 happyReduce_549,
 happyReduce_550,
 happyReduce_551,
 happyReduce_552,
 happyReduce_553,
 happyReduce_554,
 happyReduce_555,
 happyReduce_556,
 happyReduce_557,
 happyReduce_558,
 happyReduce_559,
 happyReduce_560,
 happyReduce_561,
 happyReduce_562,
 happyReduce_563,
 happyReduce_564,
 happyReduce_565,
 happyReduce_566,
 happyReduce_567,
 happyReduce_568,
 happyReduce_569,
 happyReduce_570,
 happyReduce_571,
 happyReduce_572,
 happyReduce_573,
 happyReduce_574,
 happyReduce_575,
 happyReduce_576,
 happyReduce_577,
 happyReduce_578,
 happyReduce_579,
 happyReduce_580,
 happyReduce_581,
 happyReduce_582,
 happyReduce_583,
 happyReduce_584,
 happyReduce_585,
 happyReduce_586,
 happyReduce_587,
 happyReduce_588,
 happyReduce_589,
 happyReduce_590,
 happyReduce_591,
 happyReduce_592,
 happyReduce_593,
 happyReduce_594,
 happyReduce_595,
 happyReduce_596,
 happyReduce_597,
 happyReduce_598,
 happyReduce_599,
 happyReduce_600,
 happyReduce_601,
 happyReduce_602,
 happyReduce_603,
 happyReduce_604,
 happyReduce_605,
 happyReduce_606,
 happyReduce_607,
 happyReduce_608,
 happyReduce_609,
 happyReduce_610,
 happyReduce_611,
 happyReduce_612,
 happyReduce_613,
 happyReduce_614,
 happyReduce_615,
 happyReduce_616,
 happyReduce_617,
 happyReduce_618,
 happyReduce_619,
 happyReduce_620,
 happyReduce_621,
 happyReduce_622,
 happyReduce_623,
 happyReduce_624,
 happyReduce_625,
 happyReduce_626,
 happyReduce_627,
 happyReduce_628,
 happyReduce_629,
 happyReduce_630,
 happyReduce_631,
 happyReduce_632,
 happyReduce_633,
 happyReduce_634,
 happyReduce_635,
 happyReduce_636,
 happyReduce_637,
 happyReduce_638,
 happyReduce_639,
 happyReduce_640,
 happyReduce_641,
 happyReduce_642,
 happyReduce_643,
 happyReduce_644,
 happyReduce_645,
 happyReduce_646,
 happyReduce_647,
 happyReduce_648,
 happyReduce_649,
 happyReduce_650,
 happyReduce_651,
 happyReduce_652,
 happyReduce_653,
 happyReduce_654,
 happyReduce_655,
 happyReduce_656,
 happyReduce_657 :: () => Ranges -> ({-HappyReduction (P) = -}
	   (Int, Ranges) 
	-> ((Token,Ranges))
	-> HappyState ((Token,Ranges)) (HappyStk HappyAbsSyn -> (P) HappyAbsSyn)
	-> [HappyState ((Token,Ranges)) (HappyStk HappyAbsSyn -> (P) HappyAbsSyn)] 
	-> HappyStk HappyAbsSyn 
	-> (P) HappyAbsSyn)

action_0 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_0 state_indent (4, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_3 state_indent')
action_0 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_4 state_indent')
action_0 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_0 state_indent tok = 
	 happyFail

action_1 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_2 state_indent')
action_1 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_1 state_indent tok = 
	 happyFail

action_2 state_indent (365, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_21 state_indent')
action_2 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_2 state_indent tok = 
	 happyFail

action_3 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyAccept
action_3 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_3 state_indent tok = 
	 happyFail

action_4 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_20 state_indent')
action_4 state_indent (365, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_21 state_indent')
action_4 state_indent (377, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_13 state_indent'
action_4 state_indent (6, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_5 state_indent')
action_4 state_indent (8, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_6 state_indent')
action_4 state_indent (154, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_7 state_indent')
action_4 state_indent (155, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_8 state_indent')
action_4 state_indent (156, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_9 state_indent')
action_4 state_indent (157, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_10 state_indent')
action_4 state_indent (165, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_11 state_indent')
action_4 state_indent (166, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_12 state_indent')
action_4 state_indent (167, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_13 state_indent')
action_4 state_indent (177, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_14 state_indent')
action_4 state_indent (179, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_15 state_indent')
action_4 state_indent (188, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_16 state_indent')
action_4 state_indent (196, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_17 state_indent')
action_4 state_indent (199, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_18 state_indent')
action_4 state_indent (293, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_19 state_indent')
action_4 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_4 state_indent tok = 
	 happyFail

action_5 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_2 state_indent'
action_5 state_indent tok = 
	 happyFail

action_6 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_5 state_indent'
action_6 state_indent tok = 
	 happyFail

action_7 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_11 state_indent'
action_7 state_indent (155, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_82 state_indent')
action_7 state_indent (156, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_83 state_indent')
action_7 state_indent (157, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_10 state_indent')
action_7 state_indent (165, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_11 state_indent')
action_7 state_indent (166, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_12 state_indent')
action_7 state_indent (167, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_13 state_indent')
action_7 state_indent (177, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_14 state_indent')
action_7 state_indent (179, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_15 state_indent')
action_7 state_indent (188, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_16 state_indent')
action_7 state_indent (196, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_17 state_indent')
action_7 state_indent (199, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_18 state_indent')
action_7 state_indent (293, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_19 state_indent')
action_7 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_7 state_indent tok = 
	 happyFail

action_8 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_8 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_8 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_80 state_indent')
action_8 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_81 state_indent')
action_8 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_8 state_indent tok = 
	 happyFail

action_9 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_12 state_indent'
action_9 state_indent tok = 
	 happyFail

action_10 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_10 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_10 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_78 state_indent')
action_10 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_79 state_indent')
action_10 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_10 state_indent tok = 
	 happyFail

action_11 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_337 state_indent'
action_11 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_337 state_indent'
action_11 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_337 state_indent'
action_11 state_indent tok = 
	 happyFail

action_12 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_338 state_indent'
action_12 state_indent (157, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_76 state_indent')
action_12 state_indent (167, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_13 state_indent')
action_12 state_indent (177, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_14 state_indent')
action_12 state_indent (179, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_15 state_indent')
action_12 state_indent (188, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_16 state_indent')
action_12 state_indent (196, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_17 state_indent')
action_12 state_indent (199, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_18 state_indent')
action_12 state_indent (293, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_77 state_indent')
action_12 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_12 state_indent tok = 
	 happyFail

action_13 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_340 state_indent'
action_13 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_340 state_indent'
action_13 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_340 state_indent'
action_13 state_indent tok = 
	 happyFail

action_14 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_369 state_indent'
action_14 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_369 state_indent'
action_14 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_369 state_indent'
action_14 state_indent tok = 
	 happyFail

action_15 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_368 state_indent'
action_15 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_368 state_indent'
action_15 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_368 state_indent'
action_15 state_indent tok = 
	 happyFail

action_16 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_387 state_indent'
action_16 state_indent tok = 
	 happyFail

action_17 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_386 state_indent'
action_17 state_indent tok = 
	 happyFail

action_18 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_385 state_indent'
action_18 state_indent tok = 
	 happyFail

action_19 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_44 state_indent')
action_19 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_45 state_indent')
action_19 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_46 state_indent')
action_19 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_47 state_indent')
action_19 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_48 state_indent')
action_19 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_49 state_indent')
action_19 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_50 state_indent')
action_19 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_51 state_indent')
action_19 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_52 state_indent')
action_19 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_53 state_indent')
action_19 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_54 state_indent')
action_19 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_55 state_indent')
action_19 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_56 state_indent')
action_19 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_57 state_indent')
action_19 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_58 state_indent')
action_19 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_59 state_indent')
action_19 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_60 state_indent')
action_19 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_61 state_indent')
action_19 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_62 state_indent')
action_19 state_indent (358, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_63 state_indent')
action_19 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_64 state_indent')
action_19 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_65 state_indent')
action_19 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_66 state_indent')
action_19 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_67 state_indent')
action_19 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_68 state_indent')
action_19 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_69 state_indent')
action_19 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_70 state_indent')
action_19 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_71 state_indent')
action_19 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_72 state_indent')
action_19 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_73 state_indent')
action_19 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_74 state_indent')
action_19 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_75 state_indent')
action_19 state_indent (191, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_28 state_indent')
action_19 state_indent (215, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_29 state_indent')
action_19 state_indent (227, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_30 state_indent')
action_19 state_indent (230, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_31 state_indent')
action_19 state_indent (237, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_32 state_indent')
action_19 state_indent (243, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_33 state_indent')
action_19 state_indent (244, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_34 state_indent')
action_19 state_indent (254, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_35 state_indent')
action_19 state_indent (259, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_36 state_indent')
action_19 state_indent (262, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_37 state_indent')
action_19 state_indent (263, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_38 state_indent')
action_19 state_indent (268, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_39 state_indent')
action_19 state_indent (277, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_40 state_indent')
action_19 state_indent (282, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_41 state_indent')
action_19 state_indent (284, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_42 state_indent')
action_19 state_indent (288, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_43 state_indent')
action_19 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_19 state_indent tok = 
	 happyFail

action_20 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_20 state_indent (7, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_25 state_indent')
action_20 state_indent (9, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_26 state_indent')
action_20 state_indent (10, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_27 state_indent')
action_20 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_20 state_indent tok = 
	 happyFail

action_21 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_23 state_indent')
action_21 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_24 state_indent')
action_21 state_indent (148, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_22 state_indent')
action_21 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_21 state_indent tok = 
	 happyFail

action_22 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_241 state_indent')
action_22 state_indent (11, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_239 state_indent')
action_22 state_indent (12, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_240 state_indent')
action_22 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_18 state_indent'
action_22 state_indent tok = 
	 happyFail

action_23 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_327 state_indent'
action_23 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_327 state_indent'
action_23 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_327 state_indent'
action_23 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_327 state_indent'
action_23 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_327 state_indent'
action_23 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_327 state_indent'
action_23 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_327 state_indent'
action_23 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_327 state_indent'
action_23 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_327 state_indent'
action_23 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_327 state_indent'
action_23 state_indent tok = 
	 happyFail

action_24 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_328 state_indent'
action_24 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_328 state_indent'
action_24 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_328 state_indent'
action_24 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_328 state_indent'
action_24 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_328 state_indent'
action_24 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_328 state_indent'
action_24 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_328 state_indent'
action_24 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_328 state_indent'
action_24 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_328 state_indent'
action_24 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_328 state_indent'
action_24 state_indent tok = 
	 happyFail

action_25 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_238 state_indent')
action_25 state_indent (147, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_237 state_indent')
action_25 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_25 state_indent tok = 
	 happyFail

action_26 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_26 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_26 state_indent tok = 
	 happyFail

action_27 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_27 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_9 state_indent'
action_27 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_231 state_indent')
action_27 state_indent (17, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_232 state_indent')
action_27 state_indent (18, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_233 state_indent')
action_27 state_indent (28, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_89 state_indent')
action_27 state_indent (32, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_234 state_indent')
action_27 state_indent (35, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_235 state_indent')
action_27 state_indent (36, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_236 state_indent')
action_27 state_indent (40, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_91 state_indent')
action_27 state_indent (45, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_92 state_indent')
action_27 state_indent (47, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_93 state_indent')
action_27 state_indent (79, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_94 state_indent')
action_27 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_27 state_indent tok = 
	 happyFail

action_28 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_28 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_28 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_201 state_indent')
action_28 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_202 state_indent')
action_28 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_227 state_indent')
action_28 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_206 state_indent')
action_28 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_228 state_indent')
action_28 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_28 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_28 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_216 state_indent')
action_28 state_indent (81, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_217 state_indent')
action_28 state_indent (82, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_218 state_indent')
action_28 state_indent (83, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_219 state_indent')
action_28 state_indent (127, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_220 state_indent')
action_28 state_indent (130, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_221 state_indent')
action_28 state_indent (132, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_222 state_indent')
action_28 state_indent (134, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_223 state_indent')
action_28 state_indent (139, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_192 state_indent')
action_28 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_193 state_indent')
action_28 state_indent (141, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_224 state_indent')
action_28 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_196 state_indent')
action_28 state_indent (145, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_225 state_indent')
action_28 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_28 state_indent tok = 
	 happyFail

action_29 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_415 state_indent'
action_29 state_indent tok = 
	 happyFail

action_30 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_215 state_indent')
action_30 state_indent (29, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_214 state_indent')
action_30 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_58 state_indent'
action_30 state_indent tok = 
	 happyFail

action_31 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_212 state_indent')
action_31 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_213 state_indent')
action_31 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_31 state_indent tok = 
	 happyFail

action_32 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_32 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_32 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_32 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_32 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_32 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_32 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_32 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_32 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_32 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_32 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_32 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_32 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_32 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_32 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_32 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_32 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_32 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_32 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_211 state_indent')
action_32 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_32 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_32 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_32 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_32 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_32 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_32 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_32 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_32 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_32 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_32 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_461 state_indent'
action_32 state_indent tok = 
	 happyFail

action_33 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_624 state_indent'
action_33 state_indent tok = 
	 happyFail

action_34 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_516 state_indent'
action_34 state_indent tok = 
	 happyFail

action_35 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_506 state_indent'
action_35 state_indent tok = 
	 happyFail

action_36 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_487 state_indent'
action_36 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_487 state_indent'
action_36 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_210 state_indent')
action_36 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_36 state_indent tok = 
	 happyFail

action_37 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_549 state_indent'
action_37 state_indent tok = 
	 happyFail

action_38 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_555 state_indent'
action_38 state_indent tok = 
	 happyFail

action_39 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_209 state_indent')
action_39 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_542 state_indent'
action_39 state_indent tok = 
	 happyFail

action_40 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_573 state_indent'
action_40 state_indent tok = 
	 happyFail

action_41 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_589 state_indent'
action_41 state_indent tok = 
	 happyFail

action_42 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_590 state_indent'
action_42 state_indent tok = 
	 happyFail

action_43 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_612 state_indent'
action_43 state_indent tok = 
	 happyFail

action_44 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_557 state_indent'
action_44 state_indent tok = 
	 happyFail

action_45 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_556 state_indent'
action_45 state_indent tok = 
	 happyFail

action_46 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_518 state_indent'
action_46 state_indent tok = 
	 happyFail

action_47 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_517 state_indent'
action_47 state_indent tok = 
	 happyFail

action_48 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_615 state_indent'
action_48 state_indent tok = 
	 happyFail

action_49 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_617 state_indent'
action_49 state_indent tok = 
	 happyFail

action_50 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_616 state_indent'
action_50 state_indent tok = 
	 happyFail

action_51 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_618 state_indent'
action_51 state_indent tok = 
	 happyFail

action_52 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_52 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_52 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_52 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_52 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_199 state_indent')
action_52 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_52 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_201 state_indent')
action_52 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_202 state_indent')
action_52 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_52 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_52 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_52 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_52 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_52 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_203 state_indent')
action_52 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_52 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_204 state_indent')
action_52 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_52 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_205 state_indent')
action_52 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_206 state_indent')
action_52 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_52 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_52 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_207 state_indent')
action_52 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_208 state_indent')
action_52 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_52 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_52 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_52 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_52 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_52 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_52 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_52 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_52 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_52 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_52 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_184 state_indent')
action_52 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_52 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_52 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_185 state_indent')
action_52 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_52 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_52 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_52 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_52 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_52 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_52 state_indent (96, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_186 state_indent')
action_52 state_indent (97, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_187 state_indent')
action_52 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_52 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_52 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_52 state_indent (128, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_188 state_indent')
action_52 state_indent (130, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_189 state_indent')
action_52 state_indent (133, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_190 state_indent')
action_52 state_indent (134, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_191 state_indent')
action_52 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_52 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_52 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_52 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_52 state_indent (139, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_192 state_indent')
action_52 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_193 state_indent')
action_52 state_indent (141, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_194 state_indent')
action_52 state_indent (142, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_195 state_indent')
action_52 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_196 state_indent')
action_52 state_indent (144, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_197 state_indent')
action_52 state_indent (145, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_198 state_indent')
action_52 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_52 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_52 state_indent tok = 
	 happyFail

action_53 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_53 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_53 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_53 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_53 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_53 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_53 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_53 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_53 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_53 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_53 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_183 state_indent')
action_53 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_53 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_53 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_53 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_53 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_53 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_53 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_53 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_53 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_53 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_53 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_53 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_53 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_53 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_53 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_180 state_indent')
action_53 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_53 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_53 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_53 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_53 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_53 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_53 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_53 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_53 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_53 state_indent (98, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_181 state_indent')
action_53 state_indent (99, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_182 state_indent')
action_53 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_53 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_53 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_53 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_53 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_53 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_53 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_53 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_53 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_53 state_indent tok = 
	 happyFail

action_54 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_596 state_indent'
action_54 state_indent tok = 
	 happyFail

action_55 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_55 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_55 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_55 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_55 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_55 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_55 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_55 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_55 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_55 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_55 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_55 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_55 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_55 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_55 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_55 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_55 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_55 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_55 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_179 state_indent')
action_55 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_55 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_55 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_55 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_55 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_55 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_55 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_55 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_55 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_55 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_55 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_55 state_indent tok = 
	 happyFail

action_56 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_56 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_56 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_56 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_56 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_56 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_56 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_56 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_56 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_56 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_56 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_56 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_56 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_56 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_56 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_56 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_56 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_56 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_56 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_178 state_indent')
action_56 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_56 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_56 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_56 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_56 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_56 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_56 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_56 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_56 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_56 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_56 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_56 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_56 state_indent tok = 
	 happyFail

action_57 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_57 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_57 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_57 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_57 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_57 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_57 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_57 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_57 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_57 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_57 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_57 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_57 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_57 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_57 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_57 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_57 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_57 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_57 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_57 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_57 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_57 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_57 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_57 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_57 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_146 state_indent')
action_57 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_57 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_57 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_57 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_57 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_57 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_57 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_57 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_57 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_57 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_57 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_57 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_57 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_57 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_57 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_57 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_57 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_57 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_57 state_indent tok = 
	 happyFail

action_58 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_58 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_58 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_58 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_58 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_58 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_58 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_58 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_58 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_58 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_58 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_58 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_100 state_indent')
action_58 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_101 state_indent')
action_58 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_58 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_58 state_indent (55, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_145 state_indent')
action_58 state_indent (56, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_105 state_indent')
action_58 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_58 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_58 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_58 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_58 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_58 state_indent tok = 
	 happyFail

action_59 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_59 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_59 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_59 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_59 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_59 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_59 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_59 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_59 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_59 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_59 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_59 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_100 state_indent')
action_59 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_101 state_indent')
action_59 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_59 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_59 state_indent (55, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_144 state_indent')
action_59 state_indent (56, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_105 state_indent')
action_59 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_59 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_59 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_59 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_59 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_59 state_indent tok = 
	 happyFail

action_60 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_143 state_indent')
action_60 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_60 state_indent tok = 
	 happyFail

action_61 state_indent (313, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_44 state_indent')
action_61 state_indent (314, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_45 state_indent')
action_61 state_indent (315, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_46 state_indent')
action_61 state_indent (316, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_47 state_indent')
action_61 state_indent (321, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_48 state_indent')
action_61 state_indent (322, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_49 state_indent')
action_61 state_indent (323, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_50 state_indent')
action_61 state_indent (324, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_51 state_indent')
action_61 state_indent (325, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_52 state_indent')
action_61 state_indent (327, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_138 state_indent')
action_61 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_139 state_indent')
action_61 state_indent (331, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_53 state_indent')
action_61 state_indent (334, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_54 state_indent')
action_61 state_indent (340, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_140 state_indent')
action_61 state_indent (345, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_55 state_indent')
action_61 state_indent (347, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_56 state_indent')
action_61 state_indent (349, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_57 state_indent')
action_61 state_indent (354, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_61 state_indent')
action_61 state_indent (357, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_141 state_indent')
action_61 state_indent (364, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_142 state_indent')
action_61 state_indent (371, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_70 state_indent')
action_61 state_indent (372, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_71 state_indent')
action_61 state_indent (373, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_72 state_indent')
action_61 state_indent (374, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_73 state_indent')
action_61 state_indent (375, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_74 state_indent')
action_61 state_indent (376, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_75 state_indent')
action_61 state_indent (112, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_125 state_indent')
action_61 state_indent (113, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_126 state_indent')
action_61 state_indent (114, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_127 state_indent')
action_61 state_indent (161, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_128 state_indent')
action_61 state_indent (162, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_129 state_indent')
action_61 state_indent (170, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_130 state_indent')
action_61 state_indent (172, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_131 state_indent')
action_61 state_indent (190, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_132 state_indent')
action_61 state_indent (191, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_133 state_indent')
action_61 state_indent (193, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_134 state_indent')
action_61 state_indent (214, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_135 state_indent')
action_61 state_indent (215, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_29 state_indent')
action_61 state_indent (236, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_136 state_indent')
action_61 state_indent (237, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_32 state_indent')
action_61 state_indent (243, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_33 state_indent')
action_61 state_indent (244, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_34 state_indent')
action_61 state_indent (254, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_35 state_indent')
action_61 state_indent (259, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_137 state_indent')
action_61 state_indent (262, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_37 state_indent')
action_61 state_indent (263, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_38 state_indent')
action_61 state_indent (268, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_39 state_indent')
action_61 state_indent (277, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_40 state_indent')
action_61 state_indent (282, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_41 state_indent')
action_61 state_indent (284, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_42 state_indent')
action_61 state_indent (288, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_43 state_indent')
action_61 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_61 state_indent tok = 
	 happyFail

action_62 state_indent (358, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_123 state_indent')
action_62 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_124 state_indent')
action_62 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_62 state_indent tok = 
	 happyFail

action_63 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_122 state_indent')
action_63 state_indent (19, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_121 state_indent')
action_63 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_37 state_indent'
action_63 state_indent tok = 
	 happyFail

action_64 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_478 state_indent'
action_64 state_indent tok = 
	 happyFail

action_65 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_479 state_indent'
action_65 state_indent tok = 
	 happyFail

action_66 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_480 state_indent'
action_66 state_indent tok = 
	 happyFail

action_67 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_67 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_67 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_67 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_67 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_67 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_67 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_67 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_67 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_67 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_67 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_67 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_100 state_indent')
action_67 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_101 state_indent')
action_67 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_67 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_67 state_indent (55, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_120 state_indent')
action_67 state_indent (56, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_105 state_indent')
action_67 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_67 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_67 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_67 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_67 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_67 state_indent tok = 
	 happyFail

action_68 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_68 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_68 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_68 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_68 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_68 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_68 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_68 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_68 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_68 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_68 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_68 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_100 state_indent')
action_68 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_101 state_indent')
action_68 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_68 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_68 state_indent (55, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_104 state_indent')
action_68 state_indent (56, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_105 state_indent')
action_68 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_68 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_68 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_68 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_68 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_68 state_indent tok = 
	 happyFail

action_69 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_69 state_indent (58, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_96 state_indent')
action_69 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_97 state_indent')
action_69 state_indent (150, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_98 state_indent')
action_69 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_69 state_indent tok = 
	 happyFail

action_70 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_558 state_indent'
action_70 state_indent tok = 
	 happyFail

action_71 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_559 state_indent'
action_71 state_indent tok = 
	 happyFail

action_72 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_560 state_indent'
action_72 state_indent tok = 
	 happyFail

action_73 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_561 state_indent'
action_73 state_indent tok = 
	 happyFail

action_74 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_562 state_indent'
action_74 state_indent tok = 
	 happyFail

action_75 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_563 state_indent'
action_75 state_indent tok = 
	 happyFail

action_76 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_76 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_76 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_78 state_indent')
action_76 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_95 state_indent')
action_76 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_76 state_indent tok = 
	 happyFail

action_77 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_44 state_indent')
action_77 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_45 state_indent')
action_77 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_46 state_indent')
action_77 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_47 state_indent')
action_77 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_48 state_indent')
action_77 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_49 state_indent')
action_77 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_50 state_indent')
action_77 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_51 state_indent')
action_77 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_52 state_indent')
action_77 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_53 state_indent')
action_77 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_54 state_indent')
action_77 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_55 state_indent')
action_77 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_56 state_indent')
action_77 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_57 state_indent')
action_77 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_58 state_indent')
action_77 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_59 state_indent')
action_77 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_60 state_indent')
action_77 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_61 state_indent')
action_77 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_62 state_indent')
action_77 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_64 state_indent')
action_77 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_65 state_indent')
action_77 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_66 state_indent')
action_77 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_67 state_indent')
action_77 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_68 state_indent')
action_77 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_69 state_indent')
action_77 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_70 state_indent')
action_77 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_71 state_indent')
action_77 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_72 state_indent')
action_77 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_73 state_indent')
action_77 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_74 state_indent')
action_77 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_75 state_indent')
action_77 state_indent (191, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_28 state_indent')
action_77 state_indent (215, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_29 state_indent')
action_77 state_indent (227, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_30 state_indent')
action_77 state_indent (230, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_31 state_indent')
action_77 state_indent (237, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_32 state_indent')
action_77 state_indent (243, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_33 state_indent')
action_77 state_indent (244, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_34 state_indent')
action_77 state_indent (254, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_35 state_indent')
action_77 state_indent (259, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_36 state_indent')
action_77 state_indent (262, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_37 state_indent')
action_77 state_indent (263, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_38 state_indent')
action_77 state_indent (268, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_39 state_indent')
action_77 state_indent (277, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_40 state_indent')
action_77 state_indent (282, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_41 state_indent')
action_77 state_indent (284, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_42 state_indent')
action_77 state_indent (288, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_43 state_indent')
action_77 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_77 state_indent tok = 
	 happyFail

action_78 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_78 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_88 state_indent')
action_78 state_indent (28, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_89 state_indent')
action_78 state_indent (36, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_90 state_indent')
action_78 state_indent (40, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_91 state_indent')
action_78 state_indent (45, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_92 state_indent')
action_78 state_indent (47, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_93 state_indent')
action_78 state_indent (79, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_94 state_indent')
action_78 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_78 state_indent tok = 
	 happyFail

action_79 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_79 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_361 state_indent'
action_79 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_361 state_indent'
action_79 state_indent tok = 
	 happyFail

action_80 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_80 state_indent (358, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_80 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_80 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_86 state_indent')
action_80 state_indent (18, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_87 state_indent')
action_80 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_80 state_indent tok = 
	 happyFail

action_81 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_81 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_335 state_indent'
action_81 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_335 state_indent'
action_81 state_indent tok = 
	 happyFail

action_82 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_82 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_82 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_80 state_indent')
action_82 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_84 state_indent')
action_82 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_82 state_indent tok = 
	 happyFail

action_83 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_10 state_indent'
action_83 state_indent tok = 
	 happyFail

action_84 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_84 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_334 state_indent'
action_84 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_334 state_indent'
action_84 state_indent tok = 
	 happyFail

action_85 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0, Fin 0),(2, Inf)]) {-{([I1],[=])([=I2],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (358, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_14 state_indent'
action_85 state_indent tok = 
	 happyFail

action_86 state_indent (358, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_264 state_indent')
action_86 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_86 state_indent tok = 
	 happyFail

action_87 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_336 state_indent'
action_87 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_336 state_indent'
action_87 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_336 state_indent'
action_87 state_indent tok = 
	 happyFail

action_88 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_88 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_88 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_88 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_88 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_88 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_88 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_88 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_88 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_88 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_88 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_88 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_88 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_88 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_88 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_260 state_indent')
action_88 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_261 state_indent')
action_88 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_262 state_indent')
action_88 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_88 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_263 state_indent')
action_88 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_265 state_indent')
action_88 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_266 state_indent')
action_88 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_267 state_indent')
action_88 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_268 state_indent')
action_88 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_269 state_indent')
action_88 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_270 state_indent')
action_88 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_88 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_88 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_88 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_88 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_88 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_88 state_indent (30, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_256 state_indent')
action_88 state_indent (46, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_257 state_indent')
action_88 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_258 state_indent')
action_88 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_88 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_88 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_88 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_88 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_88 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_88 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_259 state_indent')
action_88 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_88 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_88 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_88 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_88 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_88 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_88 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_88 state_indent tok = 
	 happyFail

action_89 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_87 state_indent'
action_89 state_indent tok = 
	 happyFail

action_90 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_339 state_indent'
action_90 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_339 state_indent'
action_90 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_339 state_indent'
action_90 state_indent tok = 
	 happyFail

action_91 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_78 state_indent'
action_91 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_78 state_indent'
action_91 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_78 state_indent'
action_91 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_78 state_indent'
action_91 state_indent tok = 
	 happyFail

action_92 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_86 state_indent'
action_92 state_indent tok = 
	 happyFail

action_93 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_77 state_indent'
action_93 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_77 state_indent'
action_93 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_77 state_indent'
action_93 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_77 state_indent'
action_93 state_indent tok = 
	 happyFail

action_94 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_88 state_indent'
action_94 state_indent tok = 
	 happyFail

action_95 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_95 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_360 state_indent'
action_95 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_360 state_indent'
action_95 state_indent tok = 
	 happyFail

action_96 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_387 state_indent')
action_96 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_96 state_indent tok = 
	 happyFail

action_97 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_330 state_indent'
action_97 state_indent tok = 
	 happyFail

action_98 state_indent (59, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_386 state_indent')
action_98 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_130 state_indent'
action_98 state_indent tok = 
	 happyFail

action_99 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_308 state_indent'
action_99 state_indent tok = 
	 happyFail

action_100 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_124 state_indent'
action_100 state_indent tok = 
	 happyFail

action_101 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_101 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_101 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_101 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_101 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_101 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_385 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_125 state_indent'
action_101 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_101 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_101 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_101 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_101 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_101 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_101 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_384 state_indent')
action_101 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_101 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_101 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_101 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_101 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_101 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_101 state_indent tok = 
	 happyFail

action_102 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_112 state_indent'
action_102 state_indent tok = 
	 happyFail

action_103 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_113 state_indent'
action_103 state_indent tok = 
	 happyFail

action_104 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_383 state_indent')
action_104 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_104 state_indent tok = 
	 happyFail

action_105 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_382 state_indent')
action_105 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_105 state_indent tok = 
	 happyFail

action_106 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_333 state_indent'
action_106 state_indent tok = 
	 happyFail

action_107 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_118 state_indent'
action_107 state_indent tok = 
	 happyFail

action_108 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_108 state_indent tok = 
	 happyFail

action_109 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_114 state_indent'
action_109 state_indent tok = 
	 happyFail

action_110 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_299 state_indent'
action_110 state_indent tok = 
	 happyFail

action_111 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_307 state_indent'
action_111 state_indent tok = 
	 happyFail

action_112 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_112 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_112 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_112 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_112 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_380 state_indent')
action_112 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_112 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_204 state_indent')
action_112 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_381 state_indent')
action_112 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_112 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_112 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_112 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_112 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_112 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_112 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_377 state_indent')
action_112 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_112 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_112 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_112 state_indent (57, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_378 state_indent')
action_112 state_indent (96, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_379 state_indent')
action_112 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_112 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_112 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_112 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_112 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_112 state_indent tok = 
	 happyFail

action_113 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_113 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_113 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_113 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_113 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_113 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_376 state_indent')
action_113 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_113 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_113 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_113 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_113 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_113 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_113 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_375 state_indent')
action_113 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_113 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_113 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_113 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_113 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_113 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_113 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_113 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_113 state_indent tok = 
	 happyFail

action_114 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_300 state_indent'
action_114 state_indent tok = 
	 happyFail

action_115 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_301 state_indent'
action_115 state_indent tok = 
	 happyFail

action_116 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_302 state_indent'
action_116 state_indent tok = 
	 happyFail

action_117 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_303 state_indent'
action_117 state_indent tok = 
	 happyFail

action_118 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_304 state_indent'
action_118 state_indent tok = 
	 happyFail

action_119 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_305 state_indent'
action_119 state_indent tok = 
	 happyFail

action_120 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_162 state_indent'
action_120 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_374 state_indent')
action_120 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_162 state_indent'
action_120 state_indent (73, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_373 state_indent')
action_120 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_162 state_indent'
action_120 state_indent tok = 
	 happyFail

action_121 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_23 state_indent')
action_121 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_24 state_indent')
action_121 state_indent (148, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_372 state_indent')
action_121 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_121 state_indent tok = 
	 happyFail

action_122 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_36 state_indent'
action_122 state_indent tok = 
	 happyFail

action_123 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_371 state_indent')
action_123 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_123 state_indent tok = 
	 happyFail

action_124 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_370 state_indent')
action_124 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_124 state_indent tok = 
	 happyFail

action_125 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_460 state_indent'
action_125 state_indent tok = 
	 happyFail

action_126 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_249 state_indent'
action_126 state_indent tok = 
	 happyFail

action_127 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_250 state_indent'
action_127 state_indent tok = 
	 happyFail

action_128 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_252 state_indent'
action_128 state_indent tok = 
	 happyFail

action_129 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_44 state_indent')
action_129 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_45 state_indent')
action_129 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_46 state_indent')
action_129 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_47 state_indent')
action_129 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_48 state_indent')
action_129 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_49 state_indent')
action_129 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_50 state_indent')
action_129 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_51 state_indent')
action_129 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_52 state_indent')
action_129 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_138 state_indent')
action_129 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_53 state_indent')
action_129 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_54 state_indent')
action_129 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_140 state_indent')
action_129 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_55 state_indent')
action_129 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_56 state_indent')
action_129 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_57 state_indent')
action_129 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_61 state_indent')
action_129 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_141 state_indent')
action_129 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_142 state_indent')
action_129 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_70 state_indent')
action_129 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_71 state_indent')
action_129 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_72 state_indent')
action_129 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_73 state_indent')
action_129 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_74 state_indent')
action_129 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_75 state_indent')
action_129 state_indent (114, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_369 state_indent')
action_129 state_indent (161, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_128 state_indent')
action_129 state_indent (162, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_129 state_indent')
action_129 state_indent (170, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_130 state_indent')
action_129 state_indent (172, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_131 state_indent')
action_129 state_indent (190, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_132 state_indent')
action_129 state_indent (191, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_133 state_indent')
action_129 state_indent (193, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_134 state_indent')
action_129 state_indent (214, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_135 state_indent')
action_129 state_indent (215, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_29 state_indent')
action_129 state_indent (236, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_136 state_indent')
action_129 state_indent (237, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_32 state_indent')
action_129 state_indent (243, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_33 state_indent')
action_129 state_indent (244, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_34 state_indent')
action_129 state_indent (254, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_35 state_indent')
action_129 state_indent (259, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_137 state_indent')
action_129 state_indent (262, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_37 state_indent')
action_129 state_indent (263, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_38 state_indent')
action_129 state_indent (268, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_39 state_indent')
action_129 state_indent (277, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_40 state_indent')
action_129 state_indent (282, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_41 state_indent')
action_129 state_indent (284, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_42 state_indent')
action_129 state_indent (288, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_43 state_indent')
action_129 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_129 state_indent tok = 
	 happyFail

action_130 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_379 state_indent'
action_130 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_351 state_indent'
action_130 state_indent tok = 
	 happyFail

action_131 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_368 state_indent')
action_131 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_355 state_indent'
action_131 state_indent tok = 
	 happyFail

action_132 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_375 state_indent'
action_132 state_indent tok = 
	 happyFail

action_133 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_201 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_202 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_227 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_206 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_367 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_418 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent (127, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_220 state_indent')
action_133 state_indent (130, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_221 state_indent')
action_133 state_indent (132, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_366 state_indent')
action_133 state_indent (134, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_223 state_indent')
action_133 state_indent (139, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_192 state_indent')
action_133 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_193 state_indent')
action_133 state_indent (141, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_224 state_indent')
action_133 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_196 state_indent')
action_133 state_indent (145, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_225 state_indent')
action_133 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_413 state_indent'
action_133 state_indent tok = 
	 happyFail

action_134 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_365 state_indent')
action_134 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_134 state_indent tok = 
	 happyFail

action_135 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_412 state_indent'
action_135 state_indent tok = 
	 happyFail

action_136 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_457 state_indent'
action_136 state_indent tok = 
	 happyFail

action_137 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_210 state_indent')
action_137 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_588 state_indent'
action_137 state_indent tok = 
	 happyFail

action_138 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_138 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_138 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_138 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_138 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_138 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_138 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_138 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_138 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_138 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_364 state_indent')
action_138 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_138 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_138 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_138 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_138 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_138 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_138 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_138 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_138 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_359 state_indent')
action_138 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_138 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_138 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_138 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_138 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_138 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_138 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_360 state_indent')
action_138 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_138 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_138 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_354 state_indent')
action_138 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_138 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_138 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_138 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_138 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_138 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_138 state_indent (111, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_355 state_indent')
action_138 state_indent (115, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_361 state_indent')
action_138 state_indent (116, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_362 state_indent')
action_138 state_indent (117, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_363 state_indent')
action_138 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_138 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_138 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_138 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_138 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_138 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_138 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_138 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_138 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_138 state_indent tok = 
	 happyFail

action_139 state_indent (313, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_139 state_indent (314, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_139 state_indent (315, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_139 state_indent (316, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_139 state_indent (321, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_139 state_indent (322, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_139 state_indent (323, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_139 state_indent (324, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_139 state_indent (325, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_139 state_indent (327, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_358 state_indent')
action_139 state_indent (331, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_139 state_indent (334, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_139 state_indent (340, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_139 state_indent (345, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_139 state_indent (347, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_139 state_indent (349, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_139 state_indent (354, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_139 state_indent (357, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_139 state_indent (364, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_359 state_indent')
action_139 state_indent (371, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_139 state_indent (372, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_139 state_indent (373, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_139 state_indent (374, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_139 state_indent (375, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_139 state_indent (376, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_139 state_indent (84, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_353 state_indent')
action_139 state_indent (85, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_139 state_indent (86, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_139 state_indent (87, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_354 state_indent')
action_139 state_indent (88, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_139 state_indent (89, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_139 state_indent (90, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_139 state_indent (93, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_139 state_indent (94, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_139 state_indent (95, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_139 state_indent (111, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_355 state_indent')
action_139 state_indent (116, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_356 state_indent')
action_139 state_indent (117, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_357 state_indent')
action_139 state_indent (121, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_139 state_indent (123, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_139 state_indent (125, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_139 state_indent (135, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_139 state_indent (136, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_139 state_indent (137, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_139 state_indent (138, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_139 state_indent (146, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_139 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_139 state_indent tok = 
	 happyFail

action_140 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_352 state_indent')
action_140 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_140 state_indent tok = 
	 happyFail

action_141 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_141 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_141 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_141 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_141 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_141 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_141 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_141 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_141 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_141 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_141 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_141 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_141 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_141 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_141 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_141 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_141 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_141 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_141 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_141 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_141 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_141 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_141 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_141 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_141 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_351 state_indent')
action_141 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_141 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_141 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_141 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_141 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_141 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_141 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_141 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_141 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_141 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_141 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_141 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_141 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_141 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_141 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_141 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_141 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_141 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_141 state_indent tok = 
	 happyFail

action_142 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_323 state_indent')
action_142 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_142 state_indent (41, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_350 state_indent')
action_142 state_indent (43, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_317 state_indent')
action_142 state_indent (44, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_318 state_indent')
action_142 state_indent (158, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_319 state_indent')
action_142 state_indent (163, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_320 state_indent')
action_142 state_indent (164, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_321 state_indent')
action_142 state_indent (294, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_322 state_indent')
action_142 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_142 state_indent tok = 
	 happyFail

action_143 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_143 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_143 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_143 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_143 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_143 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_143 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_143 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_143 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_143 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_143 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_143 state_indent (37, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_346 state_indent')
action_143 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_347 state_indent')
action_143 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_143 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_143 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_143 state_indent (57, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_349 state_indent')
action_143 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_143 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_143 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_143 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_143 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_81 state_indent'
action_143 state_indent tok = 
	 happyFail

action_144 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_345 state_indent')
action_144 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_144 state_indent tok = 
	 happyFail

action_145 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_158 state_indent'
action_145 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_344 state_indent')
action_145 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_158 state_indent'
action_145 state_indent (72, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_343 state_indent')
action_145 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_158 state_indent'
action_145 state_indent tok = 
	 happyFail

action_146 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_342 state_indent')
action_146 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_146 state_indent tok = 
	 happyFail

action_147 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_181 state_indent'
action_147 state_indent tok = 
	 happyFail

action_148 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_182 state_indent'
action_148 state_indent tok = 
	 happyFail

action_149 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_201 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_202 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_227 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_206 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_307 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent (127, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_220 state_indent')
action_149 state_indent (130, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_221 state_indent')
action_149 state_indent (132, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_341 state_indent')
action_149 state_indent (134, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_223 state_indent')
action_149 state_indent (139, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_192 state_indent')
action_149 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_193 state_indent')
action_149 state_indent (141, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_224 state_indent')
action_149 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_196 state_indent')
action_149 state_indent (145, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_225 state_indent')
action_149 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_149 state_indent tok = 
	 happyFail

action_150 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_185 state_indent'
action_150 state_indent tok = 
	 happyFail

action_151 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_187 state_indent'
action_151 state_indent tok = 
	 happyFail

action_152 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_152 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_152 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_152 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_152 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_152 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_152 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_152 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_152 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_152 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_152 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_152 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_152 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_152 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_152 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_152 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_152 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_152 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_152 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_315 state_indent')
action_152 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_152 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_152 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_152 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_152 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_152 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_152 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_152 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_152 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_152 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_152 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_194 state_indent'
action_152 state_indent tok = 
	 happyFail

action_153 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_196 state_indent'
action_153 state_indent tok = 
	 happyFail

action_154 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_340 state_indent')
action_154 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_202 state_indent'
action_154 state_indent tok = 
	 happyFail

action_155 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_205 state_indent'
action_155 state_indent tok = 
	 happyFail

action_156 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_207 state_indent'
action_156 state_indent tok = 
	 happyFail

action_157 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_339 state_indent')
action_157 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_157 state_indent tok = 
	 happyFail

action_158 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_270 state_indent'
action_158 state_indent tok = 
	 happyFail

action_159 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_273 state_indent'
action_159 state_indent tok = 
	 happyFail

action_160 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_297 state_indent'
action_160 state_indent tok = 
	 happyFail

action_161 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_277 state_indent'
action_161 state_indent tok = 
	 happyFail

action_162 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_208 state_indent'
action_162 state_indent tok = 
	 happyFail

action_163 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_298 state_indent'
action_163 state_indent tok = 
	 happyFail

action_164 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_322 state_indent'
action_164 state_indent tok = 
	 happyFail

action_165 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_324 state_indent'
action_165 state_indent tok = 
	 happyFail

action_166 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_323 state_indent'
action_166 state_indent tok = 
	 happyFail

action_167 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_325 state_indent'
action_167 state_indent tok = 
	 happyFail

action_168 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_168 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_168 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_168 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_168 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_199 state_indent')
action_168 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_168 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_201 state_indent')
action_168 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_202 state_indent')
action_168 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_168 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_168 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_168 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_168 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_168 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_338 state_indent')
action_168 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_168 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_204 state_indent')
action_168 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_168 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_205 state_indent')
action_168 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_206 state_indent')
action_168 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_168 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_168 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_207 state_indent')
action_168 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_208 state_indent')
action_168 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_168 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_168 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_168 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_168 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_168 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_168 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_168 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_168 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_168 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_168 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_331 state_indent')
action_168 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_168 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_168 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_332 state_indent')
action_168 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_168 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_168 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_168 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_168 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_168 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_168 state_indent (96, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_333 state_indent')
action_168 state_indent (97, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_334 state_indent')
action_168 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_168 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_168 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_168 state_indent (128, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_188 state_indent')
action_168 state_indent (130, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_189 state_indent')
action_168 state_indent (133, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_335 state_indent')
action_168 state_indent (134, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_336 state_indent')
action_168 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_168 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_168 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_168 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_168 state_indent (139, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_192 state_indent')
action_168 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_193 state_indent')
action_168 state_indent (141, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_337 state_indent')
action_168 state_indent (142, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_195 state_indent')
action_168 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_196 state_indent')
action_168 state_indent (144, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_197 state_indent')
action_168 state_indent (145, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_198 state_indent')
action_168 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_168 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_168 state_indent tok = 
	 happyFail

action_169 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_169 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_169 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_169 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_169 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_169 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_169 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_169 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_169 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_169 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_169 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_330 state_indent')
action_169 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_169 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_169 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_169 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_169 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_169 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_169 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_169 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_169 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_169 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_169 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_169 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_169 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_169 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_169 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_180 state_indent')
action_169 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_169 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_169 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_169 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_169 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_169 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_169 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_169 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_169 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_169 state_indent (98, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_329 state_indent')
action_169 state_indent (99, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_182 state_indent')
action_169 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_169 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_169 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_169 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_169 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_169 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_169 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_169 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_169 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_169 state_indent tok = 
	 happyFail

action_170 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_214 state_indent'
action_170 state_indent tok = 
	 happyFail

action_171 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_328 state_indent')
action_171 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_171 state_indent tok = 
	 happyFail

action_172 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_172 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_172 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_172 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_172 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_172 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_172 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_172 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_172 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_172 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_172 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_172 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_172 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_172 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_172 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_172 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_172 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_172 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_172 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_327 state_indent')
action_172 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_172 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_172 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_172 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_172 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_172 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_172 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_172 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_172 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_172 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_172 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_172 state_indent tok = 
	 happyFail

action_173 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_173 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_173 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_173 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_173 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_173 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_173 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_173 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_173 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_173 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_173 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_173 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_173 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_173 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_173 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_173 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_173 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_173 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_173 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_296 state_indent')
action_173 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_173 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_173 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_173 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_173 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_173 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_173 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_173 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_173 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_173 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_173 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_173 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_173 state_indent tok = 
	 happyFail

action_174 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_174 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_174 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_174 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_174 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_174 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_174 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_174 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_174 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_174 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_174 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_174 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_174 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_174 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_174 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_174 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_174 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_174 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_174 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_174 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_174 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_174 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_174 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_174 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_174 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_326 state_indent')
action_174 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_174 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_174 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_174 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_174 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_174 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_174 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_174 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_174 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_174 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_174 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_174 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_174 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_174 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_174 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_174 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_174 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_174 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_174 state_indent tok = 
	 happyFail

action_175 state_indent (313, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_44 state_indent')
action_175 state_indent (314, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_45 state_indent')
action_175 state_indent (315, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_46 state_indent')
action_175 state_indent (316, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_47 state_indent')
action_175 state_indent (321, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_48 state_indent')
action_175 state_indent (322, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_49 state_indent')
action_175 state_indent (323, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_50 state_indent')
action_175 state_indent (324, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_51 state_indent')
action_175 state_indent (325, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_52 state_indent')
action_175 state_indent (327, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_138 state_indent')
action_175 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_139 state_indent')
action_175 state_indent (331, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_53 state_indent')
action_175 state_indent (334, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_54 state_indent')
action_175 state_indent (340, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_140 state_indent')
action_175 state_indent (345, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_55 state_indent')
action_175 state_indent (347, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_56 state_indent')
action_175 state_indent (349, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_57 state_indent')
action_175 state_indent (354, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_61 state_indent')
action_175 state_indent (357, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_141 state_indent')
action_175 state_indent (364, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_142 state_indent')
action_175 state_indent (371, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_70 state_indent')
action_175 state_indent (372, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_71 state_indent')
action_175 state_indent (373, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_72 state_indent')
action_175 state_indent (374, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_73 state_indent')
action_175 state_indent (375, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_74 state_indent')
action_175 state_indent (376, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_75 state_indent')
action_175 state_indent (112, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_325 state_indent')
action_175 state_indent (113, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_126 state_indent')
action_175 state_indent (114, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_127 state_indent')
action_175 state_indent (161, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_128 state_indent')
action_175 state_indent (162, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_129 state_indent')
action_175 state_indent (170, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_130 state_indent')
action_175 state_indent (172, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_131 state_indent')
action_175 state_indent (190, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_132 state_indent')
action_175 state_indent (191, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_133 state_indent')
action_175 state_indent (193, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_134 state_indent')
action_175 state_indent (214, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_135 state_indent')
action_175 state_indent (215, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_29 state_indent')
action_175 state_indent (236, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_136 state_indent')
action_175 state_indent (237, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_32 state_indent')
action_175 state_indent (243, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_33 state_indent')
action_175 state_indent (244, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_34 state_indent')
action_175 state_indent (254, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_35 state_indent')
action_175 state_indent (259, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_137 state_indent')
action_175 state_indent (262, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_37 state_indent')
action_175 state_indent (263, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_38 state_indent')
action_175 state_indent (268, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_39 state_indent')
action_175 state_indent (277, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_40 state_indent')
action_175 state_indent (282, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_41 state_indent')
action_175 state_indent (284, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_42 state_indent')
action_175 state_indent (288, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_43 state_indent')
action_175 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_175 state_indent tok = 
	 happyFail

action_176 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_176 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_176 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_176 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_176 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_176 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_176 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_176 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_176 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_176 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_176 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_176 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_176 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_176 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_176 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_176 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_176 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_176 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_176 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_176 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_176 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_176 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_176 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_176 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_176 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_324 state_indent')
action_176 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_176 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_176 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_176 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_176 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_176 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_176 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_176 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_176 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_176 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_176 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_176 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_176 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_176 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_176 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_176 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_176 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_176 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_176 state_indent tok = 
	 happyFail

action_177 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_323 state_indent')
action_177 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_177 state_indent (41, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_316 state_indent')
action_177 state_indent (43, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_317 state_indent')
action_177 state_indent (44, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_318 state_indent')
action_177 state_indent (158, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_319 state_indent')
action_177 state_indent (163, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_320 state_indent')
action_177 state_indent (164, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_321 state_indent')
action_177 state_indent (294, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_322 state_indent')
action_177 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_177 state_indent tok = 
	 happyFail

action_178 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_178 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_178 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_178 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_178 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_178 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_178 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_178 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_178 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_178 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_178 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_178 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_178 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_178 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_178 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_178 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_178 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_178 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_178 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_315 state_indent')
action_178 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_178 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_178 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_178 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_178 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_178 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_178 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_178 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_178 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_178 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_178 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_459 state_indent'
action_178 state_indent tok = 
	 happyFail

action_179 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_541 state_indent'
action_179 state_indent tok = 
	 happyFail

action_180 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_312 state_indent')
action_180 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_313 state_indent')
action_180 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_314 state_indent')
action_180 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_219 state_indent'
action_180 state_indent tok = 
	 happyFail

action_181 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_311 state_indent')
action_181 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_181 state_indent tok = 
	 happyFail

action_182 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_310 state_indent')
action_182 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_220 state_indent'
action_182 state_indent tok = 
	 happyFail

action_183 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_610 state_indent'
action_183 state_indent tok = 
	 happyFail

action_184 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_308 state_indent')
action_184 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_309 state_indent')
action_184 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_184 state_indent tok = 
	 happyFail

action_185 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_185 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_185 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_201 state_indent')
action_185 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_202 state_indent')
action_185 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_227 state_indent')
action_185 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_206 state_indent')
action_185 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_307 state_indent')
action_185 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_185 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_185 state_indent (127, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_220 state_indent')
action_185 state_indent (130, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_221 state_indent')
action_185 state_indent (132, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_306 state_indent')
action_185 state_indent (134, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_223 state_indent')
action_185 state_indent (139, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_192 state_indent')
action_185 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_193 state_indent')
action_185 state_indent (141, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_224 state_indent')
action_185 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_196 state_indent')
action_185 state_indent (145, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_225 state_indent')
action_185 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_185 state_indent tok = 
	 happyFail

action_186 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_304 state_indent')
action_186 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_305 state_indent')
action_186 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_186 state_indent tok = 
	 happyFail

action_187 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_302 state_indent')
action_187 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_303 state_indent')
action_187 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_187 state_indent tok = 
	 happyFail

action_188 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_293 state_indent'
action_188 state_indent tok = 
	 happyFail

action_189 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_294 state_indent'
action_189 state_indent tok = 
	 happyFail

action_190 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_190 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_190 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_190 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_190 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_190 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_190 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_190 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_190 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_190 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_190 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_190 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_190 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_190 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_190 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_190 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_190 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_190 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_190 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_190 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_190 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_190 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_190 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_190 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_190 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_300 state_indent')
action_190 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_190 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_301 state_indent')
action_190 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_190 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_190 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_190 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_190 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_190 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_190 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_190 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_190 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_190 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_190 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_190 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_190 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_190 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_190 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_190 state_indent tok = 
	 happyFail

action_191 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_299 state_indent')
action_191 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_287 state_indent'
action_191 state_indent tok = 
	 happyFail

action_192 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_296 state_indent'
action_192 state_indent tok = 
	 happyFail

action_193 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_309 state_indent'
action_193 state_indent tok = 
	 happyFail

action_194 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_298 state_indent')
action_194 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_194 state_indent tok = 
	 happyFail

action_195 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_283 state_indent'
action_195 state_indent tok = 
	 happyFail

action_196 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_312 state_indent'
action_196 state_indent tok = 
	 happyFail

action_197 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_314 state_indent'
action_197 state_indent tok = 
	 happyFail

action_198 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_313 state_indent'
action_198 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_315 state_indent'
action_198 state_indent tok = 
	 happyFail

action_199 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_199 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_319 state_indent'
action_199 state_indent tok = 
	 happyFail

action_200 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) }) {-{([I1],[=])([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (Ranges [(0, Fin 0),(2, Inf)]) {-{([I1],[=])([I2],[I1])([=I2],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) }) {-{([I1],[=])([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) }) {-{([I1],[=])([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_311 state_indent'
action_200 state_indent tok = 
	 happyFail

action_201 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_321 state_indent'
action_201 state_indent tok = 
	 happyFail

action_202 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_310 state_indent'
action_202 state_indent tok = 
	 happyFail

action_203 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_609 state_indent'
action_203 state_indent tok = 
	 happyFail

action_204 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_216 state_indent'
action_204 state_indent tok = 
	 happyFail

action_205 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_205 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_205 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_205 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_205 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_205 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_205 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_205 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_205 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_205 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_205 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_297 state_indent')
action_205 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_205 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_273 state_indent')
action_205 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_205 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_205 state_indent tok = 
	 happyFail

action_206 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_295 state_indent'
action_206 state_indent tok = 
	 happyFail

action_207 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_207 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_207 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_207 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_207 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_207 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_207 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_207 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_207 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_207 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_207 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_207 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_207 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_207 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_207 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_207 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_207 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_207 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_207 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_296 state_indent')
action_207 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_207 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_207 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_207 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_207 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_207 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_207 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_207 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_207 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_207 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_207 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_207 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_207 state_indent tok = 
	 happyFail

action_208 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_208 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_320 state_indent'
action_208 state_indent tok = 
	 happyFail

action_209 state_indent (313, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_209 state_indent (314, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_209 state_indent (325, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_248 state_indent')
action_209 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_238 state_indent')
action_209 state_indent (371, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_209 state_indent (372, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_209 state_indent (373, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_209 state_indent (374, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_209 state_indent (375, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_209 state_indent (376, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_209 state_indent (118, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_291 state_indent')
action_209 state_indent (119, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_292 state_indent')
action_209 state_indent (120, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_293 state_indent')
action_209 state_indent (123, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_294 state_indent')
action_209 state_indent (135, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_209 state_indent (136, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_209 state_indent (147, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_295 state_indent')
action_209 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_209 state_indent tok = 
	 happyFail

action_210 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_210 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_210 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_210 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_210 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_210 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_210 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_210 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_210 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_210 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_210 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_210 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_210 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_210 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_210 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_210 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_210 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_210 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_210 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_290 state_indent')
action_210 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_210 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_210 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_210 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_210 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_210 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_210 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_210 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_210 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_210 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_210 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_210 state_indent tok = 
	 happyFail

action_211 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_505 state_indent'
action_211 state_indent tok = 
	 happyFail

action_212 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_212 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_289 state_indent')
action_212 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_212 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_212 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_212 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_212 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_212 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_212 state_indent (122, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_287 state_indent')
action_212 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_288 state_indent')
action_212 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_212 state_indent tok = 
	 happyFail

action_213 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_213 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_213 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_213 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_213 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_213 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_213 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_213 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_213 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_213 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_213 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_213 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_100 state_indent')
action_213 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_101 state_indent')
action_213 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_213 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_213 state_indent (55, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_286 state_indent')
action_213 state_indent (56, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_105 state_indent')
action_213 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_213 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_213 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_213 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_213 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_213 state_indent tok = 
	 happyFail

action_214 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_214 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_214 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_285 state_indent')
action_214 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_214 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_214 state_indent (31, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_279 state_indent')
action_214 state_indent (126, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_280 state_indent')
action_214 state_indent (129, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_281 state_indent')
action_214 state_indent (131, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_282 state_indent')
action_214 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_283 state_indent')
action_214 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_284 state_indent')
action_214 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_214 state_indent tok = 
	 happyFail

action_215 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_59 state_indent'
action_215 state_indent tok = 
	 happyFail

action_216 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_278 state_indent')
action_216 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_216 state_indent tok = 
	 happyFail

action_217 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_277 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent (80, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_276 state_indent')
action_217 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_217 state_indent tok = 
	 happyFail

action_218 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_216 state_indent')
action_218 state_indent (83, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_275 state_indent')
action_218 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_176 state_indent'
action_218 state_indent tok = 
	 happyFail

action_219 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_178 state_indent'
action_219 state_indent tok = 
	 happyFail

action_220 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_291 state_indent'
action_220 state_indent tok = 
	 happyFail

action_221 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_292 state_indent'
action_221 state_indent tok = 
	 happyFail

action_222 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_222 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_222 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_222 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_222 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_222 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_222 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_222 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_222 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_222 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_222 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_222 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_222 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_222 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_222 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_222 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_222 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_222 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_222 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_222 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_222 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_222 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_274 state_indent')
action_222 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_222 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_222 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_222 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_222 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_222 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_222 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_222 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_222 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_222 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_222 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_222 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_222 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_222 state_indent tok = 
	 happyFail

action_223 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_287 state_indent'
action_223 state_indent tok = 
	 happyFail

action_224 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_281 state_indent'
action_224 state_indent tok = 
	 happyFail

action_225 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_313 state_indent'
action_225 state_indent tok = 
	 happyFail

action_226 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) }) {-{([I1],[=])([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (Ranges [(0, Fin 0),(2, Inf)]) {-{([I1],[=])([I2],[I1])([=I2],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) }) {-{([I1],[=])([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) }) {-{([I1],[=])([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_316 state_indent'
action_226 state_indent tok = 
	 happyFail

action_227 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_227 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_227 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_227 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_227 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_227 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_227 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_227 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_227 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_227 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_227 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_272 state_indent')
action_227 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_227 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_273 state_indent')
action_227 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_227 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_227 state_indent tok = 
	 happyFail

action_228 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_228 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_228 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_228 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_228 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_228 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_228 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_228 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_228 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_228 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_228 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_228 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_228 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_228 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_228 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_228 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_228 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_228 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_228 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_228 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_228 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_228 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_228 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_228 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_228 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_271 state_indent')
action_228 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_228 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_228 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_228 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_228 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_228 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_228 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_228 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_228 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_228 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_228 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_228 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_228 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_228 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_228 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_228 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_228 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_228 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_228 state_indent tok = 
	 happyFail

action_229 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) }) {-{([I1],[=])([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (Ranges [(0, Fin 0),(2, Inf)]) {-{([I1],[=])([I2],[I1])([=I2],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) }) {-{([I1],[=])([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) }) {-{([I1],[=])([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_317 state_indent'
action_229 state_indent tok = 
	 happyFail

action_230 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) }) {-{([I1],[=])([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (Ranges [(0, Fin 0),(2, Inf)]) {-{([I1],[=])([I2],[I1])([=I2],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) }) {-{([I1],[=])([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) }) {-{([I1],[=])([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_318 state_indent'
action_230 state_indent tok = 
	 happyFail

action_231 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_231 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_231 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_231 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_231 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_231 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_231 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_231 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_231 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_231 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_231 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_231 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_231 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_231 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_231 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_260 state_indent')
action_231 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_261 state_indent')
action_231 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_262 state_indent')
action_231 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_231 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_263 state_indent')
action_231 state_indent (358, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_264 state_indent')
action_231 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_265 state_indent')
action_231 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_266 state_indent')
action_231 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_267 state_indent')
action_231 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_268 state_indent')
action_231 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_269 state_indent')
action_231 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_270 state_indent')
action_231 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_231 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_231 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_231 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_231 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_231 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_231 state_indent (30, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_256 state_indent')
action_231 state_indent (46, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_257 state_indent')
action_231 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_258 state_indent')
action_231 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_231 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_231 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_231 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_231 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_231 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_231 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_259 state_indent')
action_231 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_231 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_231 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_231 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_231 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_231 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_231 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_231 state_indent tok = 
	 happyFail

action_232 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_232 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_254 state_indent')
action_232 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_255 state_indent')
action_232 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_232 state_indent tok = 
	 happyFail

action_233 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_34 state_indent'
action_233 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_34 state_indent'
action_233 state_indent tok = 
	 happyFail

action_234 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_7 state_indent'
action_234 state_indent tok = 
	 happyFail

action_235 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_235 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_252 state_indent')
action_235 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_253 state_indent')
action_235 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_235 state_indent tok = 
	 happyFail

action_236 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_70 state_indent'
action_236 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_70 state_indent'
action_236 state_indent tok = 
	 happyFail

action_237 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_4 state_indent'
action_237 state_indent tok = 
	 happyFail

action_238 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(0, Fin 0)]) {-{([Z],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_326 state_indent'
action_238 state_indent tok = 
	 happyFail

action_239 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_251 state_indent')
action_239 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_239 state_indent tok = 
	 happyFail

action_240 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_17 state_indent'
action_240 state_indent tok = 
	 happyFail

action_241 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_241 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_241 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_241 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_241 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_248 state_indent')
action_241 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_249 state_indent')
action_241 state_indent (365, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_250 state_indent')
action_241 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_241 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_241 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_241 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_241 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_241 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_241 state_indent (13, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_242 state_indent')
action_241 state_indent (14, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_243 state_indent')
action_241 state_indent (15, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_244 state_indent')
action_241 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_245 state_indent')
action_241 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_241 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_241 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_246 state_indent')
action_241 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_241 state_indent (151, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_247 state_indent')
action_241 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_22 state_indent'
action_241 state_indent tok = 
	 happyFail

action_242 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_517 state_indent')
action_242 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_242 state_indent tok = 
	 happyFail

action_243 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_516 state_indent')
action_243 state_indent (13, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_515 state_indent')
action_243 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_22 state_indent'
action_243 state_indent tok = 
	 happyFail

action_244 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_24 state_indent'
action_244 state_indent tok = 
	 happyFail

action_245 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_25 state_indent'
action_245 state_indent tok = 
	 happyFail

action_246 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_331 state_indent'
action_246 state_indent tok = 
	 happyFail

action_247 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_514 state_indent')
action_247 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_26 state_indent'
action_247 state_indent tok = 
	 happyFail

action_248 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_248 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_201 state_indent')
action_248 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_248 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_248 state_indent (141, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_337 state_indent')
action_248 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_196 state_indent')
action_248 state_indent (145, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_225 state_indent')
action_248 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_248 state_indent tok = 
	 happyFail

action_249 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_21 state_indent'
action_249 state_indent tok = 
	 happyFail

action_250 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_23 state_indent')
action_250 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_24 state_indent')
action_250 state_indent (148, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_513 state_indent')
action_250 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_250 state_indent tok = 
	 happyFail

action_251 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_20 state_indent')
action_251 state_indent (377, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_13 state_indent'
action_251 state_indent (6, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_512 state_indent')
action_251 state_indent (8, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_6 state_indent')
action_251 state_indent (154, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_7 state_indent')
action_251 state_indent (155, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_8 state_indent')
action_251 state_indent (156, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_9 state_indent')
action_251 state_indent (157, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_10 state_indent')
action_251 state_indent (165, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_11 state_indent')
action_251 state_indent (166, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_12 state_indent')
action_251 state_indent (167, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_13 state_indent')
action_251 state_indent (177, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_14 state_indent')
action_251 state_indent (179, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_15 state_indent')
action_251 state_indent (188, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_16 state_indent')
action_251 state_indent (196, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_17 state_indent')
action_251 state_indent (199, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_18 state_indent')
action_251 state_indent (293, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_19 state_indent')
action_251 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_251 state_indent tok = 
	 happyFail

action_252 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_252 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_252 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_88 state_indent')
action_252 state_indent (28, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_89 state_indent')
action_252 state_indent (36, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_511 state_indent')
action_252 state_indent (40, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_91 state_indent')
action_252 state_indent (45, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_92 state_indent')
action_252 state_indent (47, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_93 state_indent')
action_252 state_indent (79, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_94 state_indent')
action_252 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_252 state_indent tok = 
	 happyFail

action_253 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_253 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_65 state_indent'
action_253 state_indent tok = 
	 happyFail

action_254 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_254 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_254 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_231 state_indent')
action_254 state_indent (18, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_509 state_indent')
action_254 state_indent (28, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_89 state_indent')
action_254 state_indent (32, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_510 state_indent')
action_254 state_indent (35, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_235 state_indent')
action_254 state_indent (36, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_236 state_indent')
action_254 state_indent (40, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_91 state_indent')
action_254 state_indent (45, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_92 state_indent')
action_254 state_indent (47, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_93 state_indent')
action_254 state_indent (79, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_94 state_indent')
action_254 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_254 state_indent tok = 
	 happyFail

action_255 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_255 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_8 state_indent'
action_255 state_indent tok = 
	 happyFail

action_256 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_215 state_indent')
action_256 state_indent (29, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_508 state_indent')
action_256 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_58 state_indent'
action_256 state_indent tok = 
	 happyFail

action_257 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_506 state_indent')
action_257 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_507 state_indent')
action_257 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_257 state_indent tok = 
	 happyFail

action_258 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_258 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_258 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_201 state_indent')
action_258 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_202 state_indent')
action_258 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_227 state_indent')
action_258 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_206 state_indent')
action_258 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_228 state_indent')
action_258 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_258 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_258 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_216 state_indent')
action_258 state_indent (81, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_504 state_indent')
action_258 state_indent (82, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_218 state_indent')
action_258 state_indent (83, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_219 state_indent')
action_258 state_indent (127, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_220 state_indent')
action_258 state_indent (130, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_221 state_indent')
action_258 state_indent (132, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_505 state_indent')
action_258 state_indent (134, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_223 state_indent')
action_258 state_indent (139, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_192 state_indent')
action_258 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_193 state_indent')
action_258 state_indent (141, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_224 state_indent')
action_258 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_196 state_indent')
action_258 state_indent (145, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_225 state_indent')
action_258 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_258 state_indent tok = 
	 happyFail

action_259 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_99 state_indent'
action_259 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_99 state_indent'
action_259 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_339 state_indent')
action_259 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_206 state_indent'
action_259 state_indent tok = 
	 happyFail

action_260 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_260 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_260 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_260 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_260 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_260 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_260 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_260 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_260 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_260 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_260 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_260 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_100 state_indent')
action_260 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_101 state_indent')
action_260 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_260 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_260 state_indent (55, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_503 state_indent')
action_260 state_indent (56, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_105 state_indent')
action_260 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_260 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_260 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_260 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_260 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_260 state_indent tok = 
	 happyFail

action_261 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_261 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_261 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_261 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_261 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_261 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_261 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_261 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_261 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_261 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_261 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_261 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_100 state_indent')
action_261 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_101 state_indent')
action_261 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_261 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_261 state_indent (55, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_502 state_indent')
action_261 state_indent (56, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_105 state_indent')
action_261 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_261 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_261 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_261 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_261 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_261 state_indent tok = 
	 happyFail

action_262 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_501 state_indent')
action_262 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_262 state_indent tok = 
	 happyFail

action_263 state_indent (358, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_499 state_indent')
action_263 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_500 state_indent')
action_263 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_263 state_indent tok = 
	 happyFail

action_264 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_122 state_indent')
action_264 state_indent (19, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_498 state_indent')
action_264 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_37 state_indent'
action_264 state_indent tok = 
	 happyFail

action_265 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_60 state_indent'
action_265 state_indent tok = 
	 happyFail

action_266 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_61 state_indent'
action_266 state_indent tok = 
	 happyFail

action_267 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_62 state_indent'
action_267 state_indent tok = 
	 happyFail

action_268 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_268 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_268 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_268 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_268 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_268 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_268 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_268 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_268 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_268 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_268 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_268 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_100 state_indent')
action_268 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_101 state_indent')
action_268 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_268 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_268 state_indent (55, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_497 state_indent')
action_268 state_indent (56, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_105 state_indent')
action_268 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_268 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_268 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_268 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_268 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_268 state_indent tok = 
	 happyFail

action_269 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_269 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_269 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_269 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_269 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_269 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_269 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_269 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_269 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_269 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_269 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_269 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_100 state_indent')
action_269 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_101 state_indent')
action_269 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_269 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_269 state_indent (55, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_496 state_indent')
action_269 state_indent (56, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_105 state_indent')
action_269 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_269 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_269 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_269 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_269 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_269 state_indent tok = 
	 happyFail

action_270 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_270 state_indent (58, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_495 state_indent')
action_270 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_97 state_indent')
action_270 state_indent (150, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_98 state_indent')
action_270 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_270 state_indent tok = 
	 happyFail

action_271 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_175 state_indent'
action_271 state_indent tok = 
	 happyFail

action_272 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_494 state_indent')
action_272 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_272 state_indent tok = 
	 happyFail

action_273 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_493 state_indent')
action_273 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_273 state_indent tok = 
	 happyFail

action_274 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_414 state_indent'
action_274 state_indent tok = 
	 happyFail

action_275 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_177 state_indent'
action_275 state_indent tok = 
	 happyFail

action_276 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_410 state_indent'
action_276 state_indent tok = 
	 happyFail

action_277 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_323 state_indent')
action_277 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I1],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_277 state_indent (41, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_492 state_indent')
action_277 state_indent (43, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_317 state_indent')
action_277 state_indent (44, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_318 state_indent')
action_277 state_indent (158, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_319 state_indent')
action_277 state_indent (163, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_320 state_indent')
action_277 state_indent (164, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_321 state_indent')
action_277 state_indent (294, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_322 state_indent')
action_277 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_277 state_indent tok = 
	 happyFail

action_278 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_278 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_278 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_278 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_278 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_278 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_278 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_278 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_278 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_278 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_278 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_278 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_278 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_278 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_278 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_278 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_278 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_278 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_278 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_278 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_278 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_278 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_278 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_278 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_278 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_491 state_indent')
action_278 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_278 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_301 state_indent')
action_278 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_278 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_278 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_278 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_278 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_278 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_278 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_278 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_278 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_278 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_278 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_278 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_278 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_278 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_278 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_278 state_indent tok = 
	 happyFail

action_279 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_490 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_425 state_indent'
action_279 state_indent tok = 
	 happyFail

action_280 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_289 state_indent'
action_280 state_indent tok = 
	 happyFail

action_281 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_290 state_indent'
action_281 state_indent tok = 
	 happyFail

action_282 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_64 state_indent'
action_282 state_indent tok = 
	 happyFail

action_283 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_285 state_indent'
action_283 state_indent tok = 
	 happyFail

action_284 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_279 state_indent'
action_284 state_indent tok = 
	 happyFail

action_285 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_285 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_285 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_285 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_285 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_285 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_285 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_285 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_285 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_488 state_indent')
action_285 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_489 state_indent')
action_285 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_285 state_indent tok = 
	 happyFail

action_286 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_430 state_indent'
action_286 state_indent tok = 
	 happyFail

action_287 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_486 state_indent'
action_287 state_indent tok = 
	 happyFail

action_288 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_271 state_indent'
action_288 state_indent tok = 
	 happyFail

action_289 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_289 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_289 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_289 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_487 state_indent')
action_289 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_289 state_indent tok = 
	 happyFail

action_290 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_540 state_indent'
action_290 state_indent tok = 
	 happyFail

action_291 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_238 state_indent')
action_291 state_indent (147, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_486 state_indent')
action_291 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_291 state_indent tok = 
	 happyFail

action_292 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_485 state_indent')
action_292 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_263 state_indent'
action_292 state_indent tok = 
	 happyFail

action_293 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_265 state_indent'
action_293 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_265 state_indent'
action_293 state_indent tok = 
	 happyFail

action_294 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_484 state_indent')
action_294 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_294 state_indent tok = 
	 happyFail

action_295 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_571 state_indent'
action_295 state_indent tok = 
	 happyFail

action_296 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_296 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_296 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_296 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_296 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_296 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_296 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_296 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_296 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_296 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_296 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_296 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_296 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_296 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_296 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_296 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_296 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_296 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_296 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_315 state_indent')
action_296 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_296 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_296 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_296 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_296 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_296 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_296 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_296 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_296 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_296 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_296 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_192 state_indent'
action_296 state_indent tok = 
	 happyFail

action_297 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_483 state_indent')
action_297 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_297 state_indent tok = 
	 happyFail

action_298 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_550 state_indent'
action_298 state_indent tok = 
	 happyFail

action_299 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_625 state_indent'
action_299 state_indent tok = 
	 happyFail

action_300 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_482 state_indent')
action_300 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_300 state_indent tok = 
	 happyFail

action_301 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_301 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_301 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_201 state_indent')
action_301 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_202 state_indent')
action_301 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_227 state_indent')
action_301 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_206 state_indent')
action_301 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_301 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_301 state_indent (127, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_220 state_indent')
action_301 state_indent (130, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_221 state_indent')
action_301 state_indent (132, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_341 state_indent')
action_301 state_indent (134, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_223 state_indent')
action_301 state_indent (139, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_192 state_indent')
action_301 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_193 state_indent')
action_301 state_indent (141, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_224 state_indent')
action_301 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_196 state_indent')
action_301 state_indent (145, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_225 state_indent')
action_301 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_301 state_indent tok = 
	 happyFail

action_302 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_592 state_indent'
action_302 state_indent tok = 
	 happyFail

action_303 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_303 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_303 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_303 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_303 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_303 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_303 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_303 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_303 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_303 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_303 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_303 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_303 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_303 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_303 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_303 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_303 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_303 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_303 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_303 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_303 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_303 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_303 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_303 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_303 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_481 state_indent')
action_303 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_303 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_303 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_303 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_303 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_303 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_303 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_303 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_303 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_303 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_303 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_303 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_303 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_303 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_303 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_303 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_303 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_303 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_303 state_indent tok = 
	 happyFail

action_304 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_611 state_indent'
action_304 state_indent tok = 
	 happyFail

action_305 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_215 state_indent'
action_305 state_indent tok = 
	 happyFail

action_306 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_306 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_306 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_306 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_306 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_306 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_306 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_306 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_306 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_306 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_480 state_indent')
action_306 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_306 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_306 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_306 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_306 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_306 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_306 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_306 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_306 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_306 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_306 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_306 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_306 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_306 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_306 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_306 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_442 state_indent')
action_306 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_443 state_indent')
action_306 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_306 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_306 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_306 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_306 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_306 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_306 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_306 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_306 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_306 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_306 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_306 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_306 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_306 state_indent tok = 
	 happyFail

action_307 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_479 state_indent')
action_307 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_307 state_indent tok = 
	 happyFail

action_308 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_591 state_indent'
action_308 state_indent tok = 
	 happyFail

action_309 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_309 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_309 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_309 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_309 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_309 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_309 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_309 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_309 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_309 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_309 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_309 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_309 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_309 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_309 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_309 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_309 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_309 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_309 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_309 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_309 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_309 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_309 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_309 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_309 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_478 state_indent')
action_309 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_309 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_309 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_309 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_309 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_309 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_309 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_309 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_309 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_309 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_309 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_309 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_309 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_309 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_309 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_309 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_309 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_309 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_309 state_indent tok = 
	 happyFail

action_310 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_310 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_310 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_310 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_310 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_310 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_310 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_310 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_310 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_310 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_310 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_310 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_310 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_310 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_310 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_310 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_310 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_310 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_310 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_310 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_310 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_310 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_310 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_310 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_310 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_477 state_indent')
action_310 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_310 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_310 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_310 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_310 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_310 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_310 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_310 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_310 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_310 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_310 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_310 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_310 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_310 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_310 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_310 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_310 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_310 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_310 state_indent tok = 
	 happyFail

action_311 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_593 state_indent'
action_311 state_indent tok = 
	 happyFail

action_312 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_312 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_312 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_312 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_312 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_312 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_312 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_312 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_312 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_312 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_312 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_312 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_312 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_312 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_312 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_312 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_312 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_312 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_312 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_312 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_312 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_312 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_312 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_312 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_312 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_476 state_indent')
action_312 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_312 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_312 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_312 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_312 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_312 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_312 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_312 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_312 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_312 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_312 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_312 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_312 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_312 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_312 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_312 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_312 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_312 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_312 state_indent tok = 
	 happyFail

action_313 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_313 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_313 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_313 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_313 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_313 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_313 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_313 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_313 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_313 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_313 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_313 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_313 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_313 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_313 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_313 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_313 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_313 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_313 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_313 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_313 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_313 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_313 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_313 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_313 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_475 state_indent')
action_313 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_313 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_313 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_313 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_313 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_313 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_313 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_313 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_313 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_313 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_313 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_313 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_313 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_313 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_313 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_313 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_313 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_313 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_221 state_indent'
action_313 state_indent tok = 
	 happyFail

action_314 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_314 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_314 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_314 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_314 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_314 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_314 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_314 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_314 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_314 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_314 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_314 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_314 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_314 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_314 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_314 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_314 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_314 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_474 state_indent')
action_314 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_314 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_314 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_314 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_314 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_314 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_314 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_470 state_indent')
action_314 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_314 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_314 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_354 state_indent')
action_314 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_314 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_314 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_314 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_314 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_314 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_314 state_indent (100, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_471 state_indent')
action_314 state_indent (101, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_472 state_indent')
action_314 state_indent (111, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_473 state_indent')
action_314 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_314 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_314 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_314 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_314 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_314 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_314 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_314 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_314 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_314 state_indent tok = 
	 happyFail

action_315 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_195 state_indent'
action_315 state_indent tok = 
	 happyFail

action_316 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_469 state_indent')
action_316 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_316 state_indent tok = 
	 happyFail

action_317 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I1],[=I1])([I2],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_90 state_indent'
action_317 state_indent tok = 
	 happyFail

action_318 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(2, Inf)])) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges []) {-{([I1],[=])([I2],[=I1])([I3],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_318 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent (158, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_468 state_indent')
action_318 state_indent (163, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_320 state_indent')
action_318 state_indent (164, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_321 state_indent')
action_318 state_indent (294, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_322 state_indent')
action_318 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_94 state_indent'
action_318 state_indent tok = 
	 happyFail

action_319 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_96 state_indent'
action_319 state_indent tok = 
	 happyFail

action_320 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (Ranges [(0, Fin 0),(2, Inf)]) {-{([I1],[=])([I2],[I1])([=I2],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_358 state_indent'
action_320 state_indent tok = 
	 happyFail

action_321 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_467 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_343 state_indent'
action_321 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_464 state_indent')
action_321 state_indent (28, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_89 state_indent')
action_321 state_indent (39, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_465 state_indent')
action_321 state_indent (40, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_466 state_indent')
action_321 state_indent (45, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_92 state_indent')
action_321 state_indent (79, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_94 state_indent')
action_321 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_321 state_indent tok = 
	 happyFail

action_322 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_463 state_indent')
action_322 state_indent (168, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_460 state_indent')
action_322 state_indent (177, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_461 state_indent')
action_322 state_indent (188, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_16 state_indent')
action_322 state_indent (196, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_17 state_indent')
action_322 state_indent (199, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_18 state_indent')
action_322 state_indent (293, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_462 state_indent')
action_322 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_322 state_indent tok = 
	 happyFail

action_323 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_323 state_indent (9, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_26 state_indent')
action_323 state_indent (10, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_458 state_indent')
action_323 state_indent (38, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_459 state_indent')
action_323 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_323 state_indent tok = 
	 happyFail

action_324 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_457 state_indent')
action_324 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_324 state_indent tok = 
	 happyFail

action_325 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_193 state_indent'
action_325 state_indent tok = 
	 happyFail

action_326 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_456 state_indent')
action_326 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_326 state_indent tok = 
	 happyFail

action_327 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_201 state_indent'
action_327 state_indent tok = 
	 happyFail

action_328 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_328 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_328 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_328 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_328 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_328 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_328 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_328 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_328 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_328 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_328 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_328 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_328 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_328 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_328 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_328 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_328 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_328 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_328 state_indent (91, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_455 state_indent')
action_328 state_indent (92, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_427 state_indent')
action_328 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_428 state_indent')
action_328 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_328 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_328 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_328 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_328 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_328 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_328 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_328 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_328 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_328 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_328 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_328 state_indent tok = 
	 happyFail

action_329 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_454 state_indent')
action_329 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_329 state_indent tok = 
	 happyFail

action_330 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_268 state_indent'
action_330 state_indent tok = 
	 happyFail

action_331 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_453 state_indent')
action_331 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_309 state_indent')
action_331 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_331 state_indent tok = 
	 happyFail

action_332 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_332 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_332 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_201 state_indent')
action_332 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_202 state_indent')
action_332 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_227 state_indent')
action_332 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_206 state_indent')
action_332 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_307 state_indent')
action_332 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_332 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_332 state_indent (127, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_220 state_indent')
action_332 state_indent (130, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_221 state_indent')
action_332 state_indent (132, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_452 state_indent')
action_332 state_indent (134, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_223 state_indent')
action_332 state_indent (139, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_192 state_indent')
action_332 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_193 state_indent')
action_332 state_indent (141, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_224 state_indent')
action_332 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_196 state_indent')
action_332 state_indent (145, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_225 state_indent')
action_332 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_332 state_indent tok = 
	 happyFail

action_333 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_451 state_indent')
action_333 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_305 state_indent')
action_333 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_333 state_indent tok = 
	 happyFail

action_334 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_450 state_indent')
action_334 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_303 state_indent')
action_334 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_334 state_indent tok = 
	 happyFail

action_335 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_335 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_335 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_335 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_335 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_335 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_335 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_335 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_335 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_335 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_335 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_335 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_335 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_335 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_335 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_335 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_335 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_335 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_335 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_335 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_335 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_335 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_335 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_335 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_335 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_449 state_indent')
action_335 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_335 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_301 state_indent')
action_335 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_335 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_335 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_335 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_335 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_335 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_335 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_335 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_335 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_335 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_335 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_335 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_335 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_335 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_335 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_335 state_indent tok = 
	 happyFail

action_336 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_448 state_indent')
action_336 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_287 state_indent'
action_336 state_indent tok = 
	 happyFail

action_337 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_447 state_indent')
action_337 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_337 state_indent tok = 
	 happyFail

action_338 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_267 state_indent'
action_338 state_indent tok = 
	 happyFail

action_339 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_339 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_339 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_339 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_339 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_339 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_339 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_339 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_339 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_339 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_339 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_339 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_339 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_339 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_339 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_339 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_339 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_339 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_339 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_446 state_indent')
action_339 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_339 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_339 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_339 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_339 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_339 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_339 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_339 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_339 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_339 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_339 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_339 state_indent tok = 
	 happyFail

action_340 state_indent (313, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_340 state_indent (314, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_340 state_indent (325, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_248 state_indent')
action_340 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_238 state_indent')
action_340 state_indent (371, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_340 state_indent (372, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_340 state_indent (373, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_340 state_indent (374, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_340 state_indent (375, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_340 state_indent (376, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_340 state_indent (118, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_444 state_indent')
action_340 state_indent (119, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_292 state_indent')
action_340 state_indent (120, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_293 state_indent')
action_340 state_indent (123, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_294 state_indent')
action_340 state_indent (135, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_340 state_indent (136, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_340 state_indent (147, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_445 state_indent')
action_340 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_340 state_indent tok = 
	 happyFail

action_341 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_341 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_341 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_341 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_341 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_341 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_341 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_341 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_341 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_341 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_341 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_341 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_341 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_341 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_341 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_341 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_341 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_341 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_341 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_341 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_341 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_341 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_341 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_341 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_341 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_442 state_indent')
action_341 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_443 state_indent')
action_341 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_341 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_341 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_341 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_341 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_341 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_341 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_341 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_341 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_341 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_341 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_341 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_341 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_341 state_indent tok = 
	 happyFail

action_342 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_441 state_indent')
action_342 state_indent (102, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_435 state_indent')
action_342 state_indent (103, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_436 state_indent')
action_342 state_indent (104, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_437 state_indent')
action_342 state_indent (160, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_438 state_indent')
action_342 state_indent (171, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_439 state_indent')
action_342 state_indent (293, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_440 state_indent')
action_342 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_342 state_indent tok = 
	 happyFail

action_343 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_365 state_indent'
action_343 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_365 state_indent'
action_343 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_365 state_indent'
action_343 state_indent tok = 
	 happyFail

action_344 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_323 state_indent')
action_344 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_344 state_indent (41, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_434 state_indent')
action_344 state_indent (43, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_317 state_indent')
action_344 state_indent (44, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_318 state_indent')
action_344 state_indent (158, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_319 state_indent')
action_344 state_indent (163, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_320 state_indent')
action_344 state_indent (164, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_321 state_indent')
action_344 state_indent (294, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_322 state_indent')
action_344 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_344 state_indent tok = 
	 happyFail

action_345 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_391 state_indent')
action_345 state_indent (60, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_432 state_indent')
action_345 state_indent (61, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_433 state_indent')
action_345 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_345 state_indent tok = 
	 happyFail

action_346 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_431 state_indent')
action_346 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_346 state_indent tok = 
	 happyFail

action_347 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_399 state_indent')
action_347 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_80 state_indent'
action_347 state_indent tok = 
	 happyFail

action_348 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_348 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_348 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_348 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_348 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_348 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_385 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_348 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_348 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_348 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_348 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_348 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_348 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_384 state_indent')
action_348 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_348 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_348 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_348 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_348 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_348 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_110 state_indent'
action_348 state_indent tok = 
	 happyFail

action_349 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_397 state_indent')
action_349 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_79 state_indent'
action_349 state_indent tok = 
	 happyFail

action_350 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_377 state_indent'
action_350 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_430 state_indent')
action_350 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_377 state_indent'
action_350 state_indent tok = 
	 happyFail

action_351 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_429 state_indent')
action_351 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_351 state_indent tok = 
	 happyFail

action_352 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_352 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_352 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_352 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_352 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_352 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_352 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_352 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_352 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_352 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_352 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_352 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_352 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_352 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_352 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_352 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_352 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_352 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_352 state_indent (91, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_426 state_indent')
action_352 state_indent (92, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_427 state_indent')
action_352 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_428 state_indent')
action_352 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_352 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_352 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_352 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_352 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_352 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_352 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_352 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_352 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_352 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_352 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_352 state_indent tok = 
	 happyFail

action_353 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_353 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_353 state_indent tok = 
	 happyFail

action_354 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_201 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_202 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_227 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_206 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_307 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_247 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent (127, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_220 state_indent')
action_354 state_indent (130, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_221 state_indent')
action_354 state_indent (132, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_341 state_indent')
action_354 state_indent (134, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_223 state_indent')
action_354 state_indent (139, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_192 state_indent')
action_354 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_193 state_indent')
action_354 state_indent (141, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_224 state_indent')
action_354 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_196 state_indent')
action_354 state_indent (145, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_225 state_indent')
action_354 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_183 state_indent'
action_354 state_indent tok = 
	 happyFail

action_355 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_425 state_indent')
action_355 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_355 state_indent tok = 
	 happyFail

action_356 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_238 state_indent')
action_356 state_indent (147, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_424 state_indent')
action_356 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_356 state_indent tok = 
	 happyFail

action_357 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_423 state_indent')
action_357 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_259 state_indent'
action_357 state_indent tok = 
	 happyFail

action_358 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_358 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_358 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_358 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_358 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_358 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_358 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_358 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_358 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_358 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_358 state_indent')
action_358 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_358 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_358 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_358 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_358 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_358 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_358 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_358 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_358 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_359 state_indent')
action_358 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_358 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_358 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_358 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_358 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_358 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_358 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_353 state_indent')
action_358 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_358 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_358 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_354 state_indent')
action_358 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_358 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_358 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_358 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_358 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_358 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_358 state_indent (111, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_355 state_indent')
action_358 state_indent (116, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_420 state_indent')
action_358 state_indent (117, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_357 state_indent')
action_358 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_358 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_358 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_358 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_358 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_358 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_358 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_358 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_358 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_358 state_indent tok = 
	 happyFail

action_359 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_323 state_indent')
action_359 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_359 state_indent (41, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_422 state_indent')
action_359 state_indent (43, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_317 state_indent')
action_359 state_indent (44, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_318 state_indent')
action_359 state_indent (158, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_319 state_indent')
action_359 state_indent (163, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_320 state_indent')
action_359 state_indent (164, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_321 state_indent')
action_359 state_indent (294, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_322 state_indent')
action_359 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_359 state_indent tok = 
	 happyFail

action_360 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_262 state_indent'
action_360 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_255 state_indent'
action_360 state_indent tok = 
	 happyFail

action_361 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_350 state_indent'
action_361 state_indent tok = 
	 happyFail

action_362 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_353 state_indent'
action_362 state_indent tok = 
	 happyFail

action_363 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_421 state_indent')
action_363 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_259 state_indent'
action_363 state_indent tok = 
	 happyFail

action_364 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_364 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_364 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_364 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_364 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_364 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_364 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_364 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_364 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_364 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_364 state_indent')
action_364 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_364 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_364 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_364 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_364 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_364 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_364 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_364 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_364 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_359 state_indent')
action_364 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_364 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_364 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_364 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_364 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_364 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_364 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_360 state_indent')
action_364 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_364 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_364 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_354 state_indent')
action_364 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_364 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_364 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_364 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_364 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_364 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_364 state_indent (111, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_355 state_indent')
action_364 state_indent (115, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_419 state_indent')
action_364 state_indent (116, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_420 state_indent')
action_364 state_indent (117, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_363 state_indent')
action_364 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_364 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_364 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_364 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_364 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_364 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_364 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_364 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_364 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_364 state_indent tok = 
	 happyFail

action_365 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_418 state_indent')
action_365 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_365 state_indent tok = 
	 happyFail

action_366 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_366 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_366 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_366 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_366 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_366 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_366 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_366 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_366 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_366 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_366 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_366 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_366 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_366 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_366 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_366 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_366 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_366 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_366 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_366 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_366 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_366 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_366 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_366 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_366 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_417 state_indent')
action_366 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_274 state_indent')
action_366 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_366 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_366 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_366 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_366 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_366 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_366 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_366 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_366 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_366 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_366 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_366 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_366 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_366 state_indent tok = 
	 happyFail

action_367 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_416 state_indent')
action_367 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_367 state_indent tok = 
	 happyFail

action_368 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_368 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_368 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_368 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_368 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_368 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_368 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_368 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_368 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_368 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_364 state_indent')
action_368 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_368 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_368 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_368 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_368 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_368 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_368 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_368 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_368 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_359 state_indent')
action_368 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_368 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_368 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_368 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_368 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_368 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_368 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_360 state_indent')
action_368 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_368 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_368 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_354 state_indent')
action_368 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_368 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_368 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_368 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_368 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_368 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_368 state_indent (111, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_355 state_indent')
action_368 state_indent (115, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_414 state_indent')
action_368 state_indent (116, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_415 state_indent')
action_368 state_indent (117, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_363 state_indent')
action_368 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_368 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_368 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_368 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_368 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_368 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_368 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_368 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_368 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_354 state_indent'
action_368 state_indent tok = 
	 happyFail

action_369 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_251 state_indent'
action_369 state_indent tok = 
	 happyFail

action_370 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_413 state_indent')
action_370 state_indent (49, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_412 state_indent')
action_370 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_106 state_indent'
action_370 state_indent tok = 
	 happyFail

action_371 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_410 state_indent')
action_371 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_411 state_indent')
action_371 state_indent (48, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_409 state_indent')
action_371 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_104 state_indent'
action_371 state_indent tok = 
	 happyFail

action_372 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_39 state_indent'
action_372 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_39 state_indent'
action_372 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_408 state_indent')
action_372 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_39 state_indent'
action_372 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_39 state_indent'
action_372 state_indent (20, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_407 state_indent')
action_372 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_39 state_indent'
action_372 state_indent tok = 
	 happyFail

action_373 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_366 state_indent'
action_373 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_366 state_indent'
action_373 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_366 state_indent'
action_373 state_indent tok = 
	 happyFail

action_374 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_406 state_indent')
action_374 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_161 state_indent'
action_374 state_indent (77, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_401 state_indent')
action_374 state_indent (78, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_402 state_indent')
action_374 state_indent (159, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_403 state_indent')
action_374 state_indent (163, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_320 state_indent')
action_374 state_indent (164, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_404 state_indent')
action_374 state_indent (294, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_405 state_indent')
action_374 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_374 state_indent tok = 
	 happyFail

action_375 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_400 state_indent')
action_375 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_375 state_indent tok = 
	 happyFail

action_376 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_121 state_indent'
action_376 state_indent tok = 
	 happyFail

action_377 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_398 state_indent')
action_377 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_399 state_indent')
action_377 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_377 state_indent tok = 
	 happyFail

action_378 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_396 state_indent')
action_378 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_397 state_indent')
action_378 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_378 state_indent tok = 
	 happyFail

action_379 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_395 state_indent')
action_379 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_305 state_indent')
action_379 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_379 state_indent tok = 
	 happyFail

action_380 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_119 state_indent'
action_380 state_indent tok = 
	 happyFail

action_381 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_394 state_indent')
action_381 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_381 state_indent tok = 
	 happyFail

action_382 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_382 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_382 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_382 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_382 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_382 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_382 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_382 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_382 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_382 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_382 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_382 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_393 state_indent')
action_382 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_382 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_382 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_382 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_382 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_382 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_382 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_382 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_382 state_indent tok = 
	 happyFail

action_383 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_391 state_indent')
action_383 state_indent (61, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_392 state_indent')
action_383 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_383 state_indent tok = 
	 happyFail

action_384 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_111 state_indent'
action_384 state_indent tok = 
	 happyFail

action_385 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_385 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_385 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_385 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_385 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_385 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_385 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_385 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_385 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_385 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_385 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_385 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_390 state_indent')
action_385 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_385 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_385 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_385 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_385 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_385 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_385 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_385 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_385 state_indent tok = 
	 happyFail

action_386 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_386 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_386 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_386 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_386 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_386 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_386 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_386 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_386 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_389 state_indent')
action_386 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_128 state_indent'
action_386 state_indent tok = 
	 happyFail

action_387 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_387 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_387 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_387 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_387 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_387 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_387 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_387 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_387 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_387 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_387 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_387 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_388 state_indent')
action_387 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_387 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_387 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_387 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_387 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_387 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_387 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_387 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_387 state_indent tok = 
	 happyFail

action_388 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_362 state_indent'
action_388 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_362 state_indent'
action_388 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_362 state_indent'
action_388 state_indent tok = 
	 happyFail

action_389 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_129 state_indent'
action_389 state_indent tok = 
	 happyFail

action_390 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_109 state_indent'
action_390 state_indent tok = 
	 happyFail

action_391 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_391 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_391 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_391 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_612 state_indent')
action_391 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_391 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_613 state_indent')
action_391 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_391 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_391 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_391 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_391 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_391 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_391 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_606 state_indent')
action_391 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_391 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_391 state_indent (62, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_607 state_indent')
action_391 state_indent (63, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_608 state_indent')
action_391 state_indent (65, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_609 state_indent')
action_391 state_indent (124, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_610 state_indent')
action_391 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_391 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_391 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_611 state_indent')
action_391 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_391 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_391 state_indent tok = 
	 happyFail

action_392 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_392 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_576 state_indent')
action_392 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_392 state_indent (70, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_605 state_indent')
action_392 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_392 state_indent tok = 
	 happyFail

action_393 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_123 state_indent'
action_393 state_indent tok = 
	 happyFail

action_394 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_120 state_indent'
action_394 state_indent tok = 
	 happyFail

action_395 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_122 state_indent'
action_395 state_indent tok = 
	 happyFail

action_396 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_115 state_indent'
action_396 state_indent tok = 
	 happyFail

action_397 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_397 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_397 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_397 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_397 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_397 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_397 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_397 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_397 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_397 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_397 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_397 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_604 state_indent')
action_397 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_397 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_397 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_397 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_397 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_397 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_397 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_397 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_397 state_indent tok = 
	 happyFail

action_398 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_117 state_indent'
action_398 state_indent tok = 
	 happyFail

action_399 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_399 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_399 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_399 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_399 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_399 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_399 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_399 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_399 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_399 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_399 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_399 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_603 state_indent')
action_399 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_399 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_399 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_399 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_399 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_399 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_399 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_399 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_399 state_indent tok = 
	 happyFail

action_400 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (346, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_116 state_indent'
action_400 state_indent tok = 
	 happyFail

action_401 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_160 state_indent'
action_401 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_160 state_indent'
action_401 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_160 state_indent'
action_401 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_160 state_indent'
action_401 state_indent tok = 
	 happyFail

action_402 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_169 state_indent'
action_402 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_169 state_indent'
action_402 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_402 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_169 state_indent'
action_402 state_indent (159, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_602 state_indent')
action_402 state_indent (163, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_320 state_indent')
action_402 state_indent (164, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_404 state_indent')
action_402 state_indent (294, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_405 state_indent')
action_402 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_169 state_indent'
action_402 state_indent tok = 
	 happyFail

action_403 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges []) `unionRanges` (token_indent) {-{([I1],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_171 state_indent'
action_403 state_indent tok = 
	 happyFail

action_404 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_467 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges []) `unionRanges` (token_indent) {-{([I1],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_346 state_indent'
action_404 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_599 state_indent')
action_404 state_indent (75, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_600 state_indent')
action_404 state_indent (79, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_601 state_indent')
action_404 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_404 state_indent tok = 
	 happyFail

action_405 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_463 state_indent')
action_405 state_indent (169, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_596 state_indent')
action_405 state_indent (188, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_597 state_indent')
action_405 state_indent (293, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_598 state_indent')
action_405 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_405 state_indent tok = 
	 happyFail

action_406 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_406 state_indent (9, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_26 state_indent')
action_406 state_indent (10, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_594 state_indent')
action_406 state_indent (74, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_595 state_indent')
action_406 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_406 state_indent tok = 
	 happyFail

action_407 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_45 state_indent'
action_407 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_41 state_indent'
action_407 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_593 state_indent')
action_407 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_41 state_indent'
action_407 state_indent (21, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_590 state_indent')
action_407 state_indent (22, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_591 state_indent')
action_407 state_indent (23, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_592 state_indent')
action_407 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_41 state_indent'
action_407 state_indent tok = 
	 happyFail

action_408 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_23 state_indent')
action_408 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_24 state_indent')
action_408 state_indent (148, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_589 state_indent')
action_408 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_408 state_indent tok = 
	 happyFail

action_409 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_413 state_indent')
action_409 state_indent (49, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_588 state_indent')
action_409 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_106 state_indent'
action_409 state_indent tok = 
	 happyFail

action_410 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_102 state_indent'
action_410 state_indent tok = 
	 happyFail

action_411 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_103 state_indent'
action_411 state_indent tok = 
	 happyFail

action_412 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_586 state_indent')
action_412 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_587 state_indent')
action_412 state_indent (50, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_585 state_indent')
action_412 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_412 state_indent tok = 
	 happyFail

action_413 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_105 state_indent'
action_413 state_indent tok = 
	 happyFail

action_414 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_349 state_indent'
action_414 state_indent tok = 
	 happyFail

action_415 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_352 state_indent'
action_415 state_indent tok = 
	 happyFail

action_416 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_416 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_416 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_416 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_416 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_416 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_416 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_416 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_416 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_416 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_416 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_416 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_100 state_indent')
action_416 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_101 state_indent')
action_416 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_416 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_416 state_indent (55, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_584 state_indent')
action_416 state_indent (56, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_105 state_indent')
action_416 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_416 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_416 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_416 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_416 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_416 state_indent tok = 
	 happyFail

action_417 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_456 state_indent'
action_417 state_indent tok = 
	 happyFail

action_418 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_418 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_418 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_418 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_418 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_418 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_418 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_418 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_418 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_418 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_418 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_418 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_418 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_418 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_418 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_418 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_418 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_418 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_418 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_418 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_418 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_418 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_418 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_418 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_418 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_583 state_indent')
action_418 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_418 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_418 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_418 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_418 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_418 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_418 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_418 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_418 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_418 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_418 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_418 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_418 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_418 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_418 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_418 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_418 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_418 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_418 state_indent tok = 
	 happyFail

action_419 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_254 state_indent'
action_419 state_indent tok = 
	 happyFail

action_420 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_257 state_indent'
action_420 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_257 state_indent'
action_420 state_indent tok = 
	 happyFail

action_421 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_421 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_421 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_421 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_421 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_421 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_421 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_421 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_421 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_421 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_364 state_indent')
action_421 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_421 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_421 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_421 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_421 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_421 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_421 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_421 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_421 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_359 state_indent')
action_421 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_421 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_421 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_421 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_421 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_421 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_421 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_360 state_indent')
action_421 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_421 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_421 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_354 state_indent')
action_421 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_421 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_421 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_421 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_421 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_421 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_421 state_indent (111, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_355 state_indent')
action_421 state_indent (115, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_582 state_indent')
action_421 state_indent (116, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_581 state_indent')
action_421 state_indent (117, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_363 state_indent')
action_421 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_421 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_421 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_421 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_421 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_421 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_421 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_421 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_421 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_258 state_indent'
action_421 state_indent tok = 
	 happyFail

action_422 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_260 state_indent'
action_422 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_260 state_indent'
action_422 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_469 state_indent')
action_422 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_260 state_indent'
action_422 state_indent tok = 
	 happyFail

action_423 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_423 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_423 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_423 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_423 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_423 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_423 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_423 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_423 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_423 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_358 state_indent')
action_423 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_423 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_423 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_423 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_423 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_423 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_423 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_423 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_423 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_359 state_indent')
action_423 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_423 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_423 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_423 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_423 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_423 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_423 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_353 state_indent')
action_423 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_423 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_423 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_354 state_indent')
action_423 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_423 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_423 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_423 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_423 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_423 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_423 state_indent (111, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_355 state_indent')
action_423 state_indent (116, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_581 state_indent')
action_423 state_indent (117, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_357 state_indent')
action_423 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_423 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_423 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_423 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_423 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_423 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_423 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_423 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_423 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_258 state_indent'
action_423 state_indent tok = 
	 happyFail

action_424 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_248 state_indent'
action_424 state_indent tok = 
	 happyFail

action_425 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_580 state_indent')
action_425 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_425 state_indent tok = 
	 happyFail

action_426 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_426 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_426 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_426 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_426 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_426 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_426 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_426 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_426 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_426 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_426 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_426 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_579 state_indent')
action_426 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_426 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_426 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_426 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_426 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_426 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_426 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_426 state_indent (92, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_562 state_indent')
action_426 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_428 state_indent')
action_426 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_426 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_426 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_426 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_426 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_426 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_426 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_426 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_426 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_426 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_426 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_426 state_indent tok = 
	 happyFail

action_427 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_198 state_indent'
action_427 state_indent tok = 
	 happyFail

action_428 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_199 state_indent'
action_428 state_indent tok = 
	 happyFail

action_429 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_429 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_429 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_429 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_429 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_429 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_429 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_429 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_429 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_429 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_429 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_429 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_429 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_429 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_429 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_429 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_429 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_429 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_429 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_429 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_429 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_429 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_429 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_429 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_429 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_578 state_indent')
action_429 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_429 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_429 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_429 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_429 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_429 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_429 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_429 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_429 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_429 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_429 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_429 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_429 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_429 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_429 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_429 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_429 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_429 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_429 state_indent tok = 
	 happyFail

action_430 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_430 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_430 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_430 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_430 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_430 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_430 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_430 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_430 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_430 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_430 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_430 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_430 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_430 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_430 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_430 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_430 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_430 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_430 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_430 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_430 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_430 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_430 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_430 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_430 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_577 state_indent')
action_430 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_430 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_430 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_430 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_430 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_430 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_430 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_430 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_430 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_430 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_430 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_430 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_430 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_430 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_430 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_430 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_430 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_430 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_430 state_indent tok = 
	 happyFail

action_431 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_367 state_indent'
action_431 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_367 state_indent'
action_431 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_367 state_indent'
action_431 state_indent tok = 
	 happyFail

action_432 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_432 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_575 state_indent')
action_432 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_576 state_indent')
action_432 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_432 state_indent (70, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_574 state_indent')
action_432 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_432 state_indent tok = 
	 happyFail

action_433 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_132 state_indent'
action_433 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_132 state_indent'
action_433 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_132 state_indent'
action_433 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_132 state_indent'
action_433 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_132 state_indent'
action_433 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_132 state_indent'
action_433 state_indent tok = 
	 happyFail

action_434 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_157 state_indent'
action_434 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_157 state_indent'
action_434 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_157 state_indent'
action_434 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_157 state_indent'
action_434 state_indent tok = 
	 happyFail

action_435 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_458 state_indent'
action_435 state_indent tok = 
	 happyFail

action_436 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_234 state_indent'
action_436 state_indent tok = 
	 happyFail

action_437 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_437 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent (160, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_573 state_indent')
action_437 state_indent (171, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_439 state_indent')
action_437 state_indent (293, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_440 state_indent')
action_437 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_235 state_indent'
action_437 state_indent tok = 
	 happyFail

action_438 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_571 state_indent')
action_438 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_572 state_indent')
action_438 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_237 state_indent'
action_438 state_indent tok = 
	 happyFail

action_439 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges []) {-{([I1],[=I1])([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_348 state_indent'
action_439 state_indent tok = 
	 happyFail

action_440 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_44 state_indent')
action_440 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_45 state_indent')
action_440 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_46 state_indent')
action_440 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_47 state_indent')
action_440 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_48 state_indent')
action_440 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_49 state_indent')
action_440 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_50 state_indent')
action_440 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_51 state_indent')
action_440 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_52 state_indent')
action_440 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_53 state_indent')
action_440 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_54 state_indent')
action_440 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_55 state_indent')
action_440 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_56 state_indent')
action_440 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_57 state_indent')
action_440 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_61 state_indent')
action_440 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_70 state_indent')
action_440 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_71 state_indent')
action_440 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_72 state_indent')
action_440 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_73 state_indent')
action_440 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_74 state_indent')
action_440 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_75 state_indent')
action_440 state_indent (191, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_569 state_indent')
action_440 state_indent (193, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_570 state_indent')
action_440 state_indent (215, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_29 state_indent')
action_440 state_indent (237, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_32 state_indent')
action_440 state_indent (243, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_33 state_indent')
action_440 state_indent (244, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_34 state_indent')
action_440 state_indent (254, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_35 state_indent')
action_440 state_indent (259, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_137 state_indent')
action_440 state_indent (262, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_37 state_indent')
action_440 state_indent (263, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_38 state_indent')
action_440 state_indent (268, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_39 state_indent')
action_440 state_indent (277, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_40 state_indent')
action_440 state_indent (282, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_41 state_indent')
action_440 state_indent (284, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_42 state_indent')
action_440 state_indent (288, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_43 state_indent')
action_440 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_440 state_indent tok = 
	 happyFail

action_441 state_indent (9, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_26 state_indent')
action_441 state_indent (10, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_567 state_indent')
action_441 state_indent (105, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_568 state_indent')
action_441 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_441 state_indent tok = 
	 happyFail

action_442 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_184 state_indent'
action_442 state_indent tok = 
	 happyFail

action_443 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_186 state_indent'
action_443 state_indent tok = 
	 happyFail

action_444 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_238 state_indent')
action_444 state_indent (147, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_566 state_indent')
action_444 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_444 state_indent tok = 
	 happyFail

action_445 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_203 state_indent'
action_445 state_indent tok = 
	 happyFail

action_446 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_200 state_indent'
action_446 state_indent tok = 
	 happyFail

action_447 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (344, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_274 state_indent'
action_447 state_indent tok = 
	 happyFail

action_448 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_278 state_indent'
action_448 state_indent tok = 
	 happyFail

action_449 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_565 state_indent')
action_449 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_449 state_indent tok = 
	 happyFail

action_450 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_210 state_indent'
action_450 state_indent tok = 
	 happyFail

action_451 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_269 state_indent'
action_451 state_indent tok = 
	 happyFail

action_452 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_452 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_452 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_452 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_452 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_452 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_452 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_452 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_452 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_452 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_564 state_indent')
action_452 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_452 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_452 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_452 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_452 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_452 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_452 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_452 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_452 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_452 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_452 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_452 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_452 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_452 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_452 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_452 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_442 state_indent')
action_452 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_443 state_indent')
action_452 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_452 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_452 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_452 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_452 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_452 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_452 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_452 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_452 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_452 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_452 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_452 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_452 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_452 state_indent tok = 
	 happyFail

action_453 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_209 state_indent'
action_453 state_indent tok = 
	 happyFail

action_454 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_211 state_indent'
action_454 state_indent tok = 
	 happyFail

action_455 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_455 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_455 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_455 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_455 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_455 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_455 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_455 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_455 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_455 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_455 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_455 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_563 state_indent')
action_455 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_455 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_455 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_455 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_455 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_455 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_455 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_455 state_indent (92, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_562 state_indent')
action_455 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_428 state_indent')
action_455 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_455 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_455 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_455 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_455 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_455 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_455 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_455 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_455 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_455 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_455 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_455 state_indent tok = 
	 happyFail

action_456 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_441 state_indent')
action_456 state_indent (102, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_561 state_indent')
action_456 state_indent (103, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_436 state_indent')
action_456 state_indent (104, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_437 state_indent')
action_456 state_indent (160, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_438 state_indent')
action_456 state_indent (171, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_439 state_indent')
action_456 state_indent (293, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_440 state_indent')
action_456 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_635 state_indent'
action_456 state_indent tok = 
	 happyFail

action_457 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_457 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_457 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_457 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_457 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_457 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_457 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_457 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_457 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_457 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_457 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_457 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_457 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_457 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_457 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_457 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_457 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_457 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_457 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_457 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_457 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_457 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_457 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_457 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_457 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_560 state_indent')
action_457 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_457 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_457 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_457 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_457 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_457 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_457 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_457 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_457 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_457 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_457 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_457 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_457 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_457 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_457 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_457 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_457 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_457 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_457 state_indent tok = 
	 happyFail

action_458 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_458 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_83 state_indent'
action_458 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_464 state_indent')
action_458 state_indent (28, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_89 state_indent')
action_458 state_indent (39, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_559 state_indent')
action_458 state_indent (40, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_466 state_indent')
action_458 state_indent (45, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_92 state_indent')
action_458 state_indent (79, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_94 state_indent')
action_458 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_458 state_indent tok = 
	 happyFail

action_459 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_238 state_indent')
action_459 state_indent (147, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_558 state_indent')
action_459 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_459 state_indent tok = 
	 happyFail

action_460 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_556 state_indent')
action_460 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_557 state_indent')
action_460 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_460 state_indent tok = 
	 happyFail

action_461 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_371 state_indent'
action_461 state_indent tok = 
	 happyFail

action_462 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_44 state_indent')
action_462 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_45 state_indent')
action_462 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_46 state_indent')
action_462 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_47 state_indent')
action_462 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_48 state_indent')
action_462 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_49 state_indent')
action_462 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_50 state_indent')
action_462 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_51 state_indent')
action_462 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_52 state_indent')
action_462 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_53 state_indent')
action_462 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_54 state_indent')
action_462 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_55 state_indent')
action_462 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_56 state_indent')
action_462 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_57 state_indent')
action_462 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_61 state_indent')
action_462 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_64 state_indent')
action_462 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_65 state_indent')
action_462 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_66 state_indent')
action_462 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_70 state_indent')
action_462 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_71 state_indent')
action_462 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_72 state_indent')
action_462 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_73 state_indent')
action_462 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_74 state_indent')
action_462 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_75 state_indent')
action_462 state_indent (191, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_28 state_indent')
action_462 state_indent (215, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_29 state_indent')
action_462 state_indent (227, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_30 state_indent')
action_462 state_indent (230, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_31 state_indent')
action_462 state_indent (237, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_32 state_indent')
action_462 state_indent (243, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_33 state_indent')
action_462 state_indent (244, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_34 state_indent')
action_462 state_indent (254, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_35 state_indent')
action_462 state_indent (259, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_36 state_indent')
action_462 state_indent (262, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_37 state_indent')
action_462 state_indent (263, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_38 state_indent')
action_462 state_indent (268, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_39 state_indent')
action_462 state_indent (277, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_40 state_indent')
action_462 state_indent (282, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_41 state_indent')
action_462 state_indent (284, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_42 state_indent')
action_462 state_indent (288, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_43 state_indent')
action_462 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_462 state_indent tok = 
	 happyFail

action_463 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0, Fin 0),(2, Inf)]) {-{([I1],[=])([=I2],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_357 state_indent'
action_463 state_indent tok = 
	 happyFail

action_464 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_464 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_464 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_464 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_464 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_464 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_464 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_464 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_464 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_464 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_464 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_464 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_464 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_464 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_464 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_464 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_265 state_indent')
action_464 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_266 state_indent')
action_464 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_267 state_indent')
action_464 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_464 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_464 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_464 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_464 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_464 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_464 state_indent (30, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_256 state_indent')
action_464 state_indent (46, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_257 state_indent')
action_464 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_258 state_indent')
action_464 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_464 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_464 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_464 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_464 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_464 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_464 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_259 state_indent')
action_464 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_464 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_464 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_464 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_464 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_464 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_464 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_464 state_indent tok = 
	 happyFail

action_465 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_554 state_indent')
action_465 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_555 state_indent')
action_465 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_465 state_indent tok = 
	 happyFail

action_466 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_85 state_indent'
action_466 state_indent tok = 
	 happyFail

action_467 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0, Fin 0),(2, Inf)]) {-{([I1],[=])([=I2],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_356 state_indent'
action_467 state_indent tok = 
	 happyFail

action_468 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_95 state_indent'
action_468 state_indent tok = 
	 happyFail

action_469 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_469 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_469 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_469 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_469 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_469 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_469 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_469 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_469 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_469 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_469 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_469 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_469 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_469 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_469 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_469 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_469 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_469 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_469 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_469 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_469 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_469 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_469 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_469 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_469 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_553 state_indent')
action_469 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_469 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_469 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_469 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_469 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_469 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_469 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_469 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_469 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_469 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_469 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_469 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_469 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_469 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_469 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_469 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_469 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_469 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_469 state_indent tok = 
	 happyFail

action_470 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_231 state_indent'
action_470 state_indent tok = 
	 happyFail

action_471 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_552 state_indent')
action_471 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_225 state_indent'
action_471 state_indent tok = 
	 happyFail

action_472 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_229 state_indent'
action_472 state_indent tok = 
	 happyFail

action_473 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_551 state_indent')
action_473 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_473 state_indent tok = 
	 happyFail

action_474 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_323 state_indent')
action_474 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_474 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_474 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_91 state_indent'
action_474 state_indent (41, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_550 state_indent')
action_474 state_indent (43, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_317 state_indent')
action_474 state_indent (44, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_318 state_indent')
action_474 state_indent (158, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_319 state_indent')
action_474 state_indent (163, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_320 state_indent')
action_474 state_indent (164, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_321 state_indent')
action_474 state_indent (294, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_322 state_indent')
action_474 state_indent (tok, token_indent) 
	| item_indent <- Ranges [(minRanges state_indent + 1, Inf)]{-[I1]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_636 state_indent'
action_474 state_indent tok = 
	 happyFail

action_475 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_223 state_indent'
action_475 state_indent tok = 
	 happyFail

action_476 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_549 state_indent')
action_476 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_227 state_indent'
action_476 state_indent tok = 
	 happyFail

action_477 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_226 state_indent'
action_477 state_indent tok = 
	 happyFail

action_478 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_218 state_indent'
action_478 state_indent tok = 
	 happyFail

action_479 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_479 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_479 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_479 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_479 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_479 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_479 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_479 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_479 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_479 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_479 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_479 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_100 state_indent')
action_479 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_101 state_indent')
action_479 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_479 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_479 state_indent (55, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_548 state_indent')
action_479 state_indent (56, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_105 state_indent')
action_479 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_479 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_479 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_479 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_479 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_479 state_indent tok = 
	 happyFail

action_480 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_594 state_indent'
action_480 state_indent tok = 
	 happyFail

action_481 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_217 state_indent'
action_481 state_indent tok = 
	 happyFail

action_482 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_595 state_indent'
action_482 state_indent tok = 
	 happyFail

action_483 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_284 state_indent'
action_483 state_indent tok = 
	 happyFail

action_484 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_484 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_484 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_484 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_484 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_484 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_484 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_484 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_484 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_484 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_484 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_484 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_484 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_484 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_484 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_484 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_484 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_484 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_484 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_484 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_484 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_484 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_484 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_484 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_484 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_547 state_indent')
action_484 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_484 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_484 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_484 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_484 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_484 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_484 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_484 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_484 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_484 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_484 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_484 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_484 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_484 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_484 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_484 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_484 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_484 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_484 state_indent tok = 
	 happyFail

action_485 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_485 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_485 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_248 state_indent')
action_485 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_485 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_485 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_485 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_485 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_485 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_485 state_indent (120, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_546 state_indent')
action_485 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_294 state_indent')
action_485 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_485 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_485 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_485 state_indent tok = 
	 happyFail

action_486 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_572 state_indent'
action_486 state_indent tok = 
	 happyFail

action_487 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_545 state_indent')
action_487 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_487 state_indent tok = 
	 happyFail

action_488 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_544 state_indent')
action_488 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_488 state_indent tok = 
	 happyFail

action_489 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_543 state_indent')
action_489 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_489 state_indent tok = 
	 happyFail

action_490 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_490 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_490 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_285 state_indent')
action_490 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_490 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_490 state_indent (126, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_280 state_indent')
action_490 state_indent (129, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_281 state_indent')
action_490 state_indent (131, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_542 state_indent')
action_490 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_283 state_indent')
action_490 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_284 state_indent')
action_490 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_490 state_indent tok = 
	 happyFail

action_491 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_541 state_indent')
action_491 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_491 state_indent tok = 
	 happyFail

action_492 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I1],[=I1])([I2],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_173 state_indent'
action_492 state_indent tok = 
	 happyFail

action_493 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_288 state_indent'
action_493 state_indent tok = 
	 happyFail

action_494 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_282 state_indent'
action_494 state_indent tok = 
	 happyFail

action_495 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_540 state_indent')
action_495 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_495 state_indent tok = 
	 happyFail

action_496 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_539 state_indent')
action_496 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_496 state_indent tok = 
	 happyFail

action_497 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_162 state_indent'
action_497 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_162 state_indent'
action_497 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_374 state_indent')
action_497 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_162 state_indent'
action_497 state_indent (73, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_538 state_indent')
action_497 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_162 state_indent'
action_497 state_indent tok = 
	 happyFail

action_498 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_23 state_indent')
action_498 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_24 state_indent')
action_498 state_indent (148, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_537 state_indent')
action_498 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_498 state_indent tok = 
	 happyFail

action_499 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_536 state_indent')
action_499 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_499 state_indent tok = 
	 happyFail

action_500 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_535 state_indent')
action_500 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_500 state_indent tok = 
	 happyFail

action_501 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_501 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_501 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_501 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_501 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_501 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_501 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_501 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_501 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_501 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_501 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_501 state_indent (37, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_534 state_indent')
action_501 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_347 state_indent')
action_501 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_501 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_501 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_501 state_indent (57, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_349 state_indent')
action_501 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_501 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_501 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_501 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_501 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_81 state_indent'
action_501 state_indent tok = 
	 happyFail

action_502 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_533 state_indent')
action_502 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_502 state_indent tok = 
	 happyFail

action_503 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_158 state_indent'
action_503 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_158 state_indent'
action_503 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_344 state_indent')
action_503 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_158 state_indent'
action_503 state_indent (72, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_532 state_indent')
action_503 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_158 state_indent'
action_503 state_indent tok = 
	 happyFail

action_504 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_277 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent (80, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_531 state_indent')
action_504 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_504 state_indent tok = 
	 happyFail

action_505 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_505 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_505 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_505 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_505 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_505 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_505 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_505 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_505 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_505 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_505 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_505 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_505 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_505 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_505 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_505 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_505 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_505 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_505 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_505 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_505 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_505 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_443 state_indent')
action_505 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_505 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_505 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_505 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_505 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_505 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_505 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_505 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_505 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_505 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_505 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_505 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_505 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_505 state_indent tok = 
	 happyFail

action_506 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_506 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_289 state_indent')
action_506 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_506 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_506 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_506 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_506 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_506 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_506 state_indent (122, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_530 state_indent')
action_506 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_288 state_indent')
action_506 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_506 state_indent tok = 
	 happyFail

action_507 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_507 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_507 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_507 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_507 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_507 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_507 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_507 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_507 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_507 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_507 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_507 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_100 state_indent')
action_507 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_101 state_indent')
action_507 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_507 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_507 state_indent (55, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_529 state_indent')
action_507 state_indent (56, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_105 state_indent')
action_507 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_507 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_507 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_507 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_507 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_507 state_indent tok = 
	 happyFail

action_508 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_508 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_508 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_285 state_indent')
action_508 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_508 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_508 state_indent (31, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_528 state_indent')
action_508 state_indent (126, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_280 state_indent')
action_508 state_indent (129, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_281 state_indent')
action_508 state_indent (131, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_282 state_indent')
action_508 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_283 state_indent')
action_508 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_284 state_indent')
action_508 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_508 state_indent tok = 
	 happyFail

action_509 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_33 state_indent'
action_509 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_33 state_indent'
action_509 state_indent tok = 
	 happyFail

action_510 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_6 state_indent'
action_510 state_indent tok = 
	 happyFail

action_511 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_69 state_indent'
action_511 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_69 state_indent'
action_511 state_indent tok = 
	 happyFail

action_512 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_1 state_indent'
action_512 state_indent tok = 
	 happyFail

action_513 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_30 state_indent'
action_513 state_indent tok = 
	 happyFail

action_514 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_514 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_514 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_525 state_indent')
action_514 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_526 state_indent')
action_514 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_527 state_indent')
action_514 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_514 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_514 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_514 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_514 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_514 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_514 state_indent (26, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_520 state_indent')
action_514 state_indent (27, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_521 state_indent')
action_514 state_indent (122, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_522 state_indent')
action_514 state_indent (124, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_523 state_indent')
action_514 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_288 state_indent')
action_514 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_524 state_indent')
action_514 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_514 state_indent tok = 
	 happyFail

action_515 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_519 state_indent')
action_515 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_515 state_indent tok = 
	 happyFail

action_516 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_516 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_516 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_516 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_516 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_248 state_indent')
action_516 state_indent (365, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_250 state_indent')
action_516 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_516 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_516 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_516 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_516 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_516 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_516 state_indent (15, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_518 state_indent')
action_516 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_245 state_indent')
action_516 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_516 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_516 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_246 state_indent')
action_516 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_516 state_indent (151, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_247 state_indent')
action_516 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_21 state_indent'
action_516 state_indent tok = 
	 happyFail

action_517 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_20 state_indent'
action_517 state_indent tok = 
	 happyFail

action_518 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_23 state_indent'
action_518 state_indent tok = 
	 happyFail

action_519 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_19 state_indent'
action_519 state_indent tok = 
	 happyFail

action_520 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_667 state_indent')
action_520 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_668 state_indent')
action_520 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_520 state_indent tok = 
	 happyFail

action_521 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_54 state_indent'
action_521 state_indent tok = 
	 happyFail

action_522 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_55 state_indent'
action_522 state_indent tok = 
	 happyFail

action_523 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_56 state_indent'
action_523 state_indent tok = 
	 happyFail

action_524 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_275 state_indent'
action_524 state_indent tok = 
	 happyFail

action_525 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_525 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_525 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_525 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_525 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_615 state_indent')
action_525 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_487 state_indent')
action_525 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_525 state_indent tok = 
	 happyFail

action_526 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_28 state_indent'
action_526 state_indent tok = 
	 happyFail

action_527 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_666 state_indent')
action_527 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_527 state_indent tok = 
	 happyFail

action_528 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_490 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_57 state_indent'
action_528 state_indent tok = 
	 happyFail

action_529 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_97 state_indent'
action_529 state_indent tok = 
	 happyFail

action_530 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_98 state_indent'
action_530 state_indent tok = 
	 happyFail

action_531 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_172 state_indent'
action_531 state_indent tok = 
	 happyFail

action_532 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_74 state_indent'
action_532 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_74 state_indent'
action_532 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_74 state_indent'
action_532 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_74 state_indent'
action_532 state_indent tok = 
	 happyFail

action_533 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_391 state_indent')
action_533 state_indent (60, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_665 state_indent')
action_533 state_indent (61, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_433 state_indent')
action_533 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_533 state_indent tok = 
	 happyFail

action_534 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_664 state_indent')
action_534 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_534 state_indent tok = 
	 happyFail

action_535 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_413 state_indent')
action_535 state_indent (49, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_663 state_indent')
action_535 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_106 state_indent'
action_535 state_indent tok = 
	 happyFail

action_536 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_410 state_indent')
action_536 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_411 state_indent')
action_536 state_indent (48, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_662 state_indent')
action_536 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_104 state_indent'
action_536 state_indent tok = 
	 happyFail

action_537 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_39 state_indent'
action_537 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_39 state_indent'
action_537 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_39 state_indent'
action_537 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_408 state_indent')
action_537 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_39 state_indent'
action_537 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_39 state_indent'
action_537 state_indent (20, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_661 state_indent')
action_537 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_39 state_indent'
action_537 state_indent tok = 
	 happyFail

action_538 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_75 state_indent'
action_538 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_75 state_indent'
action_538 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_75 state_indent'
action_538 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_75 state_indent'
action_538 state_indent tok = 
	 happyFail

action_539 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_391 state_indent')
action_539 state_indent (61, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_660 state_indent')
action_539 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_539 state_indent tok = 
	 happyFail

action_540 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_540 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_540 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_540 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_540 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_540 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_540 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_540 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_540 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_540 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_540 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_540 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_659 state_indent')
action_540 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_540 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_540 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_540 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_540 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_540 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_540 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_540 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_540 state_indent tok = 
	 happyFail

action_541 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_541 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_541 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_541 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_541 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_541 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_541 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_541 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_541 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_541 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_541 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_541 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_541 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_541 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_541 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_541 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_541 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_541 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_541 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_541 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_541 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_541 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_541 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_541 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_541 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_658 state_indent')
action_541 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_541 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_541 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_541 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_541 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_541 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_541 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_541 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_541 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_541 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_541 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_541 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_541 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_541 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_541 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_541 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_541 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_541 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_541 state_indent tok = 
	 happyFail

action_542 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_63 state_indent'
action_542 state_indent tok = 
	 happyFail

action_543 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_286 state_indent'
action_543 state_indent tok = 
	 happyFail

action_544 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_280 state_indent'
action_544 state_indent tok = 
	 happyFail

action_545 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_272 state_indent'
action_545 state_indent tok = 
	 happyFail

action_546 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_264 state_indent'
action_546 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_264 state_indent'
action_546 state_indent tok = 
	 happyFail

action_547 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_266 state_indent'
action_547 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_266 state_indent'
action_547 state_indent tok = 
	 happyFail

action_548 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_180 state_indent'
action_548 state_indent tok = 
	 happyFail

action_549 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_549 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_549 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_549 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_549 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_549 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_549 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_549 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_549 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_549 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_549 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_549 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_549 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_549 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_549 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_549 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_549 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_549 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_549 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_549 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_549 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_549 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_549 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_549 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_549 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_657 state_indent')
action_549 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_549 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_549 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_549 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_549 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_549 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_549 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_549 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_549 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_549 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_549 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_549 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_549 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_549 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_549 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_549 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_549 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_549 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_222 state_indent'
action_549 state_indent tok = 
	 happyFail

action_550 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_469 state_indent')
action_550 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_232 state_indent'
action_550 state_indent tok = 
	 happyFail

action_551 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_656 state_indent')
action_551 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_551 state_indent tok = 
	 happyFail

action_552 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_552 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_552 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_552 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_552 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_552 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_552 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_552 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_552 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_552 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_552 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_552 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_552 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_552 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_552 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_552 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_552 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_552 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_474 state_indent')
action_552 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_552 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_552 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_552 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_552 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_552 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_552 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_470 state_indent')
action_552 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_552 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_552 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_354 state_indent')
action_552 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_552 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_552 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_552 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_552 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_552 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_552 state_indent (101, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_655 state_indent')
action_552 state_indent (111, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_473 state_indent')
action_552 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_552 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_552 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_552 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_552 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_552 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_552 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_552 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_552 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_552 state_indent tok = 
	 happyFail

action_553 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_189 state_indent'
action_553 state_indent tok = 
	 happyFail

action_554 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_554 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_464 state_indent')
action_554 state_indent (28, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_89 state_indent')
action_554 state_indent (40, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_654 state_indent')
action_554 state_indent (45, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_92 state_indent')
action_554 state_indent (79, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_94 state_indent')
action_554 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_554 state_indent tok = 
	 happyFail

action_555 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_341 state_indent'
action_555 state_indent tok = 
	 happyFail

action_556 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_556 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_464 state_indent')
action_556 state_indent (28, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_89 state_indent')
action_556 state_indent (40, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_653 state_indent')
action_556 state_indent (45, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_92 state_indent')
action_556 state_indent (79, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_94 state_indent')
action_556 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_556 state_indent tok = 
	 happyFail

action_557 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(2, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(3, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([I3],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_342 state_indent'
action_557 state_indent tok = 
	 happyFail

action_558 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I1],[=I1])([I2],[I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_89 state_indent'
action_558 state_indent tok = 
	 happyFail

action_559 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_559 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_554 state_indent')
action_559 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_652 state_indent')
action_559 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_559 state_indent tok = 
	 happyFail

action_560 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_651 state_indent')
action_560 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_560 state_indent tok = 
	 happyFail

action_561 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_191 state_indent'
action_561 state_indent tok = 
	 happyFail

action_562 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_197 state_indent'
action_562 state_indent tok = 
	 happyFail

action_563 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_563 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_563 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_563 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_563 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_563 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_563 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_563 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_563 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_563 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_563 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_563 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_563 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_563 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_563 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_563 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_563 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_563 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_563 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_563 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_563 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_563 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_563 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_563 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_563 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_650 state_indent')
action_563 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_563 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_563 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_563 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_563 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_563 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_563 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_563 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_563 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_563 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_563 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_563 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_563 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_563 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_563 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_563 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_563 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_563 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_563 state_indent tok = 
	 happyFail

action_564 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_212 state_indent'
action_564 state_indent tok = 
	 happyFail

action_565 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_213 state_indent'
action_565 state_indent tok = 
	 happyFail

action_566 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [((minRanges token_indent) + 1, Inf)])) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[=])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_204 state_indent'
action_566 state_indent tok = 
	 happyFail

action_567 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_567 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_640 state_indent')
action_567 state_indent (106, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_648 state_indent')
action_567 state_indent (107, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_649 state_indent')
action_567 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_567 state_indent tok = 
	 happyFail

action_568 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_238 state_indent')
action_568 state_indent (147, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_647 state_indent')
action_568 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_568 state_indent tok = 
	 happyFail

action_569 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_569 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_569 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_201 state_indent')
action_569 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_202 state_indent')
action_569 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_227 state_indent')
action_569 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_206 state_indent')
action_569 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_569 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_569 state_indent (127, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_220 state_indent')
action_569 state_indent (130, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_221 state_indent')
action_569 state_indent (132, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_222 state_indent')
action_569 state_indent (134, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_223 state_indent')
action_569 state_indent (139, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_192 state_indent')
action_569 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_193 state_indent')
action_569 state_indent (141, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_224 state_indent')
action_569 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_196 state_indent')
action_569 state_indent (145, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_225 state_indent')
action_569 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_418 state_indent'
action_569 state_indent tok = 
	 happyFail

action_570 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_646 state_indent')
action_570 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_642 state_indent')
action_570 state_indent (108, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_643 state_indent')
action_570 state_indent (109, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_644 state_indent')
action_570 state_indent (110, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_645 state_indent')
action_570 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_570 state_indent tok = 
	 happyFail

action_571 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_571 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_640 state_indent')
action_571 state_indent (107, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_641 state_indent')
action_571 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_571 state_indent tok = 
	 happyFail

action_572 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_572 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_572 state_indent tok = 
	 happyFail

action_573 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_571 state_indent')
action_573 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_572 state_indent')
action_573 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_236 state_indent'
action_573 state_indent tok = 
	 happyFail

action_574 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_363 state_indent'
action_574 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_363 state_indent'
action_574 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_363 state_indent'
action_574 state_indent tok = 
	 happyFail

action_575 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_391 state_indent')
action_575 state_indent (61, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_639 state_indent')
action_575 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_575 state_indent tok = 
	 happyFail

action_576 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_576 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_576 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_638 state_indent')
action_576 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_636 state_indent')
action_576 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_576 state_indent (152, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_637 state_indent')
action_576 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_576 state_indent tok = 
	 happyFail

action_577 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_503 state_indent'
action_577 state_indent tok = 
	 happyFail

action_578 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_635 state_indent')
action_578 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_578 state_indent tok = 
	 happyFail

action_579 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_579 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_579 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_579 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_579 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_579 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_579 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_579 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_579 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_579 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_579 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_579 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_579 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_579 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_579 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_579 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_579 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_579 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_579 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_579 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_579 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_579 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_579 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_579 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_579 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_634 state_indent')
action_579 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_579 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_579 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_579 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_579 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_579 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_579 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_579 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_579 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_579 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_579 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_579 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_579 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_579 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_579 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_579 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_579 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_579 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_579 state_indent tok = 
	 happyFail

action_580 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_580 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_580 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_580 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_580 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_580 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_580 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_580 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_580 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_580 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_580 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_580 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_580 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_580 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_580 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_580 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_580 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_580 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_580 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_580 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_580 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_580 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_580 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_580 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_580 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_633 state_indent')
action_580 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_580 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_580 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_580 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_580 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_580 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_580 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_580 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_580 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_580 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_580 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_580 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_580 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_580 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_580 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_580 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_580 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_580 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_580 state_indent tok = 
	 happyFail

action_581 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_256 state_indent'
action_581 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_256 state_indent'
action_581 state_indent tok = 
	 happyFail

action_582 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(1, Inf)] {-{([I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_253 state_indent'
action_582 state_indent tok = 
	 happyFail

action_583 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_378 state_indent'
action_583 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_378 state_indent'
action_583 state_indent tok = 
	 happyFail

action_584 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_374 state_indent'
action_584 state_indent tok = 
	 happyFail

action_585 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_632 state_indent')
action_585 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_585 state_indent tok = 
	 happyFail

action_586 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_107 state_indent'
action_586 state_indent tok = 
	 happyFail

action_587 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_587 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_587 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_587 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_631 state_indent')
action_587 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_587 state_indent tok = 
	 happyFail

action_588 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_586 state_indent')
action_588 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_587 state_indent')
action_588 state_indent (50, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_630 state_indent')
action_588 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_588 state_indent tok = 
	 happyFail

action_589 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_38 state_indent'
action_589 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_38 state_indent'
action_589 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_38 state_indent'
action_589 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_38 state_indent'
action_589 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_38 state_indent'
action_589 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_38 state_indent'
action_589 state_indent tok = 
	 happyFail

action_590 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_359 state_indent'
action_590 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_359 state_indent'
action_590 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_359 state_indent'
action_590 state_indent tok = 
	 happyFail

action_591 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_40 state_indent'
action_591 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_40 state_indent'
action_591 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_40 state_indent'
action_591 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_40 state_indent'
action_591 state_indent tok = 
	 happyFail

action_592 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_629 state_indent')
action_592 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_592 state_indent tok = 
	 happyFail

action_593 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_44 state_indent'
action_593 state_indent tok = 
	 happyFail

action_594 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_594 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_164 state_indent'
action_594 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_599 state_indent')
action_594 state_indent (75, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_628 state_indent')
action_594 state_indent (79, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_601 state_indent')
action_594 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_594 state_indent tok = 
	 happyFail

action_595 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_238 state_indent')
action_595 state_indent (147, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_627 state_indent')
action_595 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_595 state_indent tok = 
	 happyFail

action_596 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges []) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[I1])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_625 state_indent')
action_596 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_626 state_indent')
action_596 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_596 state_indent tok = 
	 happyFail

action_597 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges []) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[I1])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_373 state_indent'
action_597 state_indent tok = 
	 happyFail

action_598 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_44 state_indent')
action_598 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_45 state_indent')
action_598 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_46 state_indent')
action_598 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_47 state_indent')
action_598 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_48 state_indent')
action_598 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_49 state_indent')
action_598 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_50 state_indent')
action_598 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_51 state_indent')
action_598 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_52 state_indent')
action_598 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_53 state_indent')
action_598 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_54 state_indent')
action_598 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_55 state_indent')
action_598 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_56 state_indent')
action_598 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_57 state_indent')
action_598 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_61 state_indent')
action_598 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_70 state_indent')
action_598 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_71 state_indent')
action_598 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_72 state_indent')
action_598 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_73 state_indent')
action_598 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_74 state_indent')
action_598 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_75 state_indent')
action_598 state_indent (191, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_28 state_indent')
action_598 state_indent (215, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_29 state_indent')
action_598 state_indent (237, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_32 state_indent')
action_598 state_indent (243, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_33 state_indent')
action_598 state_indent (244, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_34 state_indent')
action_598 state_indent (254, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_35 state_indent')
action_598 state_indent (259, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_137 state_indent')
action_598 state_indent (262, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_37 state_indent')
action_598 state_indent (263, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_38 state_indent')
action_598 state_indent (268, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_39 state_indent')
action_598 state_indent (277, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_40 state_indent')
action_598 state_indent (282, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_41 state_indent')
action_598 state_indent (284, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_42 state_indent')
action_598 state_indent (288, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_43 state_indent')
action_598 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_598 state_indent tok = 
	 happyFail

action_599 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_599 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_599 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_599 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_599 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_599 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_599 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_599 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_599 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_599 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_599 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_599 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_599 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_599 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_599 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_599 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_599 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_599 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_599 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_599 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_599 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_599 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_258 state_indent')
action_599 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_599 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_599 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_599 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_599 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_599 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_599 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_599 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_599 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_599 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_599 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_599 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_599 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_599 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_599 state_indent tok = 
	 happyFail

action_600 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges []) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[I1])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_623 state_indent')
action_600 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_624 state_indent')
action_600 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_600 state_indent tok = 
	 happyFail

action_601 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges []) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[I1])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_166 state_indent'
action_601 state_indent tok = 
	 happyFail

action_602 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges []) `unionRanges` (token_indent) {-{([I1],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_170 state_indent'
action_602 state_indent tok = 
	 happyFail

action_603 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_127 state_indent'
action_603 state_indent tok = 
	 happyFail

action_604 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_126 state_indent'
action_604 state_indent tok = 
	 happyFail

action_605 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_364 state_indent'
action_605 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_364 state_indent'
action_605 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_364 state_indent'
action_605 state_indent tok = 
	 happyFail

action_606 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_606 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_606 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_606 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_143 state_indent'
action_606 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_606 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_137 state_indent'
action_606 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_137 state_indent'
action_606 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_606 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_143 state_indent'
action_606 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_137 state_indent'
action_606 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_622 state_indent')
action_606 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_137 state_indent'
action_606 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_606 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_606 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_606 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_606 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_606 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_606 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_137 state_indent'
action_606 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_384 state_indent')
action_606 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_606 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_606 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_606 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_606 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_606 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_137 state_indent'
action_606 state_indent tok = 
	 happyFail

action_607 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_133 state_indent'
action_607 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_133 state_indent'
action_607 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_133 state_indent'
action_607 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_133 state_indent'
action_607 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_133 state_indent'
action_607 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_133 state_indent'
action_607 state_indent tok = 
	 happyFail

action_608 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_608 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_608 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_608 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_608 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_138 state_indent'
action_608 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_138 state_indent'
action_608 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_608 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_138 state_indent'
action_608 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_621 state_indent')
action_608 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_138 state_indent'
action_608 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_608 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_608 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_608 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_608 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_608 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_608 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_138 state_indent'
action_608 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_619 state_indent')
action_608 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_608 state_indent (64, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_620 state_indent')
action_608 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_608 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_608 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_608 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_608 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_138 state_indent'
action_608 state_indent tok = 
	 happyFail

action_609 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_609 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_618 state_indent')
action_609 state_indent (129, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_617 state_indent')
action_609 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_283 state_indent')
action_609 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_609 state_indent tok = 
	 happyFail

action_610 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_616 state_indent')
action_610 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_610 state_indent tok = 
	 happyFail

action_611 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (328, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_275 state_indent'
action_611 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_306 state_indent'
action_611 state_indent tok = 
	 happyFail

action_612 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_612 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_612 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_612 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_612 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_612 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_380 state_indent')
action_612 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_612 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_204 state_indent')
action_612 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_381 state_indent')
action_612 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_612 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_612 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_612 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_612 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_612 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_612 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_377 state_indent')
action_612 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_612 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_612 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_612 state_indent (57, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_378 state_indent')
action_612 state_indent (96, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_379 state_indent')
action_612 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_612 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_612 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_612 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_615 state_indent')
action_612 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_612 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_612 state_indent tok = 
	 happyFail

action_613 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_613 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_613 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_613 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_613 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_613 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_613 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_613 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_613 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_613 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_613 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_613 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_614 state_indent')
action_613 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_613 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_613 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_613 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_613 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_613 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_613 state_indent tok = 
	 happyFail

action_614 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_144 state_indent'
action_614 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_144 state_indent'
action_614 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_144 state_indent'
action_614 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_144 state_indent'
action_614 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_144 state_indent'
action_614 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_144 state_indent'
action_614 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_144 state_indent'
action_614 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_144 state_indent'
action_614 state_indent tok = 
	 happyFail

action_615 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_711 state_indent')
action_615 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_615 state_indent tok = 
	 happyFail

action_616 state_indent (313, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_616 state_indent (314, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_616 state_indent (325, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_248 state_indent')
action_616 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_238 state_indent')
action_616 state_indent (371, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_616 state_indent (372, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_616 state_indent (373, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_616 state_indent (374, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_616 state_indent (375, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_616 state_indent (376, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_616 state_indent (46, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_705 state_indent')
action_616 state_indent (66, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_706 state_indent')
action_616 state_indent (67, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_707 state_indent')
action_616 state_indent (68, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_708 state_indent')
action_616 state_indent (123, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_709 state_indent')
action_616 state_indent (135, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_616 state_indent (136, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_616 state_indent (147, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_710 state_indent')
action_616 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_616 state_indent tok = 
	 happyFail

action_617 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_617 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_617 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_617 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_617 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_617 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_613 state_indent')
action_617 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_617 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_617 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_617 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_617 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_617 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_617 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_703 state_indent')
action_617 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_617 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_617 state_indent (65, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_704 state_indent')
action_617 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_617 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_617 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_617 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_617 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_617 state_indent tok = 
	 happyFail

action_618 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_618 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_489 state_indent')
action_618 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_618 state_indent tok = 
	 happyFail

action_619 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_141 state_indent'
action_619 state_indent tok = 
	 happyFail

action_620 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_140 state_indent'
action_620 state_indent tok = 
	 happyFail

action_621 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_621 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_621 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_621 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_621 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_621 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_621 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_621 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_621 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_621 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_621 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_621 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_702 state_indent')
action_621 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_621 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_621 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_621 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_621 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_621 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_621 state_indent tok = 
	 happyFail

action_622 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_622 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_622 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_622 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_622 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_622 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_622 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_622 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_622 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_622 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_622 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_622 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_701 state_indent')
action_622 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_622 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_622 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_622 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_622 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_622 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_622 state_indent tok = 
	 happyFail

action_623 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges []) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[I1])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_623 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_599 state_indent')
action_623 state_indent (79, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_700 state_indent')
action_623 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_623 state_indent tok = 
	 happyFail

action_624 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges []) `unionRanges` (token_indent) {-{([I1],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_344 state_indent'
action_624 state_indent tok = 
	 happyFail

action_625 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges []) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[I1])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_625 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_599 state_indent')
action_625 state_indent (79, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_699 state_indent')
action_625 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_625 state_indent tok = 
	 happyFail

action_626 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges []) `unionRanges` (token_indent) {-{([I1],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_345 state_indent'
action_626 state_indent tok = 
	 happyFail

action_627 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_159 state_indent'
action_627 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_159 state_indent'
action_627 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_159 state_indent'
action_627 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_159 state_indent'
action_627 state_indent tok = 
	 happyFail

action_628 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_628 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_623 state_indent')
action_628 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_698 state_indent')
action_628 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_628 state_indent tok = 
	 happyFail

action_629 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_629 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_629 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_289 state_indent')
action_629 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_249 state_indent')
action_629 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_629 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_629 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_629 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_629 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_629 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_629 state_indent (13, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_692 state_indent')
action_629 state_indent (24, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_693 state_indent')
action_629 state_indent (25, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_694 state_indent')
action_629 state_indent (122, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_695 state_indent')
action_629 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_288 state_indent')
action_629 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_696 state_indent')
action_629 state_indent (149, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_697 state_indent')
action_629 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_22 state_indent'
action_629 state_indent tok = 
	 happyFail

action_630 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_691 state_indent')
action_630 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_630 state_indent tok = 
	 happyFail

action_631 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_690 state_indent')
action_631 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_631 state_indent tok = 
	 happyFail

action_632 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_632 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_632 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_632 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_632 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_632 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_632 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_632 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_632 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_632 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_632 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_632 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_689 state_indent')
action_632 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_632 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_632 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_632 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_632 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_632 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_632 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_632 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_632 state_indent tok = 
	 happyFail

action_633 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_261 state_indent'
action_633 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_261 state_indent'
action_633 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_261 state_indent'
action_633 state_indent tok = 
	 happyFail

action_634 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_502 state_indent'
action_634 state_indent tok = 
	 happyFail

action_635 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_635 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_635 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_635 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_635 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_635 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_635 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_635 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_635 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_635 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_635 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_635 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_635 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_635 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_635 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_635 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_635 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_635 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_635 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_635 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_635 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_635 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_635 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_635 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_635 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_688 state_indent')
action_635 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_635 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_635 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_635 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_635 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_635 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_635 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_635 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_635 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_635 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_635 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_635 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_635 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_635 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_635 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_635 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_635 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_635 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_635 state_indent tok = 
	 happyFail

action_636 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_332 state_indent'
action_636 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_332 state_indent'
action_636 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_332 state_indent'
action_636 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_332 state_indent'
action_636 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_332 state_indent'
action_636 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_332 state_indent'
action_636 state_indent tok = 
	 happyFail

action_637 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_152 state_indent'
action_637 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_152 state_indent'
action_637 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_152 state_indent'
action_637 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_152 state_indent'
action_637 state_indent tok = 
	 happyFail

action_638 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_638 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_638 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_687 state_indent')
action_638 state_indent (71, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_685 state_indent')
action_638 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_636 state_indent')
action_638 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_638 state_indent (152, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_686 state_indent')
action_638 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_638 state_indent tok = 
	 happyFail

action_639 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_131 state_indent'
action_639 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_131 state_indent'
action_639 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_131 state_indent'
action_639 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_131 state_indent'
action_639 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_131 state_indent'
action_639 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_131 state_indent'
action_639 state_indent tok = 
	 happyFail

action_640 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_640 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_640 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_640 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_640 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_640 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_640 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_640 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_640 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_640 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_640 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_640 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_640 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_640 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_640 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_640 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_640 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_640 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_640 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_640 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_640 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_640 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_683 state_indent')
action_640 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_640 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_640 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_640 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_640 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_640 state_indent (111, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_684 state_indent')
action_640 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_640 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_640 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_640 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_640 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_640 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_640 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_640 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_640 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_640 state_indent tok = 
	 happyFail

action_641 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges []) {-{([I1],[=I1])([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_347 state_indent'
action_641 state_indent tok = 
	 happyFail

action_642 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_682 state_indent')
action_642 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_642 state_indent tok = 
	 happyFail

action_643 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges []) {-{([I1],[=I1])([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_277 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent (80, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_681 state_indent')
action_643 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_643 state_indent tok = 
	 happyFail

action_644 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges []) {-{([I1],[=I1])([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_642 state_indent')
action_644 state_indent (110, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_680 state_indent')
action_644 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_243 state_indent'
action_644 state_indent tok = 
	 happyFail

action_645 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges []) {-{([I1],[=I1])([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_245 state_indent'
action_645 state_indent tok = 
	 happyFail

action_646 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_646 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_646 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_646 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_646 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_646 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_646 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_646 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_646 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_646 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_646 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_646 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_646 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_646 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_646 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_646 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_646 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_646 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_646 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_646 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_646 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_646 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_646 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_646 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_646 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_679 state_indent')
action_646 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_646 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_646 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_646 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_646 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_646 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_646 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_646 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_646 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_646 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_646 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_646 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_646 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_646 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_646 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_646 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_646 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_646 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_646 state_indent tok = 
	 happyFail

action_647 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_233 state_indent'
action_647 state_indent tok = 
	 happyFail

action_648 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_648 state_indent (9, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_677 state_indent')
action_648 state_indent (10, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_678 state_indent')
action_648 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_16 state_indent'
action_648 state_indent tok = 
	 happyFail

action_649 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_240 state_indent'
action_649 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_240 state_indent'
action_649 state_indent tok = 
	 happyFail

action_650 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_188 state_indent'
action_650 state_indent tok = 
	 happyFail

action_651 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_651 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_651 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_651 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_651 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_651 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_651 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_651 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_651 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_651 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_651 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_651 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_651 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_651 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_651 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_651 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_651 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_651 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_651 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_651 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_651 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_651 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_651 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_651 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_651 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_676 state_indent')
action_651 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_651 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_651 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_651 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_651 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_651 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_651 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_651 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_651 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_651 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_651 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_651 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_651 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_651 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_651 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_651 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_651 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_651 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_651 state_indent tok = 
	 happyFail

action_652 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_652 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_82 state_indent'
action_652 state_indent tok = 
	 happyFail

action_653 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_370 state_indent'
action_653 state_indent tok = 
	 happyFail

action_654 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_84 state_indent'
action_654 state_indent tok = 
	 happyFail

action_655 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_228 state_indent'
action_655 state_indent tok = 
	 happyFail

action_656 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_656 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_656 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_656 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_656 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_656 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_656 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_656 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_656 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_656 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_656 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_656 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_656 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_656 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_656 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_656 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_656 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_656 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_656 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_656 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_656 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_656 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_656 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_656 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_656 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_675 state_indent')
action_656 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_656 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_656 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_656 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_656 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_656 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_656 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_656 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_656 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_656 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_656 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_656 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_656 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_656 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_656 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_656 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_656 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_656 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_656 state_indent tok = 
	 happyFail

action_657 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_224 state_indent'
action_657 state_indent tok = 
	 happyFail

action_658 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(2, Inf)]) `unionRanges` (Ranges [(minRanges token_indent, Inf)]) {-{([I2],[=I1])([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)]) {-{([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_179 state_indent'
action_658 state_indent tok = 
	 happyFail

action_659 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_71 state_indent'
action_659 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_71 state_indent'
action_659 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_71 state_indent'
action_659 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_71 state_indent'
action_659 state_indent tok = 
	 happyFail

action_660 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_660 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_660 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_576 state_indent')
action_660 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_660 state_indent (70, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_674 state_indent')
action_660 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_660 state_indent tok = 
	 happyFail

action_661 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_45 state_indent'
action_661 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_41 state_indent'
action_661 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_41 state_indent'
action_661 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_593 state_indent')
action_661 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_41 state_indent'
action_661 state_indent (21, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_673 state_indent')
action_661 state_indent (22, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_591 state_indent')
action_661 state_indent (23, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_592 state_indent')
action_661 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_41 state_indent'
action_661 state_indent tok = 
	 happyFail

action_662 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_413 state_indent')
action_662 state_indent (49, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_672 state_indent')
action_662 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_106 state_indent'
action_662 state_indent tok = 
	 happyFail

action_663 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_586 state_indent')
action_663 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_587 state_indent')
action_663 state_indent (50, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_671 state_indent')
action_663 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_663 state_indent tok = 
	 happyFail

action_664 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_76 state_indent'
action_664 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_76 state_indent'
action_664 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_76 state_indent'
action_664 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_76 state_indent'
action_664 state_indent tok = 
	 happyFail

action_665 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_665 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_665 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_575 state_indent')
action_665 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_576 state_indent')
action_665 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_665 state_indent (70, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_670 state_indent')
action_665 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_151 state_indent'
action_665 state_indent tok = 
	 happyFail

action_666 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_27 state_indent'
action_666 state_indent tok = 
	 happyFail

action_667 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_29 state_indent'
action_667 state_indent tok = 
	 happyFail

action_668 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_668 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_668 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_525 state_indent')
action_668 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_668 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_668 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_668 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_668 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_668 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_668 state_indent (27, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_669 state_indent')
action_668 state_indent (122, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_522 state_indent')
action_668 state_indent (124, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_523 state_indent')
action_668 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_288 state_indent')
action_668 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_524 state_indent')
action_668 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_668 state_indent tok = 
	 happyFail

action_669 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_53 state_indent'
action_669 state_indent tok = 
	 happyFail

action_670 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_72 state_indent'
action_670 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_72 state_indent'
action_670 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_72 state_indent'
action_670 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_72 state_indent'
action_670 state_indent tok = 
	 happyFail

action_671 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_726 state_indent')
action_671 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_671 state_indent tok = 
	 happyFail

action_672 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_586 state_indent')
action_672 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_587 state_indent')
action_672 state_indent (50, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_725 state_indent')
action_672 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_672 state_indent tok = 
	 happyFail

action_673 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_35 state_indent'
action_673 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_35 state_indent'
action_673 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_35 state_indent'
action_673 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_35 state_indent'
action_673 state_indent tok = 
	 happyFail

action_674 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_73 state_indent'
action_674 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_73 state_indent'
action_674 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_73 state_indent'
action_674 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_73 state_indent'
action_674 state_indent tok = 
	 happyFail

action_675 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_230 state_indent'
action_675 state_indent tok = 
	 happyFail

action_676 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_190 state_indent'
action_676 state_indent tok = 
	 happyFail

action_677 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_677 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_15 state_indent'
action_677 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_640 state_indent')
action_677 state_indent (107, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_724 state_indent')
action_677 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_677 state_indent tok = 
	 happyFail

action_678 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_678 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_238 state_indent'
action_678 state_indent tok = 
	 happyFail

action_679 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` ((case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)])) {-{([I1],[=I1])([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_242 state_indent'
action_679 state_indent tok = 
	 happyFail

action_680 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` ((case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)])) {-{([I1],[=I1])([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_244 state_indent'
action_680 state_indent tok = 
	 happyFail

action_681 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` ((case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)])) {-{([I1],[=I1])([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_376 state_indent'
action_681 state_indent tok = 
	 happyFail

action_682 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_682 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_682 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_682 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_682 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_682 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_682 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_682 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_682 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_682 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_682 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_682 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_682 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_682 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_682 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_682 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_682 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_682 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_682 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_682 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_682 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_682 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_682 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_682 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_682 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_723 state_indent')
action_682 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_682 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_301 state_indent')
action_682 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_682 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_682 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_682 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_682 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_682 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_682 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_682 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_682 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_682 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_682 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_682 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_682 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_682 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_682 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_682 state_indent tok = 
	 happyFail

action_683 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_226 state_indent')
action_683 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_200 state_indent')
action_683 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_201 state_indent')
action_683 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_202 state_indent')
action_683 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_227 state_indent')
action_683 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_206 state_indent')
action_683 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_229 state_indent')
action_683 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_230 state_indent')
action_683 state_indent (127, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_220 state_indent')
action_683 state_indent (130, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_221 state_indent')
action_683 state_indent (132, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_505 state_indent')
action_683 state_indent (134, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_223 state_indent')
action_683 state_indent (139, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_192 state_indent')
action_683 state_indent (140, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_193 state_indent')
action_683 state_indent (141, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_224 state_indent')
action_683 state_indent (143, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_196 state_indent')
action_683 state_indent (145, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_225 state_indent')
action_683 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_247 state_indent'
action_683 state_indent tok = 
	 happyFail

action_684 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_646 state_indent')
action_684 state_indent (5, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_642 state_indent')
action_684 state_indent (108, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_722 state_indent')
action_684 state_indent (109, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_644 state_indent')
action_684 state_indent (110, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_645 state_indent')
action_684 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_3 state_indent'
action_684 state_indent tok = 
	 happyFail

action_685 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_720 state_indent')
action_685 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_721 state_indent')
action_685 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_685 state_indent tok = 
	 happyFail

action_686 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_156 state_indent'
action_686 state_indent tok = 
	 happyFail

action_687 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_153 state_indent'
action_687 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_153 state_indent'
action_687 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_153 state_indent'
action_687 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_153 state_indent'
action_687 state_indent tok = 
	 happyFail

action_688 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0, Fin 0),(1, Inf)] {-{([=I1],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_504 state_indent'
action_688 state_indent tok = 
	 happyFail

action_689 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_391 state_indent'
action_689 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_391 state_indent'
action_689 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_391 state_indent'
action_689 state_indent tok = 
	 happyFail

action_690 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_108 state_indent'
action_690 state_indent tok = 
	 happyFail

action_691 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_691 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_691 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_691 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_691 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_691 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_691 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_691 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_691 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_691 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_691 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_691 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_719 state_indent')
action_691 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_691 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_691 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_691 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_691 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_691 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_691 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_691 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_691 state_indent tok = 
	 happyFail

action_692 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_718 state_indent')
action_692 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_692 state_indent tok = 
	 happyFail

action_693 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_717 state_indent')
action_693 state_indent (13, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_716 state_indent')
action_693 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_22 state_indent'
action_693 state_indent tok = 
	 happyFail

action_694 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_47 state_indent'
action_694 state_indent tok = 
	 happyFail

action_695 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_48 state_indent'
action_695 state_indent tok = 
	 happyFail

action_696 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_329 state_indent'
action_696 state_indent tok = 
	 happyFail

action_697 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_715 state_indent')
action_697 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_49 state_indent'
action_697 state_indent tok = 
	 happyFail

action_698 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_85 state_indent')
action_698 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_163 state_indent'
action_698 state_indent tok = 
	 happyFail

action_699 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges []) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[I1])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(1, Inf)]) else Ranges []) {-{([Z1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_372 state_indent'
action_699 state_indent tok = 
	 happyFail

action_700 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges []) `unionRanges` (Ranges [(0,maxRanges token_indent)]) {-{([I1],[I1])([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_165 state_indent'
action_700 state_indent tok = 
	 happyFail

action_701 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_139 state_indent'
action_701 state_indent tok = 
	 happyFail

action_702 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(0,maxRanges token_indent)] {-{([=],[=I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_142 state_indent'
action_702 state_indent tok = 
	 happyFail

action_703 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_703 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_703 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_703 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_703 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_143 state_indent'
action_703 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_143 state_indent'
action_703 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_703 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_143 state_indent'
action_703 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_143 state_indent'
action_703 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_703 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_703 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_703 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_703 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_703 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_703 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_143 state_indent'
action_703 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_384 state_indent')
action_703 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_703 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_703 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_703 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_703 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_703 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_143 state_indent'
action_703 state_indent tok = 
	 happyFail

action_704 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_134 state_indent'
action_704 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_134 state_indent'
action_704 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_134 state_indent'
action_704 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_134 state_indent'
action_704 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_134 state_indent'
action_704 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_134 state_indent'
action_704 state_indent tok = 
	 happyFail

action_705 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_506 state_indent')
action_705 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_714 state_indent')
action_705 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_705 state_indent tok = 
	 happyFail

action_706 state_indent (329, token_indent) 
	| item_indent <- Ranges [(0, Fin 0)]{-[Z]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_238 state_indent')
action_706 state_indent (147, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if 0 == minRanges token_indent then item_indent else emptyRanges) {-{([=],[Z])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_713 state_indent')
action_706 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_706 state_indent tok = 
	 happyFail

action_707 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_712 state_indent')
action_707 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_145 state_indent'
action_707 state_indent tok = 
	 happyFail

action_708 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_147 state_indent'
action_708 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_147 state_indent'
action_708 state_indent tok = 
	 happyFail

action_709 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_99 state_indent'
action_709 state_indent tok = 
	 happyFail

action_710 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_135 state_indent'
action_710 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_135 state_indent'
action_710 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_135 state_indent'
action_710 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_135 state_indent'
action_710 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_135 state_indent'
action_710 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_135 state_indent'
action_710 state_indent tok = 
	 happyFail

action_711 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_276 state_indent'
action_711 state_indent tok = 
	 happyFail

action_712 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_712 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_712 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_248 state_indent')
action_712 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_712 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_712 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_712 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_712 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_712 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_712 state_indent (46, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_705 state_indent')
action_712 state_indent (68, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_740 state_indent')
action_712 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_709 state_indent')
action_712 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_712 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_712 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_712 state_indent tok = 
	 happyFail

action_713 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_136 state_indent'
action_713 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_136 state_indent'
action_713 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_136 state_indent'
action_713 state_indent (353, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_136 state_indent'
action_713 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_136 state_indent'
action_713 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_136 state_indent'
action_713 state_indent tok = 
	 happyFail

action_714 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_714 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_714 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_714 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_714 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_714 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_739 state_indent')
action_714 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_714 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_714 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_714 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_714 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_714 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_714 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_737 state_indent')
action_714 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_714 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_714 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_714 state_indent (69, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_738 state_indent')
action_714 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_714 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_714 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_714 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_714 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_714 state_indent tok = 
	 happyFail

action_715 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_715 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_715 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_525 state_indent')
action_715 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_735 state_indent')
action_715 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_736 state_indent')
action_715 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_715 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_715 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_715 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_715 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_715 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_715 state_indent (26, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_734 state_indent')
action_715 state_indent (27, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_521 state_indent')
action_715 state_indent (122, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_522 state_indent')
action_715 state_indent (124, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_523 state_indent')
action_715 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_288 state_indent')
action_715 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_524 state_indent')
action_715 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_715 state_indent tok = 
	 happyFail

action_716 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_733 state_indent')
action_716 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_716 state_indent tok = 
	 happyFail

action_717 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_717 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_717 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_289 state_indent')
action_717 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_717 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_717 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_717 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_717 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_717 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_717 state_indent (25, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_732 state_indent')
action_717 state_indent (122, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_695 state_indent')
action_717 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_288 state_indent')
action_717 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_696 state_indent')
action_717 state_indent (149, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_697 state_indent')
action_717 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_21 state_indent'
action_717 state_indent tok = 
	 happyFail

action_718 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_43 state_indent'
action_718 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_43 state_indent'
action_718 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_43 state_indent'
action_718 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_43 state_indent'
action_718 state_indent tok = 
	 happyFail

action_719 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_390 state_indent'
action_719 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_390 state_indent'
action_719 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_390 state_indent'
action_719 state_indent tok = 
	 happyFail

action_720 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_154 state_indent'
action_720 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_154 state_indent'
action_720 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_154 state_indent'
action_720 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_154 state_indent'
action_720 state_indent tok = 
	 happyFail

action_721 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_721 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_721 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_636 state_indent')
action_721 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_721 state_indent (152, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_731 state_indent')
action_721 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_721 state_indent tok = 
	 happyFail

action_722 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges []) {-{([I1],[=I1])([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_277 state_indent')
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent (80, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_730 state_indent')
action_722 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_174 state_indent'
action_722 state_indent tok = 
	 happyFail

action_723 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_729 state_indent')
action_723 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_723 state_indent tok = 
	 happyFail

action_724 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_239 state_indent'
action_724 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_239 state_indent'
action_724 state_indent tok = 
	 happyFail

action_725 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_728 state_indent')
action_725 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_725 state_indent tok = 
	 happyFail

action_726 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_726 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_726 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_726 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_726 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_726 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_726 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_726 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_726 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_726 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_726 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_726 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_727 state_indent')
action_726 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_726 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_726 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_726 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_726 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_726 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_726 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_726 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_726 state_indent tok = 
	 happyFail

action_727 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_101 state_indent'
action_727 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_101 state_indent'
action_727 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_101 state_indent'
action_727 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_101 state_indent'
action_727 state_indent tok = 
	 happyFail

action_728 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_728 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_728 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_728 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_728 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_728 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_728 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_728 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_728 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_728 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_728 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_728 state_indent (51, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_745 state_indent')
action_728 state_indent (52, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_348 state_indent')
action_728 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_102 state_indent')
action_728 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_728 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_728 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_728 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_728 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_728 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_728 state_indent tok = 
	 happyFail

action_729 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_729 state_indent (314, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_163 state_indent')
action_729 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_729 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_729 state_indent (321, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_164 state_indent')
action_729 state_indent (322, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_165 state_indent')
action_729 state_indent (323, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_166 state_indent')
action_729 state_indent (324, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_167 state_indent')
action_729 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_168 state_indent')
action_729 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_169 state_indent')
action_729 state_indent (334, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_170 state_indent')
action_729 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_171 state_indent')
action_729 state_indent (345, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_172 state_indent')
action_729 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_173 state_indent')
action_729 state_indent (349, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_174 state_indent')
action_729 state_indent (354, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_175 state_indent')
action_729 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_176 state_indent')
action_729 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_177 state_indent')
action_729 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_729 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_729 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_729 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_729 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_729 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_729 state_indent (84, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_744 state_indent')
action_729 state_indent (85, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_147 state_indent')
action_729 state_indent (86, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_148 state_indent')
action_729 state_indent (87, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_149 state_indent')
action_729 state_indent (88, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_150 state_indent')
action_729 state_indent (89, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_151 state_indent')
action_729 state_indent (90, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_152 state_indent')
action_729 state_indent (93, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_153 state_indent')
action_729 state_indent (94, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_154 state_indent')
action_729 state_indent (95, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_155 state_indent')
action_729 state_indent (121, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_156 state_indent')
action_729 state_indent (123, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_157 state_indent')
action_729 state_indent (125, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_158 state_indent')
action_729 state_indent (135, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_159 state_indent')
action_729 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_160 state_indent')
action_729 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_161 state_indent')
action_729 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_729 state_indent (146, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_162 state_indent')
action_729 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_729 state_indent tok = 
	 happyFail

action_730 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` ((case maxRanges token_indent of Fin 0 -> Ranges []; Fin 1 -> Ranges [(0,Fin 0),(2,Inf)]; _ -> Ranges [(0, Inf)])) {-{([I1],[=I1])([=I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (token_indent) {-{([I1],[=I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_241 state_indent'
action_730 state_indent tok = 
	 happyFail

action_731 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_155 state_indent'
action_731 state_indent tok = 
	 happyFail

action_732 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_46 state_indent'
action_732 state_indent tok = 
	 happyFail

action_733 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_42 state_indent'
action_733 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_42 state_indent'
action_733 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_42 state_indent'
action_733 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_42 state_indent'
action_733 state_indent tok = 
	 happyFail

action_734 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_743 state_indent')
action_734 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_668 state_indent')
action_734 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_734 state_indent tok = 
	 happyFail

action_735 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_51 state_indent'
action_735 state_indent tok = 
	 happyFail

action_736 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_742 state_indent')
action_736 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_736 state_indent tok = 
	 happyFail

action_737 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_149 state_indent'
action_737 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_149 state_indent'
action_737 state_indent tok = 
	 happyFail

action_738 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_148 state_indent'
action_738 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_148 state_indent'
action_738 state_indent tok = 
	 happyFail

action_739 state_indent (313, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_110 state_indent')
action_739 state_indent (315, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_99 state_indent')
action_739 state_indent (316, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_111 state_indent')
action_739 state_indent (325, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_112 state_indent')
action_739 state_indent (331, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_113 state_indent')
action_739 state_indent (371, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_114 state_indent')
action_739 state_indent (372, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_115 state_indent')
action_739 state_indent (373, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_116 state_indent')
action_739 state_indent (374, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_117 state_indent')
action_739 state_indent (375, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_118 state_indent')
action_739 state_indent (376, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyShift (action_119 state_indent')
action_739 state_indent (53, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_741 state_indent')
action_739 state_indent (54, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_103 state_indent')
action_739 state_indent (136, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_106 state_indent')
action_739 state_indent (137, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_107 state_indent')
action_739 state_indent (138, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_108 state_indent')
action_739 state_indent (153, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyGoto (action_109 state_indent')
action_739 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyFail
action_739 state_indent tok = 
	 happyFail

action_740 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_146 state_indent'
action_740 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_146 state_indent'
action_740 state_indent tok = 
	 happyFail

action_741 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_150 state_indent'
action_741 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_150 state_indent'
action_741 state_indent tok = 
	 happyFail

action_742 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_50 state_indent'
action_742 state_indent tok = 
	 happyFail

action_743 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_52 state_indent'
action_743 state_indent tok = 
	 happyFail

action_744 state_indent (317, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (318, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (319, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (320, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (326, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges []) {-{([I1],[=I1])([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Inf)]) else Ranges []) {-{([ZZ1],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (332, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (333, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (335, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (336, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (337, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (338, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (339, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (340, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (341, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (342, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (343, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (347, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- ((Ranges [(1, Inf)]) `unionRanges` (if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [])) `unionRanges` (token_indent) {-{([I1],[=I1])([I2],[I1])([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (348, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (350, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (351, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (352, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (355, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (356, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (357, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (359, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(2, Inf)] else Ranges [] {-{([I2],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (360, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (361, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (362, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (363, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (364, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (366, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (367, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (368, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(1, Inf)] else Ranges [] {-{([I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (369, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (370, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- if maxRanges token_indent >= Fin 1 then Ranges [(0, Inf)] else Ranges [] {-{([=I1],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 2, Inf)]) {-{([I2],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- Ranges [(minRanges token_indent, Inf)] {-{([=I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_246 state_indent'
action_744 state_indent tok = 
	 happyFail

action_745 state_indent (327, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- case maxRanges token_indent of { Inf -> item_indent; Fin x -> (Ranges [(0, Fin (x - 1))]) } {-{([=],[I1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_100 state_indent'
action_745 state_indent (329, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (if maxRanges token_indent >= Fin 1 then (Ranges [(0, Fin 0)]) else Ranges []) {-{([Z],[Z1])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_100 state_indent'
action_745 state_indent (377, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- (Ranges [((minRanges token_indent) + 1, Inf)]) {-{([I1],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_100 state_indent'
action_745 state_indent (tok, token_indent) 
	| item_indent <- state_indent{-[=]-}
	, lookahead_indent  <- token_indent {-{([=],[=])}-}
	, state_indent' <- item_indent `intersectRange` lookahead_indent
	, notEmptyRanges state_indent'
	= happyReduce_100 state_indent'
action_745 state_indent tok = 
	 happyFail

happyReduce_1 state_indent = happyReduce 6 (4, state_indent) happyReduction_1
happyReduction_1 ((HappyAbsSyn6  happy_var_6) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn11  happy_var_4) `HappyStk`
	(HappyAbsSyn148  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn4
		 (HsModule happy_var_1 happy_var_3 happy_var_4 (fst happy_var_6) (snd happy_var_6)
	) `HappyStk` happyRest

happyReduce_2 state_indent = happySpecReduce_2  (4, state_indent) happyReduction_2
happyReduction_2 (HappyAbsSyn6  happy_var_2)
	(HappyAbsSyn5  happy_var_1)
	 =  HappyAbsSyn4
		 (HsModule happy_var_1 main_mod (Just [HsEVar (UnQual main_name)])
							(fst happy_var_2) (snd happy_var_2)
	)
happyReduction_2 _ _  = notHappyAtAll 

happyReduce_3 state_indent = happyMonadReduce 0 (5, state_indent) happyReduction_3
happyReduction_3 (happyRest) tk
	 = happyThen (( getSrcLoc)
	) (\r -> happyReturn (HappyAbsSyn5 r))

happyReduce_4 state_indent = happySpecReduce_3  (6, state_indent) happyReduction_4
happyReduction_4 _
	(HappyAbsSyn6  happy_var_2)
	_
	 =  HappyAbsSyn6
		 (happy_var_2
	)
happyReduction_4 _ _ _  = notHappyAtAll 

happyReduce_5 state_indent = happySpecReduce_1  (6, state_indent) happyReduction_5
happyReduction_5 (HappyAbsSyn6  happy_var_1)
	 =  HappyAbsSyn6
		 (happy_var_1
	)
happyReduction_5 _  = notHappyAtAll 

happyReduce_6 state_indent = happyReduce 4 (7, state_indent) happyReduction_6
happyReduction_6 ((HappyAbsSyn32  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn16  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn6
		 ((reverse happy_var_2, happy_var_4)
	) `HappyStk` happyRest

happyReduce_7 state_indent = happySpecReduce_2  (7, state_indent) happyReduction_7
happyReduction_7 (HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn6
		 (([], happy_var_2)
	)
happyReduction_7 _ _  = notHappyAtAll 

happyReduce_8 state_indent = happySpecReduce_3  (7, state_indent) happyReduction_8
happyReduction_8 _
	(HappyAbsSyn16  happy_var_2)
	_
	 =  HappyAbsSyn6
		 ((reverse happy_var_2, [])
	)
happyReduction_8 _ _ _  = notHappyAtAll 

happyReduce_9 state_indent = happySpecReduce_1  (7, state_indent) happyReduction_9
happyReduction_9 _
	 =  HappyAbsSyn6
		 (([], [])
	)

happyReduce_10 state_indent = happySpecReduce_2  (8, state_indent) happyReduction_10
happyReduction_10 (HappyAbsSyn32  happy_var_2)
	(HappyAbsSyn16  happy_var_1)
	 =  HappyAbsSyn6
		 ((reverse happy_var_1, happy_var_2)
	)
happyReduction_10 _ _  = notHappyAtAll 

happyReduce_11 state_indent = happySpecReduce_1  (8, state_indent) happyReduction_11
happyReduction_11 (HappyAbsSyn16  happy_var_1)
	 =  HappyAbsSyn6
		 ((reverse happy_var_1, [])
	)
happyReduction_11 _  = notHappyAtAll 

happyReduce_12 state_indent = happySpecReduce_1  (8, state_indent) happyReduction_12
happyReduction_12 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn6
		 (([], happy_var_1)
	)
happyReduction_12 _  = notHappyAtAll 

happyReduce_13 state_indent = happySpecReduce_0  (8, state_indent) happyReduction_13
happyReduction_13  =  HappyAbsSyn6
		 (([], [])
	)

happyReduce_14 state_indent = happySpecReduce_2  (9, state_indent) happyReduction_14
happyReduction_14 _
	_
	 =  HappyAbsSyn9
		 (()
	)

happyReduce_15 state_indent = happySpecReduce_1  (10, state_indent) happyReduction_15
happyReduction_15 _
	 =  HappyAbsSyn9
		 (()
	)

happyReduce_16 state_indent = happySpecReduce_0  (10, state_indent) happyReduction_16
happyReduction_16  =  HappyAbsSyn9
		 (()
	)

happyReduce_17 state_indent = happySpecReduce_1  (11, state_indent) happyReduction_17
happyReduction_17 (HappyAbsSyn12  happy_var_1)
	 =  HappyAbsSyn11
		 (Just happy_var_1
	)
happyReduction_17 _  = notHappyAtAll 

happyReduce_18 state_indent = happySpecReduce_0  (11, state_indent) happyReduction_18
happyReduction_18  =  HappyAbsSyn11
		 (Nothing
	)

happyReduce_19 state_indent = happyReduce 4 (12, state_indent) happyReduction_19
happyReduction_19 (_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn12  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn12
		 (reverse happy_var_2
	) `HappyStk` happyRest

happyReduce_20 state_indent = happySpecReduce_3  (12, state_indent) happyReduction_20
happyReduction_20 _
	_
	_
	 =  HappyAbsSyn12
		 ([]
	)

happyReduce_21 state_indent = happySpecReduce_1  (13, state_indent) happyReduction_21
happyReduction_21 _
	 =  HappyAbsSyn9
		 (()
	)

happyReduce_22 state_indent = happySpecReduce_0  (13, state_indent) happyReduction_22
happyReduction_22  =  HappyAbsSyn9
		 (()
	)

happyReduce_23 state_indent = happySpecReduce_3  (14, state_indent) happyReduction_23
happyReduction_23 (HappyAbsSyn15  happy_var_3)
	_
	(HappyAbsSyn12  happy_var_1)
	 =  HappyAbsSyn12
		 (happy_var_3 : happy_var_1
	)
happyReduction_23 _ _ _  = notHappyAtAll 

happyReduce_24 state_indent = happySpecReduce_1  (14, state_indent) happyReduction_24
happyReduction_24 (HappyAbsSyn15  happy_var_1)
	 =  HappyAbsSyn12
		 ([happy_var_1]
	)
happyReduction_24 _  = notHappyAtAll 

happyReduce_25 state_indent = happySpecReduce_1  (15, state_indent) happyReduction_25
happyReduction_25 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn15
		 (HsEVar happy_var_1
	)
happyReduction_25 _  = notHappyAtAll 

happyReduce_26 state_indent = happySpecReduce_1  (15, state_indent) happyReduction_26
happyReduction_26 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn15
		 (HsEAbs happy_var_1
	)
happyReduction_26 _  = notHappyAtAll 

happyReduce_27 state_indent = happyReduce 4 (15, state_indent) happyReduction_27
happyReduction_27 (_ `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn54  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn15
		 (HsEThingAll happy_var_1
	) `HappyStk` happyRest

happyReduce_28 state_indent = happySpecReduce_3  (15, state_indent) happyReduction_28
happyReduction_28 _
	_
	(HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn15
		 (HsEThingWith happy_var_1 []
	)
happyReduction_28 _ _ _  = notHappyAtAll 

happyReduce_29 state_indent = happyReduce 4 (15, state_indent) happyReduction_29
happyReduction_29 (_ `HappyStk`
	(HappyAbsSyn26  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn54  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn15
		 (HsEThingWith happy_var_1 (reverse happy_var_3)
	) `HappyStk` happyRest

happyReduce_30 state_indent = happySpecReduce_2  (15, state_indent) happyReduction_30
happyReduction_30 (HappyAbsSyn148  happy_var_2)
	_
	 =  HappyAbsSyn15
		 (HsEModuleContents happy_var_2
	)
happyReduction_30 _ _  = notHappyAtAll 

happyReduce_31 state_indent = happySpecReduce_3  (16, state_indent) happyReduction_31
happyReduction_31 _
	(HappyAbsSyn16  happy_var_2)
	(HappyAbsSyn16  happy_var_1)
	 =  HappyAbsSyn16
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_31 _ _ _  = notHappyAtAll 

happyReduce_32 state_indent = happySpecReduce_2  (16, state_indent) happyReduction_32
happyReduction_32 _
	(HappyAbsSyn16  happy_var_1)
	 =  HappyAbsSyn16
		 (happy_var_1
	)
happyReduction_32 _ _  = notHappyAtAll 

happyReduce_33 state_indent = happySpecReduce_3  (17, state_indent) happyReduction_33
happyReduction_33 (HappyAbsSyn16  happy_var_3)
	_
	(HappyAbsSyn16  happy_var_1)
	 =  HappyAbsSyn16
		 (happy_var_3 ++ happy_var_1
	)
happyReduction_33 _ _ _  = notHappyAtAll 

happyReduce_34 state_indent = happySpecReduce_1  (17, state_indent) happyReduction_34
happyReduction_34 (HappyAbsSyn16  happy_var_1)
	 =  HappyAbsSyn16
		 (happy_var_1
	)
happyReduction_34 _  = notHappyAtAll 

happyReduce_35 state_indent = happyReduce 6 (18, state_indent) happyReduction_35
happyReduction_35 ((HappyAbsSyn21  happy_var_6) `HappyStk`
	(HappyAbsSyn20  happy_var_5) `HappyStk`
	(HappyAbsSyn148  happy_var_4) `HappyStk`
	(HappyAbsSyn19  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn16
		 ([HsImportDecl happy_var_1 happy_var_4 happy_var_3 happy_var_5 happy_var_6]
	) `HappyStk` happyRest

happyReduce_36 state_indent = happySpecReduce_1  (19, state_indent) happyReduction_36
happyReduction_36 _
	 =  HappyAbsSyn19
		 (True
	)

happyReduce_37 state_indent = happySpecReduce_0  (19, state_indent) happyReduction_37
happyReduction_37  =  HappyAbsSyn19
		 (False
	)

happyReduce_38 state_indent = happySpecReduce_2  (20, state_indent) happyReduction_38
happyReduction_38 (HappyAbsSyn148  happy_var_2)
	_
	 =  HappyAbsSyn20
		 (Just happy_var_2
	)
happyReduction_38 _ _  = notHappyAtAll 

happyReduce_39 state_indent = happySpecReduce_0  (20, state_indent) happyReduction_39
happyReduction_39  =  HappyAbsSyn20
		 (Nothing
	)

happyReduce_40 state_indent = happySpecReduce_1  (21, state_indent) happyReduction_40
happyReduction_40 (HappyAbsSyn22  happy_var_1)
	 =  HappyAbsSyn21
		 (Just happy_var_1
	)
happyReduction_40 _  = notHappyAtAll 

happyReduce_41 state_indent = happySpecReduce_0  (21, state_indent) happyReduction_41
happyReduction_41  =  HappyAbsSyn21
		 (Nothing
	)

happyReduce_42 state_indent = happyReduce 5 (22, state_indent) happyReduction_42
happyReduction_42 (_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn24  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn19  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn22
		 ((happy_var_1, reverse happy_var_3)
	) `HappyStk` happyRest

happyReduce_43 state_indent = happyReduce 4 (22, state_indent) happyReduction_43
happyReduction_43 (_ `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn19  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn22
		 ((happy_var_1, [])
	) `HappyStk` happyRest

happyReduce_44 state_indent = happySpecReduce_1  (23, state_indent) happyReduction_44
happyReduction_44 _
	 =  HappyAbsSyn19
		 (True
	)

happyReduce_45 state_indent = happySpecReduce_0  (23, state_indent) happyReduction_45
happyReduction_45  =  HappyAbsSyn19
		 (False
	)

happyReduce_46 state_indent = happySpecReduce_3  (24, state_indent) happyReduction_46
happyReduction_46 (HappyAbsSyn25  happy_var_3)
	_
	(HappyAbsSyn24  happy_var_1)
	 =  HappyAbsSyn24
		 (happy_var_3 : happy_var_1
	)
happyReduction_46 _ _ _  = notHappyAtAll 

happyReduce_47 state_indent = happySpecReduce_1  (24, state_indent) happyReduction_47
happyReduction_47 (HappyAbsSyn25  happy_var_1)
	 =  HappyAbsSyn24
		 ([happy_var_1]
	)
happyReduction_47 _  = notHappyAtAll 

happyReduce_48 state_indent = happySpecReduce_1  (25, state_indent) happyReduction_48
happyReduction_48 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn25
		 (HsIVar happy_var_1
	)
happyReduction_48 _  = notHappyAtAll 

happyReduce_49 state_indent = happySpecReduce_1  (25, state_indent) happyReduction_49
happyReduction_49 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn25
		 (HsIAbs happy_var_1
	)
happyReduction_49 _  = notHappyAtAll 

happyReduce_50 state_indent = happyReduce 4 (25, state_indent) happyReduction_50
happyReduction_50 (_ `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn50  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn25
		 (HsIThingAll happy_var_1
	) `HappyStk` happyRest

happyReduce_51 state_indent = happySpecReduce_3  (25, state_indent) happyReduction_51
happyReduction_51 _
	_
	(HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn25
		 (HsIThingWith happy_var_1 []
	)
happyReduction_51 _ _ _  = notHappyAtAll 

happyReduce_52 state_indent = happyReduce 4 (25, state_indent) happyReduction_52
happyReduction_52 (_ `HappyStk`
	(HappyAbsSyn26  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn50  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn25
		 (HsIThingWith happy_var_1 (reverse happy_var_3)
	) `HappyStk` happyRest

happyReduce_53 state_indent = happySpecReduce_3  (26, state_indent) happyReduction_53
happyReduction_53 (HappyAbsSyn27  happy_var_3)
	_
	(HappyAbsSyn26  happy_var_1)
	 =  HappyAbsSyn26
		 (happy_var_3 : happy_var_1
	)
happyReduction_53 _ _ _  = notHappyAtAll 

happyReduce_54 state_indent = happySpecReduce_1  (26, state_indent) happyReduction_54
happyReduction_54 (HappyAbsSyn27  happy_var_1)
	 =  HappyAbsSyn26
		 ([happy_var_1]
	)
happyReduction_54 _  = notHappyAtAll 

happyReduce_55 state_indent = happySpecReduce_1  (27, state_indent) happyReduction_55
happyReduction_55 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn27
		 (HsVarName happy_var_1
	)
happyReduction_55 _  = notHappyAtAll 

happyReduce_56 state_indent = happySpecReduce_1  (27, state_indent) happyReduction_56
happyReduction_56 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn27
		 (HsConName happy_var_1
	)
happyReduction_56 _  = notHappyAtAll 

happyReduce_57 state_indent = happyReduce 4 (28, state_indent) happyReduction_57
happyReduction_57 ((HappyAbsSyn31  happy_var_4) `HappyStk`
	(HappyAbsSyn29  happy_var_3) `HappyStk`
	(HappyAbsSyn30  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn28
		 (HsInfixDecl happy_var_1 happy_var_2 happy_var_3 (reverse happy_var_4)
	) `HappyStk` happyRest

happyReduce_58 state_indent = happySpecReduce_0  (29, state_indent) happyReduction_58
happyReduction_58  =  HappyAbsSyn29
		 (9
	)

happyReduce_59 state_indent = happyMonadReduce 1 (29, state_indent) happyReduction_59
happyReduction_59 ((HappyTerminal ((IntTok happy_var_1), _)) `HappyStk`
	happyRest) tk
	 = happyThen (( checkPrec happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn29 r))

happyReduce_60 state_indent = happySpecReduce_1  (30, state_indent) happyReduction_60
happyReduction_60 _
	 =  HappyAbsSyn30
		 (HsAssocNone
	)

happyReduce_61 state_indent = happySpecReduce_1  (30, state_indent) happyReduction_61
happyReduction_61 _
	 =  HappyAbsSyn30
		 (HsAssocLeft
	)

happyReduce_62 state_indent = happySpecReduce_1  (30, state_indent) happyReduction_62
happyReduction_62 _
	 =  HappyAbsSyn30
		 (HsAssocRight
	)

happyReduce_63 state_indent = happySpecReduce_3  (31, state_indent) happyReduction_63
happyReduction_63 (HappyAbsSyn131  happy_var_3)
	_
	(HappyAbsSyn31  happy_var_1)
	 =  HappyAbsSyn31
		 (happy_var_3 : happy_var_1
	)
happyReduction_63 _ _ _  = notHappyAtAll 

happyReduce_64 state_indent = happySpecReduce_1  (31, state_indent) happyReduction_64
happyReduction_64 (HappyAbsSyn131  happy_var_1)
	 =  HappyAbsSyn31
		 ([happy_var_1]
	)
happyReduction_64 _  = notHappyAtAll 

happyReduce_65 state_indent = happyMonadReduce 2 (32, state_indent) happyReduction_65
happyReduction_65 (_ `HappyStk`
	(HappyAbsSyn32  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_66 state_indent = happyMonadReduce 1 (33, state_indent) happyReduction_66
happyReduction_66 ((HappyAbsSyn32  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_67 state_indent = happySpecReduce_3  (34, state_indent) happyReduction_67
happyReduction_67 _
	(HappyAbsSyn32  happy_var_2)
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_67 _ _ _  = notHappyAtAll 

happyReduce_68 state_indent = happySpecReduce_2  (34, state_indent) happyReduction_68
happyReduction_68 _
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_68 _ _  = notHappyAtAll 

happyReduce_69 state_indent = happySpecReduce_3  (35, state_indent) happyReduction_69
happyReduction_69 (HappyAbsSyn32  happy_var_3)
	_
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_3 ++ happy_var_1
	)
happyReduction_69 _ _ _  = notHappyAtAll 

happyReduce_70 state_indent = happySpecReduce_1  (35, state_indent) happyReduction_70
happyReduction_70 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_70 _  = notHappyAtAll 

happyReduce_71 state_indent = happyReduce 5 (36, state_indent) happyReduction_71
happyReduction_71 ((HappyAbsSyn51  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn58  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn32
		 ([HsTypeDecl happy_var_1 (fst happy_var_3) (snd happy_var_3) happy_var_5]
	) `HappyStk` happyRest

happyReduce_72 state_indent = happyMonadReduce 6 (36, state_indent) happyReduction_72
happyReduction_72 ((HappyAbsSyn70  happy_var_6) `HappyStk`
	(HappyAbsSyn60  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn55  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { (cs,c,t) <- checkDataHeader happy_var_3;
				return ([HsDataDecl happy_var_1 cs c t (reverse happy_var_5) happy_var_6]) })
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_73 state_indent = happyMonadReduce 6 (36, state_indent) happyReduction_73
happyReduction_73 ((HappyAbsSyn70  happy_var_6) `HappyStk`
	(HappyAbsSyn61  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn55  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { (cs,c,t) <- checkDataHeader happy_var_3;
				return ([HsNewTypeDecl happy_var_1 cs c t happy_var_5 happy_var_6]) })
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_74 state_indent = happyMonadReduce 4 (36, state_indent) happyReduction_74
happyReduction_74 ((HappyAbsSyn32  happy_var_4) `HappyStk`
	(HappyAbsSyn55  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { (cs,c,vs) <- checkClassHeader happy_var_3;
				return ([HsClassDecl happy_var_1 cs c vs happy_var_4]) })
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_75 state_indent = happyMonadReduce 4 (36, state_indent) happyReduction_75
happyReduction_75 ((HappyAbsSyn32  happy_var_4) `HappyStk`
	(HappyAbsSyn55  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { (cs,c,ts) <- checkInstHeader happy_var_3;
				return ([HsInstDecl happy_var_1 cs c ts happy_var_4]) })
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_76 state_indent = happyReduce 5 (36, state_indent) happyReduction_76
happyReduction_76 (_ `HappyStk`
	(HappyAbsSyn37  happy_var_4) `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn32
		 ([HsDefaultDecl happy_var_1 happy_var_4]
	) `HappyStk` happyRest

happyReduce_77 state_indent = happySpecReduce_1  (36, state_indent) happyReduction_77
happyReduction_77 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn32
		 ([happy_var_1]
	)
happyReduction_77 _  = notHappyAtAll 

happyReduce_78 state_indent = happySpecReduce_1  (36, state_indent) happyReduction_78
happyReduction_78 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn32
		 ([happy_var_1]
	)
happyReduction_78 _  = notHappyAtAll 

happyReduce_79 state_indent = happySpecReduce_1  (37, state_indent) happyReduction_79
happyReduction_79 (HappyAbsSyn37  happy_var_1)
	 =  HappyAbsSyn37
		 (reverse happy_var_1
	)
happyReduction_79 _  = notHappyAtAll 

happyReduce_80 state_indent = happySpecReduce_1  (37, state_indent) happyReduction_80
happyReduction_80 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn37
		 ([happy_var_1]
	)
happyReduction_80 _  = notHappyAtAll 

happyReduce_81 state_indent = happySpecReduce_0  (37, state_indent) happyReduction_81
happyReduction_81  =  HappyAbsSyn37
		 ([]
	)

happyReduce_82 state_indent = happyMonadReduce 3 (38, state_indent) happyReduction_82
happyReduction_82 (_ `HappyStk`
	(HappyAbsSyn32  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_2)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_83 state_indent = happySpecReduce_1  (38, state_indent) happyReduction_83
happyReduction_83 _
	 =  HappyAbsSyn32
		 ([]
	)

happyReduce_84 state_indent = happySpecReduce_3  (39, state_indent) happyReduction_84
happyReduction_84 (HappyAbsSyn28  happy_var_3)
	_
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_3 : happy_var_1
	)
happyReduction_84 _ _ _  = notHappyAtAll 

happyReduce_85 state_indent = happySpecReduce_1  (39, state_indent) happyReduction_85
happyReduction_85 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn32
		 ([happy_var_1]
	)
happyReduction_85 _  = notHappyAtAll 

happyReduce_86 state_indent = happySpecReduce_1  (40, state_indent) happyReduction_86
happyReduction_86 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn28
		 (happy_var_1
	)
happyReduction_86 _  = notHappyAtAll 

happyReduce_87 state_indent = happySpecReduce_1  (40, state_indent) happyReduction_87
happyReduction_87 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn28
		 (happy_var_1
	)
happyReduction_87 _  = notHappyAtAll 

happyReduce_88 state_indent = happySpecReduce_1  (40, state_indent) happyReduction_88
happyReduction_88 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn28
		 (happy_var_1
	)
happyReduction_88 _  = notHappyAtAll 

happyReduce_89 state_indent = happySpecReduce_3  (41, state_indent) happyReduction_89
happyReduction_89 _
	(HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn32
		 (happy_var_2
	)
happyReduction_89 _ _ _  = notHappyAtAll 

happyReduce_90 state_indent = happyMonadReduce 1 (41, state_indent) happyReduction_90
happyReduction_90 ((HappyAbsSyn32  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_91 state_indent = happyMonadReduce 0 (41, state_indent) happyReduction_91
happyReduction_91 (happyRest) tk
	 = happyThen (( checkRevDecls [])
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_92 state_indent = happySpecReduce_3  (42, state_indent) happyReduction_92
happyReduction_92 _
	(HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn32
		 (happy_var_2
	)
happyReduction_92 _ _ _  = notHappyAtAll 

happyReduce_93 state_indent = happySpecReduce_1  (42, state_indent) happyReduction_93
happyReduction_93 _
	 =  HappyAbsSyn32
		 ([]
	)

happyReduce_94 state_indent = happySpecReduce_1  (43, state_indent) happyReduction_94
happyReduction_94 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_94 _  = notHappyAtAll 

happyReduce_95 state_indent = happySpecReduce_2  (44, state_indent) happyReduction_95
happyReduction_95 (HappyAbsSyn32  happy_var_2)
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_95 _ _  = notHappyAtAll 

happyReduce_96 state_indent = happySpecReduce_1  (44, state_indent) happyReduction_96
happyReduction_96 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_96 _  = notHappyAtAll 

happyReduce_97 state_indent = happyReduce 4 (45, state_indent) happyReduction_97
happyReduction_97 ((HappyAbsSyn55  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn46  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn28
		 (HsTypeSig happy_var_1 (reverse happy_var_2) happy_var_4
	) `HappyStk` happyRest

happyReduce_98 state_indent = happySpecReduce_3  (46, state_indent) happyReduction_98
happyReduction_98 (HappyAbsSyn50  happy_var_3)
	_
	(HappyAbsSyn46  happy_var_1)
	 =  HappyAbsSyn46
		 (happy_var_3 : happy_var_1
	)
happyReduction_98 _ _ _  = notHappyAtAll 

happyReduce_99 state_indent = happyMonadReduce 1 (46, state_indent) happyReduction_99
happyReduction_99 ((HappyAbsSyn54  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { n <- checkUnQual happy_var_1;
						return [n] })
	) (\r -> happyReturn (HappyAbsSyn46 r))

happyReduce_100 state_indent = happyReduce 9 (47, state_indent) happyReduction_100
happyReduction_100 ((HappyAbsSyn51  happy_var_9) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn50  happy_var_7) `HappyStk`
	(HappyAbsSyn49  happy_var_6) `HappyStk`
	(HappyAbsSyn48  happy_var_5) `HappyStk`
	(HappyTerminal ((VarId happy_var_4), _)) `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn28
		 (HsForeignImport happy_var_1 happy_var_4 happy_var_5 happy_var_6 happy_var_7 happy_var_9
	) `HappyStk` happyRest

happyReduce_101 state_indent = happyReduce 8 (47, state_indent) happyReduction_101
happyReduction_101 ((HappyAbsSyn51  happy_var_8) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn50  happy_var_6) `HappyStk`
	(HappyAbsSyn49  happy_var_5) `HappyStk`
	(HappyTerminal ((VarId happy_var_4), _)) `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn28
		 (HsForeignExport happy_var_1 happy_var_4 happy_var_5 happy_var_6 happy_var_8
	) `HappyStk` happyRest

happyReduce_102 state_indent = happySpecReduce_1  (48, state_indent) happyReduction_102
happyReduction_102 _
	 =  HappyAbsSyn48
		 (HsSafe
	)

happyReduce_103 state_indent = happySpecReduce_1  (48, state_indent) happyReduction_103
happyReduction_103 _
	 =  HappyAbsSyn48
		 (HsUnsafe
	)

happyReduce_104 state_indent = happySpecReduce_0  (48, state_indent) happyReduction_104
happyReduction_104  =  HappyAbsSyn48
		 (HsSafe
	)

happyReduce_105 state_indent = happySpecReduce_1  (49, state_indent) happyReduction_105
happyReduction_105 (HappyTerminal ((StringTok happy_var_1), _))
	 =  HappyAbsSyn49
		 (happy_var_1
	)
happyReduction_105 _  = notHappyAtAll 

happyReduce_106 state_indent = happySpecReduce_0  (49, state_indent) happyReduction_106
happyReduction_106  =  HappyAbsSyn49
		 (""
	)

happyReduce_107 state_indent = happySpecReduce_1  (50, state_indent) happyReduction_107
happyReduction_107 (HappyTerminal ((VarId happy_var_1), _))
	 =  HappyAbsSyn50
		 (HsIdent happy_var_1
	)
happyReduction_107 _  = notHappyAtAll 

happyReduce_108 state_indent = happySpecReduce_3  (50, state_indent) happyReduction_108
happyReduction_108 _
	(HappyAbsSyn50  happy_var_2)
	_
	 =  HappyAbsSyn50
		 (happy_var_2
	)
happyReduction_108 _ _ _  = notHappyAtAll 

happyReduce_109 state_indent = happySpecReduce_3  (51, state_indent) happyReduction_109
happyReduction_109 (HappyAbsSyn51  happy_var_3)
	_
	(HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn51
		 (HsTyFun happy_var_1 happy_var_3
	)
happyReduction_109 _ _ _  = notHappyAtAll 

happyReduce_110 state_indent = happySpecReduce_1  (51, state_indent) happyReduction_110
happyReduction_110 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn51
		 (happy_var_1
	)
happyReduction_110 _  = notHappyAtAll 

happyReduce_111 state_indent = happySpecReduce_2  (52, state_indent) happyReduction_111
happyReduction_111 (HappyAbsSyn51  happy_var_2)
	(HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn51
		 (HsTyApp happy_var_1 happy_var_2
	)
happyReduction_111 _ _  = notHappyAtAll 

happyReduce_112 state_indent = happySpecReduce_1  (52, state_indent) happyReduction_112
happyReduction_112 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn51
		 (happy_var_1
	)
happyReduction_112 _  = notHappyAtAll 

happyReduce_113 state_indent = happySpecReduce_1  (53, state_indent) happyReduction_113
happyReduction_113 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn51
		 (HsTyCon happy_var_1
	)
happyReduction_113 _  = notHappyAtAll 

happyReduce_114 state_indent = happySpecReduce_1  (53, state_indent) happyReduction_114
happyReduction_114 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn51
		 (HsTyVar happy_var_1
	)
happyReduction_114 _  = notHappyAtAll 

happyReduce_115 state_indent = happySpecReduce_3  (53, state_indent) happyReduction_115
happyReduction_115 _
	(HappyAbsSyn37  happy_var_2)
	_
	 =  HappyAbsSyn51
		 (HsTyTuple (reverse happy_var_2)
	)
happyReduction_115 _ _ _  = notHappyAtAll 

happyReduce_116 state_indent = happySpecReduce_3  (53, state_indent) happyReduction_116
happyReduction_116 _
	(HappyAbsSyn51  happy_var_2)
	_
	 =  HappyAbsSyn51
		 (HsTyApp list_tycon happy_var_2
	)
happyReduction_116 _ _ _  = notHappyAtAll 

happyReduce_117 state_indent = happySpecReduce_3  (53, state_indent) happyReduction_117
happyReduction_117 _
	(HappyAbsSyn51  happy_var_2)
	_
	 =  HappyAbsSyn51
		 (happy_var_2
	)
happyReduction_117 _ _ _  = notHappyAtAll 

happyReduce_118 state_indent = happySpecReduce_1  (54, state_indent) happyReduction_118
happyReduction_118 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_118 _  = notHappyAtAll 

happyReduce_119 state_indent = happySpecReduce_2  (54, state_indent) happyReduction_119
happyReduction_119 _
	_
	 =  HappyAbsSyn54
		 (unit_tycon_name
	)

happyReduce_120 state_indent = happySpecReduce_3  (54, state_indent) happyReduction_120
happyReduction_120 _
	_
	_
	 =  HappyAbsSyn54
		 (fun_tycon_name
	)

happyReduce_121 state_indent = happySpecReduce_2  (54, state_indent) happyReduction_121
happyReduction_121 _
	_
	 =  HappyAbsSyn54
		 (list_tycon_name
	)

happyReduce_122 state_indent = happySpecReduce_3  (54, state_indent) happyReduction_122
happyReduction_122 _
	(HappyAbsSyn29  happy_var_2)
	_
	 =  HappyAbsSyn54
		 (tuple_tycon_name happy_var_2
	)
happyReduction_122 _ _ _  = notHappyAtAll 

happyReduce_123 state_indent = happySpecReduce_3  (55, state_indent) happyReduction_123
happyReduction_123 (HappyAbsSyn51  happy_var_3)
	_
	(HappyAbsSyn56  happy_var_1)
	 =  HappyAbsSyn55
		 (HsQualType happy_var_1 happy_var_3
	)
happyReduction_123 _ _ _  = notHappyAtAll 

happyReduce_124 state_indent = happySpecReduce_1  (55, state_indent) happyReduction_124
happyReduction_124 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn55
		 (HsQualType [] happy_var_1
	)
happyReduction_124 _  = notHappyAtAll 

happyReduce_125 state_indent = happyMonadReduce 1 (56, state_indent) happyReduction_125
happyReduction_125 ((HappyAbsSyn51  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkContext happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn56 r))

happyReduce_126 state_indent = happySpecReduce_3  (57, state_indent) happyReduction_126
happyReduction_126 (HappyAbsSyn51  happy_var_3)
	_
	(HappyAbsSyn37  happy_var_1)
	 =  HappyAbsSyn37
		 (happy_var_3 : happy_var_1
	)
happyReduction_126 _ _ _  = notHappyAtAll 

happyReduce_127 state_indent = happySpecReduce_3  (57, state_indent) happyReduction_127
happyReduction_127 (HappyAbsSyn51  happy_var_3)
	_
	(HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn37
		 ([happy_var_3, happy_var_1]
	)
happyReduction_127 _ _ _  = notHappyAtAll 

happyReduce_128 state_indent = happySpecReduce_2  (58, state_indent) happyReduction_128
happyReduction_128 (HappyAbsSyn46  happy_var_2)
	(HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn58
		 ((happy_var_1,reverse happy_var_2)
	)
happyReduction_128 _ _  = notHappyAtAll 

happyReduce_129 state_indent = happySpecReduce_2  (59, state_indent) happyReduction_129
happyReduction_129 (HappyAbsSyn50  happy_var_2)
	(HappyAbsSyn46  happy_var_1)
	 =  HappyAbsSyn46
		 (happy_var_2 : happy_var_1
	)
happyReduction_129 _ _  = notHappyAtAll 

happyReduce_130 state_indent = happySpecReduce_0  (59, state_indent) happyReduction_130
happyReduction_130  =  HappyAbsSyn46
		 ([]
	)

happyReduce_131 state_indent = happySpecReduce_3  (60, state_indent) happyReduction_131
happyReduction_131 (HappyAbsSyn61  happy_var_3)
	_
	(HappyAbsSyn60  happy_var_1)
	 =  HappyAbsSyn60
		 (happy_var_3 : happy_var_1
	)
happyReduction_131 _ _ _  = notHappyAtAll 

happyReduce_132 state_indent = happySpecReduce_1  (60, state_indent) happyReduction_132
happyReduction_132 (HappyAbsSyn61  happy_var_1)
	 =  HappyAbsSyn60
		 ([happy_var_1]
	)
happyReduction_132 _  = notHappyAtAll 

happyReduce_133 state_indent = happySpecReduce_2  (61, state_indent) happyReduction_133
happyReduction_133 (HappyAbsSyn62  happy_var_2)
	(HappyAbsSyn5  happy_var_1)
	 =  HappyAbsSyn61
		 (HsConDecl happy_var_1 (fst happy_var_2) (snd happy_var_2)
	)
happyReduction_133 _ _  = notHappyAtAll 

happyReduce_134 state_indent = happyReduce 4 (61, state_indent) happyReduction_134
happyReduction_134 ((HappyAbsSyn64  happy_var_4) `HappyStk`
	(HappyAbsSyn50  happy_var_3) `HappyStk`
	(HappyAbsSyn64  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn61
		 (HsConDecl happy_var_1 happy_var_3 [happy_var_2,happy_var_4]
	) `HappyStk` happyRest

happyReduce_135 state_indent = happyReduce 4 (61, state_indent) happyReduction_135
happyReduction_135 (_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn50  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn61
		 (HsRecDecl happy_var_1 happy_var_2 []
	) `HappyStk` happyRest

happyReduce_136 state_indent = happyReduce 5 (61, state_indent) happyReduction_136
happyReduction_136 (_ `HappyStk`
	(HappyAbsSyn66  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn50  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn61
		 (HsRecDecl happy_var_1 happy_var_2 (reverse happy_var_4)
	) `HappyStk` happyRest

happyReduce_137 state_indent = happyMonadReduce 1 (62, state_indent) happyReduction_137
happyReduction_137 ((HappyAbsSyn51  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { (c,ts) <- splitTyConApp happy_var_1;
						return (c,map HsUnBangedTy ts) })
	) (\r -> happyReturn (HappyAbsSyn62 r))

happyReduce_138 state_indent = happySpecReduce_1  (62, state_indent) happyReduction_138
happyReduction_138 (HappyAbsSyn62  happy_var_1)
	 =  HappyAbsSyn62
		 (happy_var_1
	)
happyReduction_138 _  = notHappyAtAll 

happyReduce_139 state_indent = happyMonadReduce 3 (63, state_indent) happyReduction_139
happyReduction_139 ((HappyAbsSyn51  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn51  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { (c,ts) <- splitTyConApp happy_var_1;
						return (c,map HsUnBangedTy ts++
							[HsBangedTy happy_var_3]) })
	) (\r -> happyReturn (HappyAbsSyn62 r))

happyReduce_140 state_indent = happySpecReduce_2  (63, state_indent) happyReduction_140
happyReduction_140 (HappyAbsSyn64  happy_var_2)
	(HappyAbsSyn62  happy_var_1)
	 =  HappyAbsSyn62
		 ((fst happy_var_1, snd happy_var_1 ++ [happy_var_2] )
	)
happyReduction_140 _ _  = notHappyAtAll 

happyReduce_141 state_indent = happySpecReduce_1  (64, state_indent) happyReduction_141
happyReduction_141 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn64
		 (HsUnBangedTy happy_var_1
	)
happyReduction_141 _  = notHappyAtAll 

happyReduce_142 state_indent = happySpecReduce_2  (64, state_indent) happyReduction_142
happyReduction_142 (HappyAbsSyn51  happy_var_2)
	_
	 =  HappyAbsSyn64
		 (HsBangedTy   happy_var_2
	)
happyReduction_142 _ _  = notHappyAtAll 

happyReduce_143 state_indent = happySpecReduce_1  (65, state_indent) happyReduction_143
happyReduction_143 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn64
		 (HsUnBangedTy happy_var_1
	)
happyReduction_143 _  = notHappyAtAll 

happyReduce_144 state_indent = happySpecReduce_2  (65, state_indent) happyReduction_144
happyReduction_144 (HappyAbsSyn51  happy_var_2)
	_
	 =  HappyAbsSyn64
		 (HsBangedTy   happy_var_2
	)
happyReduction_144 _ _  = notHappyAtAll 

happyReduce_145 state_indent = happySpecReduce_1  (66, state_indent) happyReduction_145
happyReduction_145 (HappyAbsSyn66  happy_var_1)
	 =  HappyAbsSyn66
		 (happy_var_1
	)
happyReduction_145 _  = notHappyAtAll 

happyReduce_146 state_indent = happySpecReduce_3  (67, state_indent) happyReduction_146
happyReduction_146 (HappyAbsSyn68  happy_var_3)
	_
	(HappyAbsSyn66  happy_var_1)
	 =  HappyAbsSyn66
		 (happy_var_3 : happy_var_1
	)
happyReduction_146 _ _ _  = notHappyAtAll 

happyReduce_147 state_indent = happySpecReduce_1  (67, state_indent) happyReduction_147
happyReduction_147 (HappyAbsSyn68  happy_var_1)
	 =  HappyAbsSyn66
		 ([happy_var_1]
	)
happyReduction_147 _  = notHappyAtAll 

happyReduce_148 state_indent = happySpecReduce_3  (68, state_indent) happyReduction_148
happyReduction_148 (HappyAbsSyn64  happy_var_3)
	_
	(HappyAbsSyn46  happy_var_1)
	 =  HappyAbsSyn68
		 ((reverse happy_var_1, happy_var_3)
	)
happyReduction_148 _ _ _  = notHappyAtAll 

happyReduce_149 state_indent = happySpecReduce_1  (69, state_indent) happyReduction_149
happyReduction_149 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn64
		 (HsUnBangedTy happy_var_1
	)
happyReduction_149 _  = notHappyAtAll 

happyReduce_150 state_indent = happySpecReduce_2  (69, state_indent) happyReduction_150
happyReduction_150 (HappyAbsSyn51  happy_var_2)
	_
	 =  HappyAbsSyn64
		 (HsBangedTy   happy_var_2
	)
happyReduction_150 _ _  = notHappyAtAll 

happyReduce_151 state_indent = happySpecReduce_0  (70, state_indent) happyReduction_151
happyReduction_151  =  HappyAbsSyn70
		 ([]
	)

happyReduce_152 state_indent = happySpecReduce_2  (70, state_indent) happyReduction_152
happyReduction_152 (HappyAbsSyn54  happy_var_2)
	_
	 =  HappyAbsSyn70
		 ([happy_var_2]
	)
happyReduction_152 _ _  = notHappyAtAll 

happyReduce_153 state_indent = happySpecReduce_3  (70, state_indent) happyReduction_153
happyReduction_153 _
	_
	_
	 =  HappyAbsSyn70
		 ([]
	)

happyReduce_154 state_indent = happyReduce 4 (70, state_indent) happyReduction_154
happyReduction_154 (_ `HappyStk`
	(HappyAbsSyn70  happy_var_3) `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn70
		 (reverse happy_var_3
	) `HappyStk` happyRest

happyReduce_155 state_indent = happySpecReduce_3  (71, state_indent) happyReduction_155
happyReduction_155 (HappyAbsSyn54  happy_var_3)
	_
	(HappyAbsSyn70  happy_var_1)
	 =  HappyAbsSyn70
		 (happy_var_3 : happy_var_1
	)
happyReduction_155 _ _ _  = notHappyAtAll 

happyReduce_156 state_indent = happySpecReduce_1  (71, state_indent) happyReduction_156
happyReduction_156 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn70
		 ([happy_var_1]
	)
happyReduction_156 _  = notHappyAtAll 

happyReduce_157 state_indent = happyMonadReduce 2 (72, state_indent) happyReduction_157
happyReduction_157 ((HappyAbsSyn32  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkClassBody happy_var_2)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_158 state_indent = happySpecReduce_0  (72, state_indent) happyReduction_158
happyReduction_158  =  HappyAbsSyn32
		 ([]
	)

happyReduce_159 state_indent = happyMonadReduce 4 (73, state_indent) happyReduction_159
happyReduction_159 (_ `HappyStk`
	(HappyAbsSyn32  happy_var_3) `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkClassBody happy_var_3)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_160 state_indent = happyMonadReduce 2 (73, state_indent) happyReduction_160
happyReduction_160 ((HappyAbsSyn32  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_2 >>= checkClassBody)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_161 state_indent = happyMonadReduce 1 (73, state_indent) happyReduction_161
happyReduction_161 (_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls [] >>= checkClassBody)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_162 state_indent = happySpecReduce_0  (73, state_indent) happyReduction_162
happyReduction_162  =  HappyAbsSyn32
		 ([]
	)

happyReduce_163 state_indent = happyMonadReduce 3 (74, state_indent) happyReduction_163
happyReduction_163 (_ `HappyStk`
	(HappyAbsSyn32  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_2)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_164 state_indent = happySpecReduce_1  (74, state_indent) happyReduction_164
happyReduction_164 _
	 =  HappyAbsSyn32
		 ([]
	)

happyReduce_165 state_indent = happySpecReduce_3  (75, state_indent) happyReduction_165
happyReduction_165 (HappyAbsSyn28  happy_var_3)
	_
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_3 : happy_var_1
	)
happyReduction_165 _ _ _  = notHappyAtAll 

happyReduce_166 state_indent = happySpecReduce_1  (75, state_indent) happyReduction_166
happyReduction_166 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn32
		 ([happy_var_1]
	)
happyReduction_166 _  = notHappyAtAll 

happyReduce_167 state_indent = happySpecReduce_3  (76, state_indent) happyReduction_167
happyReduction_167 _
	(HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn32
		 (happy_var_2
	)
happyReduction_167 _ _ _  = notHappyAtAll 

happyReduce_168 state_indent = happySpecReduce_1  (76, state_indent) happyReduction_168
happyReduction_168 _
	 =  HappyAbsSyn32
		 ([]
	)

happyReduce_169 state_indent = happySpecReduce_1  (77, state_indent) happyReduction_169
happyReduction_169 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_169 _  = notHappyAtAll 

happyReduce_170 state_indent = happySpecReduce_2  (78, state_indent) happyReduction_170
happyReduction_170 (HappyAbsSyn32  happy_var_2)
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_170 _ _  = notHappyAtAll 

happyReduce_171 state_indent = happySpecReduce_1  (78, state_indent) happyReduction_171
happyReduction_171 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_171 _  = notHappyAtAll 

happyReduce_172 state_indent = happyMonadReduce 4 (79, state_indent) happyReduction_172
happyReduction_172 ((HappyAbsSyn32  happy_var_4) `HappyStk`
	(HappyAbsSyn81  happy_var_3) `HappyStk`
	(HappyAbsSyn84  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkValDef happy_var_1 happy_var_2 happy_var_3 happy_var_4)
	) (\r -> happyReturn (HappyAbsSyn28 r))

happyReduce_173 state_indent = happySpecReduce_2  (80, state_indent) happyReduction_173
happyReduction_173 (HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn32
		 (happy_var_2
	)
happyReduction_173 _ _  = notHappyAtAll 

happyReduce_174 state_indent = happySpecReduce_0  (80, state_indent) happyReduction_174
happyReduction_174  =  HappyAbsSyn32
		 ([]
	)

happyReduce_175 state_indent = happyMonadReduce 2 (81, state_indent) happyReduction_175
happyReduction_175 ((HappyAbsSyn84  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( do { e <- checkExpr happy_var_2;
						return (HsUnGuardedRhs e) })
	) (\r -> happyReturn (HappyAbsSyn81 r))

happyReduce_176 state_indent = happySpecReduce_1  (81, state_indent) happyReduction_176
happyReduction_176 (HappyAbsSyn82  happy_var_1)
	 =  HappyAbsSyn81
		 (HsGuardedRhss  (reverse happy_var_1)
	)
happyReduction_176 _  = notHappyAtAll 

happyReduce_177 state_indent = happySpecReduce_2  (82, state_indent) happyReduction_177
happyReduction_177 (HappyAbsSyn83  happy_var_2)
	(HappyAbsSyn82  happy_var_1)
	 =  HappyAbsSyn82
		 (happy_var_2 : happy_var_1
	)
happyReduction_177 _ _  = notHappyAtAll 

happyReduce_178 state_indent = happySpecReduce_1  (82, state_indent) happyReduction_178
happyReduction_178 (HappyAbsSyn83  happy_var_1)
	 =  HappyAbsSyn82
		 ([happy_var_1]
	)
happyReduction_178 _  = notHappyAtAll 

happyReduce_179 state_indent = happyMonadReduce 5 (83, state_indent) happyReduction_179
happyReduction_179 ((HappyAbsSyn84  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { g <- checkExpr happy_var_3;
						e <- checkExpr happy_var_5;
						return (HsGuardedRhs happy_var_1 g e) })
	) (\r -> happyReturn (HappyAbsSyn83 r))

happyReduce_180 state_indent = happyReduce 4 (84, state_indent) happyReduction_180
happyReduction_180 ((HappyAbsSyn55  happy_var_4) `HappyStk`
	(HappyAbsSyn5  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsExpTypeSig happy_var_3 happy_var_1 happy_var_4
	) `HappyStk` happyRest

happyReduce_181 state_indent = happySpecReduce_1  (84, state_indent) happyReduction_181
happyReduction_181 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_181 _  = notHappyAtAll 

happyReduce_182 state_indent = happySpecReduce_1  (85, state_indent) happyReduction_182
happyReduction_182 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_182 _  = notHappyAtAll 

happyReduce_183 state_indent = happySpecReduce_1  (85, state_indent) happyReduction_183
happyReduction_183 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_183 _  = notHappyAtAll 

happyReduce_184 state_indent = happySpecReduce_3  (86, state_indent) happyReduction_184
happyReduction_184 (HappyAbsSyn84  happy_var_3)
	(HappyAbsSyn132  happy_var_2)
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsInfixApp happy_var_1 happy_var_2 happy_var_3
	)
happyReduction_184 _ _ _  = notHappyAtAll 

happyReduce_185 state_indent = happySpecReduce_1  (86, state_indent) happyReduction_185
happyReduction_185 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_185 _  = notHappyAtAll 

happyReduce_186 state_indent = happySpecReduce_3  (87, state_indent) happyReduction_186
happyReduction_186 (HappyAbsSyn84  happy_var_3)
	(HappyAbsSyn132  happy_var_2)
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsInfixApp happy_var_1 happy_var_2 happy_var_3
	)
happyReduction_186 _ _ _  = notHappyAtAll 

happyReduce_187 state_indent = happySpecReduce_1  (87, state_indent) happyReduction_187
happyReduction_187 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_187 _  = notHappyAtAll 

happyReduce_188 state_indent = happyReduce 5 (88, state_indent) happyReduction_188
happyReduction_188 ((HappyAbsSyn84  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn91  happy_var_3) `HappyStk`
	(HappyAbsSyn5  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsLambda happy_var_2 (reverse happy_var_3) happy_var_5
	) `HappyStk` happyRest

happyReduce_189 state_indent = happyReduce 4 (88, state_indent) happyReduction_189
happyReduction_189 ((HappyAbsSyn84  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn32  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsLet happy_var_2 happy_var_4
	) `HappyStk` happyRest

happyReduce_190 state_indent = happyReduce 6 (88, state_indent) happyReduction_190
happyReduction_190 ((HappyAbsSyn84  happy_var_6) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsIf happy_var_2 happy_var_4 happy_var_6
	) `HappyStk` happyRest

happyReduce_191 state_indent = happyReduce 4 (89, state_indent) happyReduction_191
happyReduction_191 ((HappyAbsSyn102  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsCase happy_var_2 happy_var_4
	) `HappyStk` happyRest

happyReduce_192 state_indent = happySpecReduce_2  (89, state_indent) happyReduction_192
happyReduction_192 (HappyAbsSyn84  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (HsNegApp happy_var_2
	)
happyReduction_192 _ _  = notHappyAtAll 

happyReduce_193 state_indent = happySpecReduce_2  (89, state_indent) happyReduction_193
happyReduction_193 (HappyAbsSyn100  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (HsDo happy_var_2
	)
happyReduction_193 _ _  = notHappyAtAll 

happyReduce_194 state_indent = happySpecReduce_1  (89, state_indent) happyReduction_194
happyReduction_194 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_194 _  = notHappyAtAll 

happyReduce_195 state_indent = happySpecReduce_2  (90, state_indent) happyReduction_195
happyReduction_195 (HappyAbsSyn84  happy_var_2)
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsApp happy_var_1 happy_var_2
	)
happyReduction_195 _ _  = notHappyAtAll 

happyReduce_196 state_indent = happySpecReduce_1  (90, state_indent) happyReduction_196
happyReduction_196 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_196 _  = notHappyAtAll 

happyReduce_197 state_indent = happySpecReduce_2  (91, state_indent) happyReduction_197
happyReduction_197 (HappyAbsSyn92  happy_var_2)
	(HappyAbsSyn91  happy_var_1)
	 =  HappyAbsSyn91
		 (happy_var_2 : happy_var_1
	)
happyReduction_197 _ _  = notHappyAtAll 

happyReduce_198 state_indent = happySpecReduce_1  (91, state_indent) happyReduction_198
happyReduction_198 (HappyAbsSyn92  happy_var_1)
	 =  HappyAbsSyn91
		 ([happy_var_1]
	)
happyReduction_198 _  = notHappyAtAll 

happyReduce_199 state_indent = happyMonadReduce 1 (92, state_indent) happyReduction_199
happyReduction_199 ((HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkPattern happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn92 r))

happyReduce_200 state_indent = happyMonadReduce 3 (93, state_indent) happyReduction_200
happyReduction_200 ((HappyAbsSyn84  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn54  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { n <- checkUnQual happy_var_1;
						return (HsAsPat n happy_var_3) })
	) (\r -> happyReturn (HappyAbsSyn84 r))

happyReduce_201 state_indent = happySpecReduce_2  (93, state_indent) happyReduction_201
happyReduction_201 (HappyAbsSyn84  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (HsIrrPat happy_var_2
	)
happyReduction_201 _ _  = notHappyAtAll 

happyReduce_202 state_indent = happySpecReduce_1  (93, state_indent) happyReduction_202
happyReduction_202 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_202 _  = notHappyAtAll 

happyReduce_203 state_indent = happyMonadReduce 3 (94, state_indent) happyReduction_203
happyReduction_203 (_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( mkRecConstrOrUpdate happy_var_1 [])
	) (\r -> happyReturn (HappyAbsSyn84 r))

happyReduce_204 state_indent = happyMonadReduce 4 (94, state_indent) happyReduction_204
happyReduction_204 (_ `HappyStk`
	(HappyAbsSyn118  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( mkRecConstrOrUpdate happy_var_1 (reverse happy_var_3))
	) (\r -> happyReturn (HappyAbsSyn84 r))

happyReduce_205 state_indent = happySpecReduce_1  (94, state_indent) happyReduction_205
happyReduction_205 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_205 _  = notHappyAtAll 

happyReduce_206 state_indent = happySpecReduce_1  (95, state_indent) happyReduction_206
happyReduction_206 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn84
		 (HsVar happy_var_1
	)
happyReduction_206 _  = notHappyAtAll 

happyReduce_207 state_indent = happySpecReduce_1  (95, state_indent) happyReduction_207
happyReduction_207 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_207 _  = notHappyAtAll 

happyReduce_208 state_indent = happySpecReduce_1  (95, state_indent) happyReduction_208
happyReduction_208 (HappyAbsSyn146  happy_var_1)
	 =  HappyAbsSyn84
		 (HsLit happy_var_1
	)
happyReduction_208 _  = notHappyAtAll 

happyReduce_209 state_indent = happySpecReduce_3  (95, state_indent) happyReduction_209
happyReduction_209 _
	(HappyAbsSyn84  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (HsParen happy_var_2
	)
happyReduction_209 _ _ _  = notHappyAtAll 

happyReduce_210 state_indent = happySpecReduce_3  (95, state_indent) happyReduction_210
happyReduction_210 _
	(HappyAbsSyn97  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (HsTuple (reverse happy_var_2)
	)
happyReduction_210 _ _ _  = notHappyAtAll 

happyReduce_211 state_indent = happySpecReduce_3  (95, state_indent) happyReduction_211
happyReduction_211 _
	(HappyAbsSyn84  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (happy_var_2
	)
happyReduction_211 _ _ _  = notHappyAtAll 

happyReduce_212 state_indent = happyReduce 4 (95, state_indent) happyReduction_212
happyReduction_212 (_ `HappyStk`
	(HappyAbsSyn132  happy_var_3) `HappyStk`
	(HappyAbsSyn84  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsLeftSection happy_var_2 happy_var_3
	) `HappyStk` happyRest

happyReduce_213 state_indent = happyReduce 4 (95, state_indent) happyReduction_213
happyReduction_213 (_ `HappyStk`
	(HappyAbsSyn84  happy_var_3) `HappyStk`
	(HappyAbsSyn132  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsRightSection happy_var_2 happy_var_3
	) `HappyStk` happyRest

happyReduce_214 state_indent = happySpecReduce_1  (95, state_indent) happyReduction_214
happyReduction_214 _
	 =  HappyAbsSyn84
		 (HsWildCard
	)

happyReduce_215 state_indent = happySpecReduce_2  (96, state_indent) happyReduction_215
happyReduction_215 _
	(HappyAbsSyn29  happy_var_1)
	 =  HappyAbsSyn29
		 (happy_var_1 + 1
	)
happyReduction_215 _ _  = notHappyAtAll 

happyReduce_216 state_indent = happySpecReduce_1  (96, state_indent) happyReduction_216
happyReduction_216 _
	 =  HappyAbsSyn29
		 (1
	)

happyReduce_217 state_indent = happySpecReduce_3  (97, state_indent) happyReduction_217
happyReduction_217 (HappyAbsSyn84  happy_var_3)
	_
	(HappyAbsSyn97  happy_var_1)
	 =  HappyAbsSyn97
		 (happy_var_3 : happy_var_1
	)
happyReduction_217 _ _ _  = notHappyAtAll 

happyReduce_218 state_indent = happySpecReduce_3  (97, state_indent) happyReduction_218
happyReduction_218 (HappyAbsSyn84  happy_var_3)
	_
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn97
		 ([happy_var_3,happy_var_1]
	)
happyReduction_218 _ _ _  = notHappyAtAll 

happyReduce_219 state_indent = happySpecReduce_1  (98, state_indent) happyReduction_219
happyReduction_219 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsList [happy_var_1]
	)
happyReduction_219 _  = notHappyAtAll 

happyReduce_220 state_indent = happySpecReduce_1  (98, state_indent) happyReduction_220
happyReduction_220 (HappyAbsSyn97  happy_var_1)
	 =  HappyAbsSyn84
		 (HsList (reverse happy_var_1)
	)
happyReduction_220 _  = notHappyAtAll 

happyReduce_221 state_indent = happySpecReduce_2  (98, state_indent) happyReduction_221
happyReduction_221 _
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsEnumFrom happy_var_1
	)
happyReduction_221 _ _  = notHappyAtAll 

happyReduce_222 state_indent = happyReduce 4 (98, state_indent) happyReduction_222
happyReduction_222 (_ `HappyStk`
	(HappyAbsSyn84  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsEnumFromThen happy_var_1 happy_var_3
	) `HappyStk` happyRest

happyReduce_223 state_indent = happySpecReduce_3  (98, state_indent) happyReduction_223
happyReduction_223 (HappyAbsSyn84  happy_var_3)
	_
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsEnumFromTo happy_var_1 happy_var_3
	)
happyReduction_223 _ _ _  = notHappyAtAll 

happyReduce_224 state_indent = happyReduce 5 (98, state_indent) happyReduction_224
happyReduction_224 ((HappyAbsSyn84  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsEnumFromThenTo happy_var_1 happy_var_3 happy_var_5
	) `HappyStk` happyRest

happyReduce_225 state_indent = happySpecReduce_3  (98, state_indent) happyReduction_225
happyReduction_225 (HappyAbsSyn100  happy_var_3)
	_
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsListComp happy_var_1 (reverse happy_var_3)
	)
happyReduction_225 _ _ _  = notHappyAtAll 

happyReduce_226 state_indent = happySpecReduce_3  (99, state_indent) happyReduction_226
happyReduction_226 (HappyAbsSyn84  happy_var_3)
	_
	(HappyAbsSyn97  happy_var_1)
	 =  HappyAbsSyn97
		 (happy_var_3 : happy_var_1
	)
happyReduction_226 _ _ _  = notHappyAtAll 

happyReduce_227 state_indent = happySpecReduce_3  (99, state_indent) happyReduction_227
happyReduction_227 (HappyAbsSyn84  happy_var_3)
	_
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn97
		 ([happy_var_3,happy_var_1]
	)
happyReduction_227 _ _ _  = notHappyAtAll 

happyReduce_228 state_indent = happySpecReduce_3  (100, state_indent) happyReduction_228
happyReduction_228 (HappyAbsSyn101  happy_var_3)
	_
	(HappyAbsSyn100  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_3 : happy_var_1
	)
happyReduction_228 _ _ _  = notHappyAtAll 

happyReduce_229 state_indent = happySpecReduce_1  (100, state_indent) happyReduction_229
happyReduction_229 (HappyAbsSyn101  happy_var_1)
	 =  HappyAbsSyn100
		 ([happy_var_1]
	)
happyReduction_229 _  = notHappyAtAll 

happyReduce_230 state_indent = happyReduce 4 (101, state_indent) happyReduction_230
happyReduction_230 ((HappyAbsSyn84  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_2) `HappyStk`
	(HappyAbsSyn92  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn101
		 (HsGenerator happy_var_2 happy_var_1 happy_var_4
	) `HappyStk` happyRest

happyReduce_231 state_indent = happySpecReduce_1  (101, state_indent) happyReduction_231
happyReduction_231 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn101
		 (HsQualifier happy_var_1
	)
happyReduction_231 _  = notHappyAtAll 

happyReduce_232 state_indent = happySpecReduce_2  (101, state_indent) happyReduction_232
happyReduction_232 (HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn101
		 (HsLetStmt happy_var_2
	)
happyReduction_232 _ _  = notHappyAtAll 

happyReduce_233 state_indent = happySpecReduce_3  (102, state_indent) happyReduction_233
happyReduction_233 _
	(HappyAbsSyn102  happy_var_2)
	_
	 =  HappyAbsSyn102
		 (happy_var_2
	)
happyReduction_233 _ _ _  = notHappyAtAll 

happyReduce_234 state_indent = happySpecReduce_1  (102, state_indent) happyReduction_234
happyReduction_234 (HappyAbsSyn102  happy_var_1)
	 =  HappyAbsSyn102
		 (reverse happy_var_1
	)
happyReduction_234 _  = notHappyAtAll 

happyReduce_235 state_indent = happySpecReduce_1  (103, state_indent) happyReduction_235
happyReduction_235 (HappyAbsSyn102  happy_var_1)
	 =  HappyAbsSyn102
		 (happy_var_1
	)
happyReduction_235 _  = notHappyAtAll 

happyReduce_236 state_indent = happySpecReduce_2  (104, state_indent) happyReduction_236
happyReduction_236 (HappyAbsSyn102  happy_var_2)
	(HappyAbsSyn102  happy_var_1)
	 =  HappyAbsSyn102
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_236 _ _  = notHappyAtAll 

happyReduce_237 state_indent = happySpecReduce_1  (104, state_indent) happyReduction_237
happyReduction_237 (HappyAbsSyn102  happy_var_1)
	 =  HappyAbsSyn102
		 (happy_var_1
	)
happyReduction_237 _  = notHappyAtAll 

happyReduce_238 state_indent = happySpecReduce_3  (105, state_indent) happyReduction_238
happyReduction_238 _
	(HappyAbsSyn102  happy_var_2)
	_
	 =  HappyAbsSyn102
		 (reverse happy_var_2
	)
happyReduction_238 _ _ _  = notHappyAtAll 

happyReduce_239 state_indent = happySpecReduce_3  (106, state_indent) happyReduction_239
happyReduction_239 (HappyAbsSyn107  happy_var_3)
	_
	(HappyAbsSyn102  happy_var_1)
	 =  HappyAbsSyn102
		 (happy_var_3 : happy_var_1
	)
happyReduction_239 _ _ _  = notHappyAtAll 

happyReduce_240 state_indent = happySpecReduce_1  (106, state_indent) happyReduction_240
happyReduction_240 (HappyAbsSyn107  happy_var_1)
	 =  HappyAbsSyn102
		 ([happy_var_1]
	)
happyReduction_240 _  = notHappyAtAll 

happyReduce_241 state_indent = happyReduce 4 (107, state_indent) happyReduction_241
happyReduction_241 ((HappyAbsSyn32  happy_var_4) `HappyStk`
	(HappyAbsSyn108  happy_var_3) `HappyStk`
	(HappyAbsSyn92  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn107
		 (HsAlt happy_var_1 happy_var_2 happy_var_3 happy_var_4
	) `HappyStk` happyRest

happyReduce_242 state_indent = happySpecReduce_2  (108, state_indent) happyReduction_242
happyReduction_242 (HappyAbsSyn84  happy_var_2)
	_
	 =  HappyAbsSyn108
		 (HsUnGuardedAlt happy_var_2
	)
happyReduction_242 _ _  = notHappyAtAll 

happyReduce_243 state_indent = happySpecReduce_1  (108, state_indent) happyReduction_243
happyReduction_243 (HappyAbsSyn109  happy_var_1)
	 =  HappyAbsSyn108
		 (HsGuardedAlts (reverse happy_var_1)
	)
happyReduction_243 _  = notHappyAtAll 

happyReduce_244 state_indent = happySpecReduce_2  (109, state_indent) happyReduction_244
happyReduction_244 (HappyAbsSyn110  happy_var_2)
	(HappyAbsSyn109  happy_var_1)
	 =  HappyAbsSyn109
		 (happy_var_2 : happy_var_1
	)
happyReduction_244 _ _  = notHappyAtAll 

happyReduce_245 state_indent = happySpecReduce_1  (109, state_indent) happyReduction_245
happyReduction_245 (HappyAbsSyn110  happy_var_1)
	 =  HappyAbsSyn109
		 ([happy_var_1]
	)
happyReduction_245 _  = notHappyAtAll 

happyReduce_246 state_indent = happyReduce 5 (110, state_indent) happyReduction_246
happyReduction_246 ((HappyAbsSyn84  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn110
		 (HsGuardedAlt happy_var_1 happy_var_3 happy_var_5
	) `HappyStk` happyRest

happyReduce_247 state_indent = happyMonadReduce 1 (111, state_indent) happyReduction_247
happyReduction_247 ((HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkPattern happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn92 r))

happyReduce_248 state_indent = happySpecReduce_3  (112, state_indent) happyReduction_248
happyReduction_248 _
	(HappyAbsSyn100  happy_var_2)
	_
	 =  HappyAbsSyn100
		 (happy_var_2
	)
happyReduction_248 _ _ _  = notHappyAtAll 

happyReduce_249 state_indent = happySpecReduce_1  (112, state_indent) happyReduction_249
happyReduction_249 (HappyAbsSyn100  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_1
	)
happyReduction_249 _  = notHappyAtAll 

happyReduce_250 state_indent = happySpecReduce_1  (113, state_indent) happyReduction_250
happyReduction_250 (HappyAbsSyn100  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_1
	)
happyReduction_250 _  = notHappyAtAll 

happyReduce_251 state_indent = happySpecReduce_2  (114, state_indent) happyReduction_251
happyReduction_251 (HappyAbsSyn100  happy_var_2)
	(HappyAbsSyn100  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_1 ++ happy_var_2
	)
happyReduction_251 _ _  = notHappyAtAll 

happyReduce_252 state_indent = happySpecReduce_1  (114, state_indent) happyReduction_252
happyReduction_252 (HappyAbsSyn100  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_1
	)
happyReduction_252 _  = notHappyAtAll 

happyReduce_253 state_indent = happySpecReduce_3  (115, state_indent) happyReduction_253
happyReduction_253 (HappyAbsSyn100  happy_var_3)
	_
	(HappyAbsSyn101  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_1 : happy_var_3
	)
happyReduction_253 _ _ _  = notHappyAtAll 

happyReduce_254 state_indent = happySpecReduce_2  (115, state_indent) happyReduction_254
happyReduction_254 (HappyAbsSyn100  happy_var_2)
	_
	 =  HappyAbsSyn100
		 (happy_var_2
	)
happyReduction_254 _ _  = notHappyAtAll 

happyReduce_255 state_indent = happySpecReduce_1  (115, state_indent) happyReduction_255
happyReduction_255 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn100
		 ([HsQualifier happy_var_1]
	)
happyReduction_255 _  = notHappyAtAll 

happyReduce_256 state_indent = happySpecReduce_3  (116, state_indent) happyReduction_256
happyReduction_256 (HappyAbsSyn100  happy_var_3)
	_
	(HappyAbsSyn101  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_1 : happy_var_3
	)
happyReduction_256 _ _ _  = notHappyAtAll 

happyReduce_257 state_indent = happySpecReduce_2  (116, state_indent) happyReduction_257
happyReduction_257 (HappyAbsSyn100  happy_var_2)
	_
	 =  HappyAbsSyn100
		 (happy_var_2
	)
happyReduction_257 _ _  = notHappyAtAll 

happyReduce_258 state_indent = happySpecReduce_2  (116, state_indent) happyReduction_258
happyReduction_258 _
	(HappyAbsSyn101  happy_var_1)
	 =  HappyAbsSyn100
		 ([happy_var_1]
	)
happyReduction_258 _ _  = notHappyAtAll 

happyReduce_259 state_indent = happySpecReduce_1  (116, state_indent) happyReduction_259
happyReduction_259 (HappyAbsSyn101  happy_var_1)
	 =  HappyAbsSyn100
		 ([happy_var_1]
	)
happyReduction_259 _  = notHappyAtAll 

happyReduce_260 state_indent = happySpecReduce_2  (117, state_indent) happyReduction_260
happyReduction_260 (HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn101
		 (HsLetStmt happy_var_2
	)
happyReduction_260 _ _  = notHappyAtAll 

happyReduce_261 state_indent = happyReduce 4 (117, state_indent) happyReduction_261
happyReduction_261 ((HappyAbsSyn84  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_2) `HappyStk`
	(HappyAbsSyn92  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn101
		 (HsGenerator happy_var_2 happy_var_1 happy_var_4
	) `HappyStk` happyRest

happyReduce_262 state_indent = happySpecReduce_1  (117, state_indent) happyReduction_262
happyReduction_262 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn101
		 (HsQualifier happy_var_1
	)
happyReduction_262 _  = notHappyAtAll 

happyReduce_263 state_indent = happySpecReduce_1  (118, state_indent) happyReduction_263
happyReduction_263 (HappyAbsSyn118  happy_var_1)
	 =  HappyAbsSyn118
		 (happy_var_1
	)
happyReduction_263 _  = notHappyAtAll 

happyReduce_264 state_indent = happySpecReduce_3  (119, state_indent) happyReduction_264
happyReduction_264 (HappyAbsSyn120  happy_var_3)
	_
	(HappyAbsSyn118  happy_var_1)
	 =  HappyAbsSyn118
		 (happy_var_3 : happy_var_1
	)
happyReduction_264 _ _ _  = notHappyAtAll 

happyReduce_265 state_indent = happySpecReduce_1  (119, state_indent) happyReduction_265
happyReduction_265 (HappyAbsSyn120  happy_var_1)
	 =  HappyAbsSyn118
		 ([happy_var_1]
	)
happyReduction_265 _  = notHappyAtAll 

happyReduce_266 state_indent = happySpecReduce_3  (120, state_indent) happyReduction_266
happyReduction_266 (HappyAbsSyn84  happy_var_3)
	_
	(HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn120
		 (HsFieldUpdate happy_var_1 happy_var_3
	)
happyReduction_266 _ _ _  = notHappyAtAll 

happyReduce_267 state_indent = happySpecReduce_2  (121, state_indent) happyReduction_267
happyReduction_267 _
	_
	 =  HappyAbsSyn84
		 (unit_con
	)

happyReduce_268 state_indent = happySpecReduce_2  (121, state_indent) happyReduction_268
happyReduction_268 _
	_
	 =  HappyAbsSyn84
		 (HsList []
	)

happyReduce_269 state_indent = happySpecReduce_3  (121, state_indent) happyReduction_269
happyReduction_269 _
	(HappyAbsSyn29  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (tuple_con happy_var_2
	)
happyReduction_269 _ _ _  = notHappyAtAll 

happyReduce_270 state_indent = happySpecReduce_1  (121, state_indent) happyReduction_270
happyReduction_270 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn84
		 (HsCon happy_var_1
	)
happyReduction_270 _  = notHappyAtAll 

happyReduce_271 state_indent = happySpecReduce_1  (122, state_indent) happyReduction_271
happyReduction_271 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_271 _  = notHappyAtAll 

happyReduce_272 state_indent = happySpecReduce_3  (122, state_indent) happyReduction_272
happyReduction_272 _
	(HappyAbsSyn50  happy_var_2)
	_
	 =  HappyAbsSyn50
		 (happy_var_2
	)
happyReduction_272 _ _ _  = notHappyAtAll 

happyReduce_273 state_indent = happySpecReduce_1  (123, state_indent) happyReduction_273
happyReduction_273 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_273 _  = notHappyAtAll 

happyReduce_274 state_indent = happySpecReduce_3  (123, state_indent) happyReduction_274
happyReduction_274 _
	(HappyAbsSyn54  happy_var_2)
	_
	 =  HappyAbsSyn54
		 (happy_var_2
	)
happyReduction_274 _ _ _  = notHappyAtAll 

happyReduce_275 state_indent = happySpecReduce_1  (124, state_indent) happyReduction_275
happyReduction_275 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_275 _  = notHappyAtAll 

happyReduce_276 state_indent = happySpecReduce_3  (124, state_indent) happyReduction_276
happyReduction_276 _
	(HappyAbsSyn50  happy_var_2)
	_
	 =  HappyAbsSyn50
		 (happy_var_2
	)
happyReduction_276 _ _ _  = notHappyAtAll 

happyReduce_277 state_indent = happySpecReduce_1  (125, state_indent) happyReduction_277
happyReduction_277 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_277 _  = notHappyAtAll 

happyReduce_278 state_indent = happySpecReduce_3  (125, state_indent) happyReduction_278
happyReduction_278 _
	(HappyAbsSyn54  happy_var_2)
	_
	 =  HappyAbsSyn54
		 (happy_var_2
	)
happyReduction_278 _ _ _  = notHappyAtAll 

happyReduce_279 state_indent = happySpecReduce_1  (126, state_indent) happyReduction_279
happyReduction_279 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_279 _  = notHappyAtAll 

happyReduce_280 state_indent = happySpecReduce_3  (126, state_indent) happyReduction_280
happyReduction_280 _
	(HappyAbsSyn50  happy_var_2)
	_
	 =  HappyAbsSyn50
		 (happy_var_2
	)
happyReduction_280 _ _ _  = notHappyAtAll 

happyReduce_281 state_indent = happySpecReduce_1  (127, state_indent) happyReduction_281
happyReduction_281 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_281 _  = notHappyAtAll 

happyReduce_282 state_indent = happySpecReduce_3  (127, state_indent) happyReduction_282
happyReduction_282 _
	(HappyAbsSyn54  happy_var_2)
	_
	 =  HappyAbsSyn54
		 (happy_var_2
	)
happyReduction_282 _ _ _  = notHappyAtAll 

happyReduce_283 state_indent = happySpecReduce_1  (128, state_indent) happyReduction_283
happyReduction_283 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_283 _  = notHappyAtAll 

happyReduce_284 state_indent = happySpecReduce_3  (128, state_indent) happyReduction_284
happyReduction_284 _
	(HappyAbsSyn54  happy_var_2)
	_
	 =  HappyAbsSyn54
		 (happy_var_2
	)
happyReduction_284 _ _ _  = notHappyAtAll 

happyReduce_285 state_indent = happySpecReduce_1  (129, state_indent) happyReduction_285
happyReduction_285 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_285 _  = notHappyAtAll 

happyReduce_286 state_indent = happySpecReduce_3  (129, state_indent) happyReduction_286
happyReduction_286 _
	(HappyAbsSyn50  happy_var_2)
	_
	 =  HappyAbsSyn50
		 (happy_var_2
	)
happyReduction_286 _ _ _  = notHappyAtAll 

happyReduce_287 state_indent = happySpecReduce_1  (130, state_indent) happyReduction_287
happyReduction_287 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_287 _  = notHappyAtAll 

happyReduce_288 state_indent = happySpecReduce_3  (130, state_indent) happyReduction_288
happyReduction_288 _
	(HappyAbsSyn54  happy_var_2)
	_
	 =  HappyAbsSyn54
		 (happy_var_2
	)
happyReduction_288 _ _ _  = notHappyAtAll 

happyReduce_289 state_indent = happySpecReduce_1  (131, state_indent) happyReduction_289
happyReduction_289 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn131
		 (HsVarOp happy_var_1
	)
happyReduction_289 _  = notHappyAtAll 

happyReduce_290 state_indent = happySpecReduce_1  (131, state_indent) happyReduction_290
happyReduction_290 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn131
		 (HsConOp happy_var_1
	)
happyReduction_290 _  = notHappyAtAll 

happyReduce_291 state_indent = happySpecReduce_1  (132, state_indent) happyReduction_291
happyReduction_291 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn132
		 (HsQVarOp happy_var_1
	)
happyReduction_291 _  = notHappyAtAll 

happyReduce_292 state_indent = happySpecReduce_1  (132, state_indent) happyReduction_292
happyReduction_292 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn132
		 (HsQConOp happy_var_1
	)
happyReduction_292 _  = notHappyAtAll 

happyReduce_293 state_indent = happySpecReduce_1  (133, state_indent) happyReduction_293
happyReduction_293 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn132
		 (HsQVarOp happy_var_1
	)
happyReduction_293 _  = notHappyAtAll 

happyReduce_294 state_indent = happySpecReduce_1  (133, state_indent) happyReduction_294
happyReduction_294 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn132
		 (HsQConOp happy_var_1
	)
happyReduction_294 _  = notHappyAtAll 

happyReduce_295 state_indent = happySpecReduce_1  (134, state_indent) happyReduction_295
happyReduction_295 _
	 =  HappyAbsSyn54
		 (list_cons_name
	)

happyReduce_296 state_indent = happySpecReduce_1  (134, state_indent) happyReduction_296
happyReduction_296 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_296 _  = notHappyAtAll 

happyReduce_297 state_indent = happySpecReduce_1  (135, state_indent) happyReduction_297
happyReduction_297 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn54
		 (UnQual happy_var_1
	)
happyReduction_297 _  = notHappyAtAll 

happyReduce_298 state_indent = happySpecReduce_1  (135, state_indent) happyReduction_298
happyReduction_298 (HappyTerminal ((QVarId happy_var_1), _))
	 =  HappyAbsSyn54
		 (Qual (Module (fst happy_var_1)) (HsIdent (snd happy_var_1))
	)
happyReduction_298 _  = notHappyAtAll 

happyReduce_299 state_indent = happySpecReduce_1  (136, state_indent) happyReduction_299
happyReduction_299 (HappyTerminal ((VarId happy_var_1), _))
	 =  HappyAbsSyn50
		 (HsIdent happy_var_1
	)
happyReduction_299 _  = notHappyAtAll 

happyReduce_300 state_indent = happySpecReduce_1  (136, state_indent) happyReduction_300
happyReduction_300 _
	 =  HappyAbsSyn50
		 (HsIdent "as"
	)

happyReduce_301 state_indent = happySpecReduce_1  (136, state_indent) happyReduction_301
happyReduction_301 _
	 =  HappyAbsSyn50
		 (HsIdent "export"
	)

happyReduce_302 state_indent = happySpecReduce_1  (136, state_indent) happyReduction_302
happyReduction_302 _
	 =  HappyAbsSyn50
		 (HsIdent "hiding"
	)

happyReduce_303 state_indent = happySpecReduce_1  (136, state_indent) happyReduction_303
happyReduction_303 _
	 =  HappyAbsSyn50
		 (HsIdent "qualified"
	)

happyReduce_304 state_indent = happySpecReduce_1  (136, state_indent) happyReduction_304
happyReduction_304 _
	 =  HappyAbsSyn50
		 (HsIdent "safe"
	)

happyReduce_305 state_indent = happySpecReduce_1  (136, state_indent) happyReduction_305
happyReduction_305 _
	 =  HappyAbsSyn50
		 (HsIdent "unsafe"
	)

happyReduce_306 state_indent = happySpecReduce_1  (137, state_indent) happyReduction_306
happyReduction_306 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn54
		 (UnQual happy_var_1
	)
happyReduction_306 _  = notHappyAtAll 

happyReduce_307 state_indent = happySpecReduce_1  (137, state_indent) happyReduction_307
happyReduction_307 (HappyTerminal ((QConId happy_var_1), _))
	 =  HappyAbsSyn54
		 (Qual (Module (fst happy_var_1)) (HsIdent (snd happy_var_1))
	)
happyReduction_307 _  = notHappyAtAll 

happyReduce_308 state_indent = happySpecReduce_1  (138, state_indent) happyReduction_308
happyReduction_308 (HappyTerminal ((ConId happy_var_1), _))
	 =  HappyAbsSyn50
		 (HsIdent happy_var_1
	)
happyReduction_308 _  = notHappyAtAll 

happyReduce_309 state_indent = happySpecReduce_1  (139, state_indent) happyReduction_309
happyReduction_309 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn54
		 (UnQual happy_var_1
	)
happyReduction_309 _  = notHappyAtAll 

happyReduce_310 state_indent = happySpecReduce_1  (139, state_indent) happyReduction_310
happyReduction_310 (HappyTerminal ((QConSym happy_var_1), _))
	 =  HappyAbsSyn54
		 (Qual (Module (fst happy_var_1)) (HsSymbol (snd happy_var_1))
	)
happyReduction_310 _  = notHappyAtAll 

happyReduce_311 state_indent = happySpecReduce_1  (140, state_indent) happyReduction_311
happyReduction_311 (HappyTerminal ((ConSym happy_var_1), _))
	 =  HappyAbsSyn50
		 (HsSymbol happy_var_1
	)
happyReduction_311 _  = notHappyAtAll 

happyReduce_312 state_indent = happySpecReduce_1  (141, state_indent) happyReduction_312
happyReduction_312 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn54
		 (UnQual happy_var_1
	)
happyReduction_312 _  = notHappyAtAll 

happyReduce_313 state_indent = happySpecReduce_1  (141, state_indent) happyReduction_313
happyReduction_313 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_313 _  = notHappyAtAll 

happyReduce_314 state_indent = happySpecReduce_1  (142, state_indent) happyReduction_314
happyReduction_314 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn54
		 (UnQual happy_var_1
	)
happyReduction_314 _  = notHappyAtAll 

happyReduce_315 state_indent = happySpecReduce_1  (142, state_indent) happyReduction_315
happyReduction_315 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_315 _  = notHappyAtAll 

happyReduce_316 state_indent = happySpecReduce_1  (143, state_indent) happyReduction_316
happyReduction_316 (HappyTerminal ((VarSym happy_var_1), _))
	 =  HappyAbsSyn50
		 (HsSymbol happy_var_1
	)
happyReduction_316 _  = notHappyAtAll 

happyReduce_317 state_indent = happySpecReduce_1  (143, state_indent) happyReduction_317
happyReduction_317 _
	 =  HappyAbsSyn50
		 (HsSymbol "-"
	)

happyReduce_318 state_indent = happySpecReduce_1  (143, state_indent) happyReduction_318
happyReduction_318 _
	 =  HappyAbsSyn50
		 (HsSymbol "!"
	)

happyReduce_319 state_indent = happySpecReduce_1  (144, state_indent) happyReduction_319
happyReduction_319 (HappyTerminal ((VarSym happy_var_1), _))
	 =  HappyAbsSyn50
		 (HsSymbol happy_var_1
	)
happyReduction_319 _  = notHappyAtAll 

happyReduce_320 state_indent = happySpecReduce_1  (144, state_indent) happyReduction_320
happyReduction_320 _
	 =  HappyAbsSyn50
		 (HsSymbol "!"
	)

happyReduce_321 state_indent = happySpecReduce_1  (145, state_indent) happyReduction_321
happyReduction_321 (HappyTerminal ((QVarSym happy_var_1), _))
	 =  HappyAbsSyn54
		 (Qual (Module (fst happy_var_1)) (HsSymbol (snd happy_var_1))
	)
happyReduction_321 _  = notHappyAtAll 

happyReduce_322 state_indent = happySpecReduce_1  (146, state_indent) happyReduction_322
happyReduction_322 (HappyTerminal ((IntTok happy_var_1), _))
	 =  HappyAbsSyn146
		 (HsInt happy_var_1
	)
happyReduction_322 _  = notHappyAtAll 

happyReduce_323 state_indent = happySpecReduce_1  (146, state_indent) happyReduction_323
happyReduction_323 (HappyTerminal ((Character happy_var_1), _))
	 =  HappyAbsSyn146
		 (HsChar happy_var_1
	)
happyReduction_323 _  = notHappyAtAll 

happyReduce_324 state_indent = happySpecReduce_1  (146, state_indent) happyReduction_324
happyReduction_324 (HappyTerminal ((FloatTok happy_var_1), _))
	 =  HappyAbsSyn146
		 (HsFrac happy_var_1
	)
happyReduction_324 _  = notHappyAtAll 

happyReduce_325 state_indent = happySpecReduce_1  (146, state_indent) happyReduction_325
happyReduction_325 (HappyTerminal ((StringTok happy_var_1), _))
	 =  HappyAbsSyn146
		 (HsString happy_var_1
	)
happyReduction_325 _  = notHappyAtAll 

happyReduce_326 state_indent = happySpecReduce_1  (147, state_indent) happyReduction_326
happyReduction_326 _
	 =  HappyAbsSyn9
		 (()
	)

happyReduce_327 state_indent = happySpecReduce_1  (148, state_indent) happyReduction_327
happyReduction_327 (HappyTerminal ((ConId happy_var_1), _))
	 =  HappyAbsSyn148
		 (Module happy_var_1
	)
happyReduction_327 _  = notHappyAtAll 

happyReduce_328 state_indent = happySpecReduce_1  (148, state_indent) happyReduction_328
happyReduction_328 (HappyTerminal ((QConId happy_var_1), _))
	 =  HappyAbsSyn148
		 (Module (fst happy_var_1 ++ '.':snd happy_var_1)
	)
happyReduction_328 _  = notHappyAtAll 

happyReduce_329 state_indent = happySpecReduce_1  (149, state_indent) happyReduction_329
happyReduction_329 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_329 _  = notHappyAtAll 

happyReduce_330 state_indent = happySpecReduce_1  (150, state_indent) happyReduction_330
happyReduction_330 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_330 _  = notHappyAtAll 

happyReduce_331 state_indent = happySpecReduce_1  (151, state_indent) happyReduction_331
happyReduction_331 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_331 _  = notHappyAtAll 

happyReduce_332 state_indent = happySpecReduce_1  (152, state_indent) happyReduction_332
happyReduction_332 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_332 _  = notHappyAtAll 

happyReduce_333 state_indent = happySpecReduce_1  (153, state_indent) happyReduction_333
happyReduction_333 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_333 _  = notHappyAtAll 

happyReduce_334 state_indent = happySpecReduce_3  (154, state_indent) happyReduction_334
happyReduction_334 _
	(HappyAbsSyn16  happy_var_2)
	(HappyAbsSyn16  happy_var_1)
	 =  HappyAbsSyn16
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_334 _ _ _  = notHappyAtAll 

happyReduce_335 state_indent = happySpecReduce_2  (154, state_indent) happyReduction_335
happyReduction_335 _
	(HappyAbsSyn16  happy_var_1)
	 =  HappyAbsSyn16
		 (happy_var_1
	)
happyReduction_335 _ _  = notHappyAtAll 

happyReduce_336 state_indent = happySpecReduce_3  (155, state_indent) happyReduction_336
happyReduction_336 (HappyAbsSyn16  happy_var_3)
	_
	(HappyAbsSyn16  happy_var_1)
	 =  HappyAbsSyn16
		 (happy_var_3 ++ happy_var_1
	)
happyReduction_336 _ _ _  = notHappyAtAll 

happyReduce_337 state_indent = happySpecReduce_1  (155, state_indent) happyReduction_337
happyReduction_337 (HappyAbsSyn16  happy_var_1)
	 =  HappyAbsSyn16
		 (happy_var_1
	)
happyReduction_337 _  = notHappyAtAll 

happyReduce_338 state_indent = happyMonadReduce 1 (156, state_indent) happyReduction_338
happyReduction_338 ((HappyAbsSyn32  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_339 state_indent = happySpecReduce_3  (157, state_indent) happyReduction_339
happyReduction_339 (HappyAbsSyn32  happy_var_3)
	_
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_3 ++ happy_var_1
	)
happyReduction_339 _ _ _  = notHappyAtAll 

happyReduce_340 state_indent = happySpecReduce_1  (157, state_indent) happyReduction_340
happyReduction_340 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_340 _  = notHappyAtAll 

happyReduce_341 state_indent = happySpecReduce_3  (158, state_indent) happyReduction_341
happyReduction_341 _
	(HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn32
		 (happy_var_2
	)
happyReduction_341 _ _ _  = notHappyAtAll 

happyReduce_342 state_indent = happySpecReduce_3  (158, state_indent) happyReduction_342
happyReduction_342 _
	(HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn32
		 (happy_var_2
	)
happyReduction_342 _ _ _  = notHappyAtAll 

happyReduce_343 state_indent = happySpecReduce_1  (158, state_indent) happyReduction_343
happyReduction_343 _
	 =  HappyAbsSyn32
		 ([]
	)

happyReduce_344 state_indent = happySpecReduce_3  (159, state_indent) happyReduction_344
happyReduction_344 _
	(HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn32
		 (happy_var_2
	)
happyReduction_344 _ _ _  = notHappyAtAll 

happyReduce_345 state_indent = happySpecReduce_3  (159, state_indent) happyReduction_345
happyReduction_345 _
	(HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn32
		 (happy_var_2
	)
happyReduction_345 _ _ _  = notHappyAtAll 

happyReduce_346 state_indent = happySpecReduce_1  (159, state_indent) happyReduction_346
happyReduction_346 _
	 =  HappyAbsSyn32
		 ([]
	)

happyReduce_347 state_indent = happySpecReduce_3  (160, state_indent) happyReduction_347
happyReduction_347 (HappyAbsSyn107  happy_var_3)
	_
	(HappyAbsSyn102  happy_var_1)
	 =  HappyAbsSyn102
		 (happy_var_3 : happy_var_1
	)
happyReduction_347 _ _ _  = notHappyAtAll 

happyReduce_348 state_indent = happySpecReduce_1  (160, state_indent) happyReduction_348
happyReduction_348 (HappyAbsSyn107  happy_var_1)
	 =  HappyAbsSyn102
		 ([happy_var_1]
	)
happyReduction_348 _  = notHappyAtAll 

happyReduce_349 state_indent = happySpecReduce_3  (161, state_indent) happyReduction_349
happyReduction_349 (HappyAbsSyn100  happy_var_3)
	_
	(HappyAbsSyn101  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_1 : happy_var_3
	)
happyReduction_349 _ _ _  = notHappyAtAll 

happyReduce_350 state_indent = happySpecReduce_2  (161, state_indent) happyReduction_350
happyReduction_350 (HappyAbsSyn100  happy_var_2)
	_
	 =  HappyAbsSyn100
		 (happy_var_2
	)
happyReduction_350 _ _  = notHappyAtAll 

happyReduce_351 state_indent = happySpecReduce_1  (161, state_indent) happyReduction_351
happyReduction_351 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn100
		 ([HsQualifier happy_var_1]
	)
happyReduction_351 _  = notHappyAtAll 

happyReduce_352 state_indent = happySpecReduce_3  (162, state_indent) happyReduction_352
happyReduction_352 (HappyAbsSyn100  happy_var_3)
	_
	(HappyAbsSyn101  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_1 : happy_var_3
	)
happyReduction_352 _ _ _  = notHappyAtAll 

happyReduce_353 state_indent = happySpecReduce_2  (162, state_indent) happyReduction_353
happyReduction_353 (HappyAbsSyn100  happy_var_2)
	_
	 =  HappyAbsSyn100
		 (happy_var_2
	)
happyReduction_353 _ _  = notHappyAtAll 

happyReduce_354 state_indent = happySpecReduce_2  (162, state_indent) happyReduction_354
happyReduction_354 _
	(HappyAbsSyn101  happy_var_1)
	 =  HappyAbsSyn100
		 ([happy_var_1]
	)
happyReduction_354 _ _  = notHappyAtAll 

happyReduce_355 state_indent = happySpecReduce_1  (162, state_indent) happyReduction_355
happyReduction_355 (HappyAbsSyn101  happy_var_1)
	 =  HappyAbsSyn100
		 ([happy_var_1]
	)
happyReduction_355 _  = notHappyAtAll 

happyReduce_356 state_indent = happySpecReduce_2  (163, state_indent) happyReduction_356
happyReduction_356 _
	_
	 =  HappyAbsSyn9
		 (()
	)

happyReduce_357 state_indent = happySpecReduce_2  (163, state_indent) happyReduction_357
happyReduction_357 _
	_
	 =  HappyAbsSyn9
		 (()
	)

happyReduce_358 state_indent = happySpecReduce_1  (164, state_indent) happyReduction_358
happyReduction_358 _
	 =  HappyAbsSyn9
		 (()
	)

happyReduce_359 state_indent = happyReduce 6 (165, state_indent) happyReduction_359
happyReduction_359 ((HappyAbsSyn21  happy_var_6) `HappyStk`
	(HappyAbsSyn20  happy_var_5) `HappyStk`
	(HappyAbsSyn148  happy_var_4) `HappyStk`
	(HappyAbsSyn19  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn16
		 ([HsImportDecl happy_var_1 happy_var_4 happy_var_3 happy_var_5 happy_var_6]
	) `HappyStk` happyRest

happyReduce_360 state_indent = happySpecReduce_3  (166, state_indent) happyReduction_360
happyReduction_360 _
	(HappyAbsSyn32  happy_var_2)
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_360 _ _ _  = notHappyAtAll 

happyReduce_361 state_indent = happySpecReduce_2  (166, state_indent) happyReduction_361
happyReduction_361 _
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_361 _ _  = notHappyAtAll 

happyReduce_362 state_indent = happyReduce 5 (167, state_indent) happyReduction_362
happyReduction_362 ((HappyAbsSyn51  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn58  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn32
		 ([HsTypeDecl happy_var_1 (fst happy_var_3) (snd happy_var_3) happy_var_5]
	) `HappyStk` happyRest

happyReduce_363 state_indent = happyMonadReduce 6 (167, state_indent) happyReduction_363
happyReduction_363 ((HappyAbsSyn70  happy_var_6) `HappyStk`
	(HappyAbsSyn60  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn55  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { (cs,c,t) <- checkDataHeader happy_var_3;
				return ([HsDataDecl happy_var_1 cs c t (reverse happy_var_5) happy_var_6]) })
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_364 state_indent = happyMonadReduce 6 (167, state_indent) happyReduction_364
happyReduction_364 ((HappyAbsSyn70  happy_var_6) `HappyStk`
	(HappyAbsSyn61  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn55  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { (cs,c,t) <- checkDataHeader happy_var_3;
				return ([HsNewTypeDecl happy_var_1 cs c t happy_var_5 happy_var_6]) })
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_365 state_indent = happyMonadReduce 4 (167, state_indent) happyReduction_365
happyReduction_365 ((HappyAbsSyn32  happy_var_4) `HappyStk`
	(HappyAbsSyn55  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { (cs,c,vs) <- checkClassHeader happy_var_3;
				return ([HsClassDecl happy_var_1 cs c vs happy_var_4]) })
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_366 state_indent = happyMonadReduce 4 (167, state_indent) happyReduction_366
happyReduction_366 ((HappyAbsSyn32  happy_var_4) `HappyStk`
	(HappyAbsSyn55  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { (cs,c,ts) <- checkInstHeader happy_var_3;
				return ([HsInstDecl happy_var_1 cs c ts happy_var_4]) })
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_367 state_indent = happyReduce 5 (167, state_indent) happyReduction_367
happyReduction_367 (_ `HappyStk`
	(HappyAbsSyn37  happy_var_4) `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn32
		 ([HsDefaultDecl happy_var_1 happy_var_4]
	) `HappyStk` happyRest

happyReduce_368 state_indent = happySpecReduce_1  (167, state_indent) happyReduction_368
happyReduction_368 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn32
		 ([happy_var_1]
	)
happyReduction_368 _  = notHappyAtAll 

happyReduce_369 state_indent = happySpecReduce_1  (167, state_indent) happyReduction_369
happyReduction_369 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn32
		 ([happy_var_1]
	)
happyReduction_369 _  = notHappyAtAll 

happyReduce_370 state_indent = happySpecReduce_3  (168, state_indent) happyReduction_370
happyReduction_370 (HappyAbsSyn28  happy_var_3)
	_
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_3 : happy_var_1
	)
happyReduction_370 _ _ _  = notHappyAtAll 

happyReduce_371 state_indent = happySpecReduce_1  (168, state_indent) happyReduction_371
happyReduction_371 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn32
		 ([happy_var_1]
	)
happyReduction_371 _  = notHappyAtAll 

happyReduce_372 state_indent = happySpecReduce_3  (169, state_indent) happyReduction_372
happyReduction_372 (HappyAbsSyn28  happy_var_3)
	_
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_3 : happy_var_1
	)
happyReduction_372 _ _ _  = notHappyAtAll 

happyReduce_373 state_indent = happySpecReduce_1  (169, state_indent) happyReduction_373
happyReduction_373 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn32
		 ([happy_var_1]
	)
happyReduction_373 _  = notHappyAtAll 

happyReduce_374 state_indent = happyReduce 4 (170, state_indent) happyReduction_374
happyReduction_374 ((HappyAbsSyn55  happy_var_4) `HappyStk`
	(HappyAbsSyn5  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsExpTypeSig happy_var_3 happy_var_1 happy_var_4
	) `HappyStk` happyRest

happyReduce_375 state_indent = happySpecReduce_1  (170, state_indent) happyReduction_375
happyReduction_375 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_375 _  = notHappyAtAll 

happyReduce_376 state_indent = happyReduce 4 (171, state_indent) happyReduction_376
happyReduction_376 ((HappyAbsSyn32  happy_var_4) `HappyStk`
	(HappyAbsSyn108  happy_var_3) `HappyStk`
	(HappyAbsSyn92  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn107
		 (HsAlt happy_var_1 happy_var_2 happy_var_3 happy_var_4
	) `HappyStk` happyRest

happyReduce_377 state_indent = happySpecReduce_2  (172, state_indent) happyReduction_377
happyReduction_377 (HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn101
		 (HsLetStmt happy_var_2
	)
happyReduction_377 _ _  = notHappyAtAll 

happyReduce_378 state_indent = happyReduce 4 (172, state_indent) happyReduction_378
happyReduction_378 ((HappyAbsSyn84  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_2) `HappyStk`
	(HappyAbsSyn92  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn101
		 (HsGenerator happy_var_2 happy_var_1 happy_var_4
	) `HappyStk` happyRest

happyReduce_379 state_indent = happySpecReduce_1  (172, state_indent) happyReduction_379
happyReduction_379 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn101
		 (HsQualifier happy_var_1
	)
happyReduction_379 _  = notHappyAtAll 

happyReduce_380 state_indent = happySpecReduce_1  (173, state_indent) happyReduction_380
happyReduction_380 _
	 =  HappyAbsSyn19
		 (True
	)

happyReduce_381 state_indent = happySpecReduce_2  (174, state_indent) happyReduction_381
happyReduction_381 (HappyAbsSyn148  happy_var_2)
	_
	 =  HappyAbsSyn20
		 (Just happy_var_2
	)
happyReduction_381 _ _  = notHappyAtAll 

happyReduce_382 state_indent = happySpecReduce_1  (175, state_indent) happyReduction_382
happyReduction_382 (HappyAbsSyn22  happy_var_1)
	 =  HappyAbsSyn21
		 (Just happy_var_1
	)
happyReduction_382 _  = notHappyAtAll 

happyReduce_383 state_indent = happySpecReduce_1  (176, state_indent) happyReduction_383
happyReduction_383 (HappyAbsSyn37  happy_var_1)
	 =  HappyAbsSyn37
		 (reverse happy_var_1
	)
happyReduction_383 _  = notHappyAtAll 

happyReduce_384 state_indent = happySpecReduce_1  (176, state_indent) happyReduction_384
happyReduction_384 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn37
		 ([happy_var_1]
	)
happyReduction_384 _  = notHappyAtAll 

happyReduce_385 state_indent = happySpecReduce_1  (177, state_indent) happyReduction_385
happyReduction_385 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn28
		 (happy_var_1
	)
happyReduction_385 _  = notHappyAtAll 

happyReduce_386 state_indent = happySpecReduce_1  (177, state_indent) happyReduction_386
happyReduction_386 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn28
		 (happy_var_1
	)
happyReduction_386 _  = notHappyAtAll 

happyReduce_387 state_indent = happySpecReduce_1  (177, state_indent) happyReduction_387
happyReduction_387 (HappyAbsSyn28  happy_var_1)
	 =  HappyAbsSyn28
		 (happy_var_1
	)
happyReduction_387 _  = notHappyAtAll 

happyReduce_388 state_indent = happySpecReduce_3  (178, state_indent) happyReduction_388
happyReduction_388 _
	(HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn32
		 (happy_var_2
	)
happyReduction_388 _ _ _  = notHappyAtAll 

happyReduce_389 state_indent = happyMonadReduce 1 (178, state_indent) happyReduction_389
happyReduction_389 ((HappyAbsSyn32  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_390 state_indent = happyReduce 9 (179, state_indent) happyReduction_390
happyReduction_390 ((HappyAbsSyn51  happy_var_9) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn50  happy_var_7) `HappyStk`
	(HappyAbsSyn49  happy_var_6) `HappyStk`
	(HappyAbsSyn48  happy_var_5) `HappyStk`
	(HappyTerminal ((VarId happy_var_4), _)) `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn28
		 (HsForeignImport happy_var_1 happy_var_4 happy_var_5 happy_var_6 happy_var_7 happy_var_9
	) `HappyStk` happyRest

happyReduce_391 state_indent = happyReduce 8 (179, state_indent) happyReduction_391
happyReduction_391 ((HappyAbsSyn51  happy_var_8) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn50  happy_var_6) `HappyStk`
	(HappyAbsSyn49  happy_var_5) `HappyStk`
	(HappyTerminal ((VarId happy_var_4), _)) `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn28
		 (HsForeignExport happy_var_1 happy_var_4 happy_var_5 happy_var_6 happy_var_8
	) `HappyStk` happyRest

happyReduce_392 state_indent = happySpecReduce_3  (180, state_indent) happyReduction_392
happyReduction_392 (HappyAbsSyn51  happy_var_3)
	_
	(HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn51
		 (HsTyFun happy_var_1 happy_var_3
	)
happyReduction_392 _ _ _  = notHappyAtAll 

happyReduce_393 state_indent = happySpecReduce_1  (180, state_indent) happyReduction_393
happyReduction_393 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn51
		 (happy_var_1
	)
happyReduction_393 _  = notHappyAtAll 

happyReduce_394 state_indent = happySpecReduce_3  (181, state_indent) happyReduction_394
happyReduction_394 (HappyAbsSyn51  happy_var_3)
	_
	(HappyAbsSyn56  happy_var_1)
	 =  HappyAbsSyn55
		 (HsQualType happy_var_1 happy_var_3
	)
happyReduction_394 _ _ _  = notHappyAtAll 

happyReduce_395 state_indent = happySpecReduce_1  (181, state_indent) happyReduction_395
happyReduction_395 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn55
		 (HsQualType [] happy_var_1
	)
happyReduction_395 _  = notHappyAtAll 

happyReduce_396 state_indent = happySpecReduce_2  (182, state_indent) happyReduction_396
happyReduction_396 (HappyAbsSyn46  happy_var_2)
	(HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn58
		 ((happy_var_1,reverse happy_var_2)
	)
happyReduction_396 _ _  = notHappyAtAll 

happyReduce_397 state_indent = happySpecReduce_3  (183, state_indent) happyReduction_397
happyReduction_397 (HappyAbsSyn61  happy_var_3)
	_
	(HappyAbsSyn60  happy_var_1)
	 =  HappyAbsSyn60
		 (happy_var_3 : happy_var_1
	)
happyReduction_397 _ _ _  = notHappyAtAll 

happyReduce_398 state_indent = happySpecReduce_1  (183, state_indent) happyReduction_398
happyReduction_398 (HappyAbsSyn61  happy_var_1)
	 =  HappyAbsSyn60
		 ([happy_var_1]
	)
happyReduction_398 _  = notHappyAtAll 

happyReduce_399 state_indent = happySpecReduce_2  (184, state_indent) happyReduction_399
happyReduction_399 (HappyAbsSyn62  happy_var_2)
	(HappyAbsSyn5  happy_var_1)
	 =  HappyAbsSyn61
		 (HsConDecl happy_var_1 (fst happy_var_2) (snd happy_var_2)
	)
happyReduction_399 _ _  = notHappyAtAll 

happyReduce_400 state_indent = happyReduce 4 (184, state_indent) happyReduction_400
happyReduction_400 ((HappyAbsSyn64  happy_var_4) `HappyStk`
	(HappyAbsSyn50  happy_var_3) `HappyStk`
	(HappyAbsSyn64  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn61
		 (HsConDecl happy_var_1 happy_var_3 [happy_var_2,happy_var_4]
	) `HappyStk` happyRest

happyReduce_401 state_indent = happyReduce 4 (184, state_indent) happyReduction_401
happyReduction_401 (_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn50  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn61
		 (HsRecDecl happy_var_1 happy_var_2 []
	) `HappyStk` happyRest

happyReduce_402 state_indent = happyReduce 5 (184, state_indent) happyReduction_402
happyReduction_402 (_ `HappyStk`
	(HappyAbsSyn66  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn50  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn61
		 (HsRecDecl happy_var_1 happy_var_2 (reverse happy_var_4)
	) `HappyStk` happyRest

happyReduce_403 state_indent = happySpecReduce_2  (185, state_indent) happyReduction_403
happyReduction_403 (HappyAbsSyn54  happy_var_2)
	_
	 =  HappyAbsSyn70
		 ([happy_var_2]
	)
happyReduction_403 _ _  = notHappyAtAll 

happyReduce_404 state_indent = happySpecReduce_3  (185, state_indent) happyReduction_404
happyReduction_404 _
	_
	_
	 =  HappyAbsSyn70
		 ([]
	)

happyReduce_405 state_indent = happyReduce 4 (185, state_indent) happyReduction_405
happyReduction_405 (_ `HappyStk`
	(HappyAbsSyn70  happy_var_3) `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn70
		 (reverse happy_var_3
	) `HappyStk` happyRest

happyReduce_406 state_indent = happyMonadReduce 2 (186, state_indent) happyReduction_406
happyReduction_406 ((HappyAbsSyn32  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkClassBody happy_var_2)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_407 state_indent = happyMonadReduce 4 (187, state_indent) happyReduction_407
happyReduction_407 (_ `HappyStk`
	(HappyAbsSyn32  happy_var_3) `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkClassBody happy_var_3)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_408 state_indent = happyMonadReduce 2 (187, state_indent) happyReduction_408
happyReduction_408 ((HappyAbsSyn32  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_2 >>= checkClassBody)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_409 state_indent = happyMonadReduce 1 (187, state_indent) happyReduction_409
happyReduction_409 (_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls [] >>= checkClassBody)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_410 state_indent = happyMonadReduce 4 (188, state_indent) happyReduction_410
happyReduction_410 ((HappyAbsSyn32  happy_var_4) `HappyStk`
	(HappyAbsSyn81  happy_var_3) `HappyStk`
	(HappyAbsSyn84  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkValDef happy_var_1 happy_var_2 happy_var_3 happy_var_4)
	) (\r -> happyReturn (HappyAbsSyn28 r))

happyReduce_411 state_indent = happySpecReduce_2  (189, state_indent) happyReduction_411
happyReduction_411 (HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn32
		 (happy_var_2
	)
happyReduction_411 _ _  = notHappyAtAll 

happyReduce_412 state_indent = happySpecReduce_1  (190, state_indent) happyReduction_412
happyReduction_412 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_412 _  = notHappyAtAll 

happyReduce_413 state_indent = happySpecReduce_1  (190, state_indent) happyReduction_413
happyReduction_413 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_413 _  = notHappyAtAll 

happyReduce_414 state_indent = happySpecReduce_3  (191, state_indent) happyReduction_414
happyReduction_414 (HappyAbsSyn84  happy_var_3)
	(HappyAbsSyn132  happy_var_2)
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsInfixApp happy_var_1 happy_var_2 happy_var_3
	)
happyReduction_414 _ _ _  = notHappyAtAll 

happyReduce_415 state_indent = happySpecReduce_1  (191, state_indent) happyReduction_415
happyReduction_415 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_415 _  = notHappyAtAll 

happyReduce_416 state_indent = happySpecReduce_2  (192, state_indent) happyReduction_416
happyReduction_416 (HappyAbsSyn84  happy_var_2)
	_
	 =  HappyAbsSyn108
		 (HsUnGuardedAlt happy_var_2
	)
happyReduction_416 _ _  = notHappyAtAll 

happyReduce_417 state_indent = happySpecReduce_1  (192, state_indent) happyReduction_417
happyReduction_417 (HappyAbsSyn109  happy_var_1)
	 =  HappyAbsSyn108
		 (HsGuardedAlts (reverse happy_var_1)
	)
happyReduction_417 _  = notHappyAtAll 

happyReduce_418 state_indent = happyMonadReduce 1 (193, state_indent) happyReduction_418
happyReduction_418 ((HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkPattern happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn92 r))

happyReduce_419 state_indent = happySpecReduce_1  (194, state_indent) happyReduction_419
happyReduction_419 (HappyTerminal ((ConId happy_var_1), _))
	 =  HappyAbsSyn148
		 (Module happy_var_1
	)
happyReduction_419 _  = notHappyAtAll 

happyReduce_420 state_indent = happySpecReduce_1  (194, state_indent) happyReduction_420
happyReduction_420 (HappyTerminal ((QConId happy_var_1), _))
	 =  HappyAbsSyn148
		 (Module (fst happy_var_1 ++ '.':snd happy_var_1)
	)
happyReduction_420 _  = notHappyAtAll 

happyReduce_421 state_indent = happyReduce 5 (195, state_indent) happyReduction_421
happyReduction_421 (_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn24  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn19  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn22
		 ((happy_var_1, reverse happy_var_3)
	) `HappyStk` happyRest

happyReduce_422 state_indent = happyReduce 5 (195, state_indent) happyReduction_422
happyReduction_422 (_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn24  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn19  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn22
		 ((happy_var_1, reverse happy_var_3)
	) `HappyStk` happyRest

happyReduce_423 state_indent = happyReduce 4 (195, state_indent) happyReduction_423
happyReduction_423 (_ `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn19  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn22
		 ((happy_var_1, [])
	) `HappyStk` happyRest

happyReduce_424 state_indent = happyReduce 4 (195, state_indent) happyReduction_424
happyReduction_424 (_ `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn19  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn22
		 ((happy_var_1, [])
	) `HappyStk` happyRest

happyReduce_425 state_indent = happyReduce 4 (196, state_indent) happyReduction_425
happyReduction_425 ((HappyAbsSyn31  happy_var_4) `HappyStk`
	(HappyAbsSyn29  happy_var_3) `HappyStk`
	(HappyAbsSyn30  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn28
		 (HsInfixDecl happy_var_1 happy_var_2 happy_var_3 (reverse happy_var_4)
	) `HappyStk` happyRest

happyReduce_426 state_indent = happyMonadReduce 3 (197, state_indent) happyReduction_426
happyReduction_426 (_ `HappyStk`
	(HappyAbsSyn32  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_2)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_427 state_indent = happyMonadReduce 3 (197, state_indent) happyReduction_427
happyReduction_427 (_ `HappyStk`
	(HappyAbsSyn32  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_2)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_428 state_indent = happySpecReduce_1  (197, state_indent) happyReduction_428
happyReduction_428 _
	 =  HappyAbsSyn32
		 ([]
	)

happyReduce_429 state_indent = happySpecReduce_1  (198, state_indent) happyReduction_429
happyReduction_429 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_429 _  = notHappyAtAll 

happyReduce_430 state_indent = happyReduce 4 (199, state_indent) happyReduction_430
happyReduction_430 ((HappyAbsSyn55  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn46  happy_var_2) `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn28
		 (HsTypeSig happy_var_1 (reverse happy_var_2) happy_var_4
	) `HappyStk` happyRest

happyReduce_431 state_indent = happySpecReduce_1  (200, state_indent) happyReduction_431
happyReduction_431 _
	 =  HappyAbsSyn48
		 (HsSafe
	)

happyReduce_432 state_indent = happySpecReduce_1  (200, state_indent) happyReduction_432
happyReduction_432 _
	 =  HappyAbsSyn48
		 (HsUnsafe
	)

happyReduce_433 state_indent = happySpecReduce_1  (201, state_indent) happyReduction_433
happyReduction_433 (HappyTerminal ((StringTok happy_var_1), _))
	 =  HappyAbsSyn49
		 (happy_var_1
	)
happyReduction_433 _  = notHappyAtAll 

happyReduce_434 state_indent = happySpecReduce_1  (202, state_indent) happyReduction_434
happyReduction_434 (HappyTerminal ((VarId happy_var_1), _))
	 =  HappyAbsSyn50
		 (HsIdent happy_var_1
	)
happyReduction_434 _  = notHappyAtAll 

happyReduce_435 state_indent = happySpecReduce_3  (202, state_indent) happyReduction_435
happyReduction_435 _
	(HappyAbsSyn50  happy_var_2)
	_
	 =  HappyAbsSyn50
		 (happy_var_2
	)
happyReduction_435 _ _ _  = notHappyAtAll 

happyReduce_436 state_indent = happySpecReduce_2  (203, state_indent) happyReduction_436
happyReduction_436 (HappyAbsSyn51  happy_var_2)
	(HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn51
		 (HsTyApp happy_var_1 happy_var_2
	)
happyReduction_436 _ _  = notHappyAtAll 

happyReduce_437 state_indent = happySpecReduce_1  (203, state_indent) happyReduction_437
happyReduction_437 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn51
		 (happy_var_1
	)
happyReduction_437 _  = notHappyAtAll 

happyReduce_438 state_indent = happyMonadReduce 1 (204, state_indent) happyReduction_438
happyReduction_438 ((HappyAbsSyn51  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkContext happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn56 r))

happyReduce_439 state_indent = happySpecReduce_3  (205, state_indent) happyReduction_439
happyReduction_439 (HappyAbsSyn51  happy_var_3)
	_
	(HappyAbsSyn37  happy_var_1)
	 =  HappyAbsSyn37
		 (happy_var_3 : happy_var_1
	)
happyReduction_439 _ _ _  = notHappyAtAll 

happyReduce_440 state_indent = happySpecReduce_3  (205, state_indent) happyReduction_440
happyReduction_440 (HappyAbsSyn51  happy_var_3)
	_
	(HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn37
		 ([happy_var_3, happy_var_1]
	)
happyReduction_440 _ _ _  = notHappyAtAll 

happyReduce_441 state_indent = happySpecReduce_2  (206, state_indent) happyReduction_441
happyReduction_441 (HappyAbsSyn50  happy_var_2)
	(HappyAbsSyn46  happy_var_1)
	 =  HappyAbsSyn46
		 (happy_var_2 : happy_var_1
	)
happyReduction_441 _ _  = notHappyAtAll 

happyReduce_442 state_indent = happySpecReduce_2  (206, state_indent) happyReduction_442
happyReduction_442 (HappyAbsSyn50  happy_var_2)
	(HappyAbsSyn46  happy_var_1)
	 =  HappyAbsSyn46
		 (happy_var_2 : happy_var_1
	)
happyReduction_442 _ _  = notHappyAtAll 

happyReduce_443 state_indent = happyMonadReduce 1 (207, state_indent) happyReduction_443
happyReduction_443 ((HappyAbsSyn51  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { (c,ts) <- splitTyConApp happy_var_1;
						return (c,map HsUnBangedTy ts) })
	) (\r -> happyReturn (HappyAbsSyn62 r))

happyReduce_444 state_indent = happySpecReduce_1  (207, state_indent) happyReduction_444
happyReduction_444 (HappyAbsSyn62  happy_var_1)
	 =  HappyAbsSyn62
		 (happy_var_1
	)
happyReduction_444 _  = notHappyAtAll 

happyReduce_445 state_indent = happySpecReduce_1  (208, state_indent) happyReduction_445
happyReduction_445 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn64
		 (HsUnBangedTy happy_var_1
	)
happyReduction_445 _  = notHappyAtAll 

happyReduce_446 state_indent = happySpecReduce_2  (208, state_indent) happyReduction_446
happyReduction_446 (HappyAbsSyn51  happy_var_2)
	_
	 =  HappyAbsSyn64
		 (HsBangedTy   happy_var_2
	)
happyReduction_446 _ _  = notHappyAtAll 

happyReduce_447 state_indent = happySpecReduce_1  (209, state_indent) happyReduction_447
happyReduction_447 (HappyAbsSyn66  happy_var_1)
	 =  HappyAbsSyn66
		 (happy_var_1
	)
happyReduction_447 _  = notHappyAtAll 

happyReduce_448 state_indent = happySpecReduce_3  (210, state_indent) happyReduction_448
happyReduction_448 (HappyAbsSyn54  happy_var_3)
	_
	(HappyAbsSyn70  happy_var_1)
	 =  HappyAbsSyn70
		 (happy_var_3 : happy_var_1
	)
happyReduction_448 _ _ _  = notHappyAtAll 

happyReduce_449 state_indent = happySpecReduce_1  (210, state_indent) happyReduction_449
happyReduction_449 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn70
		 ([happy_var_1]
	)
happyReduction_449 _  = notHappyAtAll 

happyReduce_450 state_indent = happyMonadReduce 3 (211, state_indent) happyReduction_450
happyReduction_450 (_ `HappyStk`
	(HappyAbsSyn32  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_2)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_451 state_indent = happyMonadReduce 3 (211, state_indent) happyReduction_451
happyReduction_451 (_ `HappyStk`
	(HappyAbsSyn32  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_2)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_452 state_indent = happySpecReduce_1  (211, state_indent) happyReduction_452
happyReduction_452 _
	 =  HappyAbsSyn32
		 ([]
	)

happyReduce_453 state_indent = happySpecReduce_1  (212, state_indent) happyReduction_453
happyReduction_453 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_453 _  = notHappyAtAll 

happyReduce_454 state_indent = happyMonadReduce 2 (213, state_indent) happyReduction_454
happyReduction_454 ((HappyAbsSyn84  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest) tk
	 = happyThen (( do { e <- checkExpr happy_var_2;
						return (HsUnGuardedRhs e) })
	) (\r -> happyReturn (HappyAbsSyn81 r))

happyReduce_455 state_indent = happySpecReduce_1  (213, state_indent) happyReduction_455
happyReduction_455 (HappyAbsSyn82  happy_var_1)
	 =  HappyAbsSyn81
		 (HsGuardedRhss  (reverse happy_var_1)
	)
happyReduction_455 _  = notHappyAtAll 

happyReduce_456 state_indent = happySpecReduce_3  (214, state_indent) happyReduction_456
happyReduction_456 (HappyAbsSyn84  happy_var_3)
	(HappyAbsSyn132  happy_var_2)
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsInfixApp happy_var_1 happy_var_2 happy_var_3
	)
happyReduction_456 _ _ _  = notHappyAtAll 

happyReduce_457 state_indent = happySpecReduce_1  (214, state_indent) happyReduction_457
happyReduction_457 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_457 _  = notHappyAtAll 

happyReduce_458 state_indent = happyReduce 4 (215, state_indent) happyReduction_458
happyReduction_458 ((HappyAbsSyn102  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsCase happy_var_2 happy_var_4
	) `HappyStk` happyRest

happyReduce_459 state_indent = happySpecReduce_2  (215, state_indent) happyReduction_459
happyReduction_459 (HappyAbsSyn84  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (HsNegApp happy_var_2
	)
happyReduction_459 _ _  = notHappyAtAll 

happyReduce_460 state_indent = happySpecReduce_2  (215, state_indent) happyReduction_460
happyReduction_460 (HappyAbsSyn100  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (HsDo happy_var_2
	)
happyReduction_460 _ _  = notHappyAtAll 

happyReduce_461 state_indent = happySpecReduce_1  (215, state_indent) happyReduction_461
happyReduction_461 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_461 _  = notHappyAtAll 

happyReduce_462 state_indent = happySpecReduce_2  (216, state_indent) happyReduction_462
happyReduction_462 (HappyAbsSyn110  happy_var_2)
	(HappyAbsSyn109  happy_var_1)
	 =  HappyAbsSyn109
		 (happy_var_2 : happy_var_1
	)
happyReduction_462 _ _  = notHappyAtAll 

happyReduce_463 state_indent = happySpecReduce_1  (216, state_indent) happyReduction_463
happyReduction_463 (HappyAbsSyn110  happy_var_1)
	 =  HappyAbsSyn109
		 ([happy_var_1]
	)
happyReduction_463 _  = notHappyAtAll 

happyReduce_464 state_indent = happySpecReduce_1  (217, state_indent) happyReduction_464
happyReduction_464 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_464 _  = notHappyAtAll 

happyReduce_465 state_indent = happySpecReduce_3  (217, state_indent) happyReduction_465
happyReduction_465 _
	(HappyAbsSyn50  happy_var_2)
	_
	 =  HappyAbsSyn50
		 (happy_var_2
	)
happyReduction_465 _ _ _  = notHappyAtAll 

happyReduce_466 state_indent = happySpecReduce_1  (218, state_indent) happyReduction_466
happyReduction_466 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_466 _  = notHappyAtAll 

happyReduce_467 state_indent = happySpecReduce_3  (218, state_indent) happyReduction_467
happyReduction_467 _
	(HappyAbsSyn50  happy_var_2)
	_
	 =  HappyAbsSyn50
		 (happy_var_2
	)
happyReduction_467 _ _ _  = notHappyAtAll 

happyReduce_468 state_indent = happySpecReduce_1  (219, state_indent) happyReduction_468
happyReduction_468 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn132
		 (HsQVarOp happy_var_1
	)
happyReduction_468 _  = notHappyAtAll 

happyReduce_469 state_indent = happySpecReduce_1  (219, state_indent) happyReduction_469
happyReduction_469 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn132
		 (HsQConOp happy_var_1
	)
happyReduction_469 _  = notHappyAtAll 

happyReduce_470 state_indent = happySpecReduce_1  (220, state_indent) happyReduction_470
happyReduction_470 _
	 =  HappyAbsSyn9
		 (()
	)

happyReduce_471 state_indent = happySpecReduce_1  (221, state_indent) happyReduction_471
happyReduction_471 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_471 _  = notHappyAtAll 

happyReduce_472 state_indent = happySpecReduce_1  (222, state_indent) happyReduction_472
happyReduction_472 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_472 _  = notHappyAtAll 

happyReduce_473 state_indent = happySpecReduce_1  (223, state_indent) happyReduction_473
happyReduction_473 _
	 =  HappyAbsSyn9
		 (()
	)

happyReduce_474 state_indent = happySpecReduce_1  (224, state_indent) happyReduction_474
happyReduction_474 _
	 =  HappyAbsSyn19
		 (True
	)

happyReduce_475 state_indent = happySpecReduce_3  (225, state_indent) happyReduction_475
happyReduction_475 (HappyAbsSyn25  happy_var_3)
	_
	(HappyAbsSyn24  happy_var_1)
	 =  HappyAbsSyn24
		 (happy_var_3 : happy_var_1
	)
happyReduction_475 _ _ _  = notHappyAtAll 

happyReduce_476 state_indent = happySpecReduce_1  (225, state_indent) happyReduction_476
happyReduction_476 (HappyAbsSyn25  happy_var_1)
	 =  HappyAbsSyn24
		 ([happy_var_1]
	)
happyReduction_476 _  = notHappyAtAll 

happyReduce_477 state_indent = happyMonadReduce 1 (226, state_indent) happyReduction_477
happyReduction_477 ((HappyTerminal ((IntTok happy_var_1), _)) `HappyStk`
	happyRest) tk
	 = happyThen (( checkPrec happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn29 r))

happyReduce_478 state_indent = happySpecReduce_1  (227, state_indent) happyReduction_478
happyReduction_478 _
	 =  HappyAbsSyn30
		 (HsAssocNone
	)

happyReduce_479 state_indent = happySpecReduce_1  (227, state_indent) happyReduction_479
happyReduction_479 _
	 =  HappyAbsSyn30
		 (HsAssocLeft
	)

happyReduce_480 state_indent = happySpecReduce_1  (227, state_indent) happyReduction_480
happyReduction_480 _
	 =  HappyAbsSyn30
		 (HsAssocRight
	)

happyReduce_481 state_indent = happySpecReduce_3  (228, state_indent) happyReduction_481
happyReduction_481 (HappyAbsSyn131  happy_var_3)
	_
	(HappyAbsSyn31  happy_var_1)
	 =  HappyAbsSyn31
		 (happy_var_3 : happy_var_1
	)
happyReduction_481 _ _ _  = notHappyAtAll 

happyReduce_482 state_indent = happySpecReduce_1  (228, state_indent) happyReduction_482
happyReduction_482 (HappyAbsSyn131  happy_var_1)
	 =  HappyAbsSyn31
		 ([happy_var_1]
	)
happyReduction_482 _  = notHappyAtAll 

happyReduce_483 state_indent = happySpecReduce_2  (229, state_indent) happyReduction_483
happyReduction_483 (HappyAbsSyn32  happy_var_2)
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_483 _ _  = notHappyAtAll 

happyReduce_484 state_indent = happySpecReduce_2  (229, state_indent) happyReduction_484
happyReduction_484 (HappyAbsSyn32  happy_var_2)
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_484 _ _  = notHappyAtAll 

happyReduce_485 state_indent = happySpecReduce_1  (229, state_indent) happyReduction_485
happyReduction_485 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_485 _  = notHappyAtAll 

happyReduce_486 state_indent = happySpecReduce_3  (230, state_indent) happyReduction_486
happyReduction_486 (HappyAbsSyn50  happy_var_3)
	_
	(HappyAbsSyn46  happy_var_1)
	 =  HappyAbsSyn46
		 (happy_var_3 : happy_var_1
	)
happyReduction_486 _ _ _  = notHappyAtAll 

happyReduce_487 state_indent = happyMonadReduce 1 (230, state_indent) happyReduction_487
happyReduction_487 ((HappyAbsSyn54  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { n <- checkUnQual happy_var_1;
						return [n] })
	) (\r -> happyReturn (HappyAbsSyn46 r))

happyReduce_488 state_indent = happySpecReduce_1  (231, state_indent) happyReduction_488
happyReduction_488 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn51
		 (HsTyCon happy_var_1
	)
happyReduction_488 _  = notHappyAtAll 

happyReduce_489 state_indent = happySpecReduce_1  (231, state_indent) happyReduction_489
happyReduction_489 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn51
		 (HsTyVar happy_var_1
	)
happyReduction_489 _  = notHappyAtAll 

happyReduce_490 state_indent = happySpecReduce_3  (231, state_indent) happyReduction_490
happyReduction_490 _
	(HappyAbsSyn37  happy_var_2)
	_
	 =  HappyAbsSyn51
		 (HsTyTuple (reverse happy_var_2)
	)
happyReduction_490 _ _ _  = notHappyAtAll 

happyReduce_491 state_indent = happySpecReduce_3  (231, state_indent) happyReduction_491
happyReduction_491 _
	(HappyAbsSyn51  happy_var_2)
	_
	 =  HappyAbsSyn51
		 (HsTyApp list_tycon happy_var_2
	)
happyReduction_491 _ _ _  = notHappyAtAll 

happyReduce_492 state_indent = happySpecReduce_3  (231, state_indent) happyReduction_492
happyReduction_492 _
	(HappyAbsSyn51  happy_var_2)
	_
	 =  HappyAbsSyn51
		 (happy_var_2
	)
happyReduction_492 _ _ _  = notHappyAtAll 

happyReduce_493 state_indent = happyMonadReduce 3 (232, state_indent) happyReduction_493
happyReduction_493 ((HappyAbsSyn51  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn51  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { (c,ts) <- splitTyConApp happy_var_1;
						return (c,map HsUnBangedTy ts++
							[HsBangedTy happy_var_3]) })
	) (\r -> happyReturn (HappyAbsSyn62 r))

happyReduce_494 state_indent = happySpecReduce_2  (232, state_indent) happyReduction_494
happyReduction_494 (HappyAbsSyn64  happy_var_2)
	(HappyAbsSyn62  happy_var_1)
	 =  HappyAbsSyn62
		 ((fst happy_var_1, snd happy_var_1 ++ [happy_var_2] )
	)
happyReduction_494 _ _  = notHappyAtAll 

happyReduce_495 state_indent = happySpecReduce_3  (233, state_indent) happyReduction_495
happyReduction_495 (HappyAbsSyn68  happy_var_3)
	_
	(HappyAbsSyn66  happy_var_1)
	 =  HappyAbsSyn66
		 (happy_var_3 : happy_var_1
	)
happyReduction_495 _ _ _  = notHappyAtAll 

happyReduce_496 state_indent = happySpecReduce_1  (233, state_indent) happyReduction_496
happyReduction_496 (HappyAbsSyn68  happy_var_1)
	 =  HappyAbsSyn66
		 ([happy_var_1]
	)
happyReduction_496 _  = notHappyAtAll 

happyReduce_497 state_indent = happySpecReduce_2  (234, state_indent) happyReduction_497
happyReduction_497 (HappyAbsSyn32  happy_var_2)
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_497 _ _  = notHappyAtAll 

happyReduce_498 state_indent = happySpecReduce_2  (234, state_indent) happyReduction_498
happyReduction_498 (HappyAbsSyn32  happy_var_2)
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_498 _ _  = notHappyAtAll 

happyReduce_499 state_indent = happySpecReduce_1  (234, state_indent) happyReduction_499
happyReduction_499 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_499 _  = notHappyAtAll 

happyReduce_500 state_indent = happySpecReduce_2  (235, state_indent) happyReduction_500
happyReduction_500 (HappyAbsSyn83  happy_var_2)
	(HappyAbsSyn82  happy_var_1)
	 =  HappyAbsSyn82
		 (happy_var_2 : happy_var_1
	)
happyReduction_500 _ _  = notHappyAtAll 

happyReduce_501 state_indent = happySpecReduce_1  (235, state_indent) happyReduction_501
happyReduction_501 (HappyAbsSyn83  happy_var_1)
	 =  HappyAbsSyn82
		 ([happy_var_1]
	)
happyReduction_501 _  = notHappyAtAll 

happyReduce_502 state_indent = happyReduce 5 (236, state_indent) happyReduction_502
happyReduction_502 ((HappyAbsSyn84  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn91  happy_var_3) `HappyStk`
	(HappyAbsSyn5  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsLambda happy_var_2 (reverse happy_var_3) happy_var_5
	) `HappyStk` happyRest

happyReduce_503 state_indent = happyReduce 4 (236, state_indent) happyReduction_503
happyReduction_503 ((HappyAbsSyn84  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn32  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsLet happy_var_2 happy_var_4
	) `HappyStk` happyRest

happyReduce_504 state_indent = happyReduce 6 (236, state_indent) happyReduction_504
happyReduction_504 ((HappyAbsSyn84  happy_var_6) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsIf happy_var_2 happy_var_4 happy_var_6
	) `HappyStk` happyRest

happyReduce_505 state_indent = happySpecReduce_2  (237, state_indent) happyReduction_505
happyReduction_505 (HappyAbsSyn84  happy_var_2)
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsApp happy_var_1 happy_var_2
	)
happyReduction_505 _ _  = notHappyAtAll 

happyReduce_506 state_indent = happySpecReduce_1  (237, state_indent) happyReduction_506
happyReduction_506 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_506 _  = notHappyAtAll 

happyReduce_507 state_indent = happySpecReduce_3  (238, state_indent) happyReduction_507
happyReduction_507 _
	(HappyAbsSyn102  happy_var_2)
	_
	 =  HappyAbsSyn102
		 (happy_var_2
	)
happyReduction_507 _ _ _  = notHappyAtAll 

happyReduce_508 state_indent = happySpecReduce_1  (238, state_indent) happyReduction_508
happyReduction_508 (HappyAbsSyn102  happy_var_1)
	 =  HappyAbsSyn102
		 (reverse happy_var_1
	)
happyReduction_508 _  = notHappyAtAll 

happyReduce_509 state_indent = happyReduce 5 (239, state_indent) happyReduction_509
happyReduction_509 ((HappyAbsSyn84  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn110
		 (HsGuardedAlt happy_var_1 happy_var_3 happy_var_5
	) `HappyStk` happyRest

happyReduce_510 state_indent = happySpecReduce_3  (240, state_indent) happyReduction_510
happyReduction_510 _
	(HappyAbsSyn100  happy_var_2)
	_
	 =  HappyAbsSyn100
		 (happy_var_2
	)
happyReduction_510 _ _ _  = notHappyAtAll 

happyReduce_511 state_indent = happySpecReduce_1  (240, state_indent) happyReduction_511
happyReduction_511 (HappyAbsSyn100  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_1
	)
happyReduction_511 _  = notHappyAtAll 

happyReduce_512 state_indent = happySpecReduce_1  (241, state_indent) happyReduction_512
happyReduction_512 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_512 _  = notHappyAtAll 

happyReduce_513 state_indent = happySpecReduce_3  (241, state_indent) happyReduction_513
happyReduction_513 _
	(HappyAbsSyn54  happy_var_2)
	_
	 =  HappyAbsSyn54
		 (happy_var_2
	)
happyReduction_513 _ _ _  = notHappyAtAll 

happyReduce_514 state_indent = happySpecReduce_1  (242, state_indent) happyReduction_514
happyReduction_514 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_514 _  = notHappyAtAll 

happyReduce_515 state_indent = happySpecReduce_3  (242, state_indent) happyReduction_515
happyReduction_515 _
	(HappyAbsSyn54  happy_var_2)
	_
	 =  HappyAbsSyn54
		 (happy_var_2
	)
happyReduction_515 _ _ _  = notHappyAtAll 

happyReduce_516 state_indent = happySpecReduce_1  (243, state_indent) happyReduction_516
happyReduction_516 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn54
		 (UnQual happy_var_1
	)
happyReduction_516 _  = notHappyAtAll 

happyReduce_517 state_indent = happySpecReduce_1  (243, state_indent) happyReduction_517
happyReduction_517 (HappyTerminal ((QConId happy_var_1), _))
	 =  HappyAbsSyn54
		 (Qual (Module (fst happy_var_1)) (HsIdent (snd happy_var_1))
	)
happyReduction_517 _  = notHappyAtAll 

happyReduce_518 state_indent = happySpecReduce_1  (244, state_indent) happyReduction_518
happyReduction_518 (HappyTerminal ((ConId happy_var_1), _))
	 =  HappyAbsSyn50
		 (HsIdent happy_var_1
	)
happyReduction_518 _  = notHappyAtAll 

happyReduce_519 state_indent = happySpecReduce_1  (245, state_indent) happyReduction_519
happyReduction_519 (HappyTerminal ((ConSym happy_var_1), _))
	 =  HappyAbsSyn50
		 (HsSymbol happy_var_1
	)
happyReduction_519 _  = notHappyAtAll 

happyReduce_520 state_indent = happySpecReduce_1  (246, state_indent) happyReduction_520
happyReduction_520 (HappyTerminal ((VarSym happy_var_1), _))
	 =  HappyAbsSyn50
		 (HsSymbol happy_var_1
	)
happyReduction_520 _  = notHappyAtAll 

happyReduce_521 state_indent = happySpecReduce_1  (246, state_indent) happyReduction_521
happyReduction_521 _
	 =  HappyAbsSyn50
		 (HsSymbol "-"
	)

happyReduce_522 state_indent = happySpecReduce_1  (246, state_indent) happyReduction_522
happyReduction_522 _
	 =  HappyAbsSyn50
		 (HsSymbol "!"
	)

happyReduce_523 state_indent = happySpecReduce_1  (247, state_indent) happyReduction_523
happyReduction_523 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_523 _  = notHappyAtAll 

happyReduce_524 state_indent = happySpecReduce_1  (248, state_indent) happyReduction_524
happyReduction_524 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn25
		 (HsIVar happy_var_1
	)
happyReduction_524 _  = notHappyAtAll 

happyReduce_525 state_indent = happySpecReduce_1  (248, state_indent) happyReduction_525
happyReduction_525 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn25
		 (HsIAbs happy_var_1
	)
happyReduction_525 _  = notHappyAtAll 

happyReduce_526 state_indent = happyReduce 4 (248, state_indent) happyReduction_526
happyReduction_526 (_ `HappyStk`
	_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn50  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn25
		 (HsIThingAll happy_var_1
	) `HappyStk` happyRest

happyReduce_527 state_indent = happySpecReduce_3  (248, state_indent) happyReduction_527
happyReduction_527 _
	_
	(HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn25
		 (HsIThingWith happy_var_1 []
	)
happyReduction_527 _ _ _  = notHappyAtAll 

happyReduce_528 state_indent = happyReduce 4 (248, state_indent) happyReduction_528
happyReduction_528 (_ `HappyStk`
	(HappyAbsSyn26  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn50  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn25
		 (HsIThingWith happy_var_1 (reverse happy_var_3)
	) `HappyStk` happyRest

happyReduce_529 state_indent = happySpecReduce_1  (249, state_indent) happyReduction_529
happyReduction_529 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_529 _  = notHappyAtAll 

happyReduce_530 state_indent = happySpecReduce_2  (249, state_indent) happyReduction_530
happyReduction_530 _
	_
	 =  HappyAbsSyn54
		 (unit_tycon_name
	)

happyReduce_531 state_indent = happySpecReduce_3  (249, state_indent) happyReduction_531
happyReduction_531 _
	_
	_
	 =  HappyAbsSyn54
		 (fun_tycon_name
	)

happyReduce_532 state_indent = happySpecReduce_2  (249, state_indent) happyReduction_532
happyReduction_532 _
	_
	 =  HappyAbsSyn54
		 (list_tycon_name
	)

happyReduce_533 state_indent = happySpecReduce_3  (249, state_indent) happyReduction_533
happyReduction_533 _
	(HappyAbsSyn29  happy_var_2)
	_
	 =  HappyAbsSyn54
		 (tuple_tycon_name happy_var_2
	)
happyReduction_533 _ _ _  = notHappyAtAll 

happyReduce_534 state_indent = happySpecReduce_1  (250, state_indent) happyReduction_534
happyReduction_534 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn64
		 (HsUnBangedTy happy_var_1
	)
happyReduction_534 _  = notHappyAtAll 

happyReduce_535 state_indent = happySpecReduce_2  (250, state_indent) happyReduction_535
happyReduction_535 (HappyAbsSyn51  happy_var_2)
	_
	 =  HappyAbsSyn64
		 (HsBangedTy   happy_var_2
	)
happyReduction_535 _ _  = notHappyAtAll 

happyReduce_536 state_indent = happySpecReduce_3  (251, state_indent) happyReduction_536
happyReduction_536 (HappyAbsSyn64  happy_var_3)
	_
	(HappyAbsSyn46  happy_var_1)
	 =  HappyAbsSyn68
		 ((reverse happy_var_1, happy_var_3)
	)
happyReduction_536 _ _ _  = notHappyAtAll 

happyReduce_537 state_indent = happyMonadReduce 5 (252, state_indent) happyReduction_537
happyReduction_537 ((HappyAbsSyn84  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { g <- checkExpr happy_var_3;
						e <- checkExpr happy_var_5;
						return (HsGuardedRhs happy_var_1 g e) })
	) (\r -> happyReturn (HappyAbsSyn83 r))

happyReduce_538 state_indent = happySpecReduce_2  (253, state_indent) happyReduction_538
happyReduction_538 (HappyAbsSyn92  happy_var_2)
	(HappyAbsSyn91  happy_var_1)
	 =  HappyAbsSyn91
		 (happy_var_2 : happy_var_1
	)
happyReduction_538 _ _  = notHappyAtAll 

happyReduce_539 state_indent = happySpecReduce_1  (253, state_indent) happyReduction_539
happyReduction_539 (HappyAbsSyn92  happy_var_1)
	 =  HappyAbsSyn91
		 ([happy_var_1]
	)
happyReduction_539 _  = notHappyAtAll 

happyReduce_540 state_indent = happyMonadReduce 3 (254, state_indent) happyReduction_540
happyReduction_540 ((HappyAbsSyn84  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn54  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( do { n <- checkUnQual happy_var_1;
						return (HsAsPat n happy_var_3) })
	) (\r -> happyReturn (HappyAbsSyn84 r))

happyReduce_541 state_indent = happySpecReduce_2  (254, state_indent) happyReduction_541
happyReduction_541 (HappyAbsSyn84  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (HsIrrPat happy_var_2
	)
happyReduction_541 _ _  = notHappyAtAll 

happyReduce_542 state_indent = happySpecReduce_1  (254, state_indent) happyReduction_542
happyReduction_542 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_542 _  = notHappyAtAll 

happyReduce_543 state_indent = happySpecReduce_1  (255, state_indent) happyReduction_543
happyReduction_543 (HappyAbsSyn102  happy_var_1)
	 =  HappyAbsSyn102
		 (happy_var_1
	)
happyReduction_543 _  = notHappyAtAll 

happyReduce_544 state_indent = happySpecReduce_3  (256, state_indent) happyReduction_544
happyReduction_544 _
	(HappyAbsSyn102  happy_var_2)
	_
	 =  HappyAbsSyn102
		 (reverse happy_var_2
	)
happyReduction_544 _ _ _  = notHappyAtAll 

happyReduce_545 state_indent = happySpecReduce_3  (256, state_indent) happyReduction_545
happyReduction_545 _
	(HappyAbsSyn102  happy_var_2)
	_
	 =  HappyAbsSyn102
		 (reverse happy_var_2
	)
happyReduction_545 _ _ _  = notHappyAtAll 

happyReduce_546 state_indent = happySpecReduce_1  (257, state_indent) happyReduction_546
happyReduction_546 (HappyAbsSyn100  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_1
	)
happyReduction_546 _  = notHappyAtAll 

happyReduce_547 state_indent = happySpecReduce_1  (258, state_indent) happyReduction_547
happyReduction_547 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_547 _  = notHappyAtAll 

happyReduce_548 state_indent = happySpecReduce_3  (258, state_indent) happyReduction_548
happyReduction_548 _
	(HappyAbsSyn50  happy_var_2)
	_
	 =  HappyAbsSyn50
		 (happy_var_2
	)
happyReduction_548 _ _ _  = notHappyAtAll 

happyReduce_549 state_indent = happySpecReduce_1  (259, state_indent) happyReduction_549
happyReduction_549 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_549 _  = notHappyAtAll 

happyReduce_550 state_indent = happySpecReduce_3  (259, state_indent) happyReduction_550
happyReduction_550 _
	(HappyAbsSyn54  happy_var_2)
	_
	 =  HappyAbsSyn54
		 (happy_var_2
	)
happyReduction_550 _ _ _  = notHappyAtAll 

happyReduce_551 state_indent = happySpecReduce_1  (260, state_indent) happyReduction_551
happyReduction_551 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn131
		 (HsVarOp happy_var_1
	)
happyReduction_551 _  = notHappyAtAll 

happyReduce_552 state_indent = happySpecReduce_1  (260, state_indent) happyReduction_552
happyReduction_552 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn131
		 (HsConOp happy_var_1
	)
happyReduction_552 _  = notHappyAtAll 

happyReduce_553 state_indent = happySpecReduce_1  (261, state_indent) happyReduction_553
happyReduction_553 _
	 =  HappyAbsSyn54
		 (list_cons_name
	)

happyReduce_554 state_indent = happySpecReduce_1  (261, state_indent) happyReduction_554
happyReduction_554 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_554 _  = notHappyAtAll 

happyReduce_555 state_indent = happySpecReduce_1  (262, state_indent) happyReduction_555
happyReduction_555 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn54
		 (UnQual happy_var_1
	)
happyReduction_555 _  = notHappyAtAll 

happyReduce_556 state_indent = happySpecReduce_1  (262, state_indent) happyReduction_556
happyReduction_556 (HappyTerminal ((QVarId happy_var_1), _))
	 =  HappyAbsSyn54
		 (Qual (Module (fst happy_var_1)) (HsIdent (snd happy_var_1))
	)
happyReduction_556 _  = notHappyAtAll 

happyReduce_557 state_indent = happySpecReduce_1  (263, state_indent) happyReduction_557
happyReduction_557 (HappyTerminal ((VarId happy_var_1), _))
	 =  HappyAbsSyn50
		 (HsIdent happy_var_1
	)
happyReduction_557 _  = notHappyAtAll 

happyReduce_558 state_indent = happySpecReduce_1  (263, state_indent) happyReduction_558
happyReduction_558 _
	 =  HappyAbsSyn50
		 (HsIdent "as"
	)

happyReduce_559 state_indent = happySpecReduce_1  (263, state_indent) happyReduction_559
happyReduction_559 _
	 =  HappyAbsSyn50
		 (HsIdent "export"
	)

happyReduce_560 state_indent = happySpecReduce_1  (263, state_indent) happyReduction_560
happyReduction_560 _
	 =  HappyAbsSyn50
		 (HsIdent "hiding"
	)

happyReduce_561 state_indent = happySpecReduce_1  (263, state_indent) happyReduction_561
happyReduction_561 _
	 =  HappyAbsSyn50
		 (HsIdent "qualified"
	)

happyReduce_562 state_indent = happySpecReduce_1  (263, state_indent) happyReduction_562
happyReduction_562 _
	 =  HappyAbsSyn50
		 (HsIdent "safe"
	)

happyReduce_563 state_indent = happySpecReduce_1  (263, state_indent) happyReduction_563
happyReduction_563 _
	 =  HappyAbsSyn50
		 (HsIdent "unsafe"
	)

happyReduce_564 state_indent = happySpecReduce_1  (264, state_indent) happyReduction_564
happyReduction_564 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn54
		 (UnQual happy_var_1
	)
happyReduction_564 _  = notHappyAtAll 

happyReduce_565 state_indent = happySpecReduce_1  (264, state_indent) happyReduction_565
happyReduction_565 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_565 _  = notHappyAtAll 

happyReduce_566 state_indent = happySpecReduce_3  (265, state_indent) happyReduction_566
happyReduction_566 (HappyAbsSyn27  happy_var_3)
	_
	(HappyAbsSyn26  happy_var_1)
	 =  HappyAbsSyn26
		 (happy_var_3 : happy_var_1
	)
happyReduction_566 _ _ _  = notHappyAtAll 

happyReduce_567 state_indent = happySpecReduce_1  (265, state_indent) happyReduction_567
happyReduction_567 (HappyAbsSyn27  happy_var_1)
	 =  HappyAbsSyn26
		 ([happy_var_1]
	)
happyReduction_567 _  = notHappyAtAll 

happyReduce_568 state_indent = happySpecReduce_1  (266, state_indent) happyReduction_568
happyReduction_568 (HappyAbsSyn51  happy_var_1)
	 =  HappyAbsSyn64
		 (HsUnBangedTy happy_var_1
	)
happyReduction_568 _  = notHappyAtAll 

happyReduce_569 state_indent = happySpecReduce_2  (266, state_indent) happyReduction_569
happyReduction_569 (HappyAbsSyn51  happy_var_2)
	_
	 =  HappyAbsSyn64
		 (HsBangedTy   happy_var_2
	)
happyReduction_569 _ _  = notHappyAtAll 

happyReduce_570 state_indent = happyMonadReduce 1 (267, state_indent) happyReduction_570
happyReduction_570 ((HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkPattern happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn92 r))

happyReduce_571 state_indent = happyMonadReduce 3 (268, state_indent) happyReduction_571
happyReduction_571 (_ `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( mkRecConstrOrUpdate happy_var_1 [])
	) (\r -> happyReturn (HappyAbsSyn84 r))

happyReduce_572 state_indent = happyMonadReduce 4 (268, state_indent) happyReduction_572
happyReduction_572 (_ `HappyStk`
	(HappyAbsSyn118  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( mkRecConstrOrUpdate happy_var_1 (reverse happy_var_3))
	) (\r -> happyReturn (HappyAbsSyn84 r))

happyReduce_573 state_indent = happySpecReduce_1  (268, state_indent) happyReduction_573
happyReduction_573 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_573 _  = notHappyAtAll 

happyReduce_574 state_indent = happySpecReduce_2  (269, state_indent) happyReduction_574
happyReduction_574 _
	(HappyAbsSyn29  happy_var_1)
	 =  HappyAbsSyn29
		 (happy_var_1 + 1
	)
happyReduction_574 _ _  = notHappyAtAll 

happyReduce_575 state_indent = happySpecReduce_1  (269, state_indent) happyReduction_575
happyReduction_575 _
	 =  HappyAbsSyn29
		 (1
	)

happyReduce_576 state_indent = happySpecReduce_2  (270, state_indent) happyReduction_576
happyReduction_576 (HappyAbsSyn102  happy_var_2)
	(HappyAbsSyn102  happy_var_1)
	 =  HappyAbsSyn102
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_576 _ _  = notHappyAtAll 

happyReduce_577 state_indent = happySpecReduce_1  (270, state_indent) happyReduction_577
happyReduction_577 (HappyAbsSyn102  happy_var_1)
	 =  HappyAbsSyn102
		 (happy_var_1
	)
happyReduction_577 _  = notHappyAtAll 

happyReduce_578 state_indent = happySpecReduce_2  (271, state_indent) happyReduction_578
happyReduction_578 (HappyAbsSyn100  happy_var_2)
	(HappyAbsSyn100  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_1 ++ happy_var_2
	)
happyReduction_578 _ _  = notHappyAtAll 

happyReduce_579 state_indent = happySpecReduce_1  (271, state_indent) happyReduction_579
happyReduction_579 (HappyAbsSyn100  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_1
	)
happyReduction_579 _  = notHappyAtAll 

happyReduce_580 state_indent = happySpecReduce_1  (272, state_indent) happyReduction_580
happyReduction_580 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_580 _  = notHappyAtAll 

happyReduce_581 state_indent = happySpecReduce_3  (272, state_indent) happyReduction_581
happyReduction_581 _
	(HappyAbsSyn50  happy_var_2)
	_
	 =  HappyAbsSyn50
		 (happy_var_2
	)
happyReduction_581 _ _ _  = notHappyAtAll 

happyReduce_582 state_indent = happySpecReduce_1  (273, state_indent) happyReduction_582
happyReduction_582 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn54
		 (UnQual happy_var_1
	)
happyReduction_582 _  = notHappyAtAll 

happyReduce_583 state_indent = happySpecReduce_1  (273, state_indent) happyReduction_583
happyReduction_583 (HappyTerminal ((QConSym happy_var_1), _))
	 =  HappyAbsSyn54
		 (Qual (Module (fst happy_var_1)) (HsSymbol (snd happy_var_1))
	)
happyReduction_583 _  = notHappyAtAll 

happyReduce_584 state_indent = happySpecReduce_1  (274, state_indent) happyReduction_584
happyReduction_584 (HappyTerminal ((QVarSym happy_var_1), _))
	 =  HappyAbsSyn54
		 (Qual (Module (fst happy_var_1)) (HsSymbol (snd happy_var_1))
	)
happyReduction_584 _  = notHappyAtAll 

happyReduce_585 state_indent = happySpecReduce_1  (275, state_indent) happyReduction_585
happyReduction_585 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn50
		 (happy_var_1
	)
happyReduction_585 _  = notHappyAtAll 

happyReduce_586 state_indent = happySpecReduce_1  (276, state_indent) happyReduction_586
happyReduction_586 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn27
		 (HsVarName happy_var_1
	)
happyReduction_586 _  = notHappyAtAll 

happyReduce_587 state_indent = happySpecReduce_1  (276, state_indent) happyReduction_587
happyReduction_587 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn27
		 (HsConName happy_var_1
	)
happyReduction_587 _  = notHappyAtAll 

happyReduce_588 state_indent = happySpecReduce_1  (277, state_indent) happyReduction_588
happyReduction_588 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn84
		 (HsVar happy_var_1
	)
happyReduction_588 _  = notHappyAtAll 

happyReduce_589 state_indent = happySpecReduce_1  (277, state_indent) happyReduction_589
happyReduction_589 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (happy_var_1
	)
happyReduction_589 _  = notHappyAtAll 

happyReduce_590 state_indent = happySpecReduce_1  (277, state_indent) happyReduction_590
happyReduction_590 (HappyAbsSyn146  happy_var_1)
	 =  HappyAbsSyn84
		 (HsLit happy_var_1
	)
happyReduction_590 _  = notHappyAtAll 

happyReduce_591 state_indent = happySpecReduce_3  (277, state_indent) happyReduction_591
happyReduction_591 _
	(HappyAbsSyn84  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (HsParen happy_var_2
	)
happyReduction_591 _ _ _  = notHappyAtAll 

happyReduce_592 state_indent = happySpecReduce_3  (277, state_indent) happyReduction_592
happyReduction_592 _
	(HappyAbsSyn97  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (HsTuple (reverse happy_var_2)
	)
happyReduction_592 _ _ _  = notHappyAtAll 

happyReduce_593 state_indent = happySpecReduce_3  (277, state_indent) happyReduction_593
happyReduction_593 _
	(HappyAbsSyn84  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (happy_var_2
	)
happyReduction_593 _ _ _  = notHappyAtAll 

happyReduce_594 state_indent = happyReduce 4 (277, state_indent) happyReduction_594
happyReduction_594 (_ `HappyStk`
	(HappyAbsSyn132  happy_var_3) `HappyStk`
	(HappyAbsSyn84  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsLeftSection happy_var_2 happy_var_3
	) `HappyStk` happyRest

happyReduce_595 state_indent = happyReduce 4 (277, state_indent) happyReduction_595
happyReduction_595 (_ `HappyStk`
	(HappyAbsSyn84  happy_var_3) `HappyStk`
	(HappyAbsSyn132  happy_var_2) `HappyStk`
	_ `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsRightSection happy_var_2 happy_var_3
	) `HappyStk` happyRest

happyReduce_596 state_indent = happySpecReduce_1  (277, state_indent) happyReduction_596
happyReduction_596 _
	 =  HappyAbsSyn84
		 (HsWildCard
	)

happyReduce_597 state_indent = happySpecReduce_1  (278, state_indent) happyReduction_597
happyReduction_597 (HappyAbsSyn118  happy_var_1)
	 =  HappyAbsSyn118
		 (happy_var_1
	)
happyReduction_597 _  = notHappyAtAll 

happyReduce_598 state_indent = happySpecReduce_3  (279, state_indent) happyReduction_598
happyReduction_598 (HappyAbsSyn84  happy_var_3)
	_
	(HappyAbsSyn97  happy_var_1)
	 =  HappyAbsSyn97
		 (happy_var_3 : happy_var_1
	)
happyReduction_598 _ _ _  = notHappyAtAll 

happyReduce_599 state_indent = happySpecReduce_3  (279, state_indent) happyReduction_599
happyReduction_599 (HappyAbsSyn84  happy_var_3)
	_
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn97
		 ([happy_var_3,happy_var_1]
	)
happyReduction_599 _ _ _  = notHappyAtAll 

happyReduce_600 state_indent = happySpecReduce_1  (280, state_indent) happyReduction_600
happyReduction_600 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsList [happy_var_1]
	)
happyReduction_600 _  = notHappyAtAll 

happyReduce_601 state_indent = happySpecReduce_1  (280, state_indent) happyReduction_601
happyReduction_601 (HappyAbsSyn97  happy_var_1)
	 =  HappyAbsSyn84
		 (HsList (reverse happy_var_1)
	)
happyReduction_601 _  = notHappyAtAll 

happyReduce_602 state_indent = happySpecReduce_2  (280, state_indent) happyReduction_602
happyReduction_602 _
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsEnumFrom happy_var_1
	)
happyReduction_602 _ _  = notHappyAtAll 

happyReduce_603 state_indent = happyReduce 4 (280, state_indent) happyReduction_603
happyReduction_603 (_ `HappyStk`
	(HappyAbsSyn84  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsEnumFromThen happy_var_1 happy_var_3
	) `HappyStk` happyRest

happyReduce_604 state_indent = happySpecReduce_3  (280, state_indent) happyReduction_604
happyReduction_604 (HappyAbsSyn84  happy_var_3)
	_
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsEnumFromTo happy_var_1 happy_var_3
	)
happyReduction_604 _ _ _  = notHappyAtAll 

happyReduce_605 state_indent = happyReduce 5 (280, state_indent) happyReduction_605
happyReduction_605 ((HappyAbsSyn84  happy_var_5) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_3) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn84  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn84
		 (HsEnumFromThenTo happy_var_1 happy_var_3 happy_var_5
	) `HappyStk` happyRest

happyReduce_606 state_indent = happySpecReduce_3  (280, state_indent) happyReduction_606
happyReduction_606 (HappyAbsSyn100  happy_var_3)
	_
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn84
		 (HsListComp happy_var_1 (reverse happy_var_3)
	)
happyReduction_606 _ _ _  = notHappyAtAll 

happyReduce_607 state_indent = happySpecReduce_3  (281, state_indent) happyReduction_607
happyReduction_607 (HappyAbsSyn120  happy_var_3)
	_
	(HappyAbsSyn118  happy_var_1)
	 =  HappyAbsSyn118
		 (happy_var_3 : happy_var_1
	)
happyReduction_607 _ _ _  = notHappyAtAll 

happyReduce_608 state_indent = happySpecReduce_1  (281, state_indent) happyReduction_608
happyReduction_608 (HappyAbsSyn120  happy_var_1)
	 =  HappyAbsSyn118
		 ([happy_var_1]
	)
happyReduction_608 _  = notHappyAtAll 

happyReduce_609 state_indent = happySpecReduce_2  (282, state_indent) happyReduction_609
happyReduction_609 _
	_
	 =  HappyAbsSyn84
		 (unit_con
	)

happyReduce_610 state_indent = happySpecReduce_2  (282, state_indent) happyReduction_610
happyReduction_610 _
	_
	 =  HappyAbsSyn84
		 (HsList []
	)

happyReduce_611 state_indent = happySpecReduce_3  (282, state_indent) happyReduction_611
happyReduction_611 _
	(HappyAbsSyn29  happy_var_2)
	_
	 =  HappyAbsSyn84
		 (tuple_con happy_var_2
	)
happyReduction_611 _ _ _  = notHappyAtAll 

happyReduce_612 state_indent = happySpecReduce_1  (282, state_indent) happyReduction_612
happyReduction_612 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn84
		 (HsCon happy_var_1
	)
happyReduction_612 _  = notHappyAtAll 

happyReduce_613 state_indent = happySpecReduce_1  (283, state_indent) happyReduction_613
happyReduction_613 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn132
		 (HsQVarOp happy_var_1
	)
happyReduction_613 _  = notHappyAtAll 

happyReduce_614 state_indent = happySpecReduce_1  (283, state_indent) happyReduction_614
happyReduction_614 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn132
		 (HsQConOp happy_var_1
	)
happyReduction_614 _  = notHappyAtAll 

happyReduce_615 state_indent = happySpecReduce_1  (284, state_indent) happyReduction_615
happyReduction_615 (HappyTerminal ((IntTok happy_var_1), _))
	 =  HappyAbsSyn146
		 (HsInt happy_var_1
	)
happyReduction_615 _  = notHappyAtAll 

happyReduce_616 state_indent = happySpecReduce_1  (284, state_indent) happyReduction_616
happyReduction_616 (HappyTerminal ((Character happy_var_1), _))
	 =  HappyAbsSyn146
		 (HsChar happy_var_1
	)
happyReduction_616 _  = notHappyAtAll 

happyReduce_617 state_indent = happySpecReduce_1  (284, state_indent) happyReduction_617
happyReduction_617 (HappyTerminal ((FloatTok happy_var_1), _))
	 =  HappyAbsSyn146
		 (HsFrac happy_var_1
	)
happyReduction_617 _  = notHappyAtAll 

happyReduce_618 state_indent = happySpecReduce_1  (284, state_indent) happyReduction_618
happyReduction_618 (HappyTerminal ((StringTok happy_var_1), _))
	 =  HappyAbsSyn146
		 (HsString happy_var_1
	)
happyReduction_618 _  = notHappyAtAll 

happyReduce_619 state_indent = happySpecReduce_3  (285, state_indent) happyReduction_619
happyReduction_619 (HappyAbsSyn84  happy_var_3)
	_
	(HappyAbsSyn97  happy_var_1)
	 =  HappyAbsSyn97
		 (happy_var_3 : happy_var_1
	)
happyReduction_619 _ _ _  = notHappyAtAll 

happyReduce_620 state_indent = happySpecReduce_3  (285, state_indent) happyReduction_620
happyReduction_620 (HappyAbsSyn84  happy_var_3)
	_
	(HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn97
		 ([happy_var_3,happy_var_1]
	)
happyReduction_620 _ _ _  = notHappyAtAll 

happyReduce_621 state_indent = happySpecReduce_3  (286, state_indent) happyReduction_621
happyReduction_621 (HappyAbsSyn101  happy_var_3)
	_
	(HappyAbsSyn100  happy_var_1)
	 =  HappyAbsSyn100
		 (happy_var_3 : happy_var_1
	)
happyReduction_621 _ _ _  = notHappyAtAll 

happyReduce_622 state_indent = happySpecReduce_1  (286, state_indent) happyReduction_622
happyReduction_622 (HappyAbsSyn101  happy_var_1)
	 =  HappyAbsSyn100
		 ([happy_var_1]
	)
happyReduction_622 _  = notHappyAtAll 

happyReduce_623 state_indent = happySpecReduce_3  (287, state_indent) happyReduction_623
happyReduction_623 (HappyAbsSyn84  happy_var_3)
	_
	(HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn120
		 (HsFieldUpdate happy_var_1 happy_var_3
	)
happyReduction_623 _ _ _  = notHappyAtAll 

happyReduce_624 state_indent = happySpecReduce_1  (288, state_indent) happyReduction_624
happyReduction_624 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_624 _  = notHappyAtAll 

happyReduce_625 state_indent = happySpecReduce_3  (288, state_indent) happyReduction_625
happyReduction_625 _
	(HappyAbsSyn54  happy_var_2)
	_
	 =  HappyAbsSyn54
		 (happy_var_2
	)
happyReduction_625 _ _ _  = notHappyAtAll 

happyReduce_626 state_indent = happySpecReduce_1  (289, state_indent) happyReduction_626
happyReduction_626 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_626 _  = notHappyAtAll 

happyReduce_627 state_indent = happySpecReduce_3  (289, state_indent) happyReduction_627
happyReduction_627 _
	(HappyAbsSyn54  happy_var_2)
	_
	 =  HappyAbsSyn54
		 (happy_var_2
	)
happyReduction_627 _ _ _  = notHappyAtAll 

happyReduce_628 state_indent = happyReduce 4 (290, state_indent) happyReduction_628
happyReduction_628 ((HappyAbsSyn84  happy_var_4) `HappyStk`
	_ `HappyStk`
	(HappyAbsSyn5  happy_var_2) `HappyStk`
	(HappyAbsSyn92  happy_var_1) `HappyStk`
	happyRest)
	 = HappyAbsSyn101
		 (HsGenerator happy_var_2 happy_var_1 happy_var_4
	) `HappyStk` happyRest

happyReduce_629 state_indent = happySpecReduce_1  (290, state_indent) happyReduction_629
happyReduction_629 (HappyAbsSyn84  happy_var_1)
	 =  HappyAbsSyn101
		 (HsQualifier happy_var_1
	)
happyReduction_629 _  = notHappyAtAll 

happyReduce_630 state_indent = happySpecReduce_2  (290, state_indent) happyReduction_630
happyReduction_630 (HappyAbsSyn32  happy_var_2)
	_
	 =  HappyAbsSyn101
		 (HsLetStmt happy_var_2
	)
happyReduction_630 _ _  = notHappyAtAll 

happyReduce_631 state_indent = happySpecReduce_1  (291, state_indent) happyReduction_631
happyReduction_631 (HappyAbsSyn50  happy_var_1)
	 =  HappyAbsSyn54
		 (UnQual happy_var_1
	)
happyReduction_631 _  = notHappyAtAll 

happyReduce_632 state_indent = happySpecReduce_1  (291, state_indent) happyReduction_632
happyReduction_632 (HappyAbsSyn54  happy_var_1)
	 =  HappyAbsSyn54
		 (happy_var_1
	)
happyReduction_632 _  = notHappyAtAll 

happyReduce_633 state_indent = happySpecReduce_1  (292, state_indent) happyReduction_633
happyReduction_633 (HappyTerminal ((VarSym happy_var_1), _))
	 =  HappyAbsSyn50
		 (HsSymbol happy_var_1
	)
happyReduction_633 _  = notHappyAtAll 

happyReduce_634 state_indent = happySpecReduce_1  (292, state_indent) happyReduction_634
happyReduction_634 _
	 =  HappyAbsSyn50
		 (HsSymbol "!"
	)

happyReduce_635 state_indent = happyMonadReduce 0 (293, state_indent) happyReduction_635
happyReduction_635 (happyRest) tk
	 = happyThen (( getSrcLoc)
	) (\r -> happyReturn (HappyAbsSyn5 r))

happyReduce_636 state_indent = happySpecReduce_0  (294, state_indent) happyReduction_636
happyReduction_636  =  HappyAbsSyn9
		 (()
	)

happyReduce_637 state_indent = happySpecReduce_0  (295, state_indent) happyReduction_637
happyReduction_637  =  HappyAbsSyn9
		 (()
	)

happyReduce_638 state_indent = happySpecReduce_0  (296, state_indent) happyReduction_638
happyReduction_638  =  HappyAbsSyn19
		 (False
	)

happyReduce_639 state_indent = happySpecReduce_0  (297, state_indent) happyReduction_639
happyReduction_639  =  HappyAbsSyn20
		 (Nothing
	)

happyReduce_640 state_indent = happySpecReduce_0  (298, state_indent) happyReduction_640
happyReduction_640  =  HappyAbsSyn19
		 (False
	)

happyReduce_641 state_indent = happySpecReduce_0  (299, state_indent) happyReduction_641
happyReduction_641  =  HappyAbsSyn29
		 (9
	)

happyReduce_642 state_indent = happySpecReduce_0  (300, state_indent) happyReduction_642
happyReduction_642  =  HappyAbsSyn37
		 ([]
	)

happyReduce_643 state_indent = happySpecReduce_1  (301, state_indent) happyReduction_643
happyReduction_643 _
	 =  HappyAbsSyn32
		 ([]
	)

happyReduce_644 state_indent = happyMonadReduce 1 (302, state_indent) happyReduction_644
happyReduction_644 ((HappyAbsSyn32  happy_var_1) `HappyStk`
	happyRest) tk
	 = happyThen (( checkRevDecls happy_var_1)
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_645 state_indent = happyMonadReduce 0 (302, state_indent) happyReduction_645
happyReduction_645 (happyRest) tk
	 = happyThen (( checkRevDecls [])
	) (\r -> happyReturn (HappyAbsSyn32 r))

happyReduce_646 state_indent = happySpecReduce_2  (303, state_indent) happyReduction_646
happyReduction_646 (HappyAbsSyn32  happy_var_2)
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_646 _ _  = notHappyAtAll 

happyReduce_647 state_indent = happySpecReduce_1  (303, state_indent) happyReduction_647
happyReduction_647 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_647 _  = notHappyAtAll 

happyReduce_648 state_indent = happySpecReduce_0  (304, state_indent) happyReduction_648
happyReduction_648  =  HappyAbsSyn48
		 (HsSafe
	)

happyReduce_649 state_indent = happySpecReduce_0  (305, state_indent) happyReduction_649
happyReduction_649  =  HappyAbsSyn49
		 (""
	)

happyReduce_650 state_indent = happySpecReduce_0  (306, state_indent) happyReduction_650
happyReduction_650  =  HappyAbsSyn46
		 ([]
	)

happyReduce_651 state_indent = happySpecReduce_1  (307, state_indent) happyReduction_651
happyReduction_651 _
	 =  HappyAbsSyn32
		 ([]
	)

happyReduce_652 state_indent = happySpecReduce_2  (308, state_indent) happyReduction_652
happyReduction_652 (HappyAbsSyn32  happy_var_2)
	(HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_2 ++ happy_var_1
	)
happyReduction_652 _ _  = notHappyAtAll 

happyReduce_653 state_indent = happySpecReduce_1  (308, state_indent) happyReduction_653
happyReduction_653 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_653 _  = notHappyAtAll 

happyReduce_654 state_indent = happySpecReduce_1  (309, state_indent) happyReduction_654
happyReduction_654 _
	 =  HappyAbsSyn32
		 ([]
	)

happyReduce_655 state_indent = happySpecReduce_1  (310, state_indent) happyReduction_655
happyReduction_655 (HappyAbsSyn32  happy_var_1)
	 =  HappyAbsSyn32
		 (happy_var_1
	)
happyReduction_655 _  = notHappyAtAll 

happyReduce_656 state_indent = happySpecReduce_1  (311, state_indent) happyReduction_656
happyReduction_656 _
	 =  HappyAbsSyn32
		 ([]
	)

happyReduce_657 state_indent = happySpecReduce_0  (312, state_indent) happyReduction_657
happyReduction_657  =  HappyAbsSyn32
		 ([]
	)

happyNewToken action sts stk
	= lexer(\(tk, indent) -> 
	let cont i = action i i (tk, indent) (HappyState action) sts stk in
	case tk of {
	EOF -> let eofIndent = Ranges [(0, Fin 0)] in
	action (377, eofIndent) (377, eofIndent) (tk, eofIndent) (HappyState action) sts stk;
	VarId happy_dollar_dollar -> cont (313, indent);
	QVarId happy_dollar_dollar -> cont (314, indent);
	ConId happy_dollar_dollar -> cont (315, indent);
	QConId happy_dollar_dollar -> cont (316, indent);
	VarSym happy_dollar_dollar -> cont (317, indent);
	ConSym happy_dollar_dollar -> cont (318, indent);
	QVarSym happy_dollar_dollar -> cont (319, indent);
	QConSym happy_dollar_dollar -> cont (320, indent);
	IntTok happy_dollar_dollar -> cont (321, indent);
	FloatTok happy_dollar_dollar -> cont (322, indent);
	Character happy_dollar_dollar -> cont (323, indent);
	StringTok happy_dollar_dollar -> cont (324, indent);
	LeftParen -> cont (325, indent);
	RightParen -> cont (326, indent);
	SemiColon -> cont (327, indent);
	LeftCurly -> cont (328, indent);
	RightCurly -> cont (329, indent);
	VRightCurly -> cont (330, indent);
	LeftSquare -> cont (331, indent);
	RightSquare -> cont (332, indent);
	Comma -> cont (333, indent);
	Underscore -> cont (334, indent);
	BackQuote -> cont (335, indent);
	DotDot -> cont (336, indent);
	Colon -> cont (337, indent);
	DoubleColon -> cont (338, indent);
	Equals -> cont (339, indent);
	Backslash -> cont (340, indent);
	Bar -> cont (341, indent);
	LeftArrow -> cont (342, indent);
	RightArrow -> cont (343, indent);
	At -> cont (344, indent);
	Tilde -> cont (345, indent);
	DoubleArrow -> cont (346, indent);
	Minus -> cont (347, indent);
	Exclamation -> cont (348, indent);
	KW_Case -> cont (349, indent);
	KW_Class -> cont (350, indent);
	KW_Data -> cont (351, indent);
	KW_Default -> cont (352, indent);
	KW_Deriving -> cont (353, indent);
	KW_Do -> cont (354, indent);
	KW_Else -> cont (355, indent);
	KW_Foreign -> cont (356, indent);
	KW_If -> cont (357, indent);
	KW_Import -> cont (358, indent);
	KW_In -> cont (359, indent);
	KW_Infix -> cont (360, indent);
	KW_InfixL -> cont (361, indent);
	KW_InfixR -> cont (362, indent);
	KW_Instance -> cont (363, indent);
	KW_Let -> cont (364, indent);
	KW_Module -> cont (365, indent);
	KW_NewType -> cont (366, indent);
	KW_Of -> cont (367, indent);
	KW_Then -> cont (368, indent);
	KW_Type -> cont (369, indent);
	KW_Where -> cont (370, indent);
	KW_As -> cont (371, indent);
	KW_Export -> cont (372, indent);
	KW_Hiding -> cont (373, indent);
	KW_Qualified -> cont (374, indent);
	KW_Safe -> cont (375, indent);
	KW_Unsafe -> cont (376, indent);
	_ -> happyError' (tk,indent)
	})

happyError_ (377, eofIndent) tk = happyError' tk
happyError_ _ tk = happyError' tk

happyThen :: () => P a -> (a -> P b) -> P b
happyThen = (>>=)
happyReturn :: () => a -> P a
happyReturn = (return)
happyThen1 = happyThen
happyReturn1 :: () => a -> P a
happyReturn1 = happyReturn
happyError' :: () => ((Token,Ranges)) -> P a
happyError' tk = (\token -> happyError) tk

parse = happySomeParser where
  happySomeParser = happyThen (happyParse action_0) (\x -> case x of {HappyAbsSyn4 z -> happyReturn z; _other -> notHappyAtAll })

happySeq = happyDontSeq


happyError :: P a
happyError = fail "Parse error"

-- | Parse of a string, which should contain a complete Haskell 98 module.
parseModule :: String -> ParseResult HsModule
parseModule = runParser parse

-- | Parse of a string, which should contain a complete Haskell 98 module.
parseModuleWithMode :: ParseMode -> String -> ParseResult HsModule
parseModuleWithMode mode = runParserWithMode mode parse
{-# LINE 1 "templates/GenericTemplate.hs" #-}
{-# LINE 1 "templates/GenericTemplate.hs" #-}
{-# LINE 1 "<built-in>" #-}
{-# LINE 1 "<command-line>" #-}
{-# LINE 1 "templates/GenericTemplate.hs" #-}
-- Id: GenericTemplate.hs,v 1.26 2005/01/14 14:47:22 simonmar Exp 

{-# LINE 30 "templates/GenericTemplate.hs" #-}








{-# LINE 51 "templates/GenericTemplate.hs" #-}

{-# LINE 61 "templates/GenericTemplate.hs" #-}

{-# LINE 70 "templates/GenericTemplate.hs" #-}

infixr 9 `HappyStk`
data HappyStk a = HappyStk a (HappyStk a)

-----------------------------------------------------------------------------
-- starting the parse

happyParse start_state = happyNewToken (start_state (Ranges [(0, Fin 0)])) notHappyAtAll notHappyAtAll

-----------------------------------------------------------------------------
-- Accepting the parse

-- If the current token is ((1), errorRange), it means we've just accepted a partial
-- parse (a %partial parser).  We must ignore the saved token on the top of
-- the stack in this case.
happyAccept ((1), errorRange) tk st sts (_ `HappyStk` ans `HappyStk` _) =
	happyReturn1 ans
happyAccept j tk st sts (HappyStk ans _) = 
	 (happyReturn1 ans)

-----------------------------------------------------------------------------
-- Arrays only: do the next action

{-# LINE 148 "templates/GenericTemplate.hs" #-}

-----------------------------------------------------------------------------
-- HappyState data type (not arrays)



newtype HappyState b c = HappyState
        ((Int, Ranges) ->                    -- token number
         (Int, Ranges) ->                    -- token number (yes, again)
         b ->                           -- token semantic value
         HappyState b c ->              -- current state
         [HappyState b c] ->            -- state stack
         c)



-----------------------------------------------------------------------------
-- Shifting a token

happyShift new_state ((1), errorRange) tk st sts stk@(x `HappyStk` _) =
     let (i) = (case x of { HappyErrorToken (i) -> i }) in
--     trace "shifting the error token" $
     new_state i i tk (HappyState (new_state)) ((st):(sts)) (stk)

happyShift new_state i tk st sts stk =
     happyNewToken new_state ((st):(sts)) ((HappyTerminal (tk))`HappyStk`stk)

-- happyReduce is specialised for the common cases.

happySpecReduce_0 i fn ((1), errorRange) tk st sts stk
     = happyFail ((1), errorRange) tk st sts stk
happySpecReduce_0 nt fn j tk st@((HappyState (action))) sts stk
     = action nt j tk st ((st):(sts)) (fn `HappyStk` stk)

happySpecReduce_1 i fn ((1), errorRange) tk st sts stk
     = happyFail ((1), errorRange) tk st sts stk
happySpecReduce_1 nt fn j tk _ sts@(((st@(HappyState (action))):(_))) (v1`HappyStk`stk')
     = let r = fn v1 in
       happySeq r (action nt j tk st sts (r `HappyStk` stk'))

happySpecReduce_2 i fn ((1), errorRange) tk st sts stk
     = happyFail ((1), errorRange) tk st sts stk
happySpecReduce_2 nt fn j tk _ ((_):(sts@(((st@(HappyState (action))):(_))))) (v1`HappyStk`v2`HappyStk`stk')
     = let r = fn v1 v2 in
       happySeq r (action nt j tk st sts (r `HappyStk` stk'))

happySpecReduce_3 i fn ((1), errorRange) tk st sts stk
     = happyFail ((1), errorRange) tk st sts stk
happySpecReduce_3 nt fn j tk _ ((_):(((_):(sts@(((st@(HappyState (action))):(_))))))) (v1`HappyStk`v2`HappyStk`v3`HappyStk`stk')
     = let r = fn v1 v2 v3 in
       happySeq r (action nt j tk st sts (r `HappyStk` stk'))

happyReduce k i fn ((1), errorRange) tk st sts stk
     = happyFail ((1), errorRange) tk st sts stk
happyReduce k nt fn j tk st sts stk
     = case happyDrop (k - ((1) :: Int)) sts of
	 sts1@(((st1@(HappyState (action))):(_))) ->
        	let r = fn stk in  -- it doesn't hurt to always seq here...
       		happyDoSeq r (action nt j tk st1 sts1 r)

happyMonadReduce k nt fn ((1), errorRange) tk st sts stk
     = happyFail ((1), errorRange) tk st sts stk
happyMonadReduce k nt fn j tk st sts stk =
        happyThen1 (fn stk tk) (\r -> action nt j tk st1 sts1 (r `HappyStk` drop_stk))
       where (sts1@(((st1@(HappyState (action))):(_)))) = happyDrop k ((st):(sts))
             drop_stk = happyDropStk k stk

happyMonad2Reduce k nt fn ((1), errorRange) tk st sts stk
     = happyFail ((1), errorRange) tk st sts stk
happyMonad2Reduce k nt fn j tk st sts stk =
       happyThen1 (fn stk tk) (\r -> happyNewToken new_state sts1 (r `HappyStk` drop_stk))
       where (sts1@(((st1@(HappyState (action))):(_)))) = happyDrop k ((st):(sts))
             drop_stk = happyDropStk k stk





             new_state = action


happyDrop (0) l = l
happyDrop n ((_):(t)) = happyDrop (n - ((1) :: Int)) t

happyDropStk (0) l = l
happyDropStk n (x `HappyStk` xs) = happyDropStk (n - ((1)::Int)) xs

-----------------------------------------------------------------------------
-- Moving to a new state after a reduction

{-# LINE 246 "templates/GenericTemplate.hs" #-}
happyGoto action j tk st = action j j tk (HappyState action)


-----------------------------------------------------------------------------
-- Error recovery (((1), errorRange) is the error token)

-- parse error if we are in recovery and we fail again
happyFail ((1), errorRange) tk old_st _ stk@(x `HappyStk` _) =
     let (i) = (case x of { HappyErrorToken (i) -> i }) in
--	trace "failing" $ 
        happyError_ i tk

{-  We don't need state discarding for our restricted implementation of
    "error".  In fact, it can cause some bogus parses, so I've disabled it
    for now --SDM

-- discard a state
happyFail  ((1), errorRange) tk old_st (((HappyState (action))):(sts)) 
						(saved_tok `HappyStk` _ `HappyStk` stk) =
--	trace ("discarding state, depth " ++ show (length stk))  $
	action ((1), errorRange) ((1), errorRange) tk (HappyState (action)) sts ((saved_tok`HappyStk`stk))
-}

-- Enter error recovery: generate an error token,
--                       save the old token and carry on.
happyFail  i tk (HappyState (action)) sts stk =
--      trace "entering error recovery" $
        let (_, errorRange) = i in
	action ((1), errorRange) ((1), errorRange) tk (HappyState (action)) sts ( (HappyErrorToken (i)) `HappyStk` stk)

-- Internal happy errors:

notHappyAtAll :: a
notHappyAtAll = error "Internal Happy error\n"

-----------------------------------------------------------------------------
-- Hack to get the typechecker to accept our action functions







-----------------------------------------------------------------------------
-- Seq-ing.  If the --strict flag is given, then Happy emits 
--	happySeq = happyDoSeq
-- otherwise it emits
-- 	happySeq = happyDontSeq

happyDoSeq, happyDontSeq :: a -> b -> b
happyDoSeq   a b = a `seq` b
happyDontSeq a b = b

-----------------------------------------------------------------------------
-- Don't inline any functions from the template.  GHC has a nasty habit
-- of deciding to inline happyGoto everywhere, which increases the size of
-- the generated parser quite a bit.

{-# LINE 313 "templates/GenericTemplate.hs" #-}
{-# NOINLINE happyShift #-}
{-# NOINLINE happySpecReduce_0 #-}
{-# NOINLINE happySpecReduce_1 #-}
{-# NOINLINE happySpecReduce_2 #-}
{-# NOINLINE happySpecReduce_3 #-}
{-# NOINLINE happyReduce #-}
{-# NOINLINE happyMonadReduce #-}
{-# NOINLINE happyGoto #-}
{-# NOINLINE happyFail #-}

--data Nat = Fin Int | Inf deriving (Ord, Eq, Show)
--data Ranges = Ranges [(Int, Nat)] deriving (Ord, Eq, Show) -- Bounds are inclusive
emptyRanges = Ranges []

unionRanges :: Ranges -> Ranges -> Ranges
unionRanges (Ranges x) (Ranges y) = Ranges $ fn (sort (x ++ y)) where
  fn [] = []
  fn ((x_low, Inf):_) = [(x_low, Inf)]
  fn [x] = [x]
  fn ((x_low, Fin x_high):(y_low, y_high):ys)
     | y_low > x_high + 1 = (x_low, Fin x_high) : fn ((y_low, y_high):ys)
     | otherwise = fn ((x_low, max (Fin x_high) y_high) : ys)

minRanges (Ranges rs) = minimum (map fst rs)
maxRanges (Ranges rs) = maximum (map snd rs)
intersectRange :: Ranges -> Ranges -> Ranges
intersectRange (Ranges x) (Ranges y) = Ranges $ fn x y where
  fn [] _ = []
  fn _ [] = []
  fn ((x_low, Fin x_hi):xs) ((y_low, y_hi):ys)
     | x_hi < y_low = fn xs ((y_low, y_hi):ys)
  fn ((x_low, x_hi):xs) ((y_low, Fin y_hi):ys)
     | y_hi < x_low = fn ((x_low, x_hi):xs) ys
  fn ((x_low, x_hi):xs) ((y_low, y_hi):ys)
     | otherwise = (max x_low y_low, min x_hi y_hi) : if x_hi < y_hi
                                                      then fn xs ((y_low, y_hi):ys)
                                                      else fn ((x_low, x_hi):xs) ys

notEmptyRanges (Ranges []) = False
notEmptyRanges _ = True

foo_def = happyFail

-- end of Happy Template.
